/*

*********************************************************************
*********************************************************************

  file bfs.cpp :source code of bfs utility

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/


#include "config.h"
#include "comuni.h"
#include "vector"
using namespace std;

void bfs(char* graphMultiFileName, char* tempMultiFileName, char* outputMultiFileName, int firstNode,long available_RAM, bool printScreen, bool visitPred, bool visitSucc){

  //aux variable to store string name
//  char auxBuffer[DIM_TEXT_FILENAME];

  //file info (graph)
  char infoGraphMultiFile[DIM_TEXT_FILENAME];sprintf(infoGraphMultiFile,"%s.info",graphMultiFileName);

  //file succ (graoh)
  char succGraphMultiFile[DIM_TEXT_FILENAME];sprintf(succGraphMultiFile,"%s.succ",graphMultiFileName);

  //file pred (graoh)
  char predGraphMultiFile[DIM_TEXT_FILENAME];sprintf(predGraphMultiFile,"%s.pred",graphMultiFileName);

  //report multifile
  char reportMultiFile[DIM_TEXT_FILENAME];sprintf(reportMultiFile,"%s.report",outputMultiFileName);

  //conta num di nodi del grafo
  long numNodes = getTotalSizeMultiFiles(infoGraphMultiFile) / sizeof(infoNodo);

  //not a node of the graph
  if (firstNode>=numNodes || firstNode<0) {cout<<"Graph does not contain starting node "<<firstNode<<"\n";return;}

  //nodi da visitare nel prox livello 
  char temp_UnsortedLevelNodes[DIM_TEXT_FILENAME];sprintf(temp_UnsortedLevelNodes,"%s.unsortedLevelNode",tempMultiFileName);//(non ordinati)
  char temp_SortedLevelNodes[DIM_TEXT_FILENAME];  sprintf(temp_SortedLevelNodes,  "%s.sortedLevelNode",tempMultiFileName);//(ordinati)
  char temp_Sort[DIM_TEXT_FILENAME];              sprintf(temp_Sort,"%s.externalSort",tempMultiFileName);//(ordinati)

  //nodo visitato se ho incontrato il nodo e ho considerato tutti i successori - addedCurrentLevel = 1 se il nodo  gia' stato aggiunto nel prox livello
  vector<bool> visitedNodes (numNodes,false); //vettore di bit dei nodi visitati (di cui ho preso i successori)
  vector<bool> enqueuedNodes(numNodes,false); //vettore di bit dei nodi che sono stati aggiunti nella coda (gia' incontrati ma non ancora visitati nel senso di sopra)

  //aux nodo
  infoNodo appoInfoNodo;

  //aux string
  char aux[DIM_TEXT_FILENAME];

  //aux var
  long auxFileSize;

  //total alayzed nodes
  long totalVisitedNodes=0;

  #ifdef PRINT_DEBUG
  printScreen=true;
  #endif

  //aggiungo il primo nodo da visitare...  
  multiFileWriter faux=multiFileWriter(temp_UnsortedLevelNodes,DIM_MAX);
  faux.write(&firstNode,sizeof(firstNode),1);
  enqueuedNodes[firstNode]=true;
  faux.close();

  //file dei risultati
  multiFileWriter fout=multiFileWriter(outputMultiFileName,DIM_MAX);

  //file report
  multiFileWriter freport=multiFileWriter(reportMultiFile,DIM_MAX);

  //open graph files
  multiFileReader fGrafoInfo=multiFileReader(infoGraphMultiFile,-1,false);
  multiFileReader fGrafoSucc=multiFileReader(succGraphMultiFile,-1,false);
  multiFileReader fGrafoPred=multiFileReader(predGraphMultiFile,-1,false);

  //write report header
  sprintf(aux,"%s\t%s\n","level","number of nodes in level");
  if (freport.write(aux,sizeof(char),strlen(aux) )<1){fprintf(stderr,"Program error 11");FINE;}

  if (printScreen) {
    printf("BFS progress for %ld nodes.\n\n",numNodes);    
    printf("|---------+---------------++-------+-------------+-----------------|\n");
    printf("| %c total | visited nodes || level | level nodes | tot level nodes |\n",'%');
    printf("|---------+---------------++-------+-------------+-----------------|\n");
  }
  
  
  //while there are nodes to be used
  // ... FOR EACH LEVEL .......................................................................
  for (long currentLevel=0; (auxFileSize=getTotalSizeMultiFiles(temp_UnsortedLevelNodes))>0 ;currentLevel++){

    long currentLevelVisitedNodes=0;
    long numberNodesCurrentLevel=auxFileSize/sizeof(long);
    timer t=timer(2);

    #ifdef PRINT_DEBUG
    printf("LIVELLO %d\n",currentLevel);
    #endif

    //print
    if (printScreen) {
       t.forcePrint("%c| %6.2f%c | %13ld || %5ld | %11s | %15s |",
                  13,100.0*totalVisitedNodes/numNodes,'%',totalVisitedNodes, currentLevel, "--","sorting nodes");
    }

    //sort current level nodes
    //printf("step 1 - ram %d\n",available_RAM);fflush(stdout);
    externalSort_long(temp_UnsortedLevelNodes, temp_Sort, temp_SortedLevelNodes,
                      true,(char*)"Sorting nodes",available_RAM,64*1024,4,
                      64*1024,10,30,false);
    
    //open for reading level nodes multifile (current level reading, next level writing)
    multiFileReader fCurrentLevelNodes=multiFileReader(temp_SortedLevelNodes,-1,true);
    multiFileWriter fNextLevelNodes=multiFileWriter(temp_UnsortedLevelNodes,DIM_MAX);
    long currentNode;

    //mentre ho nodi da visitare
    while (fCurrentLevelNodes.read(&currentNode,sizeof(currentNode),1)>0){

    
      #ifdef PRINT_DEBUG
      printf("Nodo corrente = %d ",currentNode);
      #endif

      
      /////////////////////////////////////////////////////////////////////////////////////////////////////
      // NON SERVE QUESTO CONTROLLO !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      /////////////////////////////////////////////////////////////////////////////////////////////////////
      //checks if node is already visited (visited all successors) - should be always false because each node is inserted in the queue only once
      bool visited=visitedNodes[currentNode];
      if (visited) {
        printf("Program error 2/bis\n");
        FINE;
        continue;//if the node is visited skip to the next
      }
      /////////////////////////////////////////////////////////////////////////////////////////////////////

      

      //set visited node
      visitedNodes[currentNode]=true;
      totalVisitedNodes++;
      currentLevelVisitedNodes++;

      //Stores the visited node in the output levels file
      if (fout.write(&currentNode,sizeof(currentNode),1)<1){fprintf(stderr,"Program error 4");FINE;}
      

      // = = = = = = = = = = = = = = = read all successors = = = = = = = = = = = = = = = =       

      //read information about "currentNode" node
      fGrafoInfo.seek(sizeof(appoInfoNodo), currentNode);
      if (fGrafoInfo.read(&appoInfoNodo,sizeof(infoNodo),1)<1){fprintf(stderr,"Program error 5");FINE;}
      #ifdef PRINT_DEBUG
      printf(" - successori (outDeg %ld - puntOut %lld) : \n",appoInfoNodo.outDeg,appoInfoNodo.puntOut);
      #endif

      //seek to successors
      fGrafoSucc.seek( sizeof(long), appoInfoNodo.puntOut );

      
      /////////////////////////////////////////////////////////////////////////////
      //                                stampa
      /////////////////////////////////////////////////////////////////////////////
      if (printScreen) {
          if (currentLevelVisitedNodes==numberNodesCurrentLevel)
            t.forcePrint("%c| %6.2f%c | %13ld || %5ld | %11ld | %15ld |",
                  13,100.0*totalVisitedNodes/numNodes,'%',totalVisitedNodes, currentLevel, currentLevelVisitedNodes,numberNodesCurrentLevel);
          else
            t.print("%c| %6.2f%c | %13ld || %5ld | %11ld | %15ld |",
                  13,100.0*totalVisitedNodes/numNodes,'%',totalVisitedNodes, currentLevel, currentLevelVisitedNodes,numberNodesCurrentLevel);
      }
      /////////////////////////////////////////////////////////////////////////////

      
      //read all successors of current Node
      for (int conta=0;conta<appoInfoNodo.outDeg && visitSucc;conta++){
        
        long currentSucc;//read current successor node
        if (fGrafoSucc.read(&currentSucc,sizeof(long),1)<1){fprintf(stderr,"Program error 6");FINE;}
        #ifdef PRINT_DEBUG
        printf("  >>>  %d ",currentSucc);
        #endif
        
        //checks if node is already inserted in the queue
        bool visitedSucc=enqueuedNodes[currentSucc];
        if (visitedSucc) {
          #ifdef PRINT_DEBUG
          printf("(gia' in coda)\n");
          #endif
          continue;//if the node is visited skip to the next
        }
        #ifdef PRINT_DEBUG
        else
          printf("(add coda)\n");
        #endif
        

        //first insertion of node in the queue
        enqueuedNodes[currentSucc]=true;

        //insert current Successor in the queue
        if (fNextLevelNodes.write(&currentSucc,sizeof(currentSucc),1)<1){fprintf(stderr,"Program error 9");FINE;}
      }


      
      //read all pred of current Node
      for (int conta=0;conta<appoInfoNodo.inDeg && visitPred;conta++){

        long currentPred;//read current successor node
        if (fGrafoPred.read(&currentPred,sizeof(currentPred),1)<1){fprintf(stderr,"Program error 6");FINE;}
        #ifdef PRINT_DEBUG
        printf("  >>>  %d ",currentPred);
        #endif

        //checks if node is already inserted in the queue
        bool visitedPred=enqueuedNodes[currentPred];
        if (visitedPred) {
          #ifdef PRINT_DEBUG
          printf("(gia' in coda)\n");
          #endif
          continue;//if the node is visited skip to the next
        }
        #ifdef PRINT_DEBUG
        else
          printf("(add coda)\n");
        #endif

        //first insertion of node in the queue
        enqueuedNodes[currentPred]=true;

        //insert current Successor in the queue
        if (fNextLevelNodes.write(&currentPred,sizeof(currentPred),1)<1){fprintf(stderr,"Program error 9");FINE;}
      }

            
    }


    //close level files
    fNextLevelNodes.close();    
    fCurrentLevelNodes.close();

    //close current level on output multifile
    currentNode=-1;
    if (fout.write(&currentNode,sizeof(currentNode),1)<1){fprintf(stderr,"Program error 10");FINE;}

    //write current level info in the report file
    sprintf(aux,"%ld\t%ld\n",currentLevel,currentLevelVisitedNodes);
    if (freport.write(aux,sizeof(char),strlen(aux) )<1){fprintf(stderr,"Program error 11");FINE;}
    if (printScreen) printf("\n");
    
  }
  // ... END LEVEL ................................................................................

  if (printScreen)  printf("|---------+---------------++-------+-------------+-----------------|\n");

  //remove all temp files
  multiFileReader fapp=multiFileReader(temp_UnsortedLevelNodes,-1,true);fapp.close();

    
  //close graph files
  fGrafoSucc.close(); //succ files
  fGrafoPred.close(); //succ files
  fGrafoInfo.close(); //info files
  fout.close();       //result nodes
  freport.close();    //report files
  
}


