/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file bloccografo.cpp :source code of clique utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "clique.h"
#include "bloccografo.h"

//se possibile legge i successivi 'size' nodi (il blocco successivo) e ritorna il numero
// il numero di nodi letti


//#####################################################################################################
//                  PRIVATE               PRIVATE                PRIVATE
//#####################################################################################################
//legge il blocco successivo se possibile e ritorna il num di nodi NUOVI letti
long bloccoGrafo::leggiProssimoInsiemeDiNodiNeiFile(proprietaNodi* p){

  //deallocazione dei buffer dei pred e succ (INFO resta allocato e pari a size+1 elem)
  if (vetSucc!=NULL)   {delete[] vetSucc;vetSucc=NULL;}
  if (vetPred!=NULL)   {delete[] vetPred;vetPred=NULL;}

  //var di appoggio per la lettura
  infoNodo g;

  //num nodi totali per pred/succ (compresi per i nodi no FAN e non CENTER)
  long numTotSucc=0;
  long numTotPred=0;

  
  //num pred/succ effettivamente presi... (escludo quelli di nodi no FAN/CENTER)
  long numTotPredValidi=0;
  long numTotSuccValidi=0;

  //reset il conteggio dei nodi
  numNodiPresenti=0;//num nodi inseriti
  long i=0;//indice nodo corrente
  long nodoCorrente=nodoInizialeCorrente;//numero progressivo del nodo in analisi

  //per la stampa...
  long long iterazioni=MAX_ITER*20;


  // ------------ analisi di tutti i nodi di INFO fino al max possibile -----------------------------
  // ------------------------------------------------------------------------------------------------
  //leggo tutti i nodi dal file info fino ad un max di size (se non termina prima il file)
  while ( fInfo->read(&g,sizeof(g),1)>0 ){

      if (iterazioni++>MAX_ITER/10){
      	   iterazioni=0;
      	   char cc[100];
      	   sprintf(cc,"INFO node %ld",nodoCorrente);
        	 indiceProgressione(cc);
      }

      //set il puntatore come se tutti i nodi fossero validi ... poi li correggo eliminando quelli non validi
      vetNodi[i].puntIn=numTotPred;
      vetNodi[i].puntOut=numTotSucc;

      //incremento i puntatori ai successori del num di in/out nodi - solo per i nodi FAN/CENTER
      if (p->getCENTER(nodoCorrente))
         numTotPredValidi+=g.inDeg;//conta solo i pred dei nodi CENTER validi

      numTotPred+=g.inDeg;
      		

      if (p->getFAN(nodoCorrente))
		      numTotSuccValidi+=g.outDeg;//conta solo i succ dei nodi FAN validi

      numTotSucc+=g.outDeg;

      //indice successivo
      i++;

      //numero prossimo nodo
      nodoCorrente++;

      numNodiPresenti++;//un nuovo nodo letto

      //se raggiunto il massimo
      if (numNodiPresenti>=size){
          break;//fine lettura
      }
  }
  // ------------------------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------------------------

  // aggiungo un nodo fittizio VUOTO che mi serve per il calcolo dell'in/out DEGREE
  // che e' il punt al primo nodo libero
  vetNodi[i].puntIn=numTotPred;
  vetNodi[i].puntOut=numTotSucc;

  //se faccio la diff tra v[i+1] e v[i] ottengo il grado del nodo i (X costruzione)

  //ultimo nodo analizzato (compreso...)
  long ultimoNodoAnalizzato=nodoCorrente-1;
	
  //calcolo dimensioni dei due vettori IN/OUT (per i solo nodi validi)
  sizePred=numTotPredValidi;
  sizeSucc=numTotSuccValidi;

	//controllo se ho terminato senza aver letto nulla
  if (numNodiPresenti==0) return 0;

  indiceProgressione((char*)"Writing block (ALLOC)");
  if (  (vetPred=new long[sizePred])==NULL){
      printError("Memory allocation error (pred)");
      FINE
  }

  //creazione dei buffer per la lettura dei successori e dei predecessori
  if (  (vetSucc=new long[sizeSucc])==NULL){
      printError("Memory allocation error (succ)");
      FINE
  }



  //var che mi serve per eliminare i buchi dovuti ai successori non validi
  long sottraiPred=0;
  long sottraiSucc=0;

  iterazioni=MAX_ITER+1;

  //legge predecessori e successori di tutti i nodi
  int index=0;


//  cout<<"\n\n";
//  printf("Risparmio memoria blocco %ld: totPred %ld , realPred %ld  -  totSucc %ld , realSucc %ld\n",numeroBlocco, numTotPred,sizePred,numTotSucc,sizeSucc);


  // ----------- per ogni nodo analizzato trovo i predecessori e i successori ----------------------
  // ------------------------------------------------------------------------------------------------
  for (int nodo=nodoInizialeCorrente;nodo<=ultimoNodoAnalizzato;nodo++,index++){

    //stampa progressione
    if (iterazioni++>=MAX_ITER/10){
         iterazioni=0;
         char cc[100];
         sprintf(cc,"P/S node %d",nodo);
         indiceProgressione(cc);
    }

    //calcolo in/out deg del nodo in analisi (quello reale)
    long indeg =vetNodi[index+1].puntIn-vetNodi[index].puntIn;
    long outdeg=vetNodi[index+1].puntOut-vetNodi[index].puntOut;

    //aggiorno il vero punt ai predecessori e successori in base ai pred/succ eliminati prima
    vetNodi[index].puntIn  -= sottraiPred;//vera pos in cui scrivere risp. pred e succ
    vetNodi[index].puntOut -= sottraiSucc;


//    cout<<"-- nodo "<<nodo<<" - OUT "<<outdeg<<" - punt_OUT_nuovo "<<vetNodi[index].puntOut;
//    cout<<" - IN "<<indeg<<" - punt_IN_nuovo "<<vetNodi[index].puntIn<<"\n";





    //---------------------------- PREDECESSORI -----------------------
    //controlla se e' un nodo CENTER (validi i predecessori)
    if (p->getCENTER(nodo)){
		  if ( fPred->read( (vetPred+vetNodi[index].puntIn)  ,sizeof(long),indeg)<indeg ){
    		 cout<<"Number of predecessor of the node "<<nodo<<" is not enough";
     		 FINE;
		  }
    }
    //se non e' un center devo saltare i successori
    else{
      sottraiPred+=indeg;//il nodo corrente non ha predecessori e propago l'informazione

      #ifdef USA_SEEK_LETTURA_PRED_SUCC
      //salto tutti i predecessori
      fPred->seekInAvanti(sizeof(long),indeg);
      #else

      //leggi e butta via i predecessori presenti del nodo corrente
      long rimasti=indeg;
      long buffer[1000];
      while (rimasti>0){

         //num nodi da leggere in questa iterazione
         long leggi=rimasti;
         if (leggi>1000) leggi=1000;

         //lettura del buffer e controllo la lettura corretta
         long letti=fPred->read(buffer,sizeof(long),leggi);
         if (letti<leggi) {
		    		 cout<<"Number of predecessor of the node "<<nodo<<" is not enough (Pred Jump)";
    		 		 FINE;
         }

         //calcolo nuovi rimasti
         rimasti-=letti;
      }
      #endif

    }


    //---------------------------- SUCCESSORI -----------------------
    //controlla se e' un nodo FAN (validi i predecessori)
    if (p->getFAN(nodo)){
		  if ( fSucc->read( (vetSucc+vetNodi[index].puntOut)  ,sizeof(long),outdeg)<outdeg ){
    		 cout<<"Predecessor number for the node "<<nodo<<" is not enough";
     		 FINE;
		  }
    }
    //se non e' un center devo saltare i successori
    else{
      sottraiSucc+=outdeg;//il nodo corrente non ha predecessori e propago l'informazione

      #ifdef USA_SEEK_LETTURA_PRED_SUCC
      //salto tutti i successori
      fSucc->seekInAvanti(sizeof(long),outdeg);

      #else

      //leggi e butta via i predecessori presenti del nodo corrente
      long rimasti=outdeg;
      long buffer[1000];
      while (rimasti>0){

         //num nodi da leggere in questa iterazione
         long leggi=rimasti;
         if (leggi>1000) leggi=1000;

         //lettura del buffer e controllo la lettura corretta
         long letti=fSucc->read(buffer,sizeof(long),leggi);
         if (letti<leggi) {
		    		 cout<<"Successor number for the node "<<nodo<<" is not enough";
    		 		 FINE;
         }

         //calcolo nuovi rimasti
         rimasti-=letti;
      }
			#endif

    }


  }

  //aggiorna l'ultimo puntatore
  //aggiorno il vero punt ai predecessori e successori
  vetNodi[index].puntIn  -= sottraiPred;
  vetNodi[index].puntOut -= sottraiSucc;



  if (vetNodi[index].puntIn!=numTotPredValidi){
      cout<<"Pointers calculation error: PRED not valid";
      FINE;
  }


  if (vetNodi[index].puntOut!=numTotSuccValidi){
      cout<<"Pointers calculation error: SUCC not valid";
      FINE;
  }

  return numNodiPresenti;
}





//#####################################################################################################
//#####################################################################################################
void bloccoGrafo::cancella(){

  if (cancellato) return;
  if (vetNodi!=NULL)   {delete[] vetNodi;vetNodi=NULL;}
  if (vetSucc!=NULL)   {delete[] vetSucc;vetSucc=NULL;vetSucc=NULL;}
  if (vetPred!=NULL)   {delete[] vetPred;vetPred=NULL;vetPred=NULL;}
  if (fInfo!=NULL)     {fInfo->cancella(); delete fInfo;fInfo=NULL;}
  if (fPred!=NULL)     {fPred->cancella(); delete fPred;fPred=NULL;}
  if (fSucc!=NULL)     {fSucc->cancella(); delete fSucc;fSucc=NULL;}
  cancellato=true;

//REPORT e PROPRIETANODI NON VANN CANCELLATI X CHE' E' PASSATO ESTERNO

}





//#####################################################################################################
//#####################################################################################################
//legge il prossimo blocco e ritorna il numero di blocco corrispondente letto (da 0)
//-1 se finito
long bloccoGrafo::leggiBloccoSuccessivo(proprietaNodi* p){
    if (cancellato    )  return -1;

    //controllo se e' stato inizializzato
    if (numeroBlocco<0)  return -1;

    //aggiorno il nodo iniziale
    nodoInizialeCorrente+=numNodiPresenti;

    //se ho finito il grafo, ritorno -1
    if (leggiProssimoInsiemeDiNodiNeiFile(p)<=0){
       numeroBlocco=-1;
       numNodiPresenti=0;
    }
    else
       numeroBlocco++;
    return numeroBlocco;
}





//#####################################################################################################
//#####################################################################################################
//legge il prossimo blocco e ritorna il numero di blocco corrispondente letto (da 0)
//-1 se finito
long bloccoGrafo::leggiBloccoIniziale(proprietaNodi* p){

    if (cancellato) return -1;

    //se sono gia' state eseguite letture, chiudo tutti  i files
    if (fInfo!=NULL)     {fInfo->cancella(); delete fInfo;fInfo=NULL;}
    if (fPred!=NULL)     {fPred->cancella(); delete fPred;fPred=NULL;}
    if (fSucc!=NULL)     {fSucc->cancella(); delete fSucc;fSucc=NULL;}

    nodoInizialeCorrente=0;
    numeroBlocco=0;

    //apertura dei files INFO/PRED/SUCC
    char aux[1000];
    int n=r->getNomeAndNumFilesInfo(aux);
    fInfo=new bufferedThreadFileInputStream(aux,n,DIM_BUFFER_LETTURA_GENERICA,NUM_BUFFER_LETTURA_GENERICA );
    check(fInfo);

    n=r->getNomeAndNumFilesPred(aux);
    #ifndef USA_SEEK_LETTURA_PRED_SUCC
    fPred=new bufferedThreadFileInputStream(aux,n,DIM_BUFFER_LETTURA_GENERICA,NUM_BUFFER_LETTURA_GENERICA );
    #else
    fPred=new multiFileReader(aux,n);
    #endif
    check(fPred);

    n=r->getNomeAndNumFilesSucc(aux);
    #ifndef USA_SEEK_LETTURA_PRED_SUCC
    fSucc=new bufferedThreadFileInputStream(aux,n,DIM_BUFFER_LETTURA_GENERICA,NUM_BUFFER_LETTURA_GENERICA );
    #else
    fSucc=new multiFileReader(aux,n);
    #endif
    check(fSucc);

    //se ho finito il grafo gia' ora, vuol dire che il grafo e' vuoto
    //leggo i successivi nodi nel file grafo
    if (leggiProssimoInsiemeDiNodiNeiFile(p)<=0){
        numeroBlocco=-1;
        numNodiPresenti=0;
    }

    //ritorno il num di blocco corrente (-1 se il file grafo e' vuoto)
    return numeroBlocco;
}






//#####################################################################################################
//#####################################################################################################
long* bloccoGrafo::getPred(long nodo,long& numero){
  //ritorna il puntatore al pred del nodo e il loro num (DA NON MODOFICARE!!!!)
  //RITORNA NULL se il nodo no e' in lista

  if (numeroBlocco<0)	return NULL;

  if (nodo<nodoInizialeCorrente || nodo>=numNodiPresenti+nodoInizialeCorrente)
     return NULL;

  //puntatore diretto alla struttura (da 0 a n max -1)
  long puntRelativo=nodo-nodoInizialeCorrente;

  //calcolo indegree
  numero=vetNodi[puntRelativo+1].puntIn-vetNodi[puntRelativo].puntIn;

  //puntatore al vettore, a partire dalla posizione dei nodi IN
  return (vetPred+vetNodi[puntRelativo].puntIn);
}






//#####################################################################################################
//#####################################################################################################
long* bloccoGrafo::getSucc(long nodo,long& numero){
  //ritorna il puntatore al pred del nodo e il loro num (DA NON MODOFICARE!!!!)
  //RITORNA NULL se il nodo no e' in lista

  if (numeroBlocco<0)	return NULL;

  if (nodo<nodoInizialeCorrente || nodo>=numNodiPresenti+nodoInizialeCorrente)
     return NULL;

  //puntatore diretto alla struttura (da 0 a n max -1)
  long puntRelativo=nodo-nodoInizialeCorrente;

  //calcolo outdeg
  numero=vetNodi[puntRelativo+1].puntOut-vetNodi[puntRelativo].puntOut;

  //puntatore al vettore, a partire dalla posizione dei nodi OUT
  return (vetSucc+vetNodi[puntRelativo].puntOut);
}








//#####################################################################################################
//#############################################  ########################################################
bloccoGrafo::bloccoGrafo(report2* r,long sizeBlocco){

  //la struttura non e' ancora inizializzata (bloccoCorrente = -1)
  cancellato=FALSE;

  //analisi del file di report
  this->r=r;//new report(nomeFileReport);

  //dimensione in nodi di ciascun blocco
  this->size=sizeBlocco;

  //reset delle grandezze di count
  nodoInizialeCorrente=0;
  numNodiPresenti=0;
  vetPred=vetSucc=NULL;
  sizePred=sizeSucc=0;
  fInfo=NULL;
  fSucc=fPred=NULL;
  numeroBlocco=-1;

  //allocazione della memoria per i nodi (+1 per la chiusura)
  vetNodi=new nodoLocale[size+1];

  //calcolo il numero di blocchi totali che si possono ottentere
  numBlocchiTotali=r->getNumNodiTotali()/size;
  if (r->getNumNodiTotali()%size!=0)
       numBlocchiTotali++;
	
}

