/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file bloccografo.h :source code of clique utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "comuni.h"
#include "proprietanodi.h"


#ifndef BLOCCOGRAFO_H
#define BLOCCOGRAFO_H


/*
classe che memorizza le informazioni sulla porzione di grafo considerata
In particolare contiene, per ogni nodo, i nodi successivi e i predecessori
ed in piu' un riferimento alle proprieta' di fan/center di ogni nodo del grafo
*/


class nodoLocale{
  public:
    long puntIn;
    long puntOut;
};


class bloccoGrafo {

private:
  long numeroBlocco;//numero blocco corrente da 0
  long cancellato;

  proprietaNodi* p;//proprieta' FAN/CENTER di ogni nodo allocate automaticamente e gestite internamente
  report2* r;

  nodoLocale* vetNodi;//lista dei nodi con i puntatori ai successori e ai predecessori (il nodo alla fine e' fittizio e serve per la terminazione)
  long*       vetPred;
  long* 		  vetSucc;

  long        numBlocchiTotali; //num totale di blocchi necessari per contentere tutto il grafo

  long numNodiPresenti;  //numero di nodi (elementi nodolocale) attualmente memorizzati (+1 di chiusura)
  long sizePred; //numero di succ totali
  long sizeSucc; //numero di predecessori

  long nodoInizialeCorrente; //primo nodo del blocco corrente
  long size; //num nodi in analisi in questo blocco (massimi che si possono ottenere)

  //files di lettura
  bufferedThreadFileInputStream* fInfo;

  #ifndef USA_SEEK_LETTURA_PRED_SUCC
  bufferedThreadFileInputStream* fSucc;
  bufferedThreadFileInputStream* fPred;
  #else
  multiFileReader* fSucc;
  multiFileReader* fPred;
  #endif
  //nome file tempo dei blocchi
  char nomeFileOperazioni[2000];

private:
  //legge al max size nodi seguenti nel file rispetto al puntatore corrente
  long leggiProssimoInsiemeDiNodiNeiFile(proprietaNodi* p);

public:

  //costruttore della struttura a blocchi
	bloccoGrafo(report2* r,long sizeBlocco);
	
	//legge il primo blocco (-1 se errore)
	long  leggiBloccoIniziale(proprietaNodi* p);
	
	//legge i nodi successivi (-1 se finito il file - va allora ripreso con leggiBloccoIniziale)
	long  leggiBloccoSuccessivo(proprietaNodi* p);

	//ritorna il puntatore ai pred del nodo e il loro num (DA NON MODOFICARE perche' accede alla struttura interna e non e' una copia!!!!)
  long* getPred(long nodo,long& numero);
	
	//ritorna il puntatore ai succ del nodo e il loro num (DA NON MODOFICARE perche' accede alla struttura interna e non e' una copia!!!!)
	long* getSucc(long nodo,long& numero);
	
	//ritorna il num di nodi presenti in memoria (del blocco corrente)
  long  getNodiPresentiBloccoCorrente(){return numNodiPresenti;}

  //numero blocco corrente (da 0 a maxblocco-1; -1 se il blocco corrente non e' valido (es finito file o errata inizializzazione)
  long  getNumeroBloccoCorrente()  {return numeroBlocco;}
	
	//ritorna il num totale di blocchi necessari per contenere tutto il grafo
	long  getNumeroTotaleBlocchi()   {return numBlocchiTotali;}

	//numero massimo nodi per ogni blocco
	long getNumeroNodiPerBlocco() 	 { return size; }
	
	//numero tot nodi totali
  long getNumeroNodiTotali()       { return r->getNumNodiTotali(); }

  //nodo + alto
  long getMassimoNodoPresente()    { return nodoInizialeCorrente + size-1; }

  //nodo + alto
  long getMinimoNodoPresente()     { return nodoInizialeCorrente; }
	
	//dealloca le varie strutture utilizzate	
	void  cancella();

};


#endif
