/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file nodosucc.cc :source code of clique utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "nodosucc.h"

  //--------------------------------------------------------------------------------------------
  //crea nodo vuoto
  nodoSucc::nodoSucc(){
  		nodo=size=-1;
  		punt=NULL;
  		setNodoNullo();
  }



  //--------------------------------------------------------------------------------------------
  //crea nodo vuoto
  nodoSucc::nodoSucc(long nodo,bloccoGrafo* b,proprietaNodi* p){
     this->nodo=nodo;
     size=-1;punt=NULL;

    //aggiorno il nodo corrente (aggiungo predecessori e aggiorno FAN/CENTER)
    update(b,p);
  }



  //--------------------------------------------------------------------------------------------
  //crea nodo incompleto
  nodoSucc::nodoSucc(long nodo){
     this->nodo=nodo;
     size=-1;punt=NULL;
  }



  //--------------------------------------------------------------------------------------------
  //legge il blocco corrente dal file e aggiorna se possibile se varie informazioni
  nodoSucc::nodoSucc(bufferedThreadFileInputStream* fp,bloccoGrafo* b,proprietaNodi* p){
 		//inizializzo ip punt dei successori
 		punt=NULL;

    //legge il nodo
    if (fp->read(&nodo,sizeof(nodo),1)<0){cout<<"Reading error for a new node";	FINE; }

    //legge num predecessori
    if (fp->read(&size,sizeof(size),1)<=0){cout<<"Reading error for the node "<<nodo;FINE; }

    //se e' vuoto non vado oltre a leggere
    if ( estNullo() ) return;

    //se e' completo, aggiorno solo i valori FAN/CENTER
    if (size>0){
       //alloca nuova memoria per i predecessori
       punt=new long [size];

       //legge tutti i predecessori rimasti
       int n;
       if ( (n=(long)fp->read(punt,sizeof(long),size)) < size ){
       			cout<<"Reading error for node "<<nodo<<". Not enough successors found.";
       			FINE;
       }
    }

    //aggiorno il nodo corrente (aggiungo predecessori o aggiorno FAN/CENTER)
    update(b,p);

  }



  //--------------------------------------------------------------------------------------------
  //aggiorna la struttura (se necessario aggiunge predecessori else AGGIORNA FAN/CENTER)
  void nodoSucc::update(bloccoGrafo* b,proprietaNodi* p){
    //se nullo allora non va aggiornato
    if (estNullo()) return;

    if (estNodoCompleto())
        //se il nodo  e' completo basta solo fare l'upgrade dei vari nodi predecessori ed eventualente eliminare il nodo se non piu' CENTER
        update_FAN_CENTER(p);
    else
        //altrimenti cerco di aggiungere nuovi predecessori e controllare FAN/CENTER
        aggiungePredecessori(b,p);
  }


  //--------------------------------------------------------------------------------------------
  //aggiorna la struttura solo come FAN/CENTER
  void nodoSucc::update(proprietaNodi* p){
    //se nullo allora non va aggiornato
    if (estNullo()) return;

    if (estNodoCompleto())
        //se il nodo  e' completo basta solo fare l'upgrade dei vari nodi predecessori ed eventualente eliminare il nodo se non piu' CENTER
        update_FAN_CENTER(p);
  }


  //--------------------------------------------------------------------------------------------
  //cancella la struttura
  void nodoSucc::cancella(){
    setNodoNullo();
  }

  //--------------------------------------------------------------------------------------------
  //  //ritorna i predecessori e il loro numero
  long* nodoSucc::getPredecessori(long& num){
  		num=size;
  		if (num<0)
  					num=0;
  		return punt;
  }

  //--------------------------------------------------------------------------------------------
  //scrive il blocco corrente su file
  void nodoSucc::scriveSuFile(FILE* fp){
    //scrivo il nodo
    fwrite(&nodo,sizeof(nodo),1,fp);

    //scrive num predecessori
    fwrite(&size,sizeof(size),1,fp);

    //scrive solo se c'e' almeno un successore (no se il nodo e' nullo o non completo)
    if (size>0)
       //scrive tutti i predecessori rimasti
       fwrite(punt,sizeof(long),size,fp);
  }



  //--------------------------------------------------------------------------------------------
  //controlla se il nodo corrente e' vuoto (NULLO)
  long nodoSucc::estNullo(){return (nodo<0);}



  //--------------------------------------------------------------------------------------------
  //controlla se col nodo sono presenti anche i predecessori o no (se vuoto e' anche completo)
  long nodoSucc::estNodoCompleto(){return (size>=0)||estNullo();}



  //--------------------------------------------------------------------------------------------
  //cerca i predecessori del nodo - o nodoNullo se il nodo non puo' essere CENTER
  void nodoSucc::stampa(){
    if (size<0) { printf("  SUCC %3ld incompleto\n" ,nodo);return; }
    if (nodo<0){  cout<<"  SUCC VUOTO \n";return; }

    printf("  SUCC %3ld (InDeg %3ld): " ,nodo,size);
    for (int i=0;i<size;i++)
       cout<<" "<<punt[i];
    cout<<"\n";
  }








// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXX  PRIVATE    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXX  PRIVATE    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX




  //--------------------------------------------------------------------------------------------
  //set il nodo corrente a nodo nullo - cancellando eventualmente i nodi gia' allocati
  void nodoSucc::setNodoNullo(){
    nodo=-1;
    size=0;
    if (punt==NULL) return;
    delete[] punt;punt=NULL;
  }




  //--------------------------------------------------------------------------------------------
  //cerca i predecessori del nodo - o lo trasforma a nodoNullo se il nodo non puo' essere CENTER
  long nodoSucc::aggiungePredecessori(bloccoGrafo* b,proprietaNodi* p){

     //aggiungo solo per i nodi non completi e non vuoti
     if (estNullo()) return TRUE;
     if (estNodoCompleto() ) return TRUE;

     //controllo se il nodo corrente e' possibile CENTER - se no, non leggo i predecessori (indico 0)
     if (p->getCENTER(nodo)!=TRUE){
       setNodoNullo();
       return TRUE;
     }

     //se e' valido come center....cerco i predecessori del nodo corrente - se possibile...
     long n;
     long* l=b->getPred(nodo,n);
     if (l==NULL) return FALSE;//se NULL, nel blocco corrente in nodo non c'e'


     //conta tutti i predecessori che sono ancora possibili FAN
     size=0;
     for (int i=0;i<n;i++){
        if ( p->getFAN(l[i]) )
           size++;
     }

     if (size>0){
        //cancella eventuali altre assegnazioni
        if (punt!=NULL) delete[] punt;
        punt=NULL;

        //scrive tutti i predecessori che sono ancora possibili FAN - allocando prima la memoria
        punt=new long[size];
        int pos=0;
        for (int i=0;i<n;i++){
          if ( p->getFAN(l[i]) )//aggiunge solo i predecessori che sono FAN
            punt[pos++]=l[i];
        }
     }

     //controlla se il nodo corrente puo' essere ancora CENTER - se no lo sost con un nodo vuoto
     if (size<p->minOutDeg){
          p->setCENTER(nodo,FALSE);
					setNodoNullo();
     }

     return TRUE;
  }



  //--------------------------------------------------------------------------------------------
  //elimina i nodi predecessori non FAN e se necessario trasforma il nodo a nodoNullo se non puo' essere CENTER
  void nodoSucc::update_FAN_CENTER(proprietaNodi* p){
      if (estNullo() ) return ;

      //controllo se il nodo e' center o meno - se no annullo il nodo
      if (p->getCENTER(nodo)!=TRUE){
        setNodoNullo();
        return;
      }

      //se non e' completo mi fermo a questo controllo... non posso fare altro
      if ( !estNodoCompleto() ) return;

      //altrimenti analizzo i predecessori e controllo quelli rimasti (spostandoli all'inizio e mettendo -1 ai pred non validi)
      long contaRimasti=0;

      //controlla eventuali predecessori che non sono piu' fan
      for(int i=0;i<size;i++){

          if (punt[i]>=0){
          	if (p->getFAN(punt[i])!=TRUE){//se il pred non e' fan, viene post a -1
              punt[i]=-1;
            }
            else
              //conta i soli successori ancora validi
              contaRimasti++;
          }
      }

      //compatta eventuali vuoti e controlla se il nodo corrente puo'essere ancora CENTER
      if (contaRimasti!=size){
        long pos=0;
        for (int i=0;i<size;i++){
          if (punt[i]!=-1)
            punt[pos++]=punt[i];
        }
        size=contaRimasti;
        if (size<p->minInDeg){
          p->setCENTER(nodo,FALSE);
          setNodoNullo();
        }
      }

  }


