/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file nodosucc.h :source code of clique utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#ifndef NODO_SUCC__
#define NODO_SUCC__
#include "bloccografo.h"
#include "proprietanodi.h"
#include "comuni.h"
/*
classe che rappresenta un nodo (che deve essere CENTER valido) e tutti i suoi predecessori (limitati ai soli nodi FAN validi ).
c'e' il caso di nodo non completo (manca la lettura dei predecessori)
   e il caso di nodo NULLO (nessun nodo perche' non e' CENTER)
*/

class nodoSucc{
private:
  //set il nodo corrente a nodo nullo
  void setNodoNullo();

  //cerca i predecessori del nodo - o nodoNullo se il nodo non puo' essere CENTER
  long aggiungePredecessori(bloccoGrafo* b,proprietaNodi* p);

  //elimina i nodi predecessori non FAN
  void update_FAN_CENTER(proprietaNodi* p);

public:
  //--------------------------------------------------------------------------------------------
  long  nodo;                //nodo succ possibile CENTER /-1 altrimenti
  long  size;  //num elementi predecessori rimasti
  long* punt;                //puntatore nell'array dei predecessori


public:
  //crea nodo vuoto
  nodoSucc();

  //crea un nodo con le info specificate (nodo ed eventuali predecessori) facendo update
  nodoSucc(long nodo,bloccoGrafo* b,proprietaNodi* p);

  //crea un nodo incompleto
  nodoSucc(long nodo);

  //legge dal file il blocco corrente e lo aggiorna se possibile facendo update
  nodoSucc(bufferedThreadFileInputStream* fp,bloccoGrafo* b,proprietaNodi* p);

  //cancella la struttura
  void cancella();

  //controlla se il nodo corrente e' nullo (non e' center ed e' stato cancellato)
  long estNullo();

  //controlla se la lettura del nodo e' stata completata
  long estNodoCompleto();

  //cerca i predecessori del nodo - o nodoNullo se il nodo non puo' essere CENTER
  void stampa();

  //scrive su file
  void scriveSuFile(FILE* fp);

  //update FORZATO del nodo corrente
  void update(bloccoGrafo* b,proprietaNodi* p);

  //update solo dei FAN/CENTER senza ricerca di nuovi elementi nel pezzo di grafo corrente
  void update(proprietaNodi* p);

  //ritorna il num di predecessori e il puntatore
  long* getPredecessori(long& num);

  //ritorna il num di predecessori e il puntatore
  long getNumeroPredecessori(){return size;}
};
#endif

