/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file operazioni.cpp :source code of clique utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "operazioni.h"
//#include "comuni.h"



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
operazioni::operazioni(long nodo,bloccoGrafo* b,proprietaNodi* p){

  if (nodo<0){
		  listaNodiSuccessori=NULL;
		  setOperazioneNulla();
  		return;
  }

	//se il nodo non e' fan, lo elimino
  if (p->getFAN(nodo)!=1){
		  listaNodiSuccessori=NULL;
		  setOperazioneNulla();
  		return;
  }

  //chiedo i successori (ritornati GIA' ordinati in ordine crescente...)
  long* l=b->getSucc(nodo,numeroSuccessori);
  if (l==NULL){
     nodoCorrente=-1;
     return;//se il nodo non e' nel blocco creo un nodo nullo
  }

  //set il nodo corrente
  nodoCorrente=nodo;


  //alloca la struttura per i nodi successori
  listaNodiSuccessori=new nodoSucc[numeroSuccessori] ();//tutti nodi NULLI

  //crea i vari nodi incompleti
  for (int i=0;i<numeroSuccessori;i++){
         listaNodiSuccessori[i]=nodoSucc(l[i]);
  }

  //aggiunge eventuali altri predecessori e aggiorna FAN/CENTER
  update(b,p);
}



//-------------------------------------------------------------------------------------
//legge l'operazione da file
//-------------------------------------------------------------------------------------
operazioni::operazioni(bufferedThreadFileInputStream* fp,bloccoGrafo* b,proprietaNodi* p){

    //legge il nodo
    fp->read(&nodoCorrente,sizeof(nodoCorrente),1);

    //legge num predecessori
    fp->read(&numeroSuccessori,sizeof(numeroSuccessori),1);

    //controlla se vanno letti i successori e se si li legge
    listaNodiSuccessori=NULL;

    //controlla se e' una operazione nulla - se si termino la lettura (NOTA:il controllo e' fatto senza aver aggiornato la struttura)
    if (estOperazioneNulla() ) return;

    if (numeroSuccessori>0){
      listaNodiSuccessori=new nodoSucc [numeroSuccessori];

      //legge - DAL FILE - tutti i predecessori rimasti
      for (int i=0;i<numeroSuccessori;i++)
          listaNodiSuccessori[i]=nodoSucc(fp,b,p);
    }

    //aggiunge eventuali altri predecessori
    update(b,p);
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::scriveSuFile(FILE* fp){

    //scrivo il nodo
    fwrite(&nodoCorrente,sizeof(nodoCorrente),1,fp);

    //scrive num predecessori
    fwrite(&numeroSuccessori,sizeof(numeroSuccessori),1,fp);

    //se e' una operazione nulla termino qui
    if (estOperazioneNulla()) return;

    //scrive tutti i predecessori rimasti
    for (int i=0;i<numeroSuccessori;i++)
        listaNodiSuccessori[i].scriveSuFile(fp);
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::setOperazioneNulla(){

  if (listaNodiSuccessori!=NULL){

	  //elimina tutti i successori (le singole strutture)
  	for (int i=0;i<numeroSuccessori;i++)
      listaNodiSuccessori[i].cancella();

	  //e poi elimina la struttura di memorizzazione dei successori
  	if (listaNodiSuccessori!=NULL){
		  delete[]listaNodiSuccessori;
		}
  }

  listaNodiSuccessori=NULL;

  nodoCorrente=-1;//caratteristica di operazione nulla
  numeroSuccessori=-1;
  numeroSuccessoriDisponibili=0;
  numeroSuccessoriIncompleti=0;
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::stampa(proprietaNodi* p){
  if (estOperazioneNulla()){cout<<"Operation NULL \n";return;}
  printf("NODE %ld (OutDeg %ld - FAN:outdeg>=%ld  - CENTER:indeg>=%ld)\n",nodoCorrente,numeroSuccessori,p->minOutDeg,p->minInDeg);
  for (int i=0;i<numeroSuccessori;i++){
      listaNodiSuccessori[i].stampa();

  }
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::cancella(){
  setOperazioneNulla();
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::update(proprietaNodi* p){
//  aggiorna le proprieta' FAN/CENTER dei nodi successori (NON AGGIUNGE NESSUN NODO )

  //se l'operazione e' nulla esco
  if (estOperazioneNulla())
      return;

  //se il nodo corrente non e' FAN, non vado avanti
  if (!p->getFAN(nodoCorrente) ){
      setOperazioneNulla();
      return;
  }

  //ricalcolo in num di successori ancora validi e dei nuovi nodi predecessori da aggiungere
  long num=0;

  //conteggio numero di nodi disponibili
  numeroSuccessoriDisponibili=0;

  //conto i nuovi nodi da aggiungere ed elimino i successori non piu' CENTER possibili (sostit con -1)
  for (int i=0;i<numeroSuccessori;i++){

     //x ogni successore se non e' completo aggiungo i predecessori o aggiorno solo FAN/CENTER
     listaNodiSuccessori[i].update(p);

     //se non e' un successore nullo lo conto
     if ( !listaNodiSuccessori[i].estNullo() ){
        num++;
        //conto se il nodo e' completo e non nullo
     		if (listaNodiSuccessori[i].estNodoCompleto())
     		    numeroSuccessoriDisponibili++;
     }
  }

  //se il nodo corrente ha un num di successori <del minimo allora non e' piu' valido x fan
  if (num<p->minOutDeg){
    		
    	 if (nodoCorrente>=0)
    	 		p->setFAN(nodoCorrente,FALSE);
	
	     //annulla l'operazione perche' non ci sono un numero sufficiente di successori ( <j )
  	   setOperazioneNulla();
  }

  //calcolo del numero di nodi non ancora completi
  numeroSuccessoriIncompleti=num-numeroSuccessoriDisponibili;
}



//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
void operazioni::update(bloccoGrafo* b,proprietaNodi* p){
//  aggiunge i nuovi predecessori (se necessario) presi dal blocco del grafo corrente e aggiorna le proprieta' FAN/CENTER

  //se l'operazione e' nulla esco
  if (estOperazioneNulla())
      return;

  //se il nodo corrente non e' FAN, non vado avanti
  if (!p->getFAN(nodoCorrente) ){
      setOperazioneNulla();
      return;
  }

  //ricalcolo in num di successori ancora validi e dei nuovi nodi predecessori da aggiungere
  long num=0;

  numeroSuccessoriDisponibili=0;

  //conto i nuovi nodi da aggiungere ed elimino i successori non piu' CENTER possibili (sostit con -1)
  for (int i=0;i<numeroSuccessori;i++){

     //x ogni successore se non e' completo aggiungo i predecessori o aggiorno solo FAN/CENTER
     listaNodiSuccessori[i].update(b,p);

     //se non e' un successore nullo lo conto
     if ( !listaNodiSuccessori[i].estNullo() ){
        num++;

        //conto se il nodo e' completo e non nullo
     		if (listaNodiSuccessori[i].estNodoCompleto() )
     		    numeroSuccessoriDisponibili++ ;
     }
  }

  //se il nodo corrente ha un num di successori <del minimo allora non e' piu' valido x fan
  if (num<p->minOutDeg){
    		if (nodoCorrente>=0)
    		    p->setFAN(nodoCorrente,FALSE);
    		
	     //annulla l'operazione perche' non ci sono un numero sufficiente di successori ( <j )
  	   setOperazioneNulla();
  }

  //calcolo del numero di nodi non ancora completi
  numeroSuccessoriIncompleti=num-numeroSuccessoriDisponibili;
}





//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
long operazioni::cercaClique(proprietaNodi* p){
//ritorna il num di cliques (-1 se non calcolate)

	//ricavo il num i,j
  long val_i=p->getValoreI();
  long val_j=p->getValoreJ();


  //i e j devono essere almeno 2 altrimenti non ha senso il calcolo
  if (val_j<=1 || val_i<=1) return -1;

  //se l'operazione e' nulla termino
  if (estOperazioneNulla())
     return -1;

  //se il nodo corrente non e' FAN, non vado avanti
  if (!p->getFAN(nodoCorrente) ){
      setOperazioneNulla();
      return -1;
  }

  //controlla se l'operazione e' completa e ha troppo pochi successori
  if (estOperazioneCompleta()&&numeroSuccessoriDisponibili<val_j){
			p->setFAN(nodoCorrente,FALSE);
      setOperazioneNulla();
      return -1;
  }else //controllo che ci siano sufficienti successori disponibili, altrimenti non inizio
  	 if (numeroSuccessoriDisponibili<val_j)
    		 return -1;

    		
#ifdef ESEGUI_SOLO_SE_COMPLETO
  //controlla che l'operazione sia completa
  if (numeroSuccessoriIncompleti>0)
      return -1;

#endif


  /****    ALLOCAZIONE DI MEMORIA DINAMICA     ****/

  //array che contiene il numero dei nodi validi
  long* nodiDaAnalizzare=new long[this->numeroSuccessoriDisponibili];

  /****  -----------------------------------  ****/



  //calcolo il num max di predecessori di ogni nodo successore (dovra' essere almeno i) e memorizzo i nodi da analizzare
  long max=0;
  long pos=0;
  for (int i=0;i<numeroSuccessori;i++){
     if (!listaNodiSuccessori[i].estNullo() && listaNodiSuccessori[i].estNodoCompleto() ){
        if (pos<numeroSuccessoriDisponibili)
              nodiDaAnalizzare[pos++]=i;//il nodo corrente va preso
        else{
           //controllo non necessario, ma utile per vedere se qualcosa non va nel programma
           cout<<"Unexpected fatal error 1";
           FINE;
        }
        long  num=listaNodiSuccessori[i].getNumeroPredecessori();
        if (num>max)
            max=num;
     }
  }


  //controllo che non ci sia un errore nel programma (num nodi disponibili calcolati con update != da quelli calc ora
  if (pos!=numeroSuccessoriDisponibili){
      cout<<"Unexpected fatal error 2";
      FINE;
  }

  //flag di fine
  long finito=FALSE;

  //controllo se ha senso calcolare le clique (non ha sufficienti predecessori)
  if (max<val_i){
    	 if (nodoCorrente>=0)
    	 			p->setFAN(nodoCorrente,FALSE);
	     //annulla l'operazione perche' non ci sono un numero sufficiente di successori ( <j )
  	   setOperazioneNulla();
  	
  	   /****    DEALLOCAZIONE DI MEMORIA DINAMICA     ****/
  	   //dealloco la sola struttura dinamica creata
  	   delete[] nodiDaAnalizzare;
  	   nodiDaAnalizzare=NULL;
		   /****  -----------------------------------  ****/  	
  	   return -1;
  }


  /****    ALLOCAZIONE DI MEMORIA DINAMICA     ****/

  //creo un numero di insiemi per contenere le elaborazioni parziali per il calc delle intersezioni
  //in realta' al max eseguo l'intersez di val_j elementi (dei CENTER) - 1 per ogni elaborazione
  insiemeCondiviso* calcParziale=new insiemeCondiviso[ val_j ];

  //allocazione per il calcolo degli insiemi (dimensione max pari alla max intersezione possibile
  for (int i=0;i<val_j;i++){
      long* l=new long [max];
      calcParziale[i]=insiemeCondiviso(l,max);//punta all'array che ho passato - senza farne 1 copia
      calcParziale[i].svuota();//set num elementi a 0
  }




  //---------------------------------------------------------------------------------------
  //   DEVO TROVARE J NODI COME CENTER LA CUI INTERSEZIONE CONTENGA ALMENO I NODI (I FAN)
  //---------------------------------------------------------------------------------------

  // conta il num di clique trovate
  long numeroCliqueTrovate=0;

  // struttura che gestisce i cicli per i nodi disponibili, raggruppandoli in gruppi di al max j elementi
  //inizialmente contiene 1 solo livello(il primo) con il valore 0 (il primo valore possibile)
  combinaNodi c=combinaNodi(numeroSuccessoriDisponibili,val_j);


  // analizzo le varie combinazioni di nodi successori finche' non sono terminate o sono rimasti meno di j successori
  // e non ha senso continuare
  while ( !finito ){

     //------------------------------------------------------------------------------------------
     // ------------- calcolo nuova intersez. per il passo corrente -----------------------------
     //------------------------------------------------------------------------------------------

     //prende il nuovo numero assoluto nodo corrente da analizzare (punt diretto al nodo)
     long numNodoNuovo=nodiDaAnalizzare[c.getValoreUltimoCiclo()];

     // se il numero di ciclo e' 0 (vuol dire che ho un solo ciclo)
     // l'inters. per il passo corrente e' solo il nodo corrente
     if (c.getNumeroCicloCorrente()==0){//se e' il primo ciclo quello rimasto
		
		    //variabili d'appoggio
  			long* p_aux;
  			long n_aux;
				p_aux=listaNodiSuccessori[ numNodoNuovo ].getPredecessori(n_aux);
			  if (calcParziale[ c.getNumeroCicloCorrente() ].copiaDaArray(p_aux,n_aux)<0){
           cout<<"Unexpected fatal error 3";
			     FINE;
			  }
		 }
     else{
         //altrimenti faccio l'intersezione tra il nuovo nodo e l'intersezione precedente
		     //il nodo corrente viene trasformato in insieme
         insiemeCondiviso nuovo=insiemeCondiviso( listaNodiSuccessori[numNodoNuovo]);

         //calcola la nuova intersezione tra il valore dell'inters. del passo precedente e il nodo corrente
         insiemeCondiviso::intersezioneInsiemiOrdinati(calcParziale[c.getNumeroCicloCorrente()-1],nuovo,calcParziale[c.getNumeroCicloCorrente()] );
     }
		

     //------------------------------------------------------------------------------------------
     // -------------       cerca la successiva configurazione      -----------------------------
     //------------------------------------------------------------------------------------------

     //controlla se la dimensione della intersezione e' ancora valida (  >=i xche' nella intersez ci sono i fan)
     if (calcParziale[c.getNumeroCicloCorrente()].getNumeroElementi()>=val_i){
	     	
	     	//vedo se ho trovato l'intersezione dei j nodi (cioe' sono al livello j)
  	   	if (c.getNumeroCicloCorrente()==val_j-1){//ricorda che il livello parte da 0 !!!!!!!!!

            // SCELTA DEI NODI PER IL NUCLEO e aggiornamento dei fan/center - aggiornamento di combinaNodi
            //trovata una nuova clique
            numeroCliqueTrovate++;


            // trovo i primi val_i FAN (dall'intersezione) - mi faccio dare il puntatore alla struttura
            // che contiene i nodi della intersezione (direttamente i nodi FAN)
            long* punt;
            long size;
            size=calcParziale[c.getNumeroCicloCorrente()].getArray(punt);


            //leggo i primi val_i valori della intersezione
            //cout<<"FAN : ";
            for (int i=0;i<val_i;i++){
                if (punt[i]<0){
                    cout<<"\n";
                    stampa(p);
                    cout<<"Unexpected fatal error 4 -"<<punt<<" - punt ="<<punt[i];
                    FINE;
                }

                if (p->getFAN(punt[i])!=1){
                  cout<<"Unexpected fatal error 5: node "<<punt[i]<<" is already a fan";
                  FINE;
                }

                //PRESO NODO COME FAN
                p->setFAN(punt[i],FALSE);
            }

            //trovo da tutti i center, i primi val_j - in realta' li prendo tutti (conoscendo le posizioni relative)
            //punto alla struttura a livelli (che contiene gli indici dei nodi - non i nodi direttamente)
            size=c.getArray(punt);
            //cout<<"CENTER : ";
            for (int i=0;i<size;i++){
                long nodo=listaNodiSuccessori[nodiDaAnalizzare[punt[i]] ].nodo;

                if (nodo<0){
                    cout<<"\n";
                    stampa(p);
                    cout<<"Unexpected fatal error 5 -"<<punt<<" - punt ="<<punt[i];
                    FINE;
                }

                if (p->getCENTER(nodo)!=1){
                  cout<<"Unexpected fatal error 5: node "<<punt[i]<<" is already a center";
                  FINE;
                }

                p->setCENTER(nodo,FALSE);//PRESO NODO COME CENTER
            }

            // eseguo l'aggiornamento dell'operazione (con eliminazione dei FAN/CENTER PRESI DI RECENTE)
            // e aggiornamento dei nodi (alcuni nodi li elimina se hano pochi predecessori..)
            this->update(p);

            //controllo se ha ancora senzo fare controlli (se ci sono suff nodi disponibili)
            if (numeroSuccessoriDisponibili<val_j)
                finito=TRUE;
            else{
	            /*------------------------------------------------------------------------------------------*/
              //riporto il num di STEP a livello 0 ( 1 solo nodo ) e lo sposto all'indice successivo risp a quello preso
              //nota che ora potrei incontrare anche nodi vuoti (quelli scelti come CENTER)
    	        while (c.getNumeroCicloCorrente()>0){//da 0 a max-1
        	       	c.eliminaCicloCorrente();
      	      }
      	
      	      //controlla se ci sono ancora nodi possibili da controllare - spostandomi anche al succ
      	      if (c.prossimaIterazioneCicloCorrente()<0)
      	          finito=TRUE;      	
            }
    	 	}
     		else{// non sono ancora a livello j passo ad un nuovo ciclo (aggiungo un livello )
            if (c.aggiungiNuovoCiclo()<0){
               cout<<"Unexpected fatal error 6";
               FINE;
            }
      	}
     }
     else{// se l'intersezione non e' almeno di i elementi devo cambiare la configurazione corrente
     		  // (eventualmente tornare indietro de il livello + alto e' arrivato al massimo)
	        // fino a trovarne una valida

        //torna indietro finche' non trovo una nuova combinazione valida (se esiste)
        while (c.getNumeroCicloCorrente()>=0){//da 0 a max-1
           if (c.prossimaIterazioneCicloCorrente()<0)//se il livello corrente ha raggiunto il massimo, scendo di un livello
        	     c.eliminaCicloCorrente();
        	 else//else ho trovato una nuova configurazione valida e esco
        	     break;
        }

        //controlla se ho ancora una configurazione valida e se no esco
        if (c.getNumeroCicloCorrente()<0){
             finito=TRUE;
        }
     }

  }//FINE WHILE



  //FINE DELL'ANALISI
  //controllo se sull'operazione corrente ha ancora senso fare altre ricerche...
  if (numeroSuccessoriIncompleti==0){
       //se l'operazione e' completa (tutti i nodi sono stati trovati) non occorre fare altro. Il nodo corrente puo' essere settato a FAN usato
    	 if (nodoCorrente>=0 )
    	      p->setFAN(nodoCorrente,FALSE);
    	
	     //annulla l'operazione perche' non ci sono un numero sufficiente di successori ( <j )
  	   setOperazioneNulla();
  }
  else{// se num tot di successori (sia completi che non - ma non NULLI - cioe' quelli che sono possibili CENTER)
       // e' inferiore a j allora esco altrimenti devo ritornarci.
     long numTotSuccessoriAttivi=numeroSuccessoriIncompleti+numeroSuccessoriDisponibili;

     if (numTotSuccessoriAttivi<val_j){
        //l'operazione e' completa e non occorre fare altro. Il nodo corrente puo' essere settato a -1 e cancello tutte le altre strutture
      	if (nodoCorrente>=0 )
      			p->setFAN(nodoCorrente,FALSE);
	      //annulla l'operazione perche' non ci sono un numero sufficiente di successori ( <j )
  	    setOperazioneNulla();
  	 }
  }


  /****    DEALLOCAZIONE DI MEMORIA DINAMICA     ****/

  //dealloca tutti i vettori alla base degli insiemi
  for (int i=0;i<val_j;i++){
      delete[] calcParziale[i].punt;
      calcParziale[i].punt=NULL;
  }

  //cancello la struttura di calcolo intersezioni parziali
  delete[] calcParziale;calcParziale=NULL;


  delete[] nodiDaAnalizzare;
  nodiDaAnalizzare=NULL;
  /****  -----------------------------------  ****/

  return numeroCliqueTrovate;

}




//-------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------
long operazioni::ritornaNumeroBloccoProssimoSalvataggio(bloccoGrafo* b  ){

  //se l'operazione e' vuota, non ho altri nodi da leggere
  if ( this->estOperazioneNulla() )	return -1;

  //se i succ incompleti non ci sono, non servono altre letture
  if (numeroSuccessoriIncompleti==0) return -1;

  //var ausiliarie
  long numNodiPerBlocco=b->getNumeroNodiPerBlocco();
  long primoNodoIncontrato=-1;
  long primoNodoNonCompletoBloccoSuccessivo=-1;
  long maxNumNodoBloccoCorrente=b->getMassimoNodoPresente();

  /*cout<<"  numNodiPerBlocco = "<<numNodiPerBlocco<<
  			" / numBloccoCorrente = "<<numBloccoCorrente<<
        " / numNodiTotali = "<<numNodiTotali<<
        " / maxNumNodoBloccoCorrente = "<<b->getMassimoNodoPresente()<<"\n";
  */
  //cerco il nodo incompleto del blocco successivo

  for (int i=0;i<numeroSuccessori;i++){

    //solo i nodi incompleti-non nulli
    if ( !listaNodiSuccessori[i].estNodoCompleto() ){
       //cerco anche il primo nodo (puo' servire nel caso in cui devo tornare indietro ai primi blocchi)
       if (primoNodoIncontrato<0)
           primoNodoIncontrato=listaNodiSuccessori[i].nodo;

       if  ( listaNodiSuccessori[i].nodo>maxNumNodoBloccoCorrente ){
           primoNodoNonCompletoBloccoSuccessivo=listaNodiSuccessori[i].nodo;
           break;
       }

    }
  }

  if (primoNodoNonCompletoBloccoSuccessivo>=0)
       return primoNodoNonCompletoBloccoSuccessivo/numNodiPerBlocco;//calcolo il blocco in cui si trova

  else if (primoNodoIncontrato>=0)
            return primoNodoIncontrato/numNodiPerBlocco;

       else {
            cout<<"Unexpected fatal error 7";
            FINE;
       }

  return -1;
}



