/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file binvector.cpp :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
/***************************************************************************
                          binvector.cpp  -  description
                             -------------------
    begin                : Wed Aug 14 2002
    copyright            : (C) 2002 by root
    email                : root@travelcat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "binvector.h"

void binVector::reset(){
  if (cancellato) return;
  for (long i=0;i<sizeBuffer;i++) val[i]=0;
  this->numElementiZero=numElementi;
}


binVector::binVector(long numElem){
  if (numElem<=0) numElem=1;
  this->sizeBuffer=numElem/(sizeof(unsigned char)*8);//num elementi del vettore da allocare
  if (numElem%(sizeof(unsigned char)*8)!=0 ) sizeBuffer++;

  //alloco il buffer e lo inizializzo
  this->val=new unsigned char [sizeBuffer];
  for (long i=0;i<sizeBuffer;i++)
      val[i]=0;

  this->numElementi=numElem;
  this->numElementiZero=numElementi;
  this->cancellato=false;

}






void binVector::cancella(){
  if (cancellato) return;
  cancellato=true;
  delete[] val;
  val=NULL;
}





void binVector::set(long nodo,bool bit, long&errore){
//set a 1 il pos-esimo bit al valore bit nel vettore di numElem caratteri
//nota che sizeof(...) ritorna il num di byte (per ottenere il num di bit va moltiplicato per 8 !!!!!!!!!!!!!!!)

/*
     +----+----+----+----+----+                               |----------------|
     | ...| ...| ...|... |... |       CON UN SINGOLO BYTE --> | 0,1,2,3,4,5,6,7|
     +----+----+----+----+----+         memorizzato cosi'     |----------------|
elem  0                    N-1

*/

  //controllo i vincoli di validit del nodo (tra 0 e max-1)
  errore=0;
  if (nodo<0)         {errore=-1;return;}
  if (nodo>=numElementi)  {errore=-1;return;}

  //prima vedo se il vecchio valore del bit e' uguale a quello corrente
  bool vecchioVal=get(nodo,errore);
  if (vecchioVal==bit) return;//se il bit non va cambiato...

  //individua posizione dell'array da modificare
  long posElem=        nodo / (sizeof(char)*8);//num elemento in cui agire (char)
  long numBitRelativo= 7-nodo % (sizeof(char)*8);//peso bit corrente         (il bit 0 corrisponde a quello di peso 0...)

  //crea la maschera
  unsigned char mask=1;
  //1 solo nel bit di peso numBitRelativo
  mask=mask<<numBitRelativo;

  if (bit)  val[posElem]=val[posElem]|mask;//set a 1 il bit
  else      val[posElem]=val[posElem]&(~mask);//set a 0 il bit

  if (bit==true)      this->numElementiZero--;//aggiunto un 1
  else                this->numElementiZero++;//aggiunto uno 0
}







bool binVector::get(long nodo,long& errore){
//penso ad un array memorizzato da dx a sx
//nota che size of ritorna il num di byte (per ottenere il num di bit va moltiplicato per 8 !!!!!!!!!!!!!!!)

/*
     +----+----+----+----+----+                               |----------------|
     | ...| ...| ...|... |... |       CON UN SINGOLO BYTE --> | 0,1,2,3,4,5,6,7|
     +----+----+----+----+----+         memorizzato cosi'     |----------------|
elem  0                    N-1

*/
  errore=0;
  if (nodo<0)        {errore=-1;return false;}
  if (nodo>=numElementi) {errore=-1;return false;}

  //individua posizione dell'array da modificare
  long posElem=        nodo / (sizeof(char)*8);//num elemento in cui agire
  long numBitRelativo= 7- nodo % (sizeof(char)*8);//peso bit corrente

  //selezione del bit significativo
  unsigned char mask=1;
  mask=mask<<numBitRelativo;

  if ( (val[posElem]&(mask))!=0)
     return true;//1
  else
	   return false;  //0
}

