/*

*********************************************************************
*********************************************************************

  file binVector.cpp : common functions and classes of all graph utilities

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/

#include <stdio.h>
#include "binvectorfile.h"
#include "funzioniGenerali.h"

//--------------------------------------------------------------------------------------------------------------
//  costruttore del vettore (inizializzato a initValue)
//--------------------------------------------------------------------------------------------------------------
binVectorFile::binVectorFile(char* multiFileName, long maxDimMultifile, long long numBits_LL, bool initialValue){
  if (numBits_LL<0) {
    chiuso=true;
    return;
  }
  
  this->numBits_LL=numBits_LL;
  f=multiFile(multiFileName,maxDimMultifile);

  sizeBuffer=5000;//size del buffer
  buffer=new char[sizeBuffer];

  this->numBitBuffer=0;
  this->primoBitBuffer_LL=0;
  this->updateBuffer=false;
  chiuso=false;
  

  //inizializza il buffer con tutti 1 o 0 a seconda del valore richiesto
  for (int i=0;i<sizeBuffer;i++)
    if (initialValue)
      buffer[i]=255;//tutti 1
    else
      buffer[i]=0;  //tutti 0

  //inizializza il file con tutti 1 o 0 a seconda del valore richiesto      
  long long numBytesDaScrivere=numBits_LL/(long long)8;
  if (numBits_LL%(long long)8!=0) numBytesDaScrivere++;

  while (numBytesDaScrivere>0){

    long daScrivere;    
    if (numBytesDaScrivere>(long long)sizeBuffer) daScrivere=sizeBuffer;
    else daScrivere=(long)numBytesDaScrivere;
    
    if ( f.write(buffer,sizeof(char),daScrivere)<daScrivere){perror("Writing error - bitVectorFile");ABORT();}
    numBytesDaScrivere-=daScrivere;
  }
      
}


//--------------------------------------------------------------------------------------------------------------
// chiude il vettore su file e cancella il file temporaneo
//--------------------------------------------------------------------------------------------------------------
void binVectorFile::close(){
  if (chiuso) return;
  chiuso=true;

  //svuota l'ultimo buffer se ancora non salvato
  flushBuffer();

  //chiude il file temp e lo rimuove
  f.close();
  f.removeMultiFile(f.getMultiFileName());
}




//--------------------------------------------------------------------------------------------------------------
// get il valore nella pos specificata. False in caso di errore
//--------------------------------------------------------------------------------------------------------------
bool binVectorFile::get(long long pos_LL, bool& val){
  if (chiuso) return false;
  if (pos_LL<0 || pos_LL>=this->numBits_LL) {printf("bit %lld invalid (numBits_LL = %lld)",pos_LL,numBits_LL);return false;}

  //controllo se non  nel buffer corrente
  caricaNuovoBufferSeNecessario(pos_LL);
  

  //se la pos e' gia' nel buffer corrente
  long long maxBitBuffer_LL=this->primoBitBuffer_LL+(long long)this->numBitBuffer-(long long)1;
  if (this->primoBitBuffer_LL<=pos_LL && pos_LL<=maxBitBuffer_LL ){
    val=this->getVettore(buffer, (long)(pos_LL-this->primoBitBuffer_LL) );
    return true;
  }
  else{
    printf("Program error n.1");
    ABORT();
    return false;  
  }
  
}


//--------------------------------------------------------------------------------------------------------------
// get il valore nella pos_LL specificata. False in caso di errore
//--------------------------------------------------------------------------------------------------------------
bool binVectorFile::set(long long pos_LL, bool val){
  if (chiuso) return false;
  if (pos_LL<0 || pos_LL>=this->numBits_LL) return false;


  //controllo se non  nel buffer corrente
  caricaNuovoBufferSeNecessario(pos_LL);

  //se la pos_LL e' gia' nel buffer corrente
  long long maxBitBuffer_LL=this->primoBitBuffer_LL+(long long)this->numBitBuffer-(long long)1;
  if (this->primoBitBuffer_LL<=pos_LL && pos_LL<=maxBitBuffer_LL ){
    long oldVal=this->getVettore(buffer, (long)(pos_LL-this->primoBitBuffer_LL));
    if (oldVal==val) return true;//nessuna modifica da fare
    else{
      this->setVettore(buffer, (long)(pos_LL-this->primoBitBuffer_LL), val);
      this->updateBuffer=true;//buffer modificato
    }
    return true;
  }
  
  else{
    printf("Program error n.1 bis");
    ABORT();
  }
  return false;
}




//--------------------------------------------------------------------------------------------------------------
//set il pos_LL-esimo bit del buffer al valore bitValue. Nessun controllo sulla dim del vettore viene fatto.
//--------------------------------------------------------------------------------------------------------------
void binVectorFile::setVettore(char* buffer, long pos,bool bitValue){
  if (pos<0) return;
  
  //individua posizione dell'array da modificare
  long posElem=        pos / (sizeof(char)*8);//num elemento in cui agire (char)
  long numBitRelativo= 7-pos % (sizeof(char)*8);//peso bit corrente         (il bit 0 corrisponde a quello di peso 0...)

  //crea la maschera
  unsigned char mask=1;
  //1 solo nel bit di peso numBitRelativo
  mask=mask<<numBitRelativo;

  if (bitValue)  buffer[posElem]=buffer[posElem]|mask;//set a 1 il bit
  else      buffer[posElem]=buffer[posElem]&(~mask);//set a 0 il bit
}



//--------------------------------------------------------------------------------------------------------------
//get il pos_LL-esimo bit del buffer al valore bitValue. Nessun controllo sulla dim del vettore viene fatto.
//--------------------------------------------------------------------------------------------------------------
bool binVectorFile::getVettore(char* buffer, long pos){
  if (pos<0) return false;//ERRORE

  //individua posizione dell'array da modificare
  long posElem=        pos / (sizeof(char)*8);//num elemento in cui agire
  long numBitRelativo= 7- pos % (sizeof(char)*8);//peso bit corrente

  //selezione del bit significativo
  unsigned char mask=1;
  mask=mask<<numBitRelativo;

  if ( (buffer[posElem]&(mask))!=0)
     return true;//1
  else
	   return false;//0
}



//--------------------------------------------------------------------------------------------------------------
//controlla se il buffer deve essere riscritto su file
//--------------------------------------------------------------------------------------------------------------
void binVectorFile::flushBuffer(){

  if (this->updateBuffer){
    //calcola la pos_LL nel file in cui copiare il buffer
    long long posByte=this->primoBitBuffer_LL/8;
    if (this->primoBitBuffer_LL%(long long)8!=0){printf("Program error n.2");ABORT();}

    f.seek(posByte);
    if (f.write(buffer,sizeof(char),numBitBuffer/8)<numBitBuffer/8){perror("Writing error n.3");ABORT();}
    this->updateBuffer=false;
  }

}


//--------------------------------------------------------------------------------------------------------------
//controlla se la posizione richiesta  nel buffer corrente e se no carica un nuovo buffer che contiene la pos_LL specificata.
//Se il buffer prec risulta modificato, viene scritto su file prima di sostituirlo. Nessun controllo su pos_LL viene fatto.
//--------------------------------------------------------------------------------------------------------------
void binVectorFile::caricaNuovoBufferSeNecessario(long long pos_LL){

  //se buffer contiene posizione, return
  if (this->primoBitBuffer_LL<=pos_LL && pos_LL<=(this->primoBitBuffer_LL+this->numBitBuffer-1) )return;

  //se necessario scrive il buffer su disco
  flushBuffer();

  //legge il buffer
  long long posByteBufferDaLeggere=pos_LL/8;//pos_LL in bytes dove inizia il buffer (eventualmente precedente al bit in pos_LL)
  f.seek(posByteBufferDaLeggere);
  if ( (numBitBuffer=f.read(buffer,sizeof(char),sizeBuffer)*8)<0 ){perror("Reading error n.4");ABORT();}

  this->primoBitBuffer_LL=posByteBufferDaLeggere*8;//nuovo primo byte
}

