/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file coda.h :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//#include <istd::ostream.h>
#include <iostream>

/*================================================================================================*/
/*================================================================================================*/
//                                     FILE COMUNE
/*================================================================================================*/
/*================================================================================================*/



#ifndef _coda_
#define _coda_

#include "funzioniGenerali.h"

template<class T>
 struct nodoCoda{
   T info;
   nodoCoda<T>* next;
 };

//classe che gestisce una CODA
// CON CONDIVISIONE (cioe' facendo una copia si accede alla stessa struttura)
template <class T>
class coda{
  private:
    nodoCoda<T>* inizio;//cancello elementi
    nodoCoda<T>* fine;//aggiungo elementi

  private:
    void cancellaLista();
    long size;
    int cancellato;

  public:
    coda();
    void cancella();
    long svuotaCoda(T& elem);
    long getFirst(T&);//restituisce il valore del 1 elem
    void putLast (T& elem);//copia l'elem nella coda
    long getSize(){return size;}
    void svuotaCoda();
    void stampaCoda();

    // AG_CHANGE

    //    friend std::ostream& operator<< <T>(std::ostream& out, coda<T>& a);

};
template <class T>
std::ostream& operator<<(std::ostream& out, coda<T>& a);


//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
coda<T>::coda(){
  size=0;
  inizio=fine=0;
  cancellato=FALSE;
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
void coda<T>::cancellaLista(){
  nodoCoda<T>* aux;
  while (inizio!=NULL){
    aux=inizio;
    inizio=inizio->next;
    delete aux;aux=NULL;
  }
  inizio=fine=NULL;
  size=0;
}


//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
void coda<T>::svuotaCoda(){
  cancellaLista();//svuota il contenuto della coda
}


//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
void coda<T>::cancella(){
  if (cancellato) return;
  cancellaLista();
}


//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
long coda<T>::getFirst(T& elem){

   if (inizio==NULL)
       return -1;//coda vuota

   elem=inizio->info;
   nodoCoda<T>* aux=inizio;
   if ( (inizio=inizio->next)==NULL)
     fine=NULL;

   size--;
   delete aux;aux=NULL;

   return 1;
}



//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
void coda<T>::putLast(T& elem){

   nodoCoda<T>* aux=new struct nodoCoda<T>;//nuovo nodo
   aux->info=elem;
   aux->next=NULL;
   if (fine==NULL)
     inizio=fine=aux;
   else{
     fine->next=aux;
     fine=fine->next;
   }
   size++;
}



//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
std::ostream& operator<<(std::ostream& out, coda<T>& a){
  nodoCoda<T>* aux=a.inizio;
  out<<"< ";
  while (aux!=NULL){
    out<<aux->info<<" ";
    aux=aux->next;
  }
  return (out<<"<");
}


//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
template <class T>
void coda<T>::stampaCoda(){
  std::cout<<"size: "<<size<<"\n";
  nodoCoda<T>* aux=inizio;
  std::cout<<"PUT <= ";
  while (aux!=NULL){
    std::cout<<aux->info<<" <= ";
    aux=aux->next;
  }
  std::cout<<" GET\n";
}
#endif

