/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file multiCoda.h :source code of externalsort utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "../coda.h"
#include "../funzioniGenerali.h"


#ifndef _MULTICODA_
#define _MULTICODA_
//classe che gestisce una struttura di N code
// CON CONDIVISIONE (cioe' facendo una copia si accede alla stessa struttura)
template <class T>
class multiCoda{
  private:
      coda<T>* listaCode;//punt alle code presenti
      long numeroTotaleCode;//n tot delle code presenti (anche vuote)
      long numeroCodaCorrente;//n coda corrente
      int cancellato;
  public:
      multiCoda(int numCode);
      void cancella();//
      long getElementi(T* t);//restituisce i primi elementi di ciascuna coda
      void putElemento(T& t);//aggiunge un elemento alla coda corrente
      int  estVuoto();
      friend std::ostream& operator<<<T>(std::ostream& out, multiCoda<T>& a);
      long getNumTotaleElementi();
      long getNumCodePiene();
};


template <class T>
long multiCoda<T>::getNumCodePiene(){
//ritorna il n di elementi totali presenti nelle varie code
  if (cancellato) return -1;
  long tot=0;
  for (long i=0;i<numeroTotaleCode;i++)
    if (listaCode[i].getSize()>0)
        tot++;
  return tot;
}


template <class T>
long multiCoda<T>::getNumTotaleElementi(){
//ritorna il n di elementi totali presenti nelle varie code
  if (cancellato) return -1;
  long tot=0;
  for (long i=0;i<numeroTotaleCode;i++)
    tot+=listaCode[i].getSize();
  return tot;
}



template <class T>
std::ostream& operator<<(std::ostream& out, multiCoda<T>& a){

  // AG_CHANGE
  //if (cancellato) return out;
  if (a.cancellato) return out;
  //
  for (long i=0;i<a.numeroTotaleCode;i++){
    out<<"\n CODA :"<<i;
    out<<a.listaCode[i]<<"\n";
  }
  std::cout<<"\n";
  return out;
}



template <class T>
multiCoda<T>::multiCoda(int numCode){
  if (numCode<=0){
    fprintf(stderr,"Numero di code non valido (%d<0)",numCode);
    FINE;
  }
  this->numeroTotaleCode=numCode;
  this->listaCode=new coda<T> [numeroTotaleCode];
  this->numeroCodaCorrente=0;
  cancellato=FALSE;
}



template <class T>
void multiCoda<T>::cancella(){
  if (cancellato) return;
  cancellato=TRUE;
  delete[] listaCode;listaCode=NULL;//cancella le varie code
}


template <class T>
void multiCoda<T>::putElemento(T& t){
  if (cancellato) return ;
  listaCode[numeroCodaCorrente].putLast(t);
  if (++numeroCodaCorrente>=numeroTotaleCode)
      numeroCodaCorrente=0;
}


template <class T>
long multiCoda<T>::getElementi(T* t){
//restituisce una lista di elementi - t deve essere gia' inizializzata
//ritorna il n di elementi
  if (cancellato) return -1;
  long pos=0;
  for (long i=0;i<numeroTotaleCode;i++)
    if (listaCode[i].getFirst(t[pos])!=-1)
      pos++;
  return pos;
}


template <class T>
int multiCoda<T>::estVuoto(){
  if (cancellato) return -1;
  for (long i=0;i<numeroTotaleCode;i++){
    if (listaCode[i].size()>0 )
      return FALSE;
  }
  return TRUE;
}



#endif
