/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file funzioniGenerali.h :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE revisione 24 nov 2001
//=======================================================================
//=======================================================================


//=====================================================================================================================
//=====================================================================================================================
//
//               QUESTO FILE DEVE ESSERE INCLUSO DA TUTTI I FILES !!!!!!!!!!!!!!!!!!!!!!!!!!!
//
//=====================================================================================================================
//=====================================================================================================================
//
// GESTIONE SHARED MEMORY: NECESSARIA PER PERMETTERE UNA SICURA DEALLOCAZIONE DELLA SHMEM QUANDO NON PIU' UTILIZZATA
// NOTA: IN QUESTO CASO LA ATTACH NON VIENE EREDITATA DAI PROCESSI FIGLI (IN REALTA' ALL'AVVIO DI OGNI PROCESSO
//       VENGONO DEALLOCATI TUTTI I BLOCCHI DI MEMORIA SHMEM EREDITATI DAL PADRE - USANDO LA CLASSE shmem -
//       E DUNQUE OCCORRE ESEGUIRE UN MAPPING ESPLICITO DEI BLOCCHI NECESSARI!!!!!!!!!
//
// - richiediBlocco(...)             	:richiede e fa il mapping di una shmem (usare necessariamente questa)
// - rilasciaMemoriaCondivisa(...)   	:dealloca la memoria (se non in mapping da altri processi)
// - rilasciaMappingShmem(...)       	:rilascia solo l'attach senza eliminare lo spazio shared
// - richiediMappingShmem(...)          :crea un attach con shmem (gia' esistente)
//=====================================================================================================================
//=====================================================================================================================


#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <wait.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>

// definizione delle priorita'
#define HIGH_PRIOR       (-3)
#define NORMAL_PRIOR     (+0)
#define BACKGROUND_PRIOR (+3)
#define DAEMON_PRIOR     (+10)

//definizione della stampa per l'aggiunta del tag "finto" per indicare il num di nodi
#define PRINT_ISOLATED_NODES_SPECIFICATION printf("Adding 'isolated nodes tag'. Output file uses extended edges format.\n")


#define FALSE (0)
#define TRUE  (1)

// macro che termina tutti i processi del programma corrente: da utilizzare in caso di errore o terminazione per abort
#ifndef FINE
#define FINE (  ABORT()  );
#endif

//definizione max lunghezza nome file
#ifndef DIM_TEXT_FILENAME
#define DIM_TEXT_FILENAME (512)
#endif

#ifndef N
#define N DIM_TEXT_FILENAME
#endif

#define RAM        ( getMemoriaDisponibile() )
#define MB          (1048576)


//definizione max lunghezza (in long) del buffere dei parametri e key del buffer USATO IN PROCESSI
#define KEY_PARAMETRI  0xF0F0F0F0
#define NUM_MAX_PARAMETRI_CANALE 200000  /*n long disponibili come parametri*/

void stampaHelp(int n,char* argv[]);

/*--------------------------- GESTIONE STAMPE -----------------*/
#define STAMPA_ERRORE_PARAMETRI(nomeFile,STRINGA) {fprintf(stdout,"%s\n",STRINGA);printf("Use %s -help for help\n",nomeFile);ABORT();}
#define ERRORE(STRINGA) {fprintf(stdout,"%s\n",STRINGA);ABORT();}


void stampaLicenza(int n,char* argv[]);

/*FUNZIONE CHE VA DEFINITA NEL PROGRAMMA PRINCIPALE E CHE RIGURDA I MESSAGGI DI ERRORE*/
char* eliminaPath(char* s);
void stampaHelp(int n,char* argv[]);
void printError(const char *s);

//gestione interruzioni tastiera
void interruzioneTastiera(int p);
void stopTastiera(int p);

//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------
//gestione memoria massima disponibile
extern long long memoriaDisponibile;
void setMemoriaDisponibile(long long mem);
long long getMemoriaDisponibile();
//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------

//conta il num di multifiles totali presenti
long contaNumFilesMultiFile(char* nome);
long getFileSize(char* nomeFile);
long long getTotalSizeMultiFiles(char* nomeMulti);

// FUNZIONI UTILIZZATE PER AVVIARE/TERMINARE I PROCESSI

// funzione utilizzata implicitamente in FINE
// chiude tutti i processi, dealloca shmem e semafori
void ABORT();

//da utilizzare alla fine del main per attendere la fine di tutti i processi e per liberare tutte le risorse condivise
void EXIT_ALL(int i);

//da utilizzare alla fine del main per attendere la fine di tutti i processi e per liberare tutte le risorse condivise
void EXIT(int i);

//richiede e fa il mapping di una shmem
char* richiediBlocco(long size,long& id);

// dealloca la memoria (se non in mapping da altri processi)
void rilasciaMemoriaCondivisa(void* addr,long id);

//rilascia solo l'attach senza eliminare lo spazio shared
void rilasciaMappingShmem(char* addr);

//crea un attach con shmem (gia' esistente)
char* richiediMappingShmem(long id);

//controlla che la stringa sia nel formato printf
bool estMultiFileName(char* c);

//controlla che la stringa sia nel formato printf
long contaPrintfFormat(char* c);

//richiesta di un nuovo semaforo al SO
long getSemaforo( long key,long numero);

//accetta il semaforo
void accettaSemaforo( long id);

//elimina un semaforo
long removeSemaforo(long semid);




//utilizzato per il DEBUG di alcune
//#define DEBUG_SORT  //per mostrare messaggi procedure sorting
//#define DEBUG_LETTURA //se abilitata stampa i messaggi di stato dei buffer in/out dei thread
//#define DEBUG_SCRITTURA //se abilitata stampa i messaggi di stato dei buffer in/out dei thread
//#define DEBUG_ERR //se abilitata stampa i messaggi sull'avvio thread


#ifndef _CHECK_
#define _CHECK_
template <class T>
void check(T* a){
  if (a==NULL){perror("errore nella allocazione di memoria ");FINE;}
}

template <class T>
void check(T* a,char* s){

  if (a==NULL){fprintf(stderr,s);perror("");FINE;}
}

template <class T>
void check(T* a,char* s1,char* s2){
  if (a==NULL){fprintf(stderr,s1,s2);perror("");FINE;}
}

#endif
