/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file indeg.cpp :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   

#include "indeg.h"
//#include "config.h"
#include <stdio.h>





inDeg::inDeg(long nNodi,long COEFF_LIVELLO_0, long COEFF_LIVELLO_1, long COEFF_LIVELLO_2){

  //specifica dei vari coefficienti di rapporto tra i livelli dei vari nodi
	COEFF0=COEFF_LIVELLO_0;
	COEFF1=COEFF_LIVELLO_1;
	COEFF2=COEFF_LIVELLO_2;

	this->numNodiTot=nNodi;
	if (nNodi<=0) return;

	//raggruppa gli elementi del grafo (i nodi)
	size2=nNodi/COEFF2;
	if (nNodi%COEFF2!=0 ||size2<=0)	size2++;

	//raggruppa gli elementi del grafo (i nodi)
	size1=size2/COEFF1;
	if (size2%COEFF1!=0 ||size1<=0)	size1++;

	//raggruppa i blocchi del liv1
	size0=size1/COEFF0;
	if (size1%COEFF0!=0 || size0==0 )	size0++;

  //allocazione memoria
	livello0=new long[size0];
	livello1=new long[size1];
	livello2=new long[size2];
  inDegNodi=new long[numNodiTot];


	for (long i=0;i<size0;i++)livello0[i]=0;
	for (long i=0;i<size1;i++)livello1[i]=0;
	for (long i=0;i<size2;i++)livello2[i]=0;
	for (long i=0;i<numNodiTot;i++)inDegNodi[i]=0;
  sommaInDeg=0;


/*  //inizializza tutti i nodi a 1
  for (int nodo=0;nodo<nNodi;nodo++){
   	//calcola le posizioni nei due livelli
 	  long pos2=nodo/COEFF2;
 	  long pos1=pos2/COEFF1;
 	  long pos0=pos1/COEFF0;

 	  if (pos0>=size0){printf("Errore size0");FINE ;}
 	  if (pos1>=size1){printf("Errore size1");FINE ;}
 	  if (pos2>=size2){printf("Errore size2");FINE ;}

    inDegNodi[nodo]++;
 	  livello2[pos2]++;
 	  livello1[pos1]++;
 	  livello0[pos0]++;
    sommaInDeg++;

  }
*/

}




void inDeg::stampa(){
		printf("liv 0: ");
		for (int i=0;i<size0;i++)
			printf("%ld ",livello0[i]);
		printf("\n");

		printf("liv 1: ");
		for (int i=0;i<size1;i++)
			printf("%ld ",livello1[i]);
		printf("\n");

		printf("liv 2: ");
		for (int i=0;i<size2;i++)
			printf("%ld ",livello2[i]);
		printf("\n");

		printf("nodi : ");
		for (int i=0;i<numNodiTot;i++)
			printf("%ld ",inDegNodi[i]);
		printf("\n");
}





void inDeg::cancella(){
	if (numNodiTot<=0) return;
	delete[] livello0;livello0=NULL;
	delete[] livello1;livello1=NULL;
	delete[] livello2;livello2=NULL;
	delete[] inDegNodi;inDegNodi=NULL;
	numNodiTot=-1;
}







long inDeg::aggiungiInDeg(long nodo){
 	if (nodo<0 || nodo>=numNodiTot)	return -1;

 	//calcola le posizioni nei due livelli
 	long pos2=nodo/COEFF2;
 	long pos1=pos2/COEFF1;
 	long pos0=pos1/COEFF0;

 	if (pos0>=size0)return -1;
 	if (pos1>=size1)return -2;
 	if (pos2>=size2)return -3;

  inDegNodi[nodo]++;
 	livello2[pos2]++;
 	livello1[pos1]++;
 	livello0[pos0]++;
  sommaInDeg++;

 	//printf("liv 0 -> %ld  liv 1 -> %ld  liv 1 -> %ld\n",pos0,pos1,pos2);
 	return 0;
}







//trova il nodo (o almeno l'inizio di un blocco) in cui il valore conduce
long inDeg::trovaNodo(long valore){

  //il valore accettato va da 1 al valore massimo
  if (valore<1) return -10;
  if (valore>sommaInDeg) return -11;

	//----------------------------------------------------
  //trova il blocco del livello 0
	//----------------------------------------------------
	bool trovato0=false;
	long num0;
	for (num0=0;num0<=size0&&!trovato0;num0++){
			if (valore<=livello0[num0]) { trovato0=true; break;}
			else						              valore-=livello0[num0];
	}
	if (!trovato0 ) return -1;



	//----------------------------------------------------
	//trova il blocco del livello 1
	//----------------------------------------------------
	bool trovato1=false;
	long num1;
	for (num1 = num0*COEFF0; num1<=size1&&!trovato1; num1++){
			if (valore<=livello1[num1]) { trovato1=true; break;}
			else						           valore-=livello1[num1];
	}
	if (!trovato1 ) return -2;




	//----------------------------------------------------
	//trova il blocco del livello 2
	//----------------------------------------------------
	bool trovato2=false;
	long num2;
	for (num2 = num1*COEFF1; num2<=size2 && !trovato2; num2++){
			if (valore<=livello2[num2]) { trovato2=true; break;}
      else						           valore-=livello2[num2];
	}
	if (!trovato2 ) return -3;





	//----------------------------------------------------
  //trova nodo direttamente nella struttura finale
	//----------------------------------------------------
	//num nodo da cui ripartire con il calcolo diretto sugli indegree
	long nodoTrovato = num2*COEFF2;

	//ricerca diretta nui nodi mancanti
	bool trovato=false;
	for ( ;nodoTrovato<=numNodiTot&&!trovato;nodoTrovato++)
			if (valore<=inDegNodi[nodoTrovato]) { trovato=true; break;}
			else						 valore-=inDegNodi[nodoTrovato];
  if (!trovato) return -4;



  return nodoTrovato;

}

