/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file multiSemaforo.cc :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE   rivisto 23 nov 2001
//=======================================================================
//=======================================================================

#include "funzioniGenerali.h"
#include "multiSemaforo.h"
#include <unistd.h>

/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long multiSemaforo::signal(long num){
  if (num<0||num>=numeroSemafori) return -1;

  struct sembuf aux={num,1,0};
  if (semop(sem_id,&aux,1)==-1){
    fprintf(stderr,"Unexpected fatal error for multisemaphore number %ld, ID %ld (signal function)",num,sem_id);
    printError("");
    FINE;
    return -1;
  }
  return 0;
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long multiSemaforo::wait(long num){
  if (num<0||num>=numeroSemafori) return -1;

  struct sembuf aux={num,-1,0};
  if (semop(sem_id,&aux,1)==-1){
    fprintf(stderr,"Waiting error for multi-semaforo number %ld, ID %ld",num,sem_id);
    printError("");
    FINE;
    return -1;
  }
  return 0;
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long multiSemaforo::getId(){
  return sem_id;
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long multiSemaforo::getValue(long num){
  if (num<0||num>=numeroSemafori) return -1;//errore

  long status;
  union semun aux;//variabile ausiliaria
  if ( (status=semctl(sem_id,num,GETVAL,aux))==-1){
    fprintf(stderr,"Error: unable to get the semaphore %ld ",sem_id);
    printError("");
    FINE;
    return -1;
  }
  return status;
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void multiSemaforo::cancella(){
  //distruttore - dealloco il semaforo

      union semun aux;//variabile ausiliaria
      if(semctl(sem_id,0,IPC_RMID,aux)<0){
            fprintf(stderr,"Calculation error for semaphore %ld",sem_id);
            printError("");
            FINE;
      }
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
multiSemaforo::multiSemaforo(long oldId,long oldNum,bool inutile){
/*attath di un semaforo esistente */
    inutile=0;
    numeroSemafori=oldNum;
    sem_id=oldId;
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
multiSemaforo::multiSemaforo(long num,long init){
/*crea un nuovo semaforo con valore iniziale init*/
    numeroSemafori=num;
    sem_id=-1;

    if (init<0){//utilizzare la chiamata di sistema
        init=-init;
    }

    //******************* CREAZIONE SEMAFORO ***************************
    do {
         sem_id=getSemaforo( rand(),num );//uso la wrapper function
    }
    while ((errno==EEXIST)&&(sem_id==-1));//mentre esiste
    if (sem_id==-1){
      printError("Unexpected fatal error n.- 1 multiSemaforo ");
      FINE;
      return;
    }

    //******************* INIZIALIZZAZIONE SEMAFORO ***************************
    union semun aux;//variabile ausiliaria
    aux.val=init;//campo usato per settare il valore iniziale, senza aspettare

    for (long i=0;i<numeroSemafori;i++){
      if( semctl(sem_id,i,SETVAL,aux)==-1){
          fprintf(stderr,"Unexpected fatal error n.1 function multiSemaforo - ID %ld",sem_id);
          printError("");
          FINE;
          return;
      }
    }
}


void multiSemaforo::setValue(long num,long init){
    if (num<0||num>=numeroSemafori) return;//errore

    union semun aux;//variabile ausiliaria
    aux.val=init;//campo usato per settare il valore iniziale, senza aspettare
    if( semctl(sem_id,num,SETVAL,aux)==-1){
          fprintf(stderr,"Unexpected fatal error n.1 function setValue - ID %ld",sem_id);
          printError("");
          FINE;
          return;
    }
}
