/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file processi.cc :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE rivisto 23 nov 2001
//=======================================================================
//=======================================================================

#include "processi.h"
#include "funzioniGenerali.h"
//#include <iostream.h>
#include <iostream>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>














/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
processi::processi(){

      //IN/ERR --> IN
      //fclose(stderr);
      //dup(0);

      //DISABILITA I SEGNALI SCOMODI... definiti in funzioniGenerali.cc
      signal(SIGTSTP,  interruzioneTastiera);//interrupt from key
      signal(SIGINT, stopTastiera);
      

      //file temporaneo dei semafori istanziati
      sprintf(nomeFile,"semaphore_report (PID_%d).txt",getpid() );
      fp=fopen(nomeFile,"w");
      char app[1000];
      sprintf(app,"#-------------------------------------------------------------------------------------\n#the line '+id_sem' means that the new semaphore with the identifier id_sem is created\n#the line '-id_sem' means that the semaphore with the identifier id_sem is removed#\n#-------------------------------------------------------------------------------------\n\n");
                   
      if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);
      fflush(fp);

      /*------------------------------------------------------------------*/
      //crea un nuovo semaforo per la gestione di questa struttura
      do{sem_id=semget(rand(),1,IPC_CREAT|IPC_EXCL|0660);}
      while ((errno==EEXIST)&&(sem_id==-1));//mentre esiste
      if (sem_id==-1){printError("Unexpected fatal error: processi() \n");fflush(stderr);exit(0);}

      sprintf(app,"#  +%ld\n",sem_id);
      if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);

      sprintf(app,"ipcrm sem %ld\n",sem_id);
      if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);
      fflush(fp);


      //set il valore del semaforo in uso
      inUso=new multiSemaforo(sem_id,1,false); // attach del semaforo
      inUso->setValue(0,1);//semaforo settato a 1 (libero)

      /*------------------------------------------------------------------*/ 	    	
    	//puntatori alle varie zone di memoria per le liste

      numeroSem=new long;
      listaSem= new long;

      numeroMem=new long;    
      listaMem= new long[2000];
  	
      /*------------------------------------------------------------------*/
      //init contatori       	
      *numeroSem=0;
      *numeroMem=0;
      pid=getpid();
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                                 private
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::stampaInterna(){
    printf("SEMAFORI (%ld):",*numeroSem);
    for (long i=0;i<*numeroSem;i++)
       printf("%ld ",listaSem[i]);
    printf("\n");

    printf("SHARED MEMORY (%ld):",*numeroMem);
    for (long i=0;i<*numeroMem;i++)
       printf("%ld ",listaMem[i]);
    printf("\n");

}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::stampa(){
    inUso->wait(0);
    stampaInterna();
    inUso->signal(0);
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::exitAll(long state){
// elimina tutti i processi correntemente in esecuzione
// e tutta la memoria condivisa e semafori in uso


  inUso->wait(0);

  //messaggi di errore in null..
  fclose(stderr);
  open("/dev/null",O_WRONLY);
  fclose(stdout);
  open("/dev/null",O_WRONLY);

  for (long i=0;i<*numeroSem;i++)   eliminaSemaforo(listaSem[i]);
  for (long i=0;i<*numeroMem;i++)   eliminaShmem(listaMem[i]);

  //elimina tutti i semafori del set
  union semun aux;//variabile ausiliaria
  semctl(sem_id,0,IPC_RMID,aux);

  if (fp!=NULL) fclose(fp);
  remove(nomeFile);

  //elimina il processo e tutti i figli...
  if (pid!=getpid())
    kill(pid,15);//termina il main (sigterm signal o kill a seconda del tipo di terminazione)

  fflush(stdout);fflush(stderr);
  exit(state);//termina il processo corrente

  //IL PROGRAMMA NON RITORNA
  //non permette a nessun altro processo di accedere alla mem condivisa o temrinare con errore
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::abortAll(){
// elimina tutti i processi correntemente in esecuzione
// e tutta la memoria condivisa e semafori in uso


  inUso->wait(0);
  printf("\nAborting...OK\n");fflush(stdout);

  //messaggi di errore in null..
  fclose(stderr);
  open("/dev/null",O_WRONLY);
  fclose(stdout);
  open("/dev/null",O_WRONLY);

  for (long i=0;i<*numeroSem;i++)  eliminaSemaforo(listaSem[i]);
  for (long i=0;i<*numeroMem;i++)  eliminaShmem(listaMem[i]);

  //elimina tutti i semafori del set
  union semun aux;//variabile ausiliaria
  semctl(sem_id,0,IPC_RMID,aux);

  if (fp!=NULL) fclose(fp);
  remove(nomeFile);

  //le strutture condivise non vengono rilaasciate...
  kill(pid,9);//termina il main (sigterm signal o kill a seconda del tipo di terminazione)

  fflush(stdout);fflush(stderr);
  abort();//termina il processo corrente

  //IL PROGRAMMA NON RITORNA
  //non permette a nessun altro processo di accedere alla mem condivisa o temrinare con errore
}







/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                                 private
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long processi::cancella(char* c,long pid,long* vet, long& size){
//elimina un elemento dal vettore se c'e'
  for (long i=0;i<size;i++){
    if(vet[i]==pid){
        vet[i]=vet[size-1];
        size--;
        return 0;
    }
  }
//  cout<<" ID "<<pid<<" per "<<c<<" non trovato\n";
  c=(char*)"";
  return -1;
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                                 private
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long processi::aggiungi(char* c,long pid,long* vet, long& size,long limite){
//aggiunge un elemento al vettore
  if (size>=limite){
//    fprintf(stderr,"spazio esaurito per memorizzare il PID %ld per %s\n",pid,c);
    c=(char*)"";
    return -1;
  }
  vet[size++]=pid;
  return 0;
}




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::aggiungiSem(long semid){
//aggiunge il PID di un nuovo processo alla fine
//NB: i pid sono tutti diversi

  inUso->wait(0);

  //-------------------------------------------------------------
  //aggiunge un nuovo semaforo
  //-------------------------------------------------------------
  sprintf(app,"#  +%ld\n",semid);
  if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);

  sprintf(app,"ipcrm sem %ld\n",semid);
  if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);
  fflush(fp);
  //-------------------------------------------------------------

  aggiungi((char*)" sem ",semid,listaSem,*numeroSem,2000);
  inUso->signal(0);
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::cancellaSem(long semid){
//elimina un pid
//NB: i pid sono tutti diversi

  inUso->wait(0);
  cancella((char*)" sem ",semid,listaSem,*numeroSem);

  //-------------------------------------------------------------
  //elimina il semaforo
  //-------------------------------------------------------------
  sprintf(app,"#  -%ld\n",semid);
  if (fp!=NULL) fwrite(app,sizeof(char),strlen(app),fp);
  fflush(fp);
  //-------------------------------------------------------------

  inUso->signal(0);
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::aggiungiMem(long memid){

//aggiunge il PID di un nuovo processo alla fine
//NB: i pid sono tutti diversi

  inUso->wait(0);

  aggiungi((char*)" mem ",memid,listaMem,*numeroMem,2000);
  inUso->signal(0);
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::cancellaMem(long memid){
//elimina un pid
//NB: i pid sono tutti diversi

  inUso->wait(0);

  cancella((char*)" mem ",memid,listaMem,*numeroMem);
  inUso->signal(0);
}



/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                                 private
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::eliminaSemaforo(long id){
  union semun aux;//variabile ausiliaria
  semctl( (int)id,0,IPC_RMID,aux );
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                                 private
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void processi::eliminaShmem(long id){
  //   elimina la memoria (senza il detach) --
  //   eliminazione reale solo dopo la fine di tutti i processi
  shmctl((int)id,IPC_RMID,0);
}

