#ifndef BUFFERLETTURAFILE_H_
#define BUFFERLETTURAFILE_H_

// AG_CHANGE for abort()
#include <stdlib.h>


template <class T>  class BufferLetturaFile
{
  public:
  BufferLetturaFile();
  BufferLetturaFile(FILE * );
  BufferLetturaFile(FILE * , long dim_b);
  BufferLetturaFile(FILE * , long dim_b, T * buffer_passato); //questa versione usa il buffer passato

  ~BufferLetturaFile();
  int get(T & data); //legge un dato del tipo parametrizzato dal buffer
  long capienzaBuffer() { return capienza_buffer; }; //numero di elementi nel buffer allocato
  bool skip(long num_elem_da_saltare); //salta, leggendo il file, il numero di elementi specificato. Ritorna true se abbiamo saltato tutti gli elementi desiderati prima della fine del file, false altrimenti


  private:
  enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BUFFER= 1}; //definizione costanti per la classe

  bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
  long capienza_buffer; //numero massimo di dati contenuti nel buffer
  T * buffer; //puntatore alla memoria allocata
  long next_data; //indice prossimo dato nel buffer
  long dati_letti; //quanti dati sono stati effettivamente messi nel buffer l'ultima volta
  FILE * file_input; //il file da leggere

  void init(FILE * , long dim_b); //invocata dai costruttori
  void init(FILE * , long dim_b, T * buffer_passato); //questa versione usa il buffer passato

  BufferLetturaFile & operator=(const BufferLetturaFile &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  BufferLetturaFile(const BufferLetturaFile &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};


//********************** segue la definizione della classe ************************************************************




template<class T>
BufferLetturaFile<T>::BufferLetturaFile()
{
  init(NULL, DIM_STANDARD/sizeof(T));
}


template<class T>
BufferLetturaFile<T>::BufferLetturaFile(FILE * f_inp)
{
  init(f_inp, DIM_STANDARD/sizeof(T)); //alloca un numero di elementi che si avvicina quanto possibile alla costante fissata
}


template<class T>
BufferLetturaFile<T>::BufferLetturaFile(FILE * f_inp, long dim_b)
{
  init(f_inp, dim_b/sizeof(T)); //alloca un numero di elementi pari alla dimensione richiesta al buffer (approssimando per difetto se necessario)
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
BufferLetturaFile<T>::BufferLetturaFile(FILE * f_inp, long dim_b, T * buffer_passato)
{
  init(f_inp, dim_b/sizeof(T), buffer_passato);
}


template<class T>
BufferLetturaFile<T>::~BufferLetturaFile()
{
  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] buffer;
}


template<class T>
void BufferLetturaFile<T>::init(FILE * f_inp, long cap_b) //cap_b e' il numero di elementi del tipo generico T che il buffer deve contenere
{
  file_input= f_inp;
  capienza_buffer= cap_b;
  buffer_nostro= true;

  if(capienza_buffer < CAPIENZA_MINIMA_BUFFER)
    capienza_buffer= CAPIENZA_MINIMA_BUFFER; //forziamo ad aver un numero minimo di elementi nel buffer

  buffer= new T[capienza_buffer]; //alloca spazio per il numero richiesto di elementi del tipo parametrico T
  if(buffer == NULL)
  {
    fprintf(stderr, "BufferLetturaFile: Errore nell'allocazione della memoria.\n");
    abort();
  }

  dati_letti= capienza_buffer; //serve per la prima get(): fingiamo di aver gi letto <capienza_buffer> elementi in modo da capire che non stiamo nell'ultimo blocco del file
  next_data= capienza_buffer;//ci posizioniamo alla fine del buffer in modo da forzare il caricamento alla prima richiesta di un dato
}


//questa versione usa il buffer passato
template<class T>
void BufferLetturaFile<T>::init(FILE * f_inp, long dim_b, T * buffer_passato)
{
  file_input= f_inp;
  capienza_buffer= dim_b;
  buffer_nostro= false;

  if(capienza_buffer < 1)
  {
    fprintf(stderr, "BufferLetturaFile: La dimensione del buffer passato non e' corretta\n");
    abort();
  }

  buffer= buffer_passato; //useremo il buffer specificato

  dati_letti= capienza_buffer; //serve per la prima get(): fingiamo di aver gi letto <capienza_buffer> elementi in modo da capire che non stiamo nell'ultimo blocco del file
  next_data= capienza_buffer;//ci posizioniamo alla fine del buffer in modo da forzare il caricamento alla prima richiesta di un dato
}


template<class T>
int BufferLetturaFile<T>::get(T & data)
{
  if(file_input==NULL)
    return 0; //nessun dato restituito

  if(next_data >= dati_letti) //i dati gia' caricati nel buffer sono finiti
  {
    if(dati_letti < capienza_buffer) //l'ultima lettura ha raggiunto la fine del file
      return 0; //nessun dato restituito (finito il buffer e finito il file)
    //altrimenti possiamo effettuare una nuova lettura dal file per ricaricare il buffer

    dati_letti= fread( buffer, sizeof(T), capienza_buffer, file_input); //riempiamo il buffer di dati
    if(dati_letti < 1)
      return 0; //se era finito contemporaneamente il buffer ed il file

    next_data= 0; //ci riportiamo all'inizio del buffer
  }
  //restituiamo un dato
  data= buffer[next_data];
  next_data++;
  return 1;
}


//salta, leggendo il file, il numero di elementi specificato
//ritorna true se abbiamo saltato tutti gli elementi desiderati prima della fine del file, false altrimenti
template<class T>
bool BufferLetturaFile<T>::skip(long num_elem_da_saltare)
{
  if(file_input==NULL)
    return false; //nessun dato accessibile

  int dati_rimasti= dati_letti - next_data;
  if( num_elem_da_saltare <= dati_rimasti ) //se il numero di posizioni da saltare e' inferiore o uguale al numero di elementi ancora da leggere presenti nel buffer
  {
    next_data += num_elem_da_saltare; //facciamo come se avessimo letto <num_elem_da_saltare> elementi ed usciamo
    return true;
  }
  //else..

  num_elem_da_saltare -= dati_rimasti; //saltiamo intanto i dati che erano rimasti nel buffer
  next_data= dati_letti; //come se li avessimo letti normalmente: ci assicura uscendo se (dati_letti < capienza_buffer) che valga la condizione di EOF sul buffer

  if(dati_letti < capienza_buffer) //l'ultima lettura ha raggiunto la fine del file
    return false; //finito il buffer e finito il file: non si possono leggere altri elementi
  //else..

  int blocks_to_skip= num_elem_da_saltare/capienza_buffer; //numero di blocchi in mezzo da saltare
  if(blocks_to_skip > 0)
  {
    int ris= fseek(file_input, (sizeof(T)*capienza_buffer)*blocks_to_skip, SEEK_CUR);
    if(ris != 0) //seek non eseguita con successo
    {
      dati_letti= 0; //serve ad assicurare la condizione dati_letti < capienza_buffer, in modo che le eventuali get() successive trovino EOF
      return false; //non siamo riusciti a "leggere" tutti gli elementi che erano da saltare
    }

    num_elem_da_saltare= num_elem_da_saltare % capienza_buffer; //quanti ne restano da saltare
    if(num_elem_da_saltare==0) //abbiamo appena saltato un numero intero di blocchi, non dobbiamo leggere (i.e. saltare) altro
      return true;
  }

  //a questo punto e' sempre: capienza_buffer > num_elem_da_saltare > 0, dati_letti == capienza_buffer
  dati_letti= fread( buffer, sizeof(T), capienza_buffer, file_input); //proviamo a riempire il buffer
  next_data= num_elem_da_saltare; //saltiamo i primi elementi del buffer
  return (dati_letti >= num_elem_da_saltare); //abbiamo saltato tutti gli elementi desiderati prima della fine del file?  N.B. se viene restituito false, perche' next_data=num_elem_da_saltare > dati_letti (con dati_letti<num_elem_da_saltare<capienza_buffer) -----> in questo caso anche le get() successive restituiranno false, come e' giusto


/*  if( num_elem_da_saltare <= capienza_buffer) //in questo caso per gli elementi che restano da saltare si puo' riempire il buffer ignorando i primi elementi
  {
    dati_letti= fread( buffer, sizeof(T), capienza_buffer, file_input); //proviamo a riempire il buffer
    next_data= num_elem_da_saltare; //saltiamo i primi elementi del buffer
    return (dati_letti >= num_elem_da_saltare); //abbiamo saltato tutti gli elementi desiderati prima della fine del file?  N.B. se viene restituito false, perche' next_data=num_elem_da_saltare > dati_letti  -----> in questo caso anche le get() successive restituiranno false, come e' giusto
  }
  else //ignoriamo gli elementi che restano ancora da saltare "sorvolandoli" su disco con la fseek
  {
    int ris= fseek(file_input, sizeof(T)*num_elem_da_saltare, SEEK_CUR);
    next_data= dati_letti; //in questo modo segnaliamo che il buffer non contiene altri dati significativi e forziamo una lettura da file alla prossima get()
    return (ris == 0); //seek eseguita con successo?
  }*/
}

#endif
