/*

  16/11/2002
  Ernesto Ricci
*/

#ifndef BUFFERSCRITTURAFILE_H_
#define BUFFERSCRITTURAFILE_H_


template <class T>  class BufferScritturaFile
{
  public:
  BufferScritturaFile();
  BufferScritturaFile(FILE * );
  BufferScritturaFile(FILE * , long dim_b);
  BufferScritturaFile(FILE * , long dim_b, T * buffer_passato); //questa versione usa il buffer passato
  ~BufferScritturaFile();

  void put(T & data); //scrive un dato del tipo parametrizzato
  long capienzaBuffer() { return capienza_buffer; } //numero di elementi nel buffer allocato
  void associaFile(FILE *);
  int flush();
  void chiudi();

  private:
  enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BUFFER= 1}; //definizione costanti per la classe

  bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
  long capienza_buffer; //numero massimo di dati contenuti nel buffer
  T * buffer; //puntatore alla memoria allocata
  long next_data; //indice prossima posizione libera nel buffer
  FILE * file_output; //il file da scrivere

  void init(FILE * , long cap_b); //invocata dai costruttori
  void init(FILE * , long cap_b, T * buffer_passato); //questa versione usa il buffer passato

  BufferScritturaFile & operator=(const BufferScritturaFile &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  BufferScritturaFile(const BufferScritturaFile &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};


//********************** segue la definizione della classe ************************************************************



template<class T>
BufferScritturaFile<T>::BufferScritturaFile()
{
  init(NULL, DIM_STANDARD/sizeof(T));
}


template<class T>
BufferScritturaFile<T>::BufferScritturaFile(FILE * h_file)
{
  init(h_file, DIM_STANDARD/sizeof(T)); //alloca un numero di elementi che si avvicina quanto possibile alla costante fissata
}


template<class T>
BufferScritturaFile<T>::BufferScritturaFile(FILE * h_file, long dim_b)
{
  init(h_file, dim_b/sizeof(T)); //alloca un numero di elementi pari alla dimensione richiesta al buffer (approssimando per difetto se necessario)
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
BufferScritturaFile<T>::BufferScritturaFile(FILE * h_file, long dim_b, T * buffer_passato)
{
  init(h_file, dim_b/sizeof(T), buffer_passato);
}


template<class T>
BufferScritturaFile<T>::~BufferScritturaFile()
{
  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] buffer;
}


template<class T>
void BufferScritturaFile<T>::chiudi()
{
  if(file_output != NULL)
  {
    this->flush();
    file_output= NULL;
  }
}


template<class T>
void BufferScritturaFile<T>::init(FILE * h_file, long cap_b) //dim_b e' il numero di elementi del tipo generico T che il buffer deve contenere
{
  file_output= h_file;
  capienza_buffer= cap_b;
  buffer_nostro= true;

  if(capienza_buffer < CAPIENZA_MINIMA_BUFFER)
    capienza_buffer= CAPIENZA_MINIMA_BUFFER; //forziamo ad aver un numero minimo di elementi nel buffer

  buffer= new T[capienza_buffer]; //alloca spazio per il numero richiesto di elementi del tipo parametrico T
  if(buffer == NULL)
  {
    fprintf(stderr, "BufferScritturaFile: Errore nell'allocazione della memoria.\n");
    abort();
  }

  next_data= 0;//ci posizioniamo all'inizio del buffer
}


//questa versione usa il buffer passato
template<class T>
void BufferScritturaFile<T>::init(FILE * h_file, long dim_b, T * buffer_passato)
{
  file_output= h_file;
  capienza_buffer= dim_b;
  buffer_nostro= false;

  if(capienza_buffer < 1)
  {
    fprintf(stderr, "BufferScritturaFile: La dimensione del buffer passato a BufferScritturaFile non e' corretta\n");
    abort();
  }

  buffer= buffer_passato; //useremo il buffer specificato

  next_data= 0;//ci posizioniamo all'inizio del buffer
}


template<class T>
int BufferScritturaFile<T>::flush()
{
  if((file_output==NULL) || (next_data==0))
    return 0; //non abbiamo scritto nulla

  long dati_scritti= fwrite( buffer, sizeof(T), next_data, file_output); //copiamo su disco il contenuto del buffer
  if(dati_scritti != next_data)
  {
    perror("BufferScritturaFile: Scrittura sul file non riuscita.");
    abort();
  }
  next_data= 0; //ci riportiamo all'inizio del buffer

  fflush(file_output); //imponiamo il flush anche al file
  return dati_scritti;
}


template<class T>
void BufferScritturaFile<T>::put(T & data)
{
  if(file_output==NULL)
    return; //nulla di fatto

  if(next_data == capienza_buffer) //buffer pieno
  { //svuotiamo il buffer salvando il contenuto su disco
    long dati_scritti= fwrite( buffer, sizeof(T), capienza_buffer, file_output); //riempiamo il buffer di dati
    if(dati_scritti < capienza_buffer)
    {
      perror("BufferScritturaFile: Scrittura sul file non riuscita.");
      abort();
    }
    next_data= 0; //ci riportiamo all'inizio del buffer
  }

  buffer[next_data]= data; //salviamo nel buffer il dato passato
  next_data++; //ci posizioniamo subito dopo nel buffer
}


template<class T>
void BufferScritturaFile<T>::associaFile(FILE * h_file)
{
  chiudi();
  next_data= 0;//ci posizioniamo all'inizio del buffer
  file_output= h_file;
}


#endif

