/*

  20/01/2003
  Ernesto Ricci
*/

#ifndef BUFFERSCRITTURAMULTIFILE_H_
#define BUFFERSCRITTURAMULTIFILE_H_
#include "../streamFiles/multiFileWriter.h"


template <class T>  class BufferScritturaMultifile
{
  public:
  BufferScritturaMultifile();
  BufferScritturaMultifile(multiFileWriter * );
  BufferScritturaMultifile(multiFileWriter * , long dim_b);
  BufferScritturaMultifile(multiFileWriter * , long dim_b, T * buffer_passato); //questa versione usa il buffer passato
  ~BufferScritturaMultifile();
  
  void put(T & data); //scrive un dato del tipo parametrizzato
  long capienzaBuffer() { return capienza_buffer; } //numero di elementi nel buffer allocato
  void associaFile(multiFileWriter *);
  int flush();
  void chiudi();

  private:
  enum {DIM_STANDARD=4096, CAPIENZA_MINIMA_BUFFER= 1}; //definizione costanti per la classe

  bool buffer_nostro; //indica se il buffer lo abbiamo allocato noi o se invece ce lo hanno passato
  long capienza_buffer; //numero massimo di dati contenuti nel buffer
  T * buffer; //puntatore alla memoria allocata
  long next_data; //indice prossima posizione libera nel buffer
  multiFileWriter * mfile; //il multifile da scrivere

  void init(multiFileWriter * , long cap_b); //invocata dai costruttori
  void init(multiFileWriter * , long cap_b, T * buffer_passato); //questa versione usa il buffer passato

  BufferScritturaMultifile & operator=(const BufferScritturaMultifile &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
  BufferScritturaMultifile(const BufferScritturaMultifile &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};


//********************** segue la definizione della classe ************************************************************



template<class T>
BufferScritturaMultifile<T>::BufferScritturaMultifile()
{
  init(NULL, DIM_STANDARD/sizeof(T));
}


template<class T>
BufferScritturaMultifile<T>::BufferScritturaMultifile(multiFileWriter * mf)
{
  init(mf, DIM_STANDARD/sizeof(T)); //alloca un numero di elementi che si avvicina quanto possibile alla costante fissata
}


template<class T>
BufferScritturaMultifile<T>::BufferScritturaMultifile(multiFileWriter * mf, long dim_b)
{
  init(mf, dim_b/sizeof(T)); //alloca un numero di elementi pari alla dimensione richiesta al buffer (approssimando per difetto se necessario)
}


//questa versione usa il buffer passato
template<class T>
BufferScritturaMultifile<T>::BufferScritturaMultifile(multiFileWriter * mf, long dim_b, T * buffer_passato)
{
  init(mf, dim_b/sizeof(T), buffer_passato);
}


template<class T>
BufferScritturaMultifile<T>::~BufferScritturaMultifile()
{
  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] buffer;
}


template<class T>
void BufferScritturaMultifile<T>::chiudi()
{
  if(mfile != NULL)
  {
    this->flush();
    mfile= NULL;
  }
}


template<class T>
void BufferScritturaMultifile<T>::init(multiFileWriter * mf, long cap_b) //dim_b e' il numero di elementi del tipo generico T che il buffer deve contenere
{
  mfile= mf;
  capienza_buffer= cap_b;
  buffer_nostro= true;

  if(capienza_buffer < CAPIENZA_MINIMA_BUFFER) //forziamo ad aver almeno un elemento nel buffer
    capienza_buffer= CAPIENZA_MINIMA_BUFFER; //posto per un solo elemento nel buffer

  buffer= new T[capienza_buffer]; //alloca spazio per il numero richiesto di elementi del tipo parametrico T
  if(buffer == NULL)
  {
    fprintf(stderr, "BufferScritturaMultifile: Errore nell'allocazione della memoria.\n");
    abort();
  }

  next_data= 0;//ci posizioniamo all'inizio del buffer
}


//questa versione usa il buffer passato
//N.B. la memoria fornita deve essere sufficiente ad allocare almeno un elemento del tipo T, altrimenti il buffer accederà oltre la memoria passata
template<class T>
void BufferScritturaMultifile<T>::init(multiFileWriter * mf, long cap_b, T * buffer_passato)
{
  mfile= mf;
  capienza_buffer= cap_b;
  buffer_nostro= false;

  if(capienza_buffer < 1)
  {
    fprintf(stderr, "BufferScritturaMultifile: La dimensione del buffer passato non e' corretta\n");
    abort();
  }

  buffer= buffer_passato; //useremo il buffer specificato

  next_data= 0;//ci posizioniamo all'inizio del buffer
}


template<class T>
int BufferScritturaMultifile<T>::flush()
{
  if((mfile==NULL) || (next_data==0))
    return 0; //non abbiamo scritto nulla

  long dati_scritti= mfile->write( buffer, sizeof(T), next_data); //copiamo su disco il contenuto del buffer
  if(dati_scritti != next_data)
  {
    perror("BufferScritturaMultifile: Scrittura sul file non riuscita.");
    abort();
  }
  next_data= 0; //ci riportiamo all'inizio del buffer

//  fflush(file_output); //imponiamo il flush anche al file
  return dati_scritti;
}


template<class T>
void BufferScritturaMultifile<T>::put(T & data)
{
  if(mfile==NULL)
    return; //nulla di fatto

  if(next_data == capienza_buffer) //buffer pieno
  { //svuotiamo il buffer salvando il contenuto su disco
    long dati_scritti= mfile->write( buffer, sizeof(T), capienza_buffer); //riempiamo il buffer di dati
    if(dati_scritti < capienza_buffer)
    {
      perror("BufferScritturaMultifile: Scrittura sul file non riuscita.");
      abort();
    }
    next_data= 0; //ci riportiamo all'inizio del buffer
  }

  buffer[next_data]= data; //salviamo nel buffer il dato passato
  next_data++; //ci posizioniamo subito dopo nel buffer
}


template<class T>
void BufferScritturaMultifile<T>::associaFile(multiFileWriter * mf)
{
  chiudi();
  mfile= mf;
  next_data= 0; //ci posizioniamo all'inizio del buffer
}


#endif

