/* Implementazione ricorsiva del Merge Sort.
   Usiamo la classe BufferAppoggio nella fase di merge
   (che va quindi inclusa prima di questa nel file che user�il Merge Sort).
   Gli elementi su cui viene fatto l'ordinamento devono essere di un tipo per cui
   sono definiti almeno gli operatori < e <=.

   1/4/2003
   Ernesto Ricci
*/

#ifndef MERGESORT_H
#define MERGESORT_H


#define DIM_BUFFER_AUX 128*1024 //memoria utilizzata per default per il buffer, se non viene specificato altrimenti chiamando mergesort()



template <class T>
  void scambia(T& a, T& b){
    T aux=a;
    a=b;
    b=aux;
  }



template <class T> //***** solo per il debug
bool check_ordinamento_ok(T * vettore, unsigned long indice_1, unsigned long indice_fine2)
{
  for(unsigned long i=indice_1; i<indice_fine2 ; i++)
  {
    //consideriamo due elementi successivi: quello alla posizione 'j'-1 ("prima") e quello alla posizione 'j' ("dopo")
    if( !(vettore[i] <= vettore[i+1]) ) //se l'elemento "dopo" e' piu' pesante dell'elemento "prima" li scambiamo
      return false;
  }

  return true;
}


//N.B. passiamo il buffer (per puntatore) invece di dichiararlo localmente, per evitare di allocare e deallocare memoria ad ogni chiamata della funzione
//sono stati fatti dei piccoli miglioramenti rispetto alla versione standard, in modo da non copiare inutilmente nel buffer
template <class T>
void merge_buffer(T * vettore, unsigned long indice_1, unsigned long indice_fine1, unsigned long indice_fine2, BufferAppoggio<T> * buffer_aux)
{
  unsigned long indice_2= indice_fine1 + 1;

  while( (indice_1 <= indice_fine1) && ( vettore[indice_1] <= vettore[indice_2] ) ) //vediamo gli elementi all'estrema SX che sono gia' al loro posto
  {
    indice_1++;
  }

  if(indice_1 > indice_fine1)
    return; //le due parti della lista erano gia' tra loro ben ordinate

  unsigned long punto_inserimento= indice_1; //ci ricordiamo a partire da dove dovremo inserire gli elementi che verranno messi nel buffer: quello che si trova a SX di questo punto e' gia' al suo posto

  while( (indice_1 <= indice_fine1) && (indice_2 <= indice_fine2) ) //finche' restano elementi in entrambe le sottoliste copiamo gli elementi piu' piccoli nel buffer
    if( vettore[indice_1] <= vettore[indice_2] )
    {
      buffer_aux->depositaValore( vettore[indice_1] );
      indice_1++;
    }
    else
    {
      buffer_aux->depositaValore( vettore[indice_2] );
      indice_2++;
    }

  if( buffer_aux->getSize() != (indice_1 - punto_inserimento) + (indice_2 - (indice_fine1 + 1) ) )
  {
    fprintf(stderr, "Impossibile scrivere tutti gli elementi nel buffer d'appoggio");
    abort();
  }
//  accertati_che( (buffer_aux->getSize() == (indice_1 - punto_inserimento) + (indice_2 - (indice_fine1 + 1) ) ), "Impossibile scrivere tutti gli elementi nel buffer d'appoggio"); //somma degli elementi della parte SX e della parte DX copiati nel buffer


  if(indice_1 > indice_fine1) //parte a SX terminata, quello che resta alla fine della parte DX e' gia' al suo posto
  {
    while( punto_inserimento < indice_2 ) //scriviamo al loro posto i nodi salvati nel buffer
    {
      buffer_aux->recuperaValore( vettore[punto_inserimento] );
      punto_inserimento++;
    }
  }
  else
  if(indice_2 > indice_fine2) //parte a DX terminata: cio' che resta alla fine della prima parte va accodato alla fine della lista e in mezzo vanno caricati i nodi salvati nel buffer
  {
    unsigned long larghezza_intervallo= indice_fine1 - indice_1 + 1;  //quanti elementi restano da considerare nella parte 1 della lista
    unsigned long copy_from= indice_1; //qua iniziano gli elementi da copiare
    unsigned long copy_to= copy_from + (indice_fine2 - indice_fine1); //l'indice a partire dal quale verranno copiato gli elementi da trasferire (notare che dobbiamo accodare gli elementi restanti nella parte 1 alla fine della lista, percio' trasliamo l'intervallo della dimensione della seconda parte della lista)

    for(unsigned long conta_up= 0; conta_up < larghezza_intervallo ; conta_up++) //dobbiamo spostare a DX, spostando prima gli elementi all'estrema DX
    {
      //(!)N.B. gli elementi devono essere copiati a partire dagli ultimi e solo alla fine i primi, se non sovrascriviamo gli elementi da copiare se i blocchi sorgente-destinazione "overlappano"
      unsigned long conta_down= larghezza_intervallo - 1 - conta_up; //'i' va da (larghezza_intervallo - 1) a 0
      vettore[ copy_to + conta_down ] = vettore[ copy_from + conta_down ];
    }

    while( punto_inserimento < copy_to )  //scriviamo al loro posto i nodi salvati nel buffer
    {
      buffer_aux->recuperaValore( vettore[punto_inserimento] );
      punto_inserimento++;
    }
  }

  if(buffer_aux->getSize() != 0)
  {
    fprintf(stderr, "Impossibile leggere tutti gli elementi dal buffer d'appoggio");
    abort();
  }
//  accertati_che( (buffer_aux->getSize() == 0), "Impossibile leggere tutti gli elementi dal buffer d'appoggio");

}


template <class T>
void mergesort_ric(T * vettore, unsigned long indice_inf, unsigned long indice_sup, BufferAppoggio<T> * buffer_aux)
{
  if(indice_sup == indice_inf) //1 elemento
    return;

  if(indice_sup == indice_inf + 1) //2 elem.
  {
    if( vettore[indice_sup] < vettore[indice_inf] ) //se l'elemento "dopo" e' piu' pesante dell'elemento "prima" li scambiamo
      scambia(vettore[indice_sup], vettore[indice_inf]);

    return;
  }

  unsigned long indice_med= (indice_inf + indice_sup) / 2;


  mergesort_ric(vettore, indice_inf, indice_med, buffer_aux);
  mergesort_ric(vettore, indice_med + 1, indice_sup, buffer_aux);

  merge_buffer(vettore, indice_inf, indice_med, indice_sup, buffer_aux);
}



//ordina gli elementi del vettore dagli indici 'indice_inf' a 'indice_sup' compresi
template <class T>
void mergesort_ricorsivo(T * vettore, unsigned long indice_inf, unsigned long indice_sup, unsigned long dim_memoria_buffer= DIM_BUFFER_AUX)
{
  BufferAppoggio<T> buffer_aux(dim_memoria_buffer);

  mergesort_ric(vettore, indice_inf, indice_sup, &buffer_aux);
}





//-------------MergeSort iterativo e con vettore "gemello": si evitano le copie scambiando tra loro i vettori -------------


// Sort a[0:n-1] using merge sort.
template<class T>
void MergeSort_iterativo(T * vettore, int n, T * aux)
{
  int s= 1; //segment size
  while(s < n)
  {
    MergePass(vettore, aux, s, n); // merge from a to b
    s += s;
    MergePass(aux, vettore, s, n); // merge from b to a
    s += s;
  }
//******
#ifdef TEST_CHECKSORT
  printf("\ncheck sort\n");
  if( !check_ordinamento_ok(vettore, 0, n-1))
  {
    fprintf(stderr, "errore nell'ordinamento di MergeSort_iterativo\n");
  }
#endif
//******
}




// Merge adjacent segments of size s.
template<class T>
void MergePass(T x[], T y[], int s, int n)
{
  int i= 0;
  while (i + 2 * s <= n)
  {
    // merge two adjacent segments of size s
    Merge_standard(x, y, i, i+s-1, i+2*s-1);
    i = i + 2 * s;
  }

  // fewer than 2s elements remain
  if(i + s < n)
    Merge_standard(x, y, i, i+s-1, n-1);
  else
    for (int j = i; j <= n-1; j++) // copy last segment to y
      y[j] = x[j];
}




// Merge c[l:m]] and c[m:r] to d[l:r].
template<class T>
void Merge_standard(T c[], T d[], int l, int m, int r)
{
  int i= l; // cursor for first segment
  int j= m + 1; // cursor for second
  int k = l; // cursor for result

  // merge until i or j exits its segment
  while((i <= m) && (j <= r))
    if(c[i] <= c[j])
      d[k++] = c[i++];
    else
      d[k++] = c[j++];

  // take care of left overs
  if(i > m)
  {
    for( ; j <= r; j++)
      d[k++] = c[j];
  }
  else
    for( ; i <= m; i++)
      d[k++] = c[i];
}



#endif
