#include "VettoreBits.h"
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>


const char * MSG_ERR_INCOMPATIBILI= "VettoreBits: operazione tra vettori di dimensione diversa non ammessa";


void VettoreBits::reset(bool valore)
{
  if(valore) //resettare ad 1
    for(unsigned long i=0; i<sizeBuffer; i++)
      buffer[i]= ~0; //NOT(00000..0) --> 1111..11
  else //resettare a 0
    for(unsigned long i=0; i<sizeBuffer; i++)
      buffer[i]= 0;
}



VettoreBits::~VettoreBits()
{
  if(buffer_nostro) //se ci e' stato passato non dobbiamo deallocarlo noi
    delete [] buffer;
}



void VettoreBits::init(unsigned long num_el)
{
  numElementi= num_el;
  buffer_nostro= true;

  sizeBuffer= numElementi / (sizeof(TIPO_BASE)*8); //num elementi del vettore da allocare

  if( (numElementi % (sizeof(TIPO_BASE)*8)) != 0 ) //se c'e' resto..
    sizeBuffer++; //..un elemento in piu'

  buffer= new TIPO_BASE[sizeBuffer]; //alloco il buffer

  if(buffer==NULL)
  {
    fprintf(stderr, "VettoreBits: Errore nell'allocazione della memoria.\n");
    abort();
  }

  //inizializziamo gli indici per la ricerca di bits=1/0
  nodo_next= 0; //il primo bit dell'array di bits
  posBuffer_next= 0; //partiamo per default dal primo elemento di <buffer>
  posBit_next= 0; //partiamo dal primo bit del primo elemento di <buffer>, quindi dal primo bit dell'array di bits
}



//questa versione usa il buffer passato
void VettoreBits::init(unsigned long num_el, unsigned long dim_b, TIPO_BASE * buffer_passato)
{
  numElementi= num_el;
  buffer_nostro= false;

  sizeBuffer= numElementi / (sizeof(TIPO_BASE)*8); //num elementi del vettore da allocare

  if( (numElementi % (sizeof(TIPO_BASE)*8)) != 0 ) //se c'e' resto..
    sizeBuffer++; //..un elemento in piu'

  if(dim_b < sizeBuffer*sizeof(TIPO_BASE))
  {
    fprintf(stderr, "VettoreBits: La dimensione del buffer passato non e' corretta\n");
    abort();
  }
  //else..
  buffer= buffer_passato; //useremo il buffer specificato

  //inizializziamo gli indici per la ricerca di bits=1/0
  nodo_next= 0; //il primo bit dell'array di bits
  posBuffer_next= 0; //partiamo per default dal primo elemento di <buffer>
  posBit_next= 0; //partiamo dal primo bit del primo elemento di <buffer>, quindi dal primo bit dell'array di bits
}



VettoreBits::VettoreBits(unsigned long num_el)
{
  init(num_el);
}



VettoreBits::VettoreBits(unsigned long num_el, bool valore_iniziale)
{
  init(num_el);
  reset(valore_iniziale); //inizializziamo i bit
}



//questa versione usa il buffer passato
VettoreBits::VettoreBits(unsigned long num_el, unsigned long dim_b, TIPO_BASE * buffer_passato)
{
  init(num_el, dim_b, buffer_passato);
}


//questa versione usa il buffer passato
VettoreBits::VettoreBits(unsigned long num_el, bool valore_iniziale, unsigned long dim_b, TIPO_BASE * buffer_passato)
{
  init(num_el, dim_b, buffer_passato);
  reset(valore_iniziale); //inizializziamo i bit
}



//stampa gli indici dei bit che sono a 1
void VettoreBits::print(const char * messaggio)
{
  unsigned long nodo_raggiunto, conta;
  bool found;

  printf("\n%s:\n", messaggio);
  conta= 0;

  //ricerca e output dei nodi marcati come visitati:
  found= findNext1(nodo_raggiunto, 0); //cerchiamo il primo nodo raggiunto, a partire dal nodo 0
  while(found)
  {
    conta++;
    printf("%lu ", nodo_raggiunto);
    found= findNext1(nodo_raggiunto);
  }
  printf("\n(tot %lu elementi)", conta);
}



bool VettoreBits::set(unsigned long nodo, bool bit, long & errore){

  //controllo i vincoli di validita' del nodo (tra 0 e max-1)
  if(nodo >= numElementi)
  {
    errore= -1; //settiamo l'errore
    return false; //valore non significativo perch�c'�stato errore
  }

  //else..nodo entro il range consentito
  errore= 0; //settiamo ok

  //individua la posizione da modificare nell'array
  unsigned long posElem= nodo / (sizeof(TIPO_BASE)*8); //num elemento del buffer in cui agire
  unsigned long numBit= nodo % (sizeof(TIPO_BASE)*8); //posizione nell'elemento di memoria del buffer

  TIPO_BASE mask= ((TIPO_BASE) 1) << numBit; //creata la maschera

  if( (bool) (buffer[posElem] & mask) == bit ) //se quel bit aveva gia' il valore richiesto non facciamo nulla
    return bit; //ritorniamo il valore precedente
  //else.. il bit in questione ha il valore opposto a quanto richiesto

  buffer[posElem] ^= mask; //faccio uno XOR solo sul bit del nodo in questione, invertendo cosi' il suo valore

  return !bit; //ritorniamo il valore precedente, opposto al valore appena settato
}



//leggiamo il valore di un certo bit
bool VettoreBits::get(unsigned long nodo, long & errore){
  //controllo i vincoli di validita' del nodo (tra 0 e max-1)
  if(nodo>=numElementi)
  {
    errore= -1; //settiamo l'errore
    return false; //questo valore non significa niente perche' viene impostato l'errore
  }

  //else..nodo entro il range consentito
  errore= 0; //settiamo ok

  //individua posizione dell'array da modificare
  unsigned long posElem= nodo / (sizeof(TIPO_BASE)*8); //num elemento del buffer in cui agire
  unsigned long numBit= nodo % (sizeof(TIPO_BASE)*8); //posizione nell'elemento di memoria del buffer

  TIPO_BASE mask= ((TIPO_BASE) 1) << numBit; //creata la maschera

  return ( buffer[posElem] & mask );  //isoliamo il bit richiesto e lo ritorniamo
}



//restituisce in "pos_trovato" il primo bit settato a 1, cercando a partire dal bit alla posizione specificata da <pos_start>. Ritorna true se e' stato trovato un tale bit, false altrimenti
//N.B. la ricerca avviene per posizioni crescenti del vettore di bits
bool VettoreBits::findNext1(unsigned long & pos_trovato, unsigned long pos_start)
{
  if(pos_start>=numElementi)
    return false; //nulla di fatto

  //else..il punto di partenza e' nel range valido

  nodo_next= pos_start; //indice del bit nell'array di bits da cui iniziare a cercare
  posBuffer_next= pos_start / (sizeof(TIPO_BASE)*8); //num elemento del buffer da cui partire
  posBit_next= pos_start % (sizeof(TIPO_BASE)*8); //posizione all'interno dell'elemento di memoria del buffer da cui iniziare a cercare

  return findNext1(pos_trovato); //deleghiamo a findNext1(long & pos_trovato) il compito di effettuare la ricerca vera e propria
}



//restituisce in "pos_trovato" il primo bit settato a 0, cercando a partire dal bit alla posizione specificata da <pos_start>. Ritorna true se e' stato trovato un tale bit, false altrimenti
//N.B. la ricerca avviene per posizioni crescenti del vettore di bits
bool VettoreBits::findNext0(unsigned long & pos_trovato, unsigned long pos_start)
{
  if(pos_start>=numElementi)
    return false; //nulla di fatto

  //else..il punto di partenza e' nel range valido

  nodo_next= pos_start; //indice del bit nell'array di bits da cui iniziare a cercare
  posBuffer_next= pos_start / (sizeof(TIPO_BASE)*8); //num elemento del buffer da cui partire
  posBit_next= pos_start % (sizeof(TIPO_BASE)*8); //posizione all'interno dell'elemento di memoria del buffer da cui iniziare a cercare

  return findNext0(pos_trovato); //deleghiamo a findNext0(long & pos_trovato) il compito di effettuare la ricerca vera e propria
}



//restituisce in "pos_trovato" il primo bit settato a 1, cercando a partire dal bit alla posizione corrente (posizione 'nodo_next'). Ritorna true se e' stato trovato un tale bit, false altrimenti
//N.B. la ricerca avviene per posizioni crescenti del vettore di bits
bool VettoreBits::findNext1(unsigned long & pos_trovato)
{
  if(nodo_next >= numElementi) //siamo gia' oltre la fine dell'array di bits
    return false; //nulla di fatto
  //else.. cerchiamo il primo elemento di <buffer>, a partire dalla posizione corrente, che non e' tutto nullo:

  TIPO_BASE mask= ~((((TIPO_BASE) 1) << posBit_next) - 1); //ovvero NOT( 2^posBit_next - 1 ). N.B.: ((1 << posBit_next) - 1) e' un del tipo 0, 1, 11, 111, 1111, ... per posBit_next=0, 1, 2, 3, 4, ... percio' il suo complemento a 1 e': 1..1, 1..10, 1..100, 1..1000, 1..10000, ecc.

  //cerco nell'elemento corrente se c'e' almeno un 1 (a partire dalla pos. di partenza, i.e. dal bit corrente)
  if( buffer[posBuffer_next] & mask ) //ignoro (maschero) la parte precedente (gia' esaminata) dell'elemento corrente per fare il test.
  {
    return scandisci_elemento_1(pos_trovato); //il residuo di questo elemento contiene degli uno: lo scandisco bit a bit
  }
  //else.. nei bit residui dell'elemento corrente di <buffer> non ci sono bit=1, inutile considerarlo --> passiamo agli elementi successivi di <buffer>


  nodo_next -= posBit_next; //mi riporto all'inizio dell'elemento corrente..
  posBit_next= 0; //..in modo che poi scandisci_elemento() inizi a cercare dal primo bit dell'elemento considerato.

  //passiamo a considerare gli elementi successivi di <buffer> fino a trovarne uno non nullo o fino ad arrivare alla fine di <buffer>
  do
  {
    posBuffer_next++; //l'elemento successivo nell'array
    nodo_next += sizeof(TIPO_BASE)*8; //avanziamo di un numero di posizioni pari ai bit contenuti in un elemento di <buffer> (in modo da trovarci sempre all'inizio, primo bit, dell'elemento corrente)

    if(posBuffer_next >= sizeBuffer) //siamo arrivati alla fine di <buffer> senza trovare alcun elemento che non fosse completamente nullo
      return false;

  } while( buffer[posBuffer_next] == 0 );

  //se non siamo usciti per la "return" nel ciclo prima allora abbiamo trovato un elemento di <buffer> non nullo: lo scandiamo bit a bit ed a partire dal primo bit
  return scandisci_elemento_1(pos_trovato); //affidiamo la risposta a scandisci_elemento() (e anche la modifica del parametro passato)
}


/*
  Trova il primo bit =1 nell'elemento corrente di <buffer>, cercando a partire dal bit numero <posBit_next>. Appena lo trova salva il numero del nodo corrispondente nel parametro di ritorno.
  Ritorna true se il bit trovato appartiene effettivamente all'array, false se invece e' uno dei bit dell'ultimo elemento di <buffer> che non fa parte del vettore di bits, ma e' inutilizzato.
  Precondizioni:  1) esiste almeno un bit =1 nell'elemento corrente di <buffer> a partire dalla posizione <posBit_next>
*/
bool VettoreBits::scandisci_elemento_1(unsigned long & pos_trovato)
{
  nodo_next -= posBit_next; //sottraggo per riportarmi all'inizio dell'elemento corrente. Serve per calcolare poi <nodo_next> facilmente tramite <posBit_next>

  TIPO_BASE mask= ((TIPO_BASE) 1) << posBit_next; //il bit selezionato

  //N.B. so per ipotesi che trovero' almeno un bit=1 a partire dalla posizione posBit_next nell'elemento, il controllo ".. && (posBit_next < sizeof(TIPO_BASE)*8)" e' superfluo
  while( !( buffer[posBuffer_next] & mask ) )
  {
    mask <<= 1; //shift a SX di una posizione: <mask> seleziona il prossimo bit piu' pesante
    posBit_next++; //aggiorno la posizione del prossimo bit da controllare
  }
  posBit_next++; //la posizione dopo il bit trovato
  nodo_next += posBit_next; //rifletto su <nodo_next> l'avanzamento di <posBit_next>: in pratica e' <nodo_next> += <posBit_next DOPO> - <posBit_next PRIMA>

  if(posBit_next == sizeof(TIPO_BASE)*8) //il prossimo bit e' nell'elemento successivo di <buffer>: faccio le dovute correzioni
  {
    posBuffer_next++; //siamo adesso nell'elemento subito successivo
    posBit_next= 0;
    //N.B. <nodo_next> invece va gia' bene cosi' e non viene modificato
  }

  //il bit trovato appartiene effettivamente all'array? (questo controllo sarebbe superfluo avendo inizializzato a zero i bit finali, inutilizzati, di <buffer>. Infatti se l'ultimo elemento di <buffer> fosse trovato non nullo i bit=1 non possono essere quelli inutilizzati)
  if(nodo_next <= numElementi) //allora "pos_trovato" sara' < "numElementi"
  {
    pos_trovato= nodo_next - 1; //salviamo il risultato: notare che <nodo_next> e' pari alla posizione DOPO l'ultimo bit=1 trovato
    return true; //N.B. <nodo_next>, <posBuffer_next>, <posBit_next> puntano alla prossima posizione da considerare, come deve essere
  }
  else
    return false; //il bit=1 non apparteneva all'array, in pratica era un bit delle posizioni finali inutilizzate dell'ultimo elemento di <buffer>
}


//restituisce in "pos_trovato" il primo bit settato a 0, cercando a partire dal bit alla posizione corrente (posizione 'nodo_next'). Ritorna true se e' stato trovato un tale bit, false altrimenti
//N.B. la ricerca avviene per posizioni crescenti del vettore di bits
bool VettoreBits::findNext0(unsigned long & pos_trovato)
{
  if(nodo_next >= numElementi) //siamo gia' oltre la fine dell'array di bits
    return false; //nulla di fatto
  //else.. cerchiamo il primo elemento di <buffer>, a partire dalla posizione corrente, che non e' tutto nullo:

  TIPO_BASE mask= (((TIPO_BASE) 1) << posBit_next) - 1; //ovvero ( 2^posBit_next - 1 ). N.B.: ((1 << posBit_next) - 1) e' un del tipo 0, 1, 11, 111, 1111, ... per posBit_next=0, 1, 2, 3, 4, ...

  //cerco nell'elemento corrente se c'e' almeno un 1 (a partire dalla pos. di partenza, i.e. dal bit corrente)
  if( (buffer[posBuffer_next] | mask) != ~((TIPO_BASE) 0)) //ignoro (sovrascrivo con 1) la parte precedente (gia' esaminata) dell'elemento corrente per fare il test. Vedo se sono o no tutti 1. Se c'e' qualche 0 chiamo la funzione ed esco
  {
    return scandisci_elemento_0(pos_trovato); //il residuo di questo elemento contiene degli uno: lo scandisco bit a bit
  }
  //else.. nei bit residui dell'elemento corrente di <buffer> non ci sono bit=1, inutile considerarlo --> passiamo agli elementi successivi di <buffer>


  nodo_next -= posBit_next; //mi riporto all'inizio dell'elemento corrente..
  posBit_next= 0; //..in modo che poi scandisci_elemento() inizi a cercare dal primo bit dell'elemento considerato.

  //passiamo a considerare gli elementi successivi di <buffer> fino a trovarne uno non nullo o fino ad arrivare alla fine di <buffer>
  do
  {
    posBuffer_next++; //l'elemento successivo nell'array
    nodo_next += sizeof(TIPO_BASE)*8; //avanziamo di un numero di posizioni pari ai bit contenuti in un elemento di <buffer> (in modo da trovarci sempre all'inizio, primo bit, dell'elemento corrente)

    if(posBuffer_next >= sizeBuffer) //siamo arrivati alla fine di <buffer> senza trovare alcun elemento che non fosse completamente nullo
      return false;

  } while( buffer[posBuffer_next] == ~((TIPO_BASE) 0) );

  //se non siamo usciti per la "return" nel ciclo prima allora abbiamo trovato un elemento di <buffer> nullo: lo scandiamo bit a bit ed a partire dal primo bit
  return scandisci_elemento_0(pos_trovato); //affidiamo la risposta a scandisci_elemento() (e anche la modifica del parametro passato)
}


/*
  Trova il primo bit =0 nell'elemento corrente di <buffer>, cercando a partire dal bit numero <posBit_next>. Appena lo trova salva il numero del nodo corrispondente nel parametro di ritorno.
  Ritorna true se il bit trovato appartiene effettivamente all'array, false se invece e' uno dei bit dell'ultimo elemento di <buffer> che non fa parte del vettore di bits, ma e' inutilizzato.
  Precondizioni:  1) esiste almeno un bit =0 nell'elemento corrente di <buffer> a partire dalla posizione <posBit_next>
*/
bool VettoreBits::scandisci_elemento_0(unsigned long & pos_trovato)
{
  nodo_next -= posBit_next; //sottraggo per riportarmi all'inizio dell'elemento corrente. Serve per calcolare poi <nodo_next> facilmente tramite <posBit_next>

  TIPO_BASE mask= ((TIPO_BASE) 1) << posBit_next; //il bit selezionato

  //N.B. so per ipotesi che trovero' almeno un bit=0 a partire dalla posizione posBit_next nell'elemento, il controllo ".. && (posBit_next < sizeof(TIPO_BASE)*8)" e' superfluo
  while( buffer[posBuffer_next] & mask )
  {
    mask <<= 1; //shift a SX di una posizione: <mask> seleziona il prossimo bit piu' pesante
    posBit_next++; //aggiorno la posizione del prossimo bit da controllare
  }
  posBit_next++; //la posizione dopo il bit trovato
  nodo_next += posBit_next; //rifletto su <nodo_next> l'avanzamento di <posBit_next>: in pratica e' <nodo_next> += <posBit_next DOPO> - <posBit_next PRIMA>

  if(posBit_next == sizeof(TIPO_BASE)*8) //il prossimo bit e' nell'elemento successivo di <buffer>: faccio le dovute correzioni
  {
    posBuffer_next++; //siamo adesso nell'elemento subito successivo
    posBit_next= 0;
    //N.B. <nodo_next> invece va gia' bene cosi' e non viene modificato
  }

  //il bit trovato appartiene effettivamente all'array? (questo controllo sarebbe superfluo avendo inizializzato a zero i bit finali, inutilizzati, di <buffer>. Infatti se l'ultimo elemento di <buffer> fosse trovato non nullo i bit=1 non possono essere quelli inutilizzati)
  if(nodo_next <= numElementi) //allora "pos_trovato" sara' < "numElementi"
  {
    pos_trovato= nodo_next - 1; //salviamo il risultato: notare che <nodo_next> e' pari alla posizione DOPO l'ultimo bit=1 trovato
    return true; //N.B. <nodo_next>, <posBuffer_next>, <posBit_next> puntano alla prossima posizione da considerare, come deve essere
  }
  else
    return false; //il bit=1 non apparteneva all'array, in pratica era un bit delle posizioni finali inutilizzate dell'ultimo elemento di <buffer>
}



//scriviamo il vettore di bits su un file. Il file conterr�il numero di byte minimo, in modo da non dipendere dalla scelta di TIPO_BASE
bool VettoreBits::salvaSuFile(const char * nomefile)
{
  FILE * file_output= fopen(nomefile, "w");
  if(file_output==NULL) //errore nell'apertura del file
    return false;

  unsigned long byte_da_scrivere= numElementi / 8;
  if(numElementi % 8)
    byte_da_scrivere++;

  unsigned long dati_scritti= fwrite( buffer, 1, byte_da_scrivere, file_output); //riempiamo il buffer di dati

  fclose(file_output);
  return (dati_scritti == byte_da_scrivere);
}



bool VettoreBits::caricaDaFile(const char * nomefile)
{
  struct stat info;
  unsigned long byte_da_leggere, dati_letti;

  byte_da_leggere= numElementi / 8;
  if(numElementi % 8)
    byte_da_leggere++;

  int err= stat(nomefile, &info);
  if(err==-1)
    return false;

  if((unsigned long)info.st_size != byte_da_leggere) //dimensione diversa dal previsto
    return false;

  FILE * file_input= fopen(nomefile, "r");
  if(file_input==NULL) //errore nell'apertura del file
    return false;

  dati_letti= fread( buffer, 1, byte_da_leggere, file_input);

  fclose(file_input);
  return (dati_letti == byte_da_leggere);
}



void VettoreBits::copia(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili per la copia
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i]= sorgente.buffer[i]; //NOT dei singoli elementi del buffer
}



void VettoreBits::NOT(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] = ~(sorgente.buffer[i]); // NOT dei singoli elementi dei buffer
}



void VettoreBits::OR(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] |= sorgente.buffer[i]; // OR dei singoli elementi dei buffer
}


void VettoreBits::XOR(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] ^= sorgente.buffer[i]; // XOR dei singoli elementi dei buffer
}



void VettoreBits::AND(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] &= sorgente.buffer[i]; // AND dei singoli elementi dei buffer
}


void VettoreBits::NOR(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] = ~(this->buffer[i] | sorgente.buffer[i]); // NOR dei singoli elementi dei buffer
}



void VettoreBits::NAND(VettoreBits & sorgente)
{
  if( (this->numElementi != sorgente.numElementi) || (this->sizeBuffer != sorgente.sizeBuffer) ) //non sono compatibili
  {
    fprintf(stderr, MSG_ERR_INCOMPATIBILI);  abort();
  }

  for(unsigned long i=0; i<sizeBuffer; i++)
    this->buffer[i] = ~(this->buffer[i] & sorgente.buffer[i]) ; // NAND dei singoli elementi dei buffer
}



//conta quanti bits a 1 ci sono nell'elemento passato
unsigned long VettoreBits::conta1_in_elemento(TIPO_BASE elemento)
{
  unsigned long risp= 0;
  const unsigned long num_bits_in_elemento= sizeof(TIPO_BASE)*8;

  if(elemento == 0)
    return 0;

  for(unsigned long j=0; j<num_bits_in_elemento; j++)
  {
    if(elemento & 1)
      risp++;

    elemento >>= 1; //shifto a DX (N.B. il parametro è passato per valore..)
  }

  return risp;
}



/*conta il numero di bits a 1 nel vettore.
  Viene scandito tutto il vettore, perciò quest'operazione è costosa e va evitata se possibile
*/
unsigned long VettoreBits::contaBits1()
{
  TIPO_BASE elemento, mask;
  const unsigned long num_bits_in_elemento= sizeof(TIPO_BASE)*8;
  unsigned long risp= 0;

  for(unsigned long i=0; i<sizeBuffer - 1; i++)
    risp += conta1_in_elemento( buffer[i] );

  //l'ultimo elemento va trattato a parte, perché i bit finali possono non essere significativi e vanno perciòignorati
  unsigned long bits_da_salvare= numElementi % num_bits_in_elemento;

  if(bits_da_salvare==0)
    elemento= buffer[sizeBuffer - 1]; //tutti i bit sono significativi
  else
  {
    mask= (((TIPO_BASE) 1) << bits_da_salvare) - 1;
    elemento= buffer[sizeBuffer - 1] & mask; //togliamo i bits che non appartengono al vettore
  }
  
  risp += conta1_in_elemento(elemento);

  return risp;
}


