/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file semaforo.cc :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE   -   RIVISTO 24 nov 2001
//=======================================================================
//=======================================================================

#include "funzioniGenerali.h"
#include "semaforo.h"
#include <unistd.h>

long semaforo::signal(){
  struct sembuf aux={0,1,0};
  if (semop(sem_id,&aux,1)==-1){
    fprintf(stderr,"Unexpected fatal error: signal() id = %ld",sem_id);
    printError("");
    FINE;
    return -1;
  }
  return 0;
}



long semaforo::wait(){
  struct sembuf aux={0,-1,0};
  if (semop(sem_id,&aux,1)==-1){
    fprintf(stderr,"Unexpected fatal error: function wait() - semaphore %ld - PID processo %ld: ",(long)sem_id,(long)getpid());
    printError("");
    FINE;
    return -1;
  }
  return 0;
}


long semaforo::getId(){
  return sem_id;
}



long semaforo::getValue(){
  long status;
  union semun aux;//variabile ausiliaria
  if ( (status=semctl(sem_id,0,GETVAL,aux))==-1){
    fprintf(stderr,"Unexpected fatal error: function getValue() - semaphore %ld ",sem_id);
    printError("");
    FINE;
    return -1;
  }
  return status;
}


void semaforo::cancella(){
  //distruttore - dealloco il semaforo
  //comunque uso la wrapper perche' tanto se il semaforo non c'e' nella lista non succede nulla

  if ( removeSemaforo(sem_id)==-1){
      fprintf(stderr,"Unexpected fatal error: function cancella() - semaphore  %ld",sem_id);
      printError("");
      FINE;
  }
}




semaforo::semaforo(long oldId,bool inutile){
//attach ad un semaforo esistente
    inutile=0;
    sem_id=oldId;
}


void semaforo::setValue(long init){
      //******************* INIZIALIZZAZIONE SEMAFORO ***************************
      union semun aux;//variabile ausiliaria
      aux.val=init;//campo usato per settare il valore iniziale, senza aspettare
      if( semctl(sem_id,0,SETVAL,aux)==-1){
        fprintf(stderr,"Unexpected fatal error: unable to init. semaphore %ld",sem_id);
        printError("");
        FINE;
        return;
      }
}



semaforo::semaforo(long init){
/*crea un nuovo semaforo con valore iniziale init*/
    sem_id=-1;

    if (init<0){//utilizzare la chiamata di sistema
        init=-init;
    }

    //******************* CREAZIONE SEMAFORO ***************************
    do {
         sem_id=getSemaforo( rand(),1 );//uso la wrapper function
    }
    while ((errno==EEXIST)&&(sem_id==-1));//mentre esiste
    
    if (sem_id==-1){
      printError("Unexpected fatal error: function semaforo(..) ");
      FINE;
      return;
    }

    accettaSemaforo(sem_id);

    //******************* INIZIALIZZAZIONE SEMAFORO ***************************
    union semun aux;//variabile ausiliaria
    aux.val=init;//campo usato per settare il valore iniziale, senza aspettare
    if( semctl(sem_id,0,SETVAL,aux)==-1){
      fprintf(stderr,"Unexpected fatal error: function semaforo(). Init semaphore %ld",sem_id);
      printError("");
      FINE;
      return;
    }
}


