/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file bufferedThreadFileInputStream.h :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE
//=======================================================================
//=======================================================================

//***********************************************
// COMPILA CON OPZIONE       -lpthread   di g++
//***********************************************


#ifndef _bufferedThreadFileInputStream_
#define _bufferedThreadFileInputStream_

#include "../semaforo.h"
#include "../multiSemaforo.h"
#include "../funzioniGenerali.h"
#include "multiFileReader.h"
#include <pthread.h>

/*NOTA GLI INTERI SONO di 4 byte, come i long!!!!!!!!!!*/

class bufferedThreadFileInputStream{
  /*bufferizzazione delle letture da file in cui il processo figlio legge da file
  e scrive in un insieme di buffer e mediante ead/readChar leggo dai buffer
  CON CONDIVISIONE
  */

  typedef struct bufAtt{
    long bufferLettura;
    long bufferScrittura;
  }bufferAttivo;


  private:
    //------------------------------------------------------------------------------
    long         numeroTotBuffer;  //n dei buffer presenti
    long         dimensioneBuffer; //dimensione in byte di ciascun buffer
    long         numeroFiles;      //N files da cui leggere l'input
    //------------------------------------------------------------------------------
    char**       bufferCondiviso;  //puntatore al buffer ( ciascun puntatore usa SHMEM )
    //------------------------------------------------------------------------------
    long*        size;             //n elementi presenti nel buffer (SHMEM)
    //------------------------------------------------------------------------------
    multiSemaforo*    lettura;     //semafori per la lettura dal buffer (LISTA SEMAFORI)
    multiSemaforo*    scrittura;   //(LISTA SEMAFORI)
    semaforo*         fine;        //semaforo che comunica al figlio di terminare (SINGOLO SEMAFORO)

    //------------------------------------------------------------------------------
    bufAtt nBufAttivo;            //punt al buffer attualmente in uso in lettura e scrittura

    //------------------------------------------------------------------------------

    char* bufferCondivisoCorrenteLettura; //puntatore al buffer condiviso correntemente in lettura (PADRE)- velocizza gli accessi

    long posCorrenteLettura;       //posizione corrente inizio lettura buffer corrente per il padre

    pthread_t tidFiglio;                //id del thread figlio

    int cancellaQuandoLetto;       //specifica se cancellare il file quando letto
    int cancellato;                //specifica se le strutture dati sono state gi cancellate
    int fineFile;                  //specifica che il file non puo' essere piu' letto perche' finito o interrotta la lettura
    char* prototipo;               //var nome prototipo multiFile

  private:
    friend void* processoFiglioInput(void* arg);
    long  minimo(long a,long b);
    void  init(char* prototipo,long numeroFiles,long dimBuffer,long numBuffer,int cancellaQuandoLetto);
    void* prova(void*){return NULL;}
  public:
    //nota che l'uso dei thread crea un altro processo di gestione dei thread e resta attivo fino alla fine del programma
    //il distruttore e il costrutt di copia on fanno nulla!!!!!!!!

    //costruttore che non fa nulla
    bufferedThreadFileInputStream(){cancellato=TRUE;}

    //lettura di un file (se il nome contiene 1 stringa del tipo %d, viene sostituita con 0)
    //(ATTENZIONE SE CI SONO + STRINGHE %d c'e' un errore)
   //lettura di una sequenza di files al variare di %d da 0 al numMax-1 (ATTENZIONE SE CI SONO + STRINGHE %d c'e' un errore)
   //se numeroFiles<0 --> il num viene calcolato in modo automatico !!!!
    bufferedThreadFileInputStream(char* prototipo,long numeroFiles,long dimBuffer,long numBuffer);
    bufferedThreadFileInputStream(char* prototipo,long numeroFiles,long dimBuffer,long numBuffer,bool cancQuandoLetto);
    
    //elimina TUTTE le strutture dati condivise e chiude il file
    //non permettendo altre letture - eventualmente prima esegue il close
    void  cancella();//come close

    //legge 8 bit e ritorna il n di bit letti (-1 se errore)
    long  readChar(char& buf);

    //legge dal multifile e ritorna il n di bit letti
    long  read(void* buf,long sizeElemento,long numElementi);

    //ritorna la qta' di memoria utilizzata
    long  getSize(){return dimensioneBuffer*numeroTotBuffer;}

    //chiude il file e non permette altre letture - SI ASSICURA CHE IL THREAD FIGLIO SIA FINITO ed elimina tutte strutture dati
    void  close();

    //ritorna il num di files
    long getNumFiles(){return numeroFiles;}
};


#endif
