/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file bufferedThreadFileOutputStream.cc :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE
//=======================================================================
//=======================================================================

//#include <iostream.h>
#include <iostream>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "bufferedThreadFileOutputStream.h"
#define MAX_BUFFER (0x4000000)
#define MAX_NUM_BUFFER (100)


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void* processoFiglioOutput(void* arg)
{
  //conversione dell'oggetto
  bufferedThreadFileOutputStream* oggetto=(bufferedThreadFileOutputStream*)arg;

  //apre il file per la scrittura
  multiFileWriter fp=multiFileWriter(oggetto->prototipo,oggetto->maxDimFiles);


#ifdef DEBUG_ERR
  fprintf(stderr,"processo figlio per SCRITTURA SU DISCO avviato...PID %d - prototipo %s\n",getpid(),prototipo);
#endif

  oggetto->nBufAttivo.bufferLettura=0;//primo buffer disponibile da cui leggere i dati da scrivere sul disco
  long n_scritti;
  do{
    //aspetto che il buffer sia libero per la scrittura
    oggetto->lettura->wait(oggetto->nBufAttivo.bufferLettura);

#ifdef DEBUG_SCRITTURA
    oggetto->bufferPieni->wait();
    printf("FIGLIO WRITE: scritto su disco buffer %d - buffer pieni da scrivere SU DISCO %d\n",nBufAttivo.bufferLettura,bufferPieni->getValue()-100);
    fflush(stdout);
#endif


    //scrivo sul file a blocchi
    n_scritti=fp.write(oggetto->bufferCondiviso[oggetto->nBufAttivo.bufferLettura],1,oggetto->size[oggetto->nBufAttivo.bufferLettura]);

    if (n_scritti==-1){
         fprintf(stderr,"Errore nella scrittura nell file");
         fprintf(stderr,oggetto->prototipo,fp.getNumeroFilesCreati());
         printError("");
         FINE;
    }

    //abilito la lettura sul buffer appena creato
    (oggetto->scrittura)->signal(oggetto->nBufAttivo.bufferLettura);

    //passo al buffer successivo
    if (++oggetto->nBufAttivo.bufferLettura>=oggetto->numeroTotBuffer)
      	oggetto->nBufAttivo.bufferLettura=0;
    	
  }//termino quando ho un blocco di dimensione <della massima
  while (n_scritti==oggetto->dimensioneBuffer);

  //chiudo quindi il file
  oggetto->numeroFilesCreati=fp.close();
  fp.cancella();
#ifdef DEBUG_ERR
  fprintf(stderr,"processo figlio SCRITTURA SU DISCO terminato...\n");
#endif
  //e termino il thread
  pthread_exit(0);
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
bufferedThreadFileOutputStream::bufferedThreadFileOutputStream(char* nomeFile,long dim,long nBuf){
//  strncpy(this->mode,"w",10);
  init(nomeFile,-1,dim,nBuf);
}




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
bufferedThreadFileOutputStream::bufferedThreadFileOutputStream(char* prototipo,long maxDimFiles,long dim,long nBuf){
//  strncpy(this->mode,"w",10);
  init(prototipo,maxDimFiles,dim,nBuf);
}




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void bufferedThreadFileOutputStream::init(char* prototipo,long maxDimFiles,long dim,long nBuf){

  if (dim<=0){
    fprintf(stderr,"Init error: dimension of the buffer too small (%ld<1)\n",dim);
    FINE;
    return;}

  if (dim>MAX_BUFFER){
    fprintf(stderr,"Init error: dimension of the buffer too high (%ld>%d)\n",dim,MAX_BUFFER);
    FINE;
    return;}

  if (nBuf<1){
    fprintf(stderr,"Init error:  buffers number too small (%ld<2)\n",nBuf);
    FINE;
    return;}

  if (nBuf>MAX_NUM_BUFFER){
    fprintf(stderr,"Init error:  buffers number too high (%ld>%ld)\n",nBuf,(long)MAX_NUM_BUFFER);
    FINE
    return;}

  //calcolo della dimensione come multiplo di un char (1 o 2 byte)
  long inc=0;
  if (nBuf%sizeof(char)!=0)
    inc=1;
  nBuf=(nBuf/sizeof(char)+inc )*sizeof(char);

  this->numeroTotBuffer=nBuf;//n buffer in uso
  this->dimensioneBuffer=dim;// dimensione dei buffer

  //variabili per la gestione multifiles
  check(this->prototipo=new char[strlen(prototipo)+1]);
  strcpy(this->prototipo,prototipo);
  this->maxDimFiles=maxDimFiles;

  check( this->lettura=new multiSemaforo (numeroTotBuffer,0) );//init dei semafori di lettura a 0
  check( this->scrittura=new multiSemaforo (numeroTotBuffer,1) );//init dei semafori di scrittura a 1
  size=new long [numeroTotBuffer];

  //creo i vari buffer...
  check(bufferCondiviso=new char* [numeroTotBuffer]); //array di puntatori ai buffer
  for (long i=0;i<numeroTotBuffer;i++){
       bufferCondiviso[i]=new char [dimensioneBuffer];
       size[i]=0;//inizializzazione della dimensione occupata
  }

  nBufAttivo.bufferLettura=0;//il padre inizia a scrivere dal 1 blocco
  nBufAttivo.bufferScrittura=0;//il figlio inizia a leggere dal 1 blocco che per ora e' bloccato
  bufferCondivisoCorrente=bufferCondiviso[nBufAttivo.bufferScrittura];//buffer corrente

  //il padre impegna il primo buffer per la lettura
  scrittura->wait(nBufAttivo.bufferLettura);//utilizzo il primo buffer per scrivere
  posCorrenteScrittura=0;//il buffer e' vuoto in scrittura




  fileChiuso=FALSE;

  #ifdef DEBUG_SCRITTURA
  check(bufferPieni=new semaforo(101));
  #endif



  tidFiglio=0;

  pthread_attr_t attr;                       //attributi del nuovo thread
  struct sched_param shedParam;              //param di scheduling
  shedParam.sched_priority=BACKGROUND_PRIOR; //fisso la priorit
  pthread_attr_init(&attr);                     //set init attrib
  pthread_attr_setschedparam(&attr,&shedParam); //set priorita'

  //avvia il nuovo thread figlio e ritorna
  if ( pthread_create(  &tidFiglio,&attr,processoFiglioOutput,this ) !=-0){
      fprintf(stderr,"Thread creation error: child thread\n");
      printError("");
      FINE
  }

}




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long bufferedThreadFileOutputStream::minimo(long a,long b){
  if (a<b) return a;
  return b;
}






/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long bufferedThreadFileOutputStream::close(){
    return cancella();
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/

long bufferedThreadFileOutputStream::write(void* buf,long sizeElemento,long numElementi){
  if (fileChiuso)
     return 0;//il file e' finito

  //n tot byte da LEGGERE
  long  numeroTotByte=sizeElemento*numElementi;
  long  numeroByteDaScrivere=numeroTotByte;
  long  posizioneBuffer=0;//si sposta sempre di un byte
  char* b=(char*)buf;//punto il buffer carattere per carattere
  long byteDaCopiare;//n byte da copiare ad ogni ciclo

  //controllo se ci sono ancora elementi nel buffer
  while (numeroByteDaScrivere>0){
    byteDaCopiare=minimo(dimensioneBuffer-posCorrenteScrittura,numeroByteDaScrivere);
    memcpy(bufferCondivisoCorrente+posCorrenteScrittura,b+posizioneBuffer,byteDaCopiare);
    posCorrenteScrittura+=byteDaCopiare;
    numeroByteDaScrivere-=byteDaCopiare;
    posizioneBuffer+=byteDaCopiare;

    if (numeroByteDaScrivere>0){
        //comunico la taglia del blocco
        size[nBufAttivo.bufferScrittura] = posCorrenteScrittura;

#ifdef DEBUG_SCRITTURA
        bufferPieni->signal();
        printf("PADRE WRITE: riempito buffer %d - buffer pieni da scrivere SU DISCO %d\n",nBufAttivo.bufferScrittura,bufferPieni->getValue()-100);
        fflush(stdout);
#endif

        //sblocco il buffer terminato per la scrittura
        lettura->signal(nBufAttivo.bufferScrittura);

        //incremento n buffer e passo al buffer successivo
        if(++nBufAttivo.bufferScrittura>=numeroTotBuffer)
             nBufAttivo.bufferScrittura=0;

        //aspetto che sia pronto per la lettura
        scrittura->wait(nBufAttivo.bufferScrittura);
        bufferCondivisoCorrente=bufferCondiviso[nBufAttivo.bufferScrittura];
        posCorrenteScrittura=0;
    }
  }//fine while
  return numElementi;
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
long bufferedThreadFileOutputStream::writeChar(char& buf){
  if (fileChiuso)
     return 0;//il file e' finito

  //controllo se c'e' ancora spazio nel buffer
  if (posCorrenteScrittura<dimensioneBuffer){
     //scrivo l'elemento
     bufferCondivisoCorrente[posCorrenteScrittura++]=buf;
     return 1;
  }
  else{

     //comunico la taglia del blocco
     size[nBufAttivo.bufferScrittura] = posCorrenteScrittura;

#ifdef DEBUG_SCRITTURA
     bufferPieni->signal();
     printf("PADRE WRITE: riempito buffer %d - buffer pieni da scrivere su disco %d\n",nBufAttivo.bufferScrittura,bufferPieni->getValue()-100);
     fflush(stdout);
#endif
     //sblocco il buffer terminato di scrivere per essere letto
     lettura->signal(nBufAttivo.bufferScrittura);

     //incremento n buffer e passo al buffer successivo
     if(++nBufAttivo.bufferScrittura>=numeroTotBuffer)
         nBufAttivo.bufferScrittura=0;
     bufferCondivisoCorrente=bufferCondiviso[nBufAttivo.bufferScrittura];

     //aspetto che sia pronto per la scrittura
     scrittura->wait(nBufAttivo.bufferScrittura);
     posCorrenteScrittura=0;

     bufferCondivisoCorrente[posCorrenteScrittura++]=buf;
     return 1;
  }
  return 1;
}


/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/


long bufferedThreadFileOutputStream::cancella(){
   if(fileChiuso) return numeroFilesCreati;

   //flag per controllare se e' necessario aggiungere un buffer vuoto
   int aggiungereVuoto=FALSE;

   //comunico la taglia dell'ultimo blocco da inviare pari a quella corrente
   size[nBufAttivo.bufferScrittura]=posCorrenteScrittura;

   //vede se il buffer e' pieno e occorre aggiungerne un'altro vuoto per la chiusura
   aggiungereVuoto=(posCorrenteScrittura==dimensioneBuffer);

   //sblocco il buffer terminato di scrivere per essere letto dal processo figlio
   lettura->signal(nBufAttivo.bufferScrittura);

   //incremento n buffer e passo al buffer successivo
   if(++nBufAttivo.bufferScrittura>=numeroTotBuffer)
         nBufAttivo.bufferScrittura=0;

   if (aggiungereVuoto){
      //aspetto che si liberi un altro buffer
      scrittura->wait(nBufAttivo.bufferScrittura);

      //fisso la taglia al valore 0<dimMax per indicare la fine
      size[nBufAttivo.bufferScrittura]=0;

      //il buffer vuoto per la segnalazione di close e' pronto per essere letto
      lettura->signal(nBufAttivo.bufferScrittura);
   }




   //attende che il processo figlio termini
   if (pthread_join(tidFiglio,NULL)!=0){
      fprintf(stderr,"Wait System call error: %ld\n",tidFiglio);
      printError("");
      FINE;
   }


   //cancello i semafori
   lettura->cancella();
   scrittura->cancella();

   #ifdef DEBUG_SCRITTURA
   bufferPieni->cancella();
   delete[] bufferPieni;bufferPieni=NULL;
   #endif

   //- ---------------- LIBERA MEMORIA ---------------- -
   delete[] prototipo;prototipo=NULL;
   delete lettura  ;lettura=NULL;//elimina i semafori di lettura
   delete scrittura;scrittura=NULL;//elimina i semafori di scrittura
   delete[] size;size=NULL;//elimina il buffer condiviso size

  //rilascia i buffer condivisi
  for (long i=0;i<numeroTotBuffer;i++){
     delete[] bufferCondiviso[i];bufferCondiviso[i]=NULL;
  }

  delete[] bufferCondiviso;bufferCondiviso=NULL;//elimino l'array di puntatori ai buffers

  fileChiuso=TRUE;//il file non e' piu' accessibile
  return numeroFilesCreati;
}
