/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file bufferedThreadFileOutputStream.h :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
//=======================================================================
//=======================================================================
//                              FILE COMUNE
//=======================================================================
//=======================================================================

//***********************************************
// COMPILA CON OPZIONE       -lpthread   di g++
//***********************************************



#ifndef _bufferedThreadFileOutputStream_
#define _bufferedThreadFileOutputStream_
#include <pthread.h>
#include "../semaforo.h"
#include "../multiSemaforo.h"
#include "../funzioniGenerali.h"
#include "multiFileWriter.h"



/*NOTA GLI INTERI SONO di 4 byte, come i long!!!!!!!!!!*/

class bufferedThreadFileOutputStream{
  /*bufferizzazione delle scritture su file*/
  //con CONDIVISIONE

  typedef struct bufAtt{
       long bufferLettura;
       long bufferScrittura;
  } bufferAttivo;

  private:
    //------------------------------------------------------------------------------
    long         numeroTotBuffer;//n dei buffer attivi
    long         dimensioneBuffer;//dimensione totale in byte di ciscun buffer
    long         maxDimFiles;//massima dimensione dei files di output
    //------------------------------------------------------------------------------
    char**       bufferCondiviso; //puntatore al buffer (array di puntatori)
    //------------------------------------------------------------------------------
    long*        size;      //n elementi presenti nei vari buffers espresso in byte
    //------------------------------------------------------------------------------
    multiSemaforo*    lettura;  //array dei semafori per la lettura dal buffer (LISTA SEMAFORI)
    multiSemaforo*    scrittura;//array dei semafori per la scrittura dal buffer (LISTA SEMAFORI)
    //------------------------------------------------------------------------------
    bufferAttivo nBufAttivo;//buffer attualmente in uso da ambo le parti (saranno 2 copie)
    //------------------------------------------------------------------------------
    pthread_t tidFiglio;//pid del processo Figlio

    //------------------------------------------------------------------------------
    long posCorrenteScrittura;   //posizione corrente scrittura nel buffer corrente per il processo padre
    int  fileChiuso;              //e' gia' stata liberata la memoria e il file di out e' stato chiuso

    //------------------------------------------------------------------------------
    long  numeroFilesCreati;     //buffer condiviso per il risultato dei files creati
    //------------------------------------------------------------------------------
    char* prototipo;//prototipo nomi file sdi scrittura
    char* bufferCondivisoCorrente;//puntatore al buffer corrente usato per maggior velocita'
    //------------------------------------------------------------------------------
    //char mode[10];

    void init(char* prototipo,long maxDimFiles,long dim,long nBuf);
  private:
    friend void* processoFiglioOutput(void* arg);
    long minimo(long a,long b);
  public:
    // gli attach delle strutture condivise non vengono ereditati dai figli (vengono in realta'
    // cancellati all'inizio in modo implicito da inizioNuovoProcesso() ... )

    /*la classe non puo' essere condivisa da + processi perche' alcune strutture dati non sono su shmem
      ma su memoria utente. Non e' assicurato che le funzioni siano atomiche....
      Per eventuali utilizzi con thread occorre occuparsi di tutte le problematiche di sincronizzazione.
      Per le operazioni di scrittura non ci sono problemi in quanto le var che si usano sono tutte shmem;
      l'unico pb e' nella chiusura del file che deve essere eseguita una sola volta da un sono processo perche'
      il controllo della chiusura si basa su una variabile locale*/

      /*I campi che vengono modificati sono:
        1) i buffer e i semafori di lettura/scrittura che sono condivisi;
        2) il flag di cancellato che pero' e' locale ai vari processi
        3) il flag di chiusura che e' locale ai vari processi
        occorre eventualmente tenere conto di cio'...*/


    //il distruttore e il costrutt di copia on fanno nulla!!!!!!!!
    bufferedThreadFileOutputStream(){fileChiuso=TRUE;}

    //scrittura di un multifile al raggiungimento del limite massimo (se il nome contiene 1 stringa del tipo %d, viene sostituita con 0)
    //(ATTENZIONE SE CI SONO + STRINGHE %d c'e' un errore)
    bufferedThreadFileOutputStream(char* prototipo,long maxDimFiles,long dim,long nBuf);

    //scrittura di un file (se il nome contiene 1 stringa del tipo %d, viene sostituita con 0)
    //(ATTENZIONE SE CI SONO + STRINGHE %d c'e' un errore)
    bufferedThreadFileOutputStream(char* nomeFile,long dim,long nBuf);

     //elimina TUTTE le strutture dati condivise e chiude il file
    //non permettendo altre letture
    long cancella();//elimina le strutture dati ed eventualmente chiude il file in scrittura

    //scrive 8 bit e ritorna il n di bit letti (-1 se errore)
    long writeChar(char& buf);

    //scrive dal multifile e ritorna il n di bit letti
    long write(void* buf,long sizeElemento,long numElementi);

    //ritorna la qta' di memoria utilizzata
    long getSize(){return dimensioneBuffer*numeroTotBuffer;}

    //chiude il file e non permette altre letture
    long  close();//come cancella
};


#endif
