/*

*********************************************************************
*********************************************************************

  file multFile.cpp : common functions and classes of all graph utilities

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/

#include "multiFile.h"
#include "../funzioniGenerali.h"
#include <stdio.h>


#define PRINT_DEBUG



/*=========================================================================================*/
/*=========================================================================================*/
/*      PRIVATE     PRIVATE     PRIVATE     PRIVATE     PRIVATE     PRIVATE     PRIVATE    */
/*=========================================================================================*/
/*=========================================================================================*/


/*=========================================================================================*/
bool multiFile::estOpenFile(){
/*=========================================================================================*/
  return fp!=NULL;
}


/*=========================================================================================*/
void multiFile::initFile(){
/*=========================================================================================*/
  numFiles=contaNumFilesMultiFile(prototipo);
  fp=NULL;
  numFileCorrente=-1;//nessun file aperto
  openFile(0,false);
}


/*=========================================================================================*/
long multiFile::sizeFile(long num){
/*=========================================================================================*/
  flushFile();//nel caso sia quello corrente
  char app[300];sprintf(app,prototipo,num);
  return getFileSize(app);  
}


/*=========================================================================================*/
void multiFile::closeFile(){
/*=========================================================================================*/
    if (fp!=NULL) fclose(fp);
    fp=NULL;numFileCorrente=-1;
}


/*=========================================================================================*/
long multiFile::readFile (void* buffer, long elemSize,long numElem){
/*=========================================================================================*/
  if (fp==NULL) return -1;
  long ris=fread(buffer,elemSize,numElem,fp);
  return ris;
}


/*=========================================================================================*/
long multiFile::writeFile (void* buffer, long elemSize,long numElem){
/*=========================================================================================*/
  if (fp==NULL) {return -1;}
  long ris=fwrite(buffer,elemSize,numElem,fp);  
  return ris;
}


/*=========================================================================================*/
long multiFile::seekFile(long pos){
/*=========================================================================================*/
  if (fp==NULL) {return -1;}
  long ris=fseek(fp,pos,SEEK_SET);  
  return ris;
}


/*=========================================================================================*/
void multiFile::flushFile(){
/*=========================================================================================*/
    if (fp==NULL)  return ;
    fflush(fp);
}


/*=========================================================================================*/
long multiFile::tellFile(){
/*=========================================================================================*/
  if (fp==NULL) {return -1;}
  return ftell(fp);
}


/*=========================================================================================*/
void multiFile::openFile(int num,bool readOnly){
/*=========================================================================================*/
  //apre il file se non e' gia' aperto, altrimenti lo lascia come gia si trova...
  if (numFileCorrente==num) return ;
  
  //chiude precedente se esiste
  closeFile();
  numFileCorrente=num;
  if (num<0) return;

  //controlla se il file gia' esiste e se no lo crea nuovo
  char app[300];sprintf(app,prototipo,num);
  fp=fopen(app,"r+");

  //se non ne puo' creare di nuovi si ferma
  if (readOnly) return;

  if (fp==NULL){
    fp=fopen(app,"w+");//crea un nuovo file
    numFiles++;
  }
  
}






/*=========================================================================================*/
/*=========================================================================================*/
/*      PUBLIC     PUBLIC     PUBLIC     PUBLIC     PUBLIC     PUBLIC     PUBLIC           */
/*=========================================================================================*/
/*=========================================================================================*/


/*=========================================================================================*/
void multiFile::removeMultiFile(char* prototipo){
/*=========================================================================================*/
  long nFiles=contaNumFilesMultiFile(prototipo);
  for (long i=0;i<nFiles;i++){
    char app[300];
    sprintf(app,prototipo,i);
    remove(app);    
  }  
}


/*=========================================================================================*/
multiFile::multiFile(char* prototipo, long maxDim){
/*=========================================================================================*/
  //elimina vecchi files
  removeMultiFile(prototipo);

  //copia prototipo nome
  strcpy(this->prototipo,prototipo);

  //fissa dim massima
  this->maxDim=maxDim;

  //inizializza files
  initFile();
}


/*=========================================================================================*/
long multiFile::read(void* buffer,long sizeElemento,long numeroElementi){
/*=========================================================================================*/
  if (!estOpenFile()) return -1;

  //NOTA: char e' sempre di un byte
  char* b=(char*)buffer;//puntatore interno al buffer (array di 1 byte)

  long numByteDaLeggere=sizeElemento*numeroElementi;//n byte da leggere in totale
  long letti;//byte letti ad ogni fread (usata nel ciclo)

  while ( (numByteDaLeggere>0) ){

     //letti=fread(b,1,numByteDaLeggere,fp);//legge quanti piu' dati possibile
     letti=readFile(b,1,numByteDaLeggere);//legge quanti piu' dati possibile
     numByteDaLeggere=numByteDaLeggere-letti;

     //se mancano alcuni dati allora e' finito il file e se posso passo al successivo
     if (numByteDaLeggere>0){
         //sposto il puntatore
         b=b+letti;

         //apro il file successivo se esiste, altrimenti il multiFile risulta terminato ma non lo chiudo
         if (numFileCorrente+1<numFiles)
           openFile(numFileCorrente+1,true);
         else
           break;//fine lettura x mancanza di dati !!! - ho raggiunto la fine dell'ultimo file
     }
  }
//  printf("Dato: ");for (int i=0;i<3;i++){printf("%08x",b[i*4]);};printf(" ---- "); for (int i=0;i<3;i++){printf("%ld;",b[i*4]);};printf("\n");
  return (sizeElemento*numeroElementi-numByteDaLeggere)/sizeElemento;
}



/*=========================================================================================*/
long multiFile::write(void* buffer,long sizeElemento,long numeroElementi){
/*=========================================================================================*/
  if (!estOpenFile()) return -1;

  //NOTA: char e' sempre di un byte
  char* b=(char*)buffer;//puntatore interno al buffer (array di 1 byte)

  long numByteDaScrivere=sizeElemento*numeroElementi;//n byte da leggere in totale
  //cout<<"Da scrivere = "<<numByteDaScrivere;
  long scritti;//byte letti ad ogni fread (usata nel ciclo)

  while ( (numByteDaScrivere>0) ){

     //calcola bytes disponibili file corrente x la scrittura
     long byteDisponibiliFileCorrente;
     if (maxDim<=0) byteDisponibiliFileCorrente=numByteDaScrivere;
     else byteDisponibiliFileCorrente=maxDim-tellFile()/*ftell(fp)*/;
     
     long daScrivereBloccoCorrente=min(byteDisponibiliFileCorrente,numByteDaScrivere);

     if (daScrivereBloccoCorrente>0){
       //scrive i dati nel file corrente
       //scritti=fwrite(b,1,daScrivereBloccoCorrente,fp);
       scritti=writeFile(b,1,daScrivereBloccoCorrente);
       if (scritti==0) break;//errore nella scrittura ritorna i bytes scritti
     }
     else
       scritti=0;
  //   cout<<"scritti = "<<scritti;
       
     numByteDaScrivere=numByteDaScrivere-scritti;

     //se mancano alcuni dati allora e' finito il file e se posso passo al successivo
     if (numByteDaScrivere>0){
         //sposto il puntatore
         b=b+scritti;

         //apro il file successivo e se non esiste ne creo uno nuovo !!!
         openFile(numFileCorrente+1,false);
     }
  }

  return (sizeElemento*numeroElementi-numByteDaScrivere)/sizeElemento;
}


/*=========================================================================================*/
long long multiFile::tell(){
/*=========================================================================================*/
  if (!estOpenFile()) return -1;
  long long dimPrec=0;

  //tutti i prec files, escluso quello corrente devono essere dimMax
  dimPrec=(long long)(numFileCorrente) * (long long)maxDim;

  //calcola pos file corrente
  long current=tellFile();//ftell(fp);
  if (current<0) return -1;//errore sulla tell del file corrente

  return dimPrec+(long long)current;
}


/*=========================================================================================*/
long long multiFile::getSize(){
/*=========================================================================================*/
  if (!estOpenFile()) return -1;

  long long totalSize;

  //se dimMax<0 le cose vanno ancora bene... x che' numFileCorrente sara' sempre 0 e il termine sara' nullo
  totalSize=maxDim*(numFiles-1)+sizeFile(numFiles-1);//size files interni + quella dell'utlimo file
  
  return totalSize;
}


/*=========================================================================================*/
void multiFile::flush(){
/*=========================================================================================*/
  flushFile();
}


/*=========================================================================================*/
void multiFile::close(){
/*=========================================================================================*/
  closeFile();
}


/*=========================================================================================*/
long multiFile::seek(long long pos){
/*=========================================================================================*/
  if (!estOpenFile()) {
    #ifdef PRINT_DEBUG
    printf("FILE CHIUSO 1 !!!");
    #endif
    return -1;}

  //sono sicuro di avere sufficienti bytes per la seek
  long numFileSeek;
  long posRelativa;

  if (maxDim>0){
    numFileSeek=(long) (pos/(long long)maxDim);
    posRelativa=pos%maxDim;}
  else {
    numFileSeek=0;
    posRelativa=(long)pos;
  }

  #ifdef PRINT_DEBUG
  if (posRelativa<0 || numFileSeek<0 ){
    char app[1000];
    sprintf(app,"\n MultiFile.cpp -- Seek program error: multifile %s - pos %lld - numFileSeek %ld - maxDim %ld \n",prototipo,pos,numFileSeek,maxDim);
    printf("Errore di approssimazione posRelativa=%ld , numFileSeek=%ld \n%s\n", posRelativa, numFileSeek,app);
    fflush(stdout);
    FINE;    
  }
  #endif

  //CONTROLLARE se NUM FILES sufficienti
  if (numFileSeek>=numFiles) return -1;//posizione inesistente

  //apro il file relativo    
  openFile(numFileSeek,true);    
  if (!estOpenFile()) {
    #ifdef PRINT_DEBUG
    char app[1000];
    printf("CIAOOOOOOO 2");
    fflush(stdout);
    sprintf(app,"\n MultiFile.cpp -- Seek program error: multifile %s - pos %lld - numFileSeek %ld - maxDim %ld \n",prototipo,pos,numFileSeek,maxDim);
    perror(app);
    #endif
    FINE;
  }//non dovrebbe accadere

  //si posiziona correttamente nel file  
  if (seekFile(posRelativa)<0){return -1;}//controlla se possibile posizionarsi all'interno del file corrente...

  return 1;//OK
}

