/*

*********************************************************************
*********************************************************************

  common functions and classes of all graph utilities

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/
#include "timer.h"

/*permette di calcolare il massimo numero rappresentabile dal numero (INTERO) passato*/
template <class T>
T maxIntValue(){

  //------------------------------
  //controllo se  un tipo unsigned
  //------------------------------
  T temp=0;
  if (~(temp=0)>=0) return ~temp;
  //------------------------------

  //------------------------------
  // e' un tipo signed
  //------------------------------
  temp=1;
  return ~(temp<<(sizeof(T)*8 - 1));
  //------------------------------

}


//calcola la differenza tra due istanti clock...
clock_t timer::calcolaDifferenza(clock_t end, clock_t start){
  clock_t diff=end-start;
  return diff<0 ? maxIntValue<clock_t>()+diff : diff;
}



timer::timer(double waitingSec=1.0){

  //calcola click da aspettare
  waitingSec=waitingSec<=0? 1: waitingSec;
  clockAttesa=(clock_t)(waitingSec*CLOCKS_PER_SEC);//num di clock da attendere prima della stampa
  if (clockAttesa<0 ) clockAttesa=CLOCKS_PER_SEC;//nel caso di overflow assegno un secondo

  reset();
}


void timer::reset(){
  //calcola istante corrente
  clockLastPrint = clock();

  //azzera contatori
  countFromLastPrint=0;
  countDown=0;
  isFirstPrint=true;
}


void timer::print(const char* format, ...){

  #ifdef ENABLE_OPTIMIZATION
  //controlla countDown prima di controllare il clock
  countFromLastPrint++;
  if (--this->countDown>0) return;
  #endif

  //calcola tempo trascorso dall'ultima chiamata
  clock_t clockTrascorsi=calcolaDifferenza(clock(), clockLastPrint);

  if ( clockAttesa > clockTrascorsi && !isFirstPrint) {


    #ifdef ENABLE_OPTIMIZATION
    /*  countFromLastPrint : clockTrascorsi == X : (clockAttesa-clockTrascorsi)  */
    if (clockTrascorsi<=0) return;
    countDown=(clock_t)(0.5 * countFromLastPrint * (clockAttesa - clockTrascorsi) / clockTrascorsi);//num di chiamate mancanti previste
/*    printf("countDown=%lld -- clock rimasti =%ld -- clockTrascorsi = %ld -- clockAttesa = %ld -- countFromLastPrint = %lld\n",
          countDown, (clockAttesa - clockTrascorsi), clockTrascorsi , clockAttesa , countFromLastPrint);
          */
    #endif
    return;
  }

  //num chiamate dall'ultima stampa
  countFromLastPrint=0;
  countDown=0;
  isFirstPrint=false;

  //stampa il formato
  va_list ap;
  va_start(ap, format);
  vprintf(format,ap);
  va_end(ap);
  fflush(stdout);

  //calcola istante corrente
  clockLastPrint = clock();
}




void timer::forcePrint(const char* format, ...){

  //stampa il formato
  va_list ap;
  va_start(ap, format);
  vprintf(format,ap);
  va_end(ap);
  fflush(stdout);
}
