/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file vettorefile.h :common functions and classes of all graph utilities
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   

#ifndef VETTOREFILE_H
#define VETTOREFILE_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/**Classe che implementa un insieme di vettori (NON inizializzati)
di dimensione indefinita (eventualmente vengono salvati su disco)

  *@author 
  */
#define NN (20)



template <class T>
class vettoreFile {
  public:
	  vettoreFile(long dimInMemoria,char* fileTempo);
	  void cancella();
	  long getElement(T& elem, long i);          //preleva un elemento
	  long putElement(T e, long i);  //scrive un elemento
	  long getSize(){return sizeInMemoria+sizeDisco;}
  private:	
    char fileTempo[1000];
  	T* buf[NN];
  	long sizeInMemoria;
	  long sizeDisco;
	  FILE* fp;
	  long sizePerBuffer;
};





template <class T>
vettoreFile<T>::vettoreFile(long dimMemoria, char* fileTempo){

  if(dimMemoria<0)dimMemoria=10;

  //calcolo dim effettiva memoria occupata
  sizePerBuffer=dimMemoria/NN;
  if (sizePerBuffer<=0)
      sizePerBuffer=1;
  dimMemoria=sizePerBuffer*NN;

  //creazione e allocazione buffers
  sizeInMemoria=dimMemoria;
  sizeDisco=0;
  for (int i=0;i<NN;i++)
  		buf[i]=new T[sizePerBuffer]();

  //apertura file d'appoggio
  strcpy(this->fileTempo,fileTempo);
  fp=fopen(fileTempo,"w+");
  if (fp==NULL) cancella();
}



template <class T>
long vettoreFile<T>::putElement(T e, long i){

  if (i<0) return -1;

  //se e' gia' disponibile in memoria
  if (i<sizeInMemoria){//da 0 a sizeInMemoria-1
  		 if (i/sizePerBuffer>=NN){std::cerr<<"Errore nel programma. Superato limite array"; FINE }
  		 (buf[i/sizePerBuffer])[i%sizePerBuffer]=e;
  }		

  //se e' gia' presente nel file
  else if (  i<(sizeInMemoria+sizeDisco)  ){//da   [sizeInMemoria , sizeInmemoria+sizeDisco-1]
          fseek(fp,sizeof(T)*(i-sizeInMemoria),SEEK_SET);
          if (fwrite(&e,sizeof(T),1,fp)<1) return -1;
       }

       //se non e' ancora presente nel file
       else{
          T aux=T();

					//si posiziona alla fine del file
          fseek(fp,0,SEEK_END);

          //aggiungo gli elementi mancanti per poi aggiungere l'ultimo...in totale devo avere i+1 elementi totali
          while ( (sizeInMemoria+sizeDisco)<i){
             if (fwrite(&aux,sizeof(T),1,fp)<1) return -1;
             sizeDisco++;
          }
          //qui (sizeInMemoria+sizeDisco)==i e allora posso aggiungere l'ultimo elemento

          //aggiungo l'elemento da insere
          if (fwrite(&e,sizeof(T),1,fp)<1) return -1;
          sizeDisco++;

          //controllo correttezza
          if (i!=sizeInMemoria+sizeDisco-1){
              std::cerr<<"Errore nel programma: gestione putElement non coretta";
              FINE;
          }

       }
  return 0;
}




template <class T>
long vettoreFile<T>::getElement(T& e, long i){

  if (i<0) return -1;

  //se e' gia' disponibile in memoria
  if (i<sizeInMemoria){
  		 if (i/sizePerBuffer>=NN){std::cerr<<"Errore nel programma. Superato limite array"; FINE }
  		 e=(buf[i/sizePerBuffer])[i%sizePerBuffer];
  }

  //se e' gia' presente nel file
  else if (i<sizeInMemoria+sizeDisco){
          fseek(fp,sizeof(T)*(i-sizeInMemoria),SEEK_SET);
          if (fread(&e,sizeof(T),1,fp)<1) return -1;
       }

       //se non e' presente nel file
       //se non e' ancora presente nel file
       else{
          T aux=T();

					//si posiziona alla fine del file
          fseek(fp,0,SEEK_END);

          //aggiungo gli elementi mancanti per poi aggiungere l'ultimo...in totale devo avere i+1 elementi totali
          while ( (sizeInMemoria+sizeDisco)<=i){
             if (fwrite(&aux,sizeof(T),1,fp)<1) return -1;
             sizeDisco++;
          }

          //qui (sizeInMemoria+sizeDisco)==i e allora posso aggiungere l'ultimo elemento
           e=aux;

          //controllo correttezza
          if (i!=sizeInMemoria+sizeDisco-1){
              std::cerr<<"Errore nel programma: gestione putElement non coretta";
              FINE;
          }

       }

  return 0;
}




template <class T>
void vettoreFile<T>::cancella(){
  if (buf[0]==NULL) return;

  for (int i=0;i<NN;i++){delete[] buf[i];buf[i]=NULL;}

  if (fp!=NULL) fclose(fp);
  fp=NULL;
  sizeInMemoria=sizeDisco=0;
  remove(fileTempo);
}



#endif
