/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file copymodel.cc :source code of copymodel utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "comuni.h"
#include <iostream>
#include "config.h"


//genera un numero compreso tra inf e sup (estremi compresi)
long generaNumeroCasuale(long inf,long sup){
   if (inf>sup) return -1;
   long intervallo = (sup-inf);

   //valore tra 1 e 10
   //long lancio=1+ (long)( (1.0*rand())/(RAND_MAX) *(10));
   //lancio=lancio%11; //da 0 a 10 e' buona

   //valore tra 0 e intervallo
   long lancio = (long)( ( 1.0*rand()/RAND_MAX  *(intervallo+1) ) )%(intervallo +1);
   return lancio+inf ;
}



void stampa(char* msg,char* z,long l){
    cout<<msg<<": ";
    for (int i=0;i<l;i++)
        cout<<"  "<<(int)z[i];
     cout<<"\n";
}



long normalizzaLancio(unsigned long lancio, long inferiore, long superiore){
    long diff=superiore-inferiore+1;
    if (diff<0) return -1;
    if (diff==0) return inferiore;
    return inferiore+lancio%diff;
}



class nodoRitornato{
  public:
    long nodoDest;
    char tipo;//spec se ho subito il nodo o un riferimento al successore di un altro nodo

    nodoRitornato(long nodo, char t){
        nodoDest=nodo;
        tipo=t;
    }

    nodoRitornato(){;}

    bool estNodo() {return tipo==1; }
    bool estNullo(){return tipo==0; }
    bool estSuccessoreDi(){return tipo==2; }
};








//funzione che prende il num di un nodo, il num del successore e mi ritorna il valore del nodo puntato o il riferimento
//il num di successore va da 0 al N_MAX -1
nodoRitornato calcolaLinkUscente(long nodoSorgente, long numSuccessore, randomGen2& r, long numeroSuccTotali,long probabilitaNuovoNodo_1_100){

  #define N_TOT_LANCI_PER_NODO ((long long) 1+2*numeroSuccTotali)

  /*  1 lancio per la scelta del nodo di riferimento (1 solo per tutti i successori)
      per ogni successore:
          1 lancio per scegliere se nodo casuale o copia e se nodo casuale 1 altro lancio...
  */


  //il nodo 0 non ha successori; il nodo 1 ne ha 1 e cosi' via
  if (numSuccessore>=nodoSorgente)
        return nodoRitornato(-1,0);//nessun nodo

  //num lanci associati ad ogni nodo = (1+2*numeroSuccTotali)

  //scelta del tipo di operazione (nuovo nodo OPPURE copia da esistente)
  float lancioSceltaTipoNodo      = 100.0*(float)r.genRand( (long long)1+(long long)nodoSorgente*N_TOT_LANCI_PER_NODO + numSuccessore*2 )/RAND_MAX;
  //cout<<"lancioSceltaTipoNodo "<<lancioSceltaTipoNodo<<" - n lancio "<<1+nodoSorgente*N_TOT_LANCI_PER_NODO + numSuccessore*2<<"\n";

  //tra 1 e  probabilitaNuovoNodo_1_100 --> nuovo nodo, altrimenti copia
  if ((lancioSceltaTipoNodo<probabilitaNuovoNodo_1_100)||(probabilitaNuovoNodo_1_100==100.0) ){
    //scelta di un nuovo nodo casuale
    long lancioSceltaNodoEffettivo =      normalizzaLancio (r.genRand( (long long)1+(long long)nodoSorgente*N_TOT_LANCI_PER_NODO + (numSuccessore*2+1)),0,nodoSorgente-1 );
    //cout<<"NODO = "<<lancioSceltaNodoEffettivo<<" - n lancio "<< 1+nodoSorgente*N_TOT_LANCI_PER_NODO + (numSuccessore*2+1) <<"\n";
    return nodoRitornato(lancioSceltaNodoEffettivo,1);
  }
  else{
    //scelta del nodo da cui copiare il successore
    long lancioNodoDiRiferimento   = normalizzaLancio (r.genRand( (long long)nodoSorgente*N_TOT_LANCI_PER_NODO),0,nodoSorgente-1 );
    //cout<<"NODO = "<<numSuccessore<<" - esimo SUCCESSORE DEL NODO "<<lancioNodoDiRiferimento<<" - n lancio "<<nodoSorgente*N_TOT_LANCI_PER_NODO<< "\n";
    //cout<<"VEDI SUCC n"<<numSuccessore<<" del nodo "<<lancioNodoDiRiferimento<<"\n";
    return nodoRitornato(lancioNodoDiRiferimento,2);
  }

}








void stampaHelp(int arg,char* argv[]){
  cout<<"This program generates a graph according to the copying model proposed by Kumar et al. The usage is as follow:\n";
	cout<<"\n";
  cout<<eliminaPath(argv[0])<<" memory num_nodes numSucc alpha seed nameOutputMultifile%d \n";
	cout<<"\n";
  cout<<"memory     memory available (MB)\n";
  cout<<"num_nodes  number of graph nodes\n";
  cout<<"numSucc    number of successor of every node (constant)\n";
  cout<<"alpha      copying parameter (in range 0-100)\n";
  cout<<"seed       seed for the random generator\n";
  cout<<"nameOutputMultifile%d  name of the output multifile (in 'extended' edges format)\n";

}



int main (int arg,char** argv){


  stampaLicenza(arg,argv);
  if (arg!=7) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");


	
  long memoria       = (long)(atof(argv[1])*1024*1024);
  long numNodi       = strtol(argv[2],NULL,10);
  long numSuccTotali = strtol(argv[3],NULL,10);		
  long probNuovo     = 100-strtol(argv[4],NULL,10);
  long seed          = strtol(argv[5],NULL,10);			
  char nomeGrafo[1000];
  strcpy(nomeGrafo,argv[6]);
	
  //gestione e controllo dei parametri
  if (!estMultiFileName(nomeGrafo)) STAMPA_ERRORE_PARAMETRI(argv[0],"Input multifile name must contain the substring '%d' ");
  if (memoria<5000000) STAMPA_ERRORE_PARAMETRI(argv[0],"Memory not enuogh");
  if (numNodi<1) STAMPA_ERRORE_PARAMETRI(argv[0],"Nodes number not valid");
  if (numSuccTotali<1) STAMPA_ERRORE_PARAMETRI(argv[0],"successors number not valid");
  if (probNuovo<0 || probNuovo>100 ) STAMPA_ERRORE_PARAMETRI(argv[0],"coping probability not valid "); 
  if (seed<0) STAMPA_ERRORE_PARAMETRI(argv[0],"seed is not valid (<0)");

  //gestione memoria
  setMemoriaDisponibile(memoria) ;
  long memoriaDispRand=(long)(getMemoriaDisponibile())-NUM_BUFFER*DIM_BUFFER;//memoria a disposizione per il calc random
  if  (memoriaDispRand<1*MB){cout<<"Memory not enough"; FINE;}

  bufferedThreadFileOutputStream fout  = bufferedThreadFileOutputStream(nomeGrafo,DIM_MAX,DIM_BUFFER,NUM_BUFFER);
  long periodo=memoriaDispRand/sizeof(unsigned long);

  //init randomizzazione
  randomGen2 r  = randomGen2(seed,periodo);
  fflush(stdout);

/*  for (int t=0;t<40;t++)
      printf("%5d - %20u\n",t,r.genRand(t));


  EXIT(0);*/

  long long conta=100000;
  long long numArchiTot=0;
  long long numArchiDistinti=0;

  long* successori = new long [numSuccTotali];
  long  sizeSuccDistinti=0;
  cout<<"\n";
  //printf(" CopyModel con outDegree = %ld e probabilita' di copia (tra 0 e 100) = %ld \n\n",numSuccTotali,100-probNuovo);

  printf("+---------------------------------------------+ +----------------------------------------------+\n");
  printf("|           GENERAL INFORMATION               | |     Iterative coping generator information   |\n");
  printf("+---------+----------+------------+-----------+ +---------------+------------------------------+\n");
  printf("|   %s     |tot nodes |    edges   | dup edges | |cicles number  |   average cicles per edge    |\n","%");
  printf("+---------+----------+------------+-----------+ +---------------+------------------------------+\n");

  long long numTentativiTot=0; //num tentativi per calcolo nodo successore
  long long numTentativiUltimiArchi=0;
  long numUltimiArchi=0;


  //per tutti i nodi e per tutti i successori
  for (long nn=0;nn<numNodi;nn++){
        sizeSuccDistinti=0;

        for (long succ=0;succ<numSuccTotali && succ<nn;succ++){

          //var di appoggio
          long nodo=nn;
          nodoRitornato ris;


          //ripeto la ricerca finche' non trovo un nodo o nulla
          do{
             //cout<<"\nPARTE DAL NODO "<<nodo<<": \n";
             ris = calcolaLinkUscente(nodo, succ, r, numSuccTotali, probNuovo);
             nodo=ris.nodoDest;//eventuale nodo da cui riprendere la ricerca - se trovato successoreDi ...
             numTentativiTot++;
             numTentativiUltimiArchi++;
          }
          while (ris.estSuccessoreDi() );


          //se e' un nodo lo inserisco in output
          if (ris.estNodo()){
             arco a;
             a.source=nn;
             a.dest=ris.nodoDest;

             //controllo se non e' gia' presente...
             bool trovato=false;
             for (long i=0;i<sizeSuccDistinti && !trovato;i++)
                 trovato = (successori[i]==a.dest);

             if (!trovato){
                 successori[sizeSuccDistinti]=a.dest;
                 sizeSuccDistinti++;
                 fout.write(&a,sizeof(a),1);
                 numArchiDistinti++;
             }
             numArchiTot++;
             numUltimiArchi++;
          }

        }


        if (conta++>50000 || nn == numNodi - 1 ){
             //printf("| %work |   nodi   |    archi   | archi dup |
             if (numUltimiArchi==0) numUltimiArchi=1;

             printf("%c| %6.2f%s |%10ld|%12lld|%11lld| | %11lld   |  %6.2f searches per edge    |%c|",
                  13,(100.0*nn/numNodi),"%",nn,numArchiDistinti,
                  (long long )(numArchiTot-numArchiDistinti),numTentativiTot,1.0*numTentativiUltimiArchi/numUltimiArchi, 13 );
             fflush(stdout);
             conta=0;
             numUltimiArchi=0;
             numTentativiUltimiArchi=0;

        }



  }


//  printf("%c| %5.2f%s |%10ld|%12lld|%11lld|",13,(100.0),"%",numNodi,numArchiDistinti,(long long )(numArchiTot-numArchiDistinti) );
  cout<<"\n";
  printf("+---------+----------+------------+-----------+ +---------------+------------------------------+\n");
  cout<<"\n";
  fflush(stdout);


  //aggiuge la label relativa al num di nodi...
  arco aux;
  aux.source=aux.dest=-numNodi;
  if (fout.write(&aux,sizeof(arco),1)<1){perror("Write error");FINE;}
  PRINT_ISOLATED_NODES_SPECIFICATION;
  
  r.cancella();
  fout.close();

  fprintf(stdout,"Number of nodes...............%ld\n"  ,numNodi);
  fprintf(stdout,"Number of edges...............%lld\n" ,numArchiDistinti);

  EXIT(0);
}



