/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file degree.cpp :source code of degree utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "comuni.h"
#include "config.h"
#include "contatore.h"

#include <map>
using namespace std;


#define STESSA_LINEA   (13)


long _min_(long a,long b){
  if (a<b) return a;
  return b;
}



FILE* apriFile(char* nomeMultiFile,long numFileDaAprire){
  char nome[DIM_TEXT_FILENAME];
  sprintf(nome,nomeMultiFile,numFileDaAprire);
  FILE* f=fopen(nome,"w");
  if (f==NULL){
      char a[200];
      sprintf(a,"Unexpected fatal error (unable to open file %s)",nome);
      printError(a);FINE;
  }
  return f;
}


// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void calcolaDegree(char* nomeFilesInfo,char* fileOutDistribuzione)
{
  cout<<" Scanning file "<<nomeFilesInfo<<endl;
  long numNodi=getTotalSizeMultiFiles(nomeFilesInfo)/sizeof(infoNodo);
  if (numNodi<=0){printf("No graph defined.");FINE}

  typedef map<long, long> mappa;
  mappa m_In;//mappa vuota
  mappa m_Out;//mappa vuota

  //file info
  multiFileReader info=multiFileReader(nomeFilesInfo,-1 );

  long letti;
  const long DIM_BUFFER = 1000 ;
  infoNodo g[DIM_BUFFER];
  long nodoCorrente=0;

  //leggo tutti i nodi
  timer t=timer(2);
  while ( (letti=info.read(g,sizeof(infoNodo),DIM_BUFFER ))>0 ){

    for (int i=0;i<letti;i++){
      #define in g[i].inDeg
      #define out g[i].outDeg
      m_In.find(in)   == m_In.end()  ? m_In[in]=1   : m_In[in]++;
      m_Out.find(out) == m_Out.end() ? m_Out[out]=1 : m_Out[out]++;
      #undef in
      #undef out
    }

    //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    nodoCorrente+=letti;
    t.print("%c[%6.2f%%] degree distribution",13,100.0*nodoCorrente/numNodi);
    
  }


  t.print("%c[%6.2f%%] degree distribution",13,100.0);
  cout<<"\n\n";
  info.close();


  // ---------- IN DEGREE ---------- 
  char nomeIn [DIM_TEXT_FILENAME];
  strcpy(nomeIn ,fileOutDistribuzione);
  strcat(nomeIn ,".in");
  FILE* ris =apriFile(nomeIn, 0);
	cout<<"Writing indegree distrib. in file "<<nomeIn<<"...";fflush(stdout);
  fprintf(ris,"Indegree\tNum nodes\n");
  for (map<long, long>::iterator i=m_In.begin(); i!=m_In.end(); i++){
    fprintf(ris,"%ld\t%ld\n",i->first,i->second);
  }
  cout<<" OK\n";fflush(stdout);
  fclose(ris);


  // ---------- OUT DEGREE ---------- 
  char nomeOut[DIM_TEXT_FILENAME];
  strcpy(nomeOut,fileOutDistribuzione);
  strcat(nomeOut,".out");
  ris =apriFile(nomeOut, 0);
	cout<<"Writing indegree distrib. in file "<<nomeIn<<"...";fflush(stdout);
  fprintf(ris,"Outdegree\tNum nodes\n");
  for (map<long, long>::iterator i=m_Out.begin(); i!=m_Out.end(); i++){
    fprintf(ris,"%ld\t%ld\n",i->first,i->second);
  }
  cout<<" OK\n";fflush(stdout);
  fclose(ris);
}

// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX




void stampaHelp(int arg,char* argv[]){  
  cout << "This program computes the indegree and outdegree distribution of the input graph, in the IPS format."
       << " The indegree and outdegree distributions are stored in two files named as specified in the command"
       << " line options, respectively with the suffices '.in' and '.out'. The usage is therefore as it follows:\n";
	cout << "\n";
  cout << eliminaPath(argv[0])<<" memory nameInputMultifile%d nameOutputMultifile\n";
	cout << "\n";
  cout << "memory           memory available (MB)\n";
  cout << "nameInputMultifile%d   is the name of the IPS multifile of the input graph \n";
  cout << "nameOutputMultifile    the name of the output text file (note that the suffices .in and .out are added to this name)\n";
}


int main(int arg,char* args[]){

  stampaLicenza(arg,args);
  if (arg!=4) STAMPA_ERRORE_PARAMETRI(args[0],"Parameters incorrect.");


  //AGGIUGNE IL SUFFISSO .info
  char infoN[300];
  strcpy(infoN,args[2]);
  strcat(infoN,".info");

  if (!estMultiFileName(infoN))
    STAMPA_ERRORE_PARAMETRI(args[0],"Input multifile name must contain the substring '%d' ");

  //************************ CONTROLLO FILES DI INPUT ***************************
  long m=(long)atof(args[1])*1024*1024;

  setMemoriaDisponibile(m);
 	calcolaDegree(infoN,args[3]);
  EXIT(0);
}


