/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file barabasi.cc :source code of evolvingmodel utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "comuni.h"
#include <iostream>
#include "config.h"


//struttura per memorizzare un lancio
class tipo_lancio{
public:
  long valore_dest;
  long nodo_sorgente;
  char tipo;//0 se nuovo nodo; 1 se valore_dest e' un lancio; 2 se valore_dest e' il nodo destinazione
};



void stampa(char* msg,char* z,long l){
    cout<<msg<<": ";
    for (int i=0;i<l;i++)
        cout<<"  "<<(int)z[i];
     cout<<"\n";
}



long normalizzaLancio(unsigned long lancio, long inferiore, long superiore){
    long diff=superiore-inferiore+1;
    if (diff<0) return -1;
    if (diff==0) return inferiore;
    return inferiore+lancio%diff;
}



void stampaHelp(int arg,char* argv[]){
  cout<<"This program generates a graph according to the Evolving Network model proposed by Albert et al. The usage is as follows:\n";
	cout<<"\n";
  cout<<eliminaPath(argv[0])<<" memory num_nodes numSucc seed nameOutputMultifile%d\n";
	cout<<"\n";
  cout<<"memory    memory available (MB)\n";
  cout<<"num-nodes number of graph's nodes \n";
  cout<<"numSucc   number of successor of every node (constant)\n";
  cout<<"seed      seed for the random number generator\n";
  cout<<"nameOutputMultifile%d  name of the output multifile (in 'extended' edges format)\n";
}




//controlla che l'arco non sia gia' presente ed aggiorna le altre strutture
bool arcoDuplicato(arco a,long& nodoInsiemeRif,insieme& insSucc,long numTotSuccessori){

  //se la sorgente e' un nodo diverso da quello corrente, ricomincio da capo
  if (a.source!=nodoInsiemeRif){
      insSucc.cancella();//svuota l'insieme precedente
      insSucc=insieme(numTotSuccessori);//nuovo insieme vuoto
      nodoInsiemeRif=a.source;
      insSucc.aggiungiElemento(a.dest);//ins il nuovo successore
//      cout<<"\n NUOVO NODO SORGENTE\n";
      return false;
  }
  else{
      //controlla se il successore e' gia' stato considerato
      if (insSucc.getPosizioneElemento(a.dest)<0){
          //nuovo successore
          insSucc.aggiungiElemento(a.dest);
//          cout<<"INS";
          return false;
      }

      //gia' presente
      return true;
  }

}





int main (int arg,char** argv){
  stampaLicenza(arg,argv);

  if (arg!=6) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");

	
	long memoria       = (long) (atof(argv[1])*1024*1024);
  
	long numNodi       = strtol(argv[2],NULL,10);
	long numSuccTotali = strtol(argv[3],NULL,10);		
	long probUniforme     = 0;
  long seed          = strtol(argv[4],NULL,10);			
  char nomeGrafo[1000];
	strcpy(nomeGrafo,argv[5]);

	
	if (!estMultiFileName(nomeGrafo)) STAMPA_ERRORE_PARAMETRI(argv[0],"Output filename of the graph must contain the substring '%d' ");
	if (memoria<5000) STAMPA_ERRORE_PARAMETRI(argv[0],"Memory not enough");
	if (numNodi<1) STAMPA_ERRORE_PARAMETRI(argv[0],"Nodes number not valid");
	if (numSuccTotali<1) STAMPA_ERRORE_PARAMETRI(argv[0],"Successors number not valid");
	if (probUniforme<0 || probUniforme>100 ) STAMPA_ERRORE_PARAMETRI(argv[0],"Probability must be between 0 and 100");
	if (seed<0) STAMPA_ERRORE_PARAMETRI(argv[0],"seed not valid (<0)");

  setMemoriaDisponibile(memoria) ;
  //cout<<"Available memory: "<<1.0*getMemoriaDisponibile()/1024/1024<<"MB \n";

  long memoriaDispRand=(long)(getMemoriaDisponibile())-NUM_BUFFER*DIM_BUFFER;//memoria a disposizione per il calc random
  if  (memoriaDispRand<1*MB){cout<<"Error: specified memory not enuogh"; FINE;}











  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //
  //                                 SUDDIVISIONE DEI NODI IN BLOCCHI E MEMORIZZAZIONE DELLE OPERAZIONI PARZIALI
  //
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  //numero blocchi necessari e num di nodi per blocco
  #define dimSingoloNodo ( (1.0*sizeof(long)+1.0*sizeof(long)/COEFF_LIV_2+1.0*sizeof(long)/COEFF_LIV_2/COEFF_LIV_1)  )
  long nodiPerBlocco=(long)( getMemoriaDisponibile() / dimSingoloNodo);
  //cout<<"Average node memory occupacy: "<<dimSingoloNodo<<" - max nodes number for each block: "<<nodiPerBlocco <<"\n";fflush(stdout);

  //se non mi servono cosi' tanti nodi ne alloco solo una parte
  if (nodiPerBlocco>numNodi)  nodiPerBlocco=numNodi;


  long numBlocchi=numNodi/nodiPerBlocco;
  if (numNodi%nodiPerBlocco!=0)      numBlocchi++;

  //lista del numero di operazioni per ogni blocco
  long long* operazioniPerBlocco=new long long [numBlocchi];


  //lista dei files associai ai blocchi TUTTI APERTI ALL'INIZIO - NE SUPPORTA FINO A 1024 CIRCA
  multiFileWriter* f = new multiFileWriter[numBlocchi]();

  for (long i=0;i<numBlocchi;i++){
      char cc[1000];sprintf(cc,nomeGrafo,i),strcat(cc,".TEMPO%d");
      f[i] = multiFileWriter(cc,DIM_MAX);
      operazioniPerBlocco[i]=0;
  }


  //gestione indegree dei vari blocchi (memoria irrisoria)
  inDeg indegBlocchi=inDeg(numBlocchi,4,4,4);

  //init generatore numeri casuali
  randGen rg=randGen(seed);
  long long conta=5000000;

  printf("%ld blocks used with %ld nodes (max)\n",numBlocchi,nodiPerBlocco) ;

  //per tutti i nodi calcolo i successori E ASSOCIO I VARI CALCOLI AI DIFFERENTI BLOCCHI
  for (long nn=0;nn<numNodi;nn++){

        //var ausiliarie
        long blocco;
        tipo_lancio t;


        //creo tutti i successori DEL NODO CORRENTE
        for (long succ=0;succ<numSuccTotali && succ<nn;succ++){

          // controlla se si tratta di un successore preso uniformemente o SECONDO INDEGREE
          if ( ((100.0*(float)rg.genrand()/RAND_MAX)<probUniforme)||(probUniforme==100.0) ){

             //calcolo il nuovo nodo direttamente
             long nuovoSucc=normalizzaLancio(rg.genrand(),0,nn-1);//scelta uniforme, QUINDI HO IL NODO DIRETTO TRA QUELLI GIA' CREATI

             //creo la struttura PER L'OPERAZIONE PARZIALE
             t.nodo_sorgente=nn;//NODO SORGENTE
             t.valore_dest=nuovoSucc;//SPECIFICO DIRETTAMENTE IL NODO DESTINAZIONE
             t.tipo=2;//TIPO NODO SUCCESSORE DIRETTAMENTE SPECIFICATO



             //calcolo il layer in cui VA MEMORIZZATA L'OPERAZIONE (CHE E' QUELLO IN CUI si trova il successore)
             blocco=nuovoSucc/nodiPerBlocco;
          }
          else{
             //lancio il dado PER LA SCELTA PER MEZZO DEGLI INDEGREE
             long lancio = normalizzaLancio( rg.genrand(),1,indegBlocchi.getInDegTotale() );//scelta pesata dall'indegree dei vari blocchi
             if(lancio<=0){cout<<"Error in throwing";FINE;}

             //scelgo il blocco in base al lancio
             blocco = indegBlocchi.trovaNodo(lancio);
             if (blocco<0){cout<<"Error in program (block choice)"; FINE;}

             //____________________________________________________________________________
             //
             //           calcolo indegree dei blocchi precedenti a questo corrente
             //____________________________________________________________________________
             long nBl;
             long sommaInDegPrec=0;
             long*v = indegBlocchi.getRifInDeg(nBl);
             for (long i=0;i<blocco;i++)
                sommaInDegPrec+=v[i];

             //calcolo il valore del lancio relativo
             long lancioRelativo=lancio-sommaInDegPrec;
             //____________________________________________________________________________
             //____________________________________________________________________________

             //creo la struttura del lancio
             t.nodo_sorgente=nn;//NODO SORGENTE
             t.valore_dest=lancioRelativo;//COME DESTINAZIONE SPECIFICO SOLO IL VALORE DEL LANCIO
             t.tipo=1;//e' un lancio

          }

          //AGGIUNGO L'INDEGREE AL BLOCCO SCELTO
          //nuovo puntatore ad uno dei nodi del blocco
          indegBlocchi.aggiungiInDeg(blocco);

          //scrivo su file temporaneo L'OPERAZIONE PARZIALE
          f[blocco].write(&t,sizeof(t),1);


          //una operazione IN PIU' nel blocco specificato
          operazioniPerBlocco[blocco]++;

//          char* tt[3];
//          tt[0]="| NUOVO  |";
//          tt[2]="| SUCC   |";
//          tt[1]="| LANCIO |";
//          cout<<" TIPO= "<<tt[(int)t.tipo ] <<" nodo= "<<t.nodo_sorgente<<" valore= "<<t.valore_dest<<" blocco= "<<blocco<<"\n";

          //STAMPA
          if (conta++>500000 || nn>=numNodi-1){
             printf("%c [%6.2f%s] preliminary analysis: %ld nodes",13,100.0*(nn+1)/numNodi,"%",nn+1);
             fflush(stdout);
             conta=0;
          }



        }

        // ------------------------------------------------
        // HO FINITO TUTTI I SUCCESSORI DEL NODO IN ANALISI
        // ------------------------------------------------


        // ...........................................
        //
        //  alla fine aggiungo il nuovo nodo corrente
        //E CREO L'OPERAZIONE PARZIALE CORRISPONDENTE
        // ...........................................

        //creo la struttura del lancio
        t.nodo_sorgente=nn;//NODO CORRENTE
        t.valore_dest=-1;
        t.tipo=0;//NUOVO NODO

        //calcolo il blocco in cui si trova il nodo NUOVO DA AGGIUNGERE
        blocco=nn/nodiPerBlocco;

        //scrivo su file temporaneo
        f[blocco].write(&t,sizeof(t),1);

        //nuovo puntatore ad uno dei nodi del blocco
        indegBlocchi.aggiungiInDeg(blocco);

        //aggiorna il num di operazioni di blocco
        operazioniPerBlocco[blocco]++;

//          char* tt[3];
//          tt[0]="| NUOVO  |";
//          tt[2]="| SUCC   |";
//          tt[1]="| LANCIO |";
//          cout<<" TIPO= "<<tt[(int)t.tipo ] <<" nodo= "<<t.nodo_sorgente<<" valore= "<<t.valore_dest<<" blocco= "<<blocco<<"\n";
//          cout<<"    --------------------------------------------------   \n";
        // -------------------------------------------

  }
  //HO FINITO DI ANALIZZARE TUTTI I NODI E DI OGNUNO HO CREATO UN FILE TEMPORANEO

  //cancella la lista indegree
  indegBlocchi.cancella();


  //lista dei files associai ai blocchi e chiusura dei files dei blocchi
  long* numFilesPerBlocco=new long[numBlocchi];

  //chiude tutti i files
  for (long i=0;i<numBlocchi;i++){
      numFilesPerBlocco[i]=f[i].close();
  }
  delete[] f;f=NULL;
  cout<<"\n";











  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //
  //                PER OGNI BLOCCO CREATO CREO I VERI ARCHI - DA NOTARE CHE ESSENDO I BLOCCHI DISGIUNTI NON CI SONO DUP DI ARCHI
  //
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  cout<<"\n";
  //printf(" NetGrowthModel con outDegree = %ld e probabilita' di scelta uniforme dei nodi (tra 0 e 100) = %ld \n\n",numSuccTotali,probUniforme);

  printf("+---------+---------+---------+----------+-----------+-----------+\n");
  printf("| %s total |  block  |  %s par  |   nodes  |   edges   | dup edges |\n","%","%");
  printf("+---------+---------+---------+----------+-----------+-----------+\n");

  //apro il file di out degli archi
  multiFileWriter fout=multiFileWriter(nomeGrafo,DIM_MAX);

  //num nodi letti
  long numTotaleNodiLetti=0;
  long long numeroArchiTot=0;
  long long numeroArchiDuplicati=0;

  //per ogni blocco ...
   for (long blocco=0;blocco<numBlocchi;blocco++ ){

      //struttura per il conteggio degli indeg dei nodi del blocco corrente
      cout<<(char)13<<"Memory allocation ..."; fflush(stdout);
      inDeg indegreeBloccoCorrente=inDeg (nodiPerBlocco,COEFF_LIV_0,COEFF_LIV_1,COEFF_LIV_2);

      char cc[1000];sprintf(cc,nomeGrafo,blocco),strcat(cc,".TEMPO%d");
      multiFileReader f = multiFileReader(cc,numFilesPerBlocco[blocco],true);

      //contatori
      long numNodiLetti=0;
      conta=50000000;

      //primo nodo del blocco corrente
      long nodoIniziale=blocco*nodiPerBlocco;

      //LEGGO TUTTE LE OPERAZIONI DEL BLOCCO CORRENTE
      tipo_lancio t;
      long long contaOperazioni=0;

      //__________________________________________________________________________________________________________________________________
      //

      //    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE    IMPORTANTE
      //__________________________________________________________________________________________________________________________________
      //  va notato che le operazioni relative allo stesso nodo sorgente in ogni blocco compaiono tutte insieme e sfruttando cio' si puo'
      //  controllare che non ci siano duplicazioni di archi
      //__________________________________________________________________________________________________________________________________
      insieme ins=insieme(numSuccTotali);
      long nodoRifInsiemeSucc=-1;//nodo di cui sto calcolando i successori

      while (f.read(&t,sizeof(t),1)>0){
         contaOperazioni++;
         switch(t.tipo){


           case 0:{//inserisci nodo
             // aggiunge un indegree per il nuovo nodo - DA RIPORTARE AL VALORE RELATIVO
             indegreeBloccoCorrente.aggiungiInDeg(t.nodo_sorgente-nodoIniziale);
             numTotaleNodiLetti++;
             numNodiLetti++;
             break;
           }


           case 1:{//ho il valore del lancio
             long nodoSucc=indegreeBloccoCorrente.trovaNodo(t.valore_dest);
             if (nodoSucc<0 || nodoSucc>=nodiPerBlocco){
                cout<<"\nError in the program ";indegreeBloccoCorrente.stampa();
                cout<<"\n";
                char* tt[3];
                tt[0]=(char*)"| NUOVO  |";
                tt[2]=(char*)"| SUCC   |";
                tt[1]=(char*)"| LANCIO |";
                cout<<" TIPO= "<<tt[(int)t.tipo ] <<" node= "<<t.nodo_sorgente<<" value= "<<t.valore_dest<<" block= "<<blocco<<"\n";
                FINE;
             }

             nodoSucc+=+nodoIniziale;

             //crea l'arco e lo memorizza
             arco a;
             a.source=t.nodo_sorgente;
             a.dest=nodoSucc;

             //CONTROLLO DUPLICAZIONI
             if (arcoDuplicato(a,nodoRifInsiemeSucc,ins,numSuccTotali))
                numeroArchiDuplicati++;
             else{
               fout.write(&a,sizeof(a),1);
               //cout<<"L "<<a.source<<" --> "<<a.dest<<"\n";
               numeroArchiTot++;
             }

             // IN OGNI CASO L'ARCO VIENE AGGIUNTO PERCHE' LE SPECIFICHE
             // AMMETTEREBBERO ARCHI MULTIPLI
             // aggiunge un nuovo indegree per il successore - DA RIPORTARE AL VALORE RELATIVO
             indegreeBloccoCorrente.aggiungiInDeg(a.dest-nodoIniziale);
             break;

           }


           case 2:{//ho il valore DIRETTO DEL NODO

             //creo l'arco
             arco a;
             a.source=t.nodo_sorgente;
             a.dest=t.valore_dest;

             //CONTROLLO DUPLICAZIONI degli archi
             if (arcoDuplicato(a,nodoRifInsiemeSucc,ins,numSuccTotali))
                numeroArchiDuplicati++;
             else{
               fout.write(&a,sizeof(a),1);
               //cout<<""<<a.source<<" --> "<<a.dest<<"\n";
               numeroArchiTot++;
             }

             // IN OGNI CASO L'ARCO VIENE AGGIUNTO PERCHE' LE SPECIFICHE
             // AMMETTEREBBERO ARCHI MULTIPLI
             //aggiunge un nuovo indegree per il successore - DA RIPORTARE AL VALORE RELATIVO
             indegreeBloccoCorrente.aggiungiInDeg(a.dest-nodoIniziale);
             break;
           }
         }


         if (conta++>500000 || contaOperazioni>=operazioniPerBlocco[blocco]){
             float percTot=100.0*numTotaleNodiLetti/numNodi;
             float percParz=100.0*contaOperazioni/operazioniPerBlocco[blocco];
             printf("%c| %6.2f%s | %7ld | %6.2f%s |%10ld| %10lld| %10lld|%c|",13,percTot,"%",blocco,percParz,"%", numTotaleNodiLetti,numeroArchiTot,numeroArchiDuplicati,13 );
             fflush(stdout);
             conta=0;
         }


      }
      // ________________________________________________________________
      //                 letto TUTTE LE OPERAZIONI
      // ________________________________________________________________

      //cancella la struttura dei successori
      ins.cancella();


      //CHIUDO IL FILE TEMPO
      f.close();

      //CONTROLLO CORRETTEZZA PROGRAMMA
      if(contaOperazioni!=operazioniPerBlocco[blocco]){cout<<"Error: not all operazions have been read"<<blocco;FINE;}

      //libera memoria
      indegreeBloccoCorrente.cancella();

      //STAMPA
      printf("\n+---------+---------+---------+----------+-----------+-----------+\n");
  }
  //FINE ANALISI DI TUTTI I BLOCCHI


  //aggiuge la label relativa al num di nodi...
  arco aux;
  aux.source=aux.dest=-numNodi;
  if (fout.write(&aux,sizeof(arco),1)<1){perror("Write error");FINE;}
  PRINT_ISOLATED_NODES_SPECIFICATION;


  long numOut=fout.close();



      // ________________________________________________________________
      //
      //                    CREA FILE DI REPORT
      // ________________________________________________________________

  //cancella num files per blocco
  delete[] numFilesPerBlocco;numFilesPerBlocco=NULL;
  delete[] operazioniPerBlocco;operazioniPerBlocco=NULL;

  cout<<"\n";
  fprintf(stdout,"Nodes number ...............%ld\n"  ,numNodi);
	fprintf(stdout,"Edges number ...............%lld\n" ,numeroArchiTot);

  EXIT(0);
}



