/*

*********************************************************************
*********************************************************************

  file main.cpp :source code of infopredsucc2dimacs utility

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/
/*
  Questo programma converte un grafo dal formato INFO-PRED-SUCC al formato DIMACS.
  Utilizza per la lettura dei multifile in input la classe MultiFileReader e la
  classe BufferLetturaMultifile. Per l'output viene utilizzato direttamente fprintf().

  Ernesto Ricci
  26/11/2002
*/


#include <stdlib.h>
#include <limits.h>

#include "comuni.h"
#include "../common/scc_utl/BufferLetturaMultifile.h"


//costante soglia avvertimento: 1 GB. Appena il file generato in output supera tale dimensione viene visualizzato un messaggio di avvertimento
#define TH_WARNING 1024*1024*1024L

struct grafoInfo{
  long inDeg;
  long outDeg;
  long long puntIn;
  long long puntOut;
};


//nomi dei multifile
char grafoINFO[300];
char grafoPRED[300];
char grafoSUCC[300];

//numero componenti dei multifile
long numINFO;
long numPRED;
long numSUCC;







void stampaHelp(int arg,char** argv){
		cout<< "This utility converts an ips format multifile in the dimacs format\n";
		cout<< " \n";
		cout<< "Usage: "<<eliminaPath(argv[0])<<" memory -savesource|-deletesource nameInputMultifile%d nameOutputFile\n";
		cout<<"\n";
		cout<< "memory                     memory available(MB)\n";
		cout<< "-savesource|-deletesource  specify whether to delete or keep the input file\n";
		cout<< "nameInputMultifile%d       name of the multifile of the input graph (ips format)\n";
		cout<< "nameOutputFile             name of the output file (dimacs format)\n";
}


//restituisce il nome del file a partire dal pathname completo
char * ricava_nome_file(char * path_name)
{
  char * p= strrchr(path_name, '/'); //cerchiamo l'ultimo slash della stringa
  if(p == NULL) // slash non trovato, la stringa e' direttamente il nome del file
    return path_name;
  else
    return p + 1; //saltiamo lo slash
}


//inserisce le info. relative ad un arco nel file: "fileout" e' l'handler del file destinazione, "buf" e' il buffer per leggere il multifile SUCC, "ID_NODO"  il numero del nodo corrente, "num_succ" e' il numero di successori del nodo corrente
int stampa_arco(FILE * fileout, BufferLetturaMultifile<long> & buf, long ID_NODO, long num_succ)
{
  long i;
  int ret= 0; //numero di byte scritti sul file

  for(i=0; i<num_succ; i++) //esploriamo tutti i successori
  {
    long nodo;
    int ok= buf.get(nodo); //leggiamo l'ID di un nodo successore del nodo corrente

    if(ok)
      ret+= fprintf(fileout, "e %ld %ld\n", ID_NODO+1, nodo+1); //scriviamo l'informazione sull'arco corrente, nel formato: "e nodo_src nodo_dest" (NB: il 1 nodo in dimacs e' 1
    else
      printf("\nwarning!: successor not found\n"); //ci aspettavamo un successore in base alle informazioni contenute nel multifile INFO (puo' indicare un'inconsistenza nel multifile grafo)
  }

  return ret; //ritorniamo il numero di caratteri scritti
}



void creaFileDIMACS( char * nome_destinazione, bool cancellaSource)
{
  long long num_nodi= getTotalSizeMultiFiles(grafoINFO) / sizeof(struct grafoInfo); //ogni record del multifile corrisponde a un nodo
  long long num_archi= getTotalSizeMultiFiles(grafoSUCC) / sizeof(long); //ogni long corrisponde ad un nodo successore cioe' ad un arco

  if(num_nodi == 0)
  {
    printf("Input graph is empty\n"); EXIT(1);
  }

  multiFileReader mf_info(grafoINFO, numINFO , cancellaSource); //apre i multifile
  multiFileReader mf_succ(grafoSUCC, numSUCC , cancellaSource);
  multiFileReader mf_pred(grafoPRED, numPRED , cancellaSource); //non viene usato, ma serve per cancellarlo

  BufferLetturaMultifile<struct grafoInfo> buf_info(&mf_info); //buffer per leggere il multifile a blocchi
  BufferLetturaMultifile<long> buf_succ(&mf_succ); //buffer per leggere il multifile a blocchi

  FILE * file_out= fopen(nome_destinazione, "w"); //file output
  check(file_out,(char*)"Fatal error in fopen function (%s)\n", nome_destinazione);

  printf("Format conversion:\n");

  long ID_nodo= 0; //l'indice del record all'interno del multifile = identificatore del nodo
  int more;
  struct grafoInfo info_corrente;
  long byte_scritti= 0;
  int iterazioni= 100;

  byte_scritti+= fprintf(file_out, "c FILE: %s\n", ricava_nome_file(nome_destinazione)); //commento iniziale
  byte_scritti+= fprintf(file_out, "c SOURCE: %s %s %s\n", ricava_nome_file(grafoINFO), ricava_nome_file(grafoPRED), ricava_nome_file(grafoSUCC)); //commento iniziale
  byte_scritti+= fprintf(file_out, "p edge %lld %lld\n", num_nodi, num_archi); //info. numero nodi/archi

  more= buf_info.get(info_corrente); // legge un record di informazioni
  while(more)
  {
    //i nodi iniziano da 0
    byte_scritti+= stampa_arco(file_out, buf_succ, ID_nodo, info_corrente.outDeg);
    ID_nodo++; //incrementiamo il conteggio dei nodi (N.B. il primo nodo ha ID=1)

    if(byte_scritti > TH_WARNING)
      printf("Warning! Maximum filesize reached (%ld bytes). Possible lost of data\n", TH_WARNING);

    if (++iterazioni >= 100) {
        iterazioni=0;
        printf("%c[%5.2f%%] ips --> dimacs",13,100.0*ID_nodo/num_nodi);
        fflush(stdout);
    }

    more= buf_info.get(info_corrente); // legge un record di informazioni
  }

  printf("%c[%5.2f%%] ips --> dimacs",13,100.0);
  fflush(stdout);
//  printf("\n\nGrafo convertito: in totale %ld bytes scritti\n", byte_scritti);

  cout<<"\nTotal nodes : "<<num_nodi<<"\n";
  cout<<"Total edges : "<<num_archi<<"\n";

  fclose(file_out);
  mf_succ.close();
  mf_pred.close(); //qui viene cancellato il multifile se richiesto dalla riga di comando
  mf_info.close();
}





int main(int argc,char** argv)
{
    long arg=argc;
    stampaLicenza(arg,argv);
    if (arg!=5) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");

    long cancellaSource;
    if (strcasecmp(argv[2],"-saveSource")==0)  cancellaSource=FALSE;
    else    if (strcasecmp(argv[2],"-deleteSource")==0)  cancellaSource=TRUE;
            else  STAMPA_ERRORE_PARAMETRI(argv[0],"option -deleteSource or -saveSource not found");


    strcpy(grafoINFO,argv[3]);strcat(grafoINFO,".info");
    strcpy(grafoPRED,argv[3]);strcat(grafoPRED,".pred");
    strcpy(grafoSUCC,argv[3]);strcat(grafoSUCC,".succ");


    numINFO=contaNumFilesMultiFile(grafoINFO);
    numPRED=contaNumFilesMultiFile(grafoPRED);
    numSUCC=contaNumFilesMultiFile(grafoSUCC);

    if (numINFO<=0)   		ERRORE("Input graph is empty");
    if (numSUCC<=0)       ERRORE("Input graph is empty");
    if (numPRED<=0)       ERRORE("Input graph is empty");

    if (!estMultiFileName(grafoINFO)) STAMPA_ERRORE_PARAMETRI(argv[0],"Input multifile name must contain the substring '%d' ");
    if (!estMultiFileName(grafoPRED)) STAMPA_ERRORE_PARAMETRI(argv[0],"Input multifile name must contain the substring '%d' ");
    if (!estMultiFileName(grafoSUCC)) STAMPA_ERRORE_PARAMETRI(argv[0],"Input multifile name must contain the substring '%d' ");


    creaFileDIMACS( argv[4], cancellaSource );

    EXIT(0);
}

