/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file copyIndegreeModel_layer.cpp :source code of layermodel utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "utilityModel.h"
#include "config.h"



// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
//
//                      crea il grafo in base all'algoritmo del copy model ma con scelta dei nuovi nodi rispetto all'indegree
//
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################


//-------------------------------------------------------------------
//                      PROBLEMI
//-------------------------------------------------------------------
//    ALLOCAZIONE DI numNodi*numSuccMassimi elementi per il grafo
//        molto grande all'aumentare del numSuccMassimi
//-------------------------------------------------------------------


//ritorna il num di archi generati
long long copyModelAndInDegree( long layer,long numTotLayer,randomGen2& generaRandom,
                                long numNodi,long numSuccMassimi,
                                char* nomeFileNodi,long numFileIn,
                                char* nomeFileArchiOut,long& numFilesOut,
                                float probNuovo,bool sceltaUniforme){

       //controllo di coerenza
       if (numSuccMassimi>numNodi) numSuccMassimi=numNodi;
       
       //apro file temporaneo archi
       multiFileWriter f_grafo = multiFileWriter(nomeFileArchiOut,DIM_MAX);
       multiFileReader f_tempo_nodi_read=multiFileReader(nomeFileNodi,numFileIn,false);//file dei nodi - da non cancellare


       long long conta=1000000;
       long long nArchiTot=0;






       //struttura per il conteggio dell'indegree dei nodi (inizialmente tutto a 0)
       //PURTROPPO DEVE GESTIRE TUTTI I NODI INSIEME
       inDeg inDegNodi=inDeg(numNodi,300,200,100);

       //struttura che memorizza il grafo
       grafo g=grafo(numNodi);

       // -------------------------------------------------------
       //    per ogni nodo del layer corrente calcolo gli archi
       // -------------------------------------------------------
       for (long nodo=0;nodo<numNodi;nodo++){

          //leggo le info corrispondenti del nodo corrente (num reale nodo e num successori da creare)
          infoNodoSucc info;
          if (f_tempo_nodi_read.read(&info,sizeof(info),1)<0){cout<<"Unexpected fatal error";FINE;}

          //limito il num di succ del nodo corrente...
          if (info.nSucc>=numNodi)info.nSucc=numNodi;

          //set il num di successori del nodo corrente
          g.setMaxSuccNodo(nodo,info.nSucc);

          //check programma
          if (info.nSucc>numSuccMassimi){cout<<"Unexpected fatal error"; FINE; }

          //scelgo a caso il nodo di riferimento tra quelli gia' inseriti (per il nodo 0 non ha senso)
          long nodoRiferimento=normalizzaLancio(generaRandom.genrand(),0,nodo-1);

          //insieme distinto dei successori del nodo corrente
          insieme succ=insieme(info.nSucc);

//          cout<<"\nNodo "<<nodo<<"("<<info.nSucc<<"):::: ";
          
          //controllo anche che il nodo 0 non abbia troppi nodi (0 e cosi' per gli altri....)
          long iterazioni=0;
          for (long n=0;n<info.nSucc && n<nodo && iterazioni<4*info.nSucc ;n++,iterazioni++){

            //nuovo nodo successore da inserire
            long nuovoSucc;

            //vede se si tratta di una copia oppure di un nuovo lancio
            
            if ( ( ((100.0*(float)generaRandom.genrand()/RAND_MAX)<probNuovo)||(probNuovo==100.0) ) ){

              if (!sceltaUniforme){
                //SCELGO IN BASE ALL'INDEGREE

                //esegue il lancio e sceglie in base all'indegree
                long lancio=normalizzaLancio(generaRandom.genrand(),1,inDegNodi.getInDegTotale());

                //cerco il nodo corrispondente
                nuovoSucc=inDegNodi.trovaNodo(lancio);

                //controllo di correttezza
                if (nuovoSucc<0 || nuovoSucc>=nodo){cout<<"Unexpected fatal error";inDegNodi.stampa(); FINE;}
              }              
              else
                //SCELGO UNIFORMEMEMNTE
                //scelta uniforme
                nuovoSucc=normalizzaLancio(generaRandom.genrand(),0,nodo-1);

            }
            else{
                //copia da un nodo precedente
                nuovoSucc=g.getSuccNodo(nodoRiferimento,n);
            }

            //inserisce se necessario il nuovo successore
            if (nuovoSucc>=0){
                

                //controllo che non ci sia gia' lo stesso successore
                if (succ.getPosizioneElemento(nuovoSucc)<0){
//                   cout<<nuovoSucc<<"  ";
                   //aggiungo alla lista degli elementi
                   succ.aggiungiElemento(nuovoSucc);

                   //aggiungo al grafo
                   g.addSucc(nodo,nuovoSucc);

                   //aggiungo alla lista degli indegree un nuovo in per il nodo
                   inDegNodi.aggiungiInDeg(nuovoSucc);
                }
                else{
                    //RIPETE SE ARCO DUPLICATO
//                    cout<<"R"<<nuovoSucc<<"  ";

//#define RIPETI_CALCOLO_SE_NODO_DUPLICATO
#ifdef RIPETI_CALCOLO_SE_NODO_DUPLICATO
                    n--;
#endif                    
                }
            }
            else{
                //RIPETE SE successore del nodo di riferimento non esistente
                // n--;
//                cout<<" M ";
            }

          }//fine inserimento successori...


          //inserisco anche il nodo corrente
          inDegNodi.aggiungiInDeg(nodo);


          //scrive l'insieme dei successori
          long* successori;
          long sizeSucc=succ.getVettore(successori);

          for (long i=0;i<sizeSucc;i++){

             //scrive gli archi e li converte automaticamente
             arco a;
             a.source=nodo;
             a.dest=successori[i];

             //controlla che non sia self-loop
             if (a.source!=a.dest){
                 f_grafo.write(&a,sizeof(a),1);
                 //aggiorna contatori archi generale e per layer
                 nArchiTot++;
             }
          }



          //cancella l'insieme

          succ.cancella();

          //STAMPA
          if ( (conta++>=300000) || (nodo>=numNodi-1)  ){
              float percParz = 100.0 * (nodo+1 )/numNodi;
              float percTot  = 100.0 * (percParz/100.0+layer )/numTotLayer;
              printf("%c|%6.2f%s |%8ld |%6.2f%s |%11ld |%12lld |",13,percTot,"%",layer,percParz,"%",nodo+1, nArchiTot);
              conta=0;fflush(stdout);
          }


       }
       // -----------------------------------------
       //    FINE CREAZIONE PER IL LAYER CORRENTE
       // -----------------------------------------


//       printf("+--------+---------+--------+------------+-------------+\n");
//       printf("| %s tot  |num layer| %s parz |  num nodi  |  num archi  |\n","%","%");
//       printf("+--------+---------+--------+------------+-------------+\n");


       //dealloca struttura per la conta degli indeg dei nodi
       inDegNodi.cancella();

       //chiude il file in lettura
       f_tempo_nodi_read.close();
       numFilesOut=f_grafo.close();

       //chiude il grafo e dealloca i vari dati
       g.cancella();

       return nArchiTot;
}



