/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file copyModel_layer.cpp :source code of layermodel utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "utilityModel.h"
#include "config.h"

// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
//
//                      crea il grafo in base all'algoritmo del copy model classico
//
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################

//funzione che prende il num di un nodo, il num del successore e mi ritorna il valore del nodo puntato o il riferimento
//il num di successore va da 0 al N_MAX -1
nodoRitornato calcolaLinkUscente(long nodoSorgente, long numSuccessore, randomGen2& r, long numeroSuccTotali,float probabilitaNuovoNodo_1_100){

  #define N_TOT_LANCI_PER_NODO ((long long) 1+2*numeroSuccTotali)

  /*  1 lancio per la scelta del nodo di riferimento (1 solo per tutti i successori)
      per ogni successore:
          1 lancio per scegliere se nodo casuale o copia e se nodo casuale 1 altro lancio...
  */


  //il nodo 0 non ha successori; il nodo 1 ne ha 1 e cosi' via
  if (numSuccessore>=nodoSorgente)
        return nodoRitornato(-1,0);//nessun nodo

  //num lanci associati ad ogni nodo = (1+2*numeroSuccTotali)

  //scelta del tipo di operazione (nuovo nodo OPPURE copia da esistente) - valore tra 0 e 100
  float lancioSceltaTipoNodo      = 100.0*(float)r.genRand( (long long)1+(long long)nodoSorgente*N_TOT_LANCI_PER_NODO + numSuccessore*2 )/RAND_MAX;
  //cout<<"lancioSceltaTipoNodo "<<lancioSceltaTipoNodo<<" - n lancio "<<1+nodoSorgente*N_TOT_LANCI_PER_NODO + numSuccessore*2<<"\n";

  //tra 1 e  probabilitaNuovoNodo_1_100 --> nuovo nodo, altrimenti copia
  if ( (lancioSceltaTipoNodo<probabilitaNuovoNodo_1_100)||(probabilitaNuovoNodo_1_100==100.0) ){
    //scelta di un nuovo nodo casuale
    long lancioSceltaNodoEffettivo =      normalizzaLancio (r.genRand( (long long)1+(long long)nodoSorgente*N_TOT_LANCI_PER_NODO + (numSuccessore*2+1)),0,nodoSorgente-1 );
    //cout<<"NODO = "<<lancioSceltaNodoEffettivo<<" - n lancio "<< 1+nodoSorgente*N_TOT_LANCI_PER_NODO + (numSuccessore*2+1) <<"\n";
    return nodoRitornato(lancioSceltaNodoEffettivo,1);
  }
  else{
    //scelta del nodo da cui copiare il successore
    long lancioNodoDiRiferimento   = normalizzaLancio (r.genRand( (long long)nodoSorgente*N_TOT_LANCI_PER_NODO),0,nodoSorgente-1 );
    //cout<<"NODO = "<<numSuccessore<<" - esimo SUCCESSORE DEL NODO "<<lancioNodoDiRiferimento<<" - n lancio "<<nodoSorgente*N_TOT_LANCI_PER_NODO<< "\n";
    //cout<<"VEDI SUCC n"<<numSuccessore<<" del nodo "<<lancioNodoDiRiferimento<<"\n";
    return nodoRitornato(lancioNodoDiRiferimento,2);
  }

}

// ##########################################################################################################################################################

// ##########################################################################################################################################################

// ##########################################################################################################################################################



//-------------------------------------------------------------------
//                            problemi 
//-------------------------------------------------------------------
//  perdita di precisione nella funzione di calcolo funzione random
//-------------------------------------------------------------------

long long  copyModel(     long layer,long numTotLayer,randomGen2& generaRandom,
                    long numNodi,long numSuccMassimi,
                    char* nomeFileNodi,long numFileIn,
                    char* nomeFileArchiOut,long& numFilesOut,
                    float probNuovo){

       //controllo di coerenza
       if (numSuccMassimi>numNodi) numSuccMassimi=numNodi;

       long long conta=1000000;
       long long nTotArchi=0;


       //apre in lettura il file tempo del layer
       multiFileWriter f_grafo = multiFileWriter(nomeFileArchiOut,DIM_MAX);
       multiFileReader f_tempo_nodi_read=multiFileReader(nomeFileNodi,numFileIn,false);//file dei nodi - da non cancellare


       // -------------------------------------------------------
       //    per ogni nodo del layer corrente calcolo gli archi
       // -------------------------------------------------------

       if (numNodi==0){
              float percParz = 100.0;
              float percTot  = 100.0 * (percParz/100.0+layer )/numTotLayer;
              printf("%c|%6.2f%s |%8ld |%6.2f%s |%11ld |%12lld |",13,percTot,"%",layer,percParz,"%",(long)0, nTotArchi);
              fflush(stdout);
       }


       for (long nodoCorrente=0;nodoCorrente<numNodi;nodoCorrente++){

          //leggo le info corrispondenti del nodo corrente (num reale nodo e num successori da creare)
          infoNodoSucc info;
          if (f_tempo_nodi_read.read(&info,sizeof(info),1)<0){cout<<"Unexpected fatal error"; FINE; }

          //limito il num di succ del nodo corrente...
          if (info.nSucc>=numNodi)info.nSucc=numNodi;

          //check programma
          if (info.nSucc>numSuccMassimi){cout<<"Unexpected fatal error"; FINE; }

          //insieme distinto dei successori del nodo corrente
          insieme succ=insieme(info.nSucc);

//          cout<<"\n-----------------------\nNodo corrente "<<nodoCorrente<<" - num succ precalc = "<<info.nSucc<<":::: ";


          //controllo anche che il nodo 0 non abbia troppi nodi (0 e cosi' per gli altri....)
          for (long n=0;n<info.nSucc && n<nodoCorrente;n++){
//              cout<<"\n succ "<<n<<"::::  ";
              //var di appoggio
              nodoRitornato ris;
              long nodo=nodoCorrente;


              //ripeto la ricerca finche' non trovo un nodo o nulla
              do{
                 //cout<<"\nPARTE DAL NODO "<<nodo<<": \n";
                 ris = calcolaLinkUscente(nodo, n, generaRandom, numSuccMassimi, probNuovo);
                 
                 nodo=ris.nodoDest;//eventuale nodo da cui riprendere la ricerca - se trovato successoreDi ...
              }
              while (ris.estSuccessoreDi() );

              //cout<<" "<<ris.nodoDest;


              //se e' un nodo lo inserisco in output
              if (ris.estNodo()){

                //nuovo nodo successore
                long nuovoSucc=ris.nodoDest;
                 //controllo se non e' gia' presente...
                //controllo che non ci sia gia' lo stesso successore
                if (succ.getPosizioneElemento(nuovoSucc)<0){

                   //aggiungo alla lista degli elementi
                   succ.aggiungiElemento(nuovoSucc);
                }
              }
          }


          //scrive l'insieme dei successori
          long* successori;
          long sizeSucc=succ.getVettore(successori);
//          cout<<" - succ trovati "<<sizeSucc<<"\n";

          for (long i=0;i<sizeSucc;i++){

             //scrive gli archi e li converte automaticamente
             arco a;
             a.source=nodoCorrente;
             a.dest=successori[i];

             if (a.source!=a.dest){
                 f_grafo.write(&a,sizeof(a),1);//scrivo sul file temporaneo
                 //aggiorna contatori archi generale e per layer
                 nTotArchi++;
             }
           }



          //cancella l'insieme
          succ.cancella();

          //STAMPA
          if ( (conta++>=300000) || (nodoCorrente>=numNodi-1)  ){
              float percParz = 100.0 * (nodoCorrente+1 )/numNodi;
              float percTot  = 100.0 * (percParz/100.0+layer )/numTotLayer;
              printf("%c|%6.2f%s |%8ld |%6.2f%s |%11ld |%12lld |",13,percTot,"%",layer,percParz,"%",nodoCorrente+1, nTotArchi);
              conta=0;fflush(stdout);
          }



       }
       // -----------------------------------------
       //    FINE CREAZIONE PER IL LAYER CORRENTE
       // -----------------------------------------

       //chiude il file in lettura
       f_tempo_nodi_read.close();

       //chiude il file tempo scritto
       numFilesOut=f_grafo.close();

       return nTotArchi;
}

