/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file distribuzionePotenza.cpp :source code of layermodel utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
/***************************************************************************
                          distribuzionePotenza.cpp  -  description
                             -------------------
    begin                : Wed Nov 27 2002
    copyright            : (C) 2002 by root
    email                : root@travelcat
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "distribuzionePotenza.h"




//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
// calcola il valore della funzione zeta limitatamente ai primi maxIter
//
//	      maxIter
//		 +----
//		  \      (esp)
//		  /    n
//		 +----
//		  n=0
//
//---------------------------------------------------------------------------------
double distrPotenza::zeta(double esp, long maxIter){
  //cout<<"Esp "<<esp<<" maxIter "<<maxIter<<"\n";
  double val=0;
  for (long i=1;i<=maxIter;  i++ ){
	val += pow( (double)i,esp);
  }
  return val;
}
//---------------------------------------------------------------------------------



//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
// costruttore - scelta per nodo
//---------------------------------------------------------------------------------
distrPotenza::distrPotenza(long numNodi, double esp, long maxDegree,long seed){
  srand(seed);

//  double zArchi = zeta(esp+1,maxDegree);
  double zNodi  = zeta(esp,maxDegree);

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //coeff necessario per avere una legge a potenza e un num di nodi (o archi) pari a quello desiderato
  coeff= 1.0*(numNodi)/zNodi;

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // stampa num valori attesi
  //long64 numArchiPrevisti   = (long64)( coeff * zArchi );

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //calcola la dim dell'array per contenere tutti i dati (NODI) in modo da avere il num di archi voluti
  sizeDistr=1;
  long numNodiAssegnati=0;
  long i=1;
  while (numNodiAssegnati<numNodi){
	sizeDistr++;
	numNodiAssegnati+=(long)ceil(coeff*pow(i,esp));
	i++;
  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //					alloca la struttura per contenere i dati
  //alloca struttura portante
  this->distr=new long[sizeDistr];
  this->degree=new long[sizeDistr];

  //calcola size dei livelli superiori
  size3=sizeDistr/RAGG3;	if (sizeDistr%RAGG3!=0 || size3<=0)size3++;
  size2=size3/RAGG2;		if (size3%RAGG2!=0 || size2<=0   )size2++;
  size1=size2/RAGG1;		if (size2%RAGG1!=0 || size1<=0   )size1++;

  //allocazione buffer livelli superiori
  liv3=new long [size3];
  liv2=new long [size2];
  liv1=new long [size1];
  for (i=0;i<size3;liv3[i++]=0);
  for (i=0;i<size2;liv2[i++]=0);
  for (i=0;i<size1;liv1[i++]=0);

  //aggiunge i vari nodi in tutte le strutture
  numElementiRimasti=0;
  degree[0]=distr[0]=0;
  for (i=1;i<sizeDistr;i++){
	  distr[i]=(long)ceil(coeff*pow(i,esp));
	  liv3[i/RAGG3            ]+=distr[i];
	  liv2[i/RAGG3/RAGG2      ]+=distr[i];
	  liv1[i/RAGG3/RAGG2/RAGG1]+=distr[i];
	  degree[i]=i;
	  numElementiRimasti+=distr[i];
  }

}
//---------------------------------------------------------------------------------








//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
// costruttore - scelta per archi
//---------------------------------------------------------------------------------
distrPotenza::distrPotenza(long numNodi, long numSuccPerNodo, double esp, long maxDegree,long seed){
  srand(seed);

  double zArchi = zeta(esp+1,maxDegree);
  double zNodi  = zeta(esp,maxDegree);

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //coeff necessario per avere una legge a potenza e un num di nodi (o archi) pari a quello desiderato
  coeff= 1.0*((long64)numNodi*numSuccPerNodo)/zArchi;

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // stampa num valori attesi
  long64 numArchiPrevisti   = (long64)( coeff * zArchi );
  //cout<<"archi previsti "<<numArchiPrevisti<<"\n";

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // calcola la dim array per contenere tutti i dati in modo da avere il num di nodi voluti
  sizeDistr=1;
  long64 numArchiAssegnati=0;
  long   numNodiPrevisti=0;
  long i=1;
  while ( numArchiAssegnati<(long64)(numNodi*numSuccPerNodo ) || (numNodiPrevisti<numNodi) ){
  	sizeDistr++;
    long val=(long64)ceil(coeff*pow(i,esp));
	  numArchiAssegnati+=val*i;
    numNodiPrevisti+=val;
  	i++;
  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //					alloca la struttura per contenere i dati

  //alloca struttura portante
  this->distr=new long[sizeDistr];
  this->degree=new long[sizeDistr];

  //calcola size dei livelli superiori
  size3=sizeDistr/RAGG3;	if (sizeDistr%RAGG3!=0 || size3<=0)size3++;
  size2=size3/RAGG2;		if (size3%RAGG2!=0 || size2<=0   )size2++;
  size1=size2/RAGG1;		if (size2%RAGG1!=0 || size1<=0   )size1++;

  //allocazione buffer livelli superiori
  liv3=new long [size3];
  liv2=new long [size2];
  liv1=new long [size1];
  for (i=0;i<size3;liv3[i++]=0);
  for (i=0;i<size2;liv2[i++]=0);
  for (i=0;i<size1;liv1[i++]=0);

  //assegna i vari nodi
  degree[0]=distr[0]=0;
  numElementiRimasti=0;
  for (i=1;i<sizeDistr;i++){
	 distr[i]=(long)ceil(coeff*pow(i,esp));
	 liv3[i/RAGG3            ]+=distr[i];
	 liv2[i/RAGG3/RAGG2      ]+=distr[i];
	 liv1[i/RAGG3/RAGG2/RAGG1]+=distr[i];
	 degree[i]=i;
	 numElementiRimasti+=distr[i];
  }
}
//---------------------------------------------------------------------------------






//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//    stampa dati
//---------------------------------------------------------------------------------
void distrPotenza::stampa(){
  if (distr==NULL) return;
  long somma=0;
  long archi=0;
  for (int j=0;j<sizeDistr;j++){
		somma+=distr[j];
		archi+=distr[j]*degree[j];
  }


  int i;
  for (i=0;i<size1;i++){
	  cout<<liv1[i]<<"\t";
  }
  cout<<"\n";

  for (i=0;i<size2;i++){
	  cout<<liv2[i]<<"\t";
  }
  cout<<"\n";

  for (i=0;i<size3;i++){
	  cout<<liv3[i]<<"\t";
  }
  cout<<"\n";

  for (i=0;i<sizeDistr;i++)cout<<distr[i]<<"\t";cout<<"\n";
  for (i=0;i<sizeDistr;i++)cout<<"["<<degree[i]<<"]\t";cout<<"\n";

  cout<<"\nAvailable nodes: "<<somma<<" - Available edges = "<<archi<<" - vector size "<<sizeDistr<<"\n";
  cout<<"\n"<<"available nodes at the end = "<<numElementiRimasti<<"\n";
}




//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//    dealloca le strutture dati
//---------------------------------------------------------------------------------
void distrPotenza::cancella(){
  if (distr==NULL) return;
  delete[] distr;distr=NULL;
  delete[] degree;degree=NULL;
  delete[] liv1;liv1=NULL;
  delete[] liv2;liv2=NULL;
  delete[] liv3;liv3=NULL;
}
//---------------------------------------------------------------------------------




//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//    ritorna l'outdeg corrente se possibile
//---------------------------------------------------------------------------------
long distrPotenza::getOutDeg(){

   // controllo che vi siano altri nodi
   if (sizeDistr==0) return 0;
   if (numElementiRimasti<=0) return 0;

   long degreeTrovato=0;

   long valoreLancio=rand()%numElementiRimasti+1;//valore tra 1 e numElementiRimasti
//   cout<<"\nLancio "<<valoreLancio<<"\n";

   //=======================================================================
   //scandisce il liv 1 (+ elevATO)
   int i=0;
   for (; i<size1 ;i++){
	   if (valoreLancio>liv1[i])  valoreLancio-=liv1[i];
   	   else					  break;//fine ricerca liv corrente
   }

   if (size1 <= i){
	   cout<<"Error: liv1 i,size = "<<i<<","<<size1<<"\n";FINE;}
//   cout<<"indice Liv 1 = "<<i<<"(rimasto "<<valoreLancio<<")\n";
   //=======================================================================


   //=======================================================================
   //scandisce il secondo livello
   i=i*RAGG1;
   for (; i<size2 ;i++){
	   if (valoreLancio>liv2[i])  valoreLancio-=liv2[i];
		else					  break;//fine ricerca liv corrente
   }

   if (size2<=i){cout<<"Error: liv2\n";FINE;}
//   cout<<"indice Liv 2 = "<<i<<"(rimasto "<<valoreLancio<<")\n";
   //=======================================================================


   //=======================================================================
   //scandisce il terzo livello
   i=i*RAGG2;
   for (; i<size2 ;i++){
	   if (valoreLancio>liv3[i])  valoreLancio-=liv3[i];
		else					  break;//fine ricerca liv corrente
   }

   if (size3<=i){cout<<"Error: liv3\n";FINE;}
//   cout<<"indice Liv 3 = "<<i<<"(rimasto "<<valoreLancio<<")\n";
   //=======================================================================


   //=======================================================================
   //scandisce l'ultimo livello
   i=i*RAGG3;
   for (; i<sizeDistr ;i++){
	   //cout<<" "<<distr[i];
	   if (valoreLancio>distr[i])  valoreLancio-=distr[i];
		else					  break;//fine ricerca liv corrente
   }

   if (sizeDistr<=i){cout<<"Error: ending liv\n";FINE;}
//   cout<<"indice Liv Finale = "<<i<<"(rimasto "<<valoreLancio<<")\n";
   //=======================================================================

   //elimina il nodo considerato
   distr[i]--;
   liv3[i/RAGG3            ]--;
   liv2[i/RAGG3/RAGG2      ]--;
   liv1[i/RAGG3/RAGG2/RAGG1]--;
   numElementiRimasti--;

   //memorizza il degree trovato
   degreeTrovato=degree[i];
//   cout<<"Degree "<<degreeTrovato<<"\n";

   //controlla se distr per l'outdeg corrente e' vuota e se si la porto alla fine
   if (distr[i]==0 && sizeDistr>0){
//	   cout<<"scambio tra "<<i<<" e "<<sizeDistr-1<<"\n";
	   long pos1=sizeDistr-1;
     long val1=distr[pos1];

	   long pos2=i;
     //long val2=distr[pos2];//vale 0 x def

	   //primo valore eleminato dalla pos1
     distr[pos1                  ]-=val1;
     liv3 [pos1/RAGG3            ]-=val1;
     liv2 [pos1/RAGG3/RAGG2      ]-=val1;
     liv1 [pos1/RAGG3/RAGG2/RAGG1]-=val1;

	   //secondo valore eliminato dalla pos2
	   //VALE 0 e non serve

	   //secondo valore scritto nella pos 1
	   //VALE 0 e non serve

	   //primo valore scritto nella pos 2
     distr[pos2                  ]+=val1;
     liv3 [pos2/RAGG3            ]+=val1;
     liv2 [pos2/RAGG3/RAGG2      ]+=val1;
     liv1 [pos2/RAGG3/RAGG2/RAGG1]+=val1;

	   //scambia i degree
	   long app=degree[pos1];
	   degree[pos1]=degree[pos2];
	   degree[pos2]=app;


	   //un elemento in meno
	   sizeDistr--;
   }


   return degreeTrovato;
}






