/*

*********************************************************************
*********************************************************************

  file funzioni.cpp :source code of utl_sccgraphconstructor utility

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi
  email      (millozzi,laura,leon)@dis.uniroma1.it

*********************************************************************
*********************************************************************

* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *

*********************************************************************
*********************************************************************

*/

#include "funzioni.h"

//#define  PRINT_INFO_NODES_SCC

/**********************************************************************************************************
***********************************************************************************************************
 Prende i multiFiles delle SCC e riporta in output una lista di record,
 che contiene <nodo, sizeSCC, tipoSCC, numSCC>
 Le scc di size < di minimumSccSize non vengono considerate nella numerazione (semplicemente vengono ignorate)
 e per i nodi che vi facevano parte si riporta scc= -1. Non tutti i nodi potrebbero essere presenti...

  FUNZIONE USATA IN etichettaNodi(....)
***********************************************************************************************************
**********************************************************************************************************/
long PRIVATE_create_LabelInfoNodeScc_Multifile(
    char* sccINFO, char* sccSUCC,//info / succ files
    multiFileWriter out,
    long minimumSccSize,
    long initialSccNumber//prima SCC da analizzare
  ){

  printf("Finding SCC from files %s and %s (single tree) and writing in %s\n",sccINFO,sccSUCC,out.getMultiFileName() );fflush(stdout);
  multiFileReader info = multiFileReader(sccINFO,-1,false);//info file
  multiFileReader succ = multiFileReader(sccSUCC,-1,false);//succ file

  infoTree auxInfoTree;//usato per leggere il campo INFO delle SCC
  long     auxNodoScc;//legge i nodi SUCC
  long     numCurrentScc=initialSccNumber;//labe scc corrente (inizia dal num passato come parametro)
  long     numeroTotaleNodiDaAssegnare=getTotalSizeMultiFiles(sccSUCC)/sizeof(long);//num tot nodi da assegnare - usato per le statistiche

  timer t=timer(2);
  long contaNodi=0;

  //mentre ci sono altre scc da leggere
  while (info.read(&auxInfoTree,sizeof(infoTree),1)>0){
    //printf("Scc %ld -- num children= %ld\n",numCurrentScc,auxInfoTree.numberOfChildren);
    contaNodi++;

    //leggo tutti i nodi della scc corrente
    for (long i=0;i<auxInfoTree.numberOfChildren;i++){
      if (succ.read(&auxNodoScc,sizeof(long),1)<=0){printf("Format error in scc multifiles. (nodes %ld in the scc %ld)\n",auxInfoTree.numberOfChildren,numCurrentScc);FINE;}

      //crea un nuovo nodo da riportare nel file temporaneo. Nel caso di componenti + piccole del minimo si riporta -1 (no componente)
      labelInfoNodeScc auxLabelInfoNodeScc;
      auxLabelInfoNodeScc.node = auxNodoScc;
      auxLabelInfoNodeScc.sizeScc = auxInfoTree.numberOfChildren;
      if (auxInfoTree.numberOfChildren<minimumSccSize) auxLabelInfoNodeScc.scc=-1;//scc troppo piccola...
      else auxLabelInfoNodeScc.scc  = numCurrentScc;//scc regolare

      //scrive in output il nuovo nodo,scc
      if (out.write(&auxLabelInfoNodeScc,sizeof(auxLabelInfoNodeScc),1)<1){perror("Write error.\n");FINE;}

      t.print("%c [%6.2f%c] Scanning %ld nodes (current scc: %ld)              ",(char)13,100.0*contaNodi/numeroTotaleNodiDaAssegnare,'%', contaNodi,numCurrentScc);
    }

    //incremento solo se la scc corrente era valida...
    if (auxInfoTree.numberOfChildren >= minimumSccSize)
      numCurrentScc++;
  }

  t.forcePrint("%c [%6.2f%c] Written %ld nodes                         \n",(char)13,100.0,'%', contaNodi);
  // IL FILE finale E' ANCORA COME LISTA DI COPPIE <nodo,scc>; va riordinato per nodo ed estratti solo i valori SCC.

  info.close();
  succ.close();
  return numCurrentScc+1;//primo num SCC libero

}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/






/**********************************************************************************************************
***********************************************************************************************************
  trasforma il formato da labelInfoNodeScc a lista di SCC
  (per ogni nodo "n" nella posizione "n" c'e' la scc in cui esso si trova)
  Nel caso di nodi non assegnati a nessuna SCC riporta -1 come scc e aggiunge eventuali nodi per
  completare quelli mancanti alla fine...

  x ogni nodo (posizione nel file) -->  <scc, sccType, sccSize>

  Tutti i nodi del grafo vengono aggiunti (eventualmente si aggiungono in mezzo e alla fine nodi con scc=-1)
  In questa fase solo i valori scc sono validi; gli altri sono inizializzati a labelSCC (CORE)

  FUNZIONE USATA IN etichettaNodi(....)
***********************************************************************************************************
**********************************************************************************************************/
void PRIVATE_trasforma_labelInfoNodeScc2InfoNodeScc(
             char* inMultiFile,
             char* temp,
             multiFile& out,//file contenente
             long numNodiGrafo,
             long available_RAM
             ){

  printf("Convert from labelInfoNodeScc to InfoNodeScc (without labels) the file %s\n\n",inMultiFile);fflush(stdout);
  char tempExternalSort[300]; sprintf(tempExternalSort,"%s_tempExternalSort",    temp);
  char tempSortLabel[300];    sprintf(tempSortLabel,   "%s_tempSortedLabelInfoSccNodes",temp);
  long numeroTotLabel=getTotalSizeMultiFiles(inMultiFile)/sizeof(labelInfoNodeScc);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   ordina le coppie nodo,scc per nodo
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  printf("Sorting %ld labels (%0.2f KB) from file %s to %s ...\n",numeroTotLabel,1.0*numeroTotLabel*sizeof(labelInfoNodeScc)/1024, inMultiFile, tempSortLabel );
  fflush(stdout);
  externalSort_labelInfoNodeScc(inMultiFile, tempExternalSort, tempSortLabel,
                      true,(char*)"Sorting nodes",available_RAM,64*1024,4,
                      64*1024,10,30,false);


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   crea la lista delle scc del nodo corrispondente alla posizione in long
   controllando allo stesso tempo che tutti i nodi siano considerati...
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  long currentNode=0;
  labelInfoNodeScc auxLabelInfoNodeScc;
  long numNodiNonAssegnati=0;
  long numNodiConPiuSCC=0;
  timer t=timer(2);

  multiFileReader label    = multiFileReader(tempSortLabel,-1,true);//label SORTED multifile
  infoNodeScc auxInfoNodeScc;

  //legge il nodo corrente
  while ( label.read(&auxLabelInfoNodeScc,sizeof(auxLabelInfoNodeScc),1)>0){

      //mentre ci sono ancora vuoti tra i nodi...
      while (currentNode<auxLabelInfoNodeScc.node){
        auxInfoNodeScc.scc=-1;
        auxInfoNodeScc.sizeScc=-1;
        auxInfoNodeScc.sccType=labelScc_NULL;//nodi nulli
        if (out.write(&auxInfoNodeScc,sizeof(infoNodeScc),1)<1){perror("Write error.\n");FINE;}
        currentNode++;
        numNodiNonAssegnati++;
      }

      //se ho trovato nel file ordinato un nodo < del corrente allora vuol dire che e' ripetuto in piu' scc !!!
      if (currentNode>auxLabelInfoNodeScc.node){
        numNodiConPiuSCC++;
        continue;//se + scc x lo stesso nodo, lo assegno solo alla prima incontrata !!!
      }

      //if (currentNode!=auxLabelInfoNodeScc.node){printf("Found  node %ld without a component.\n",currentNode);FINE;}

      auxInfoNodeScc.scc=auxLabelInfoNodeScc.scc;
      auxInfoNodeScc.sizeScc=auxLabelInfoNodeScc.sizeScc;
      auxInfoNodeScc.sccType=labelScc_SCC;

      if (out.write(&auxInfoNodeScc,sizeof(infoNodeScc),1)<1){perror("Write error.\n");FINE;}
      currentNode++;

      t.print("%c [%6.2f%c] Scanning %ld nodes                 ",(char)13,100.0*(currentNode+1)/numNodiGrafo,'%', currentNode+1);
  }

  //aggiunge altri nodi alla fine per arrivare a riempire il grafo
  while (currentNode<numNodiGrafo){
        auxInfoNodeScc.scc=-1;
        auxInfoNodeScc.sizeScc=-1;
        auxInfoNodeScc.sccType=labelScc_NULL;//nodi nulli
        if (out.write(&auxInfoNodeScc,sizeof(infoNodeScc),1)<1){perror("Write error.\n");FINE;}
        currentNode++;
        numNodiNonAssegnati++;
  }

  t.forcePrint("%c [%6.2f%c] Scanning %ld nodes\n",(char)13,100.0,'%', currentNode);

  if (numNodiConPiuSCC>0 || numNodiNonAssegnati>0)printf("Warning:\n");
  if (numNodiConPiuSCC>0)   {printf(" [x] %ld assignments to nodes already assigned to some SCC (left first assignment)\n",numNodiConPiuSCC);}
  if (numNodiNonAssegnati>0){printf(" [x] %ld nodes are not assigned to any SCC (assigned the scc number -1)\n",numNodiNonAssegnati);}


  label.close();
//  outFinal.close();
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/












/**********************************************************************************************************
***********************************************************************************************************
  Prende un file (out) nel formato InfoNodeScc con sccType non ancora scritti e per tutti i nodi
  contenuti nel set, fissa il valore del tipo a "label". Usata insieme a trasforma_labelInfoNodeScc2InfoNodeScc()
  permette di settare correttamente i type

  FUNZIONE USATA IN etichettaNodi(....)
***********************************************************************************************************
**********************************************************************************************************/
void PRIVATE_aggiungiLabel_InfoNodeScc_NEW(multiFile& fOUT ,char* setMultiFile , char* temp ,long label, long available_RAM){


    long numeroTotNodi=getTotalSizeMultiFiles(setMultiFile)/sizeof(long);
    char tempSortedNodes[300]; sprintf (tempSortedNodes ,"%s_tempSortedNodes", temp);//file ordinato dei nodi del set
    char tempExternalSort[300];sprintf (tempExternalSort,"%s_tempExternalSort",temp);//tempo external sort

    //ordina i set (di long) senza cancellarli... - NON CAMBIARE false !!!
    printf("  Sorting %ld nodes (%0.2f KB) of current set ...\n",numeroTotNodi,1.0*numeroTotNodi*sizeof(long)/1024);fflush(stdout);
    externalSort_long(setMultiFile, tempExternalSort, tempSortedNodes,
                      false,(char*)"Sorting nodes",available_RAM,64*1024,4,
                      64*1024,10,30,false);
    //printf("%c                                                        ",13);
    //printf("!!!! %ld \n",getTotalSizeMultiFiles(tempSortedNodes));

    //aggiunge labels
    multiFileReader mf = multiFileReader(tempSortedNodes,-1,true);//legge il file set ordinato

    long currentNode;//nodo nel set corrente
    long numNodi_TipoSccAssegnatoPiuVolte_SccValide=0;
    long numNodi_TipoSccAssegnatoSccNonValide=0;
    long numNodi_TipoSccAssegnatoPiuVolte_SccNonValide=0;

    //gestione a blocchi del file
    #define SIZE_BLOCCHI_InfoNodeScc (200)
    infoNodeScc buffer_InfoNodeScc[SIZE_BLOCCHI_InfoNodeScc];
    infoNodeScc* auxInfoNodeScc  = NULL;
    long long posSeekBufferCorrente=0;
    long        buffer_primoNodo = 0;
    long        buffer_dim       = 0;
    timer t=timer(2);
    long contaNodi=0;

    while (mf.read(&currentNode,sizeof(long),1)>0){

      #ifndef DEBUGGING_MODE
      contaNodi++;
      t.print("%c [%6.2f%c] Scanning %ld nodes            ",(char)13,100.0*contaNodi/numeroTotNodi,'%', contaNodi);
      #else
      printf("\n>>>>>>>> Current node %ld >>>>>>>>\n",currentNode);
      #endif

      //controlla se il nodo di trova nel blocco corrente; se no lo carica in ram
      if ( currentNode<buffer_primoNodo || currentNode >= buffer_primoNodo+buffer_dim-1 ){

        //il prec buffer lo riscrive
        if (buffer_dim>0){

          #ifdef DEBUGGING_MODE
          printf("\nSCRITTURA BUFFER ==>> ");
          for (int i=0;i<buffer_dim;i++){
            printf("[%ld] %ld,%c || ",buffer_primoNodo+i, buffer_InfoNodeScc[i].scc,buffer_InfoNodeScc[i].sccType);
          }
          printf("\n");
          #endif

          if (fOUT.seek(posSeekBufferCorrente)<0){perror("Program error 1:");FINE;}
          if (fOUT.write(buffer_InfoNodeScc, sizeof(infoNodeScc), buffer_dim )<buffer_dim){perror("Program error 2:");FINE;}

        }

        //legge il nuovo buffer
        buffer_primoNodo=currentNode;

        if (fOUT.seek( (long long)currentNode*(long long)sizeof(infoNodeScc) )<0){perror("Program error 3:");FINE;}
        posSeekBufferCorrente=fOUT.tell();
        if ( (buffer_dim = fOUT.read(buffer_InfoNodeScc, sizeof(infoNodeScc), SIZE_BLOCCHI_InfoNodeScc  )) <=0){printf("Nodo %ld ",currentNode);perror("Program error 4:");FINE;}

        #ifdef DEBUGGING_MODE
        printf("\nLETTURA BUFFER ==>> ");
        for (int i=0;i<buffer_dim;i++){
          printf("[%ld] %ld,%c || ",buffer_primoNodo+i, buffer_InfoNodeScc[i].scc,buffer_InfoNodeScc[i].sccType);
        }
        printf("\n");
        #endif

      }


      //nodo corrente
      auxInfoNodeScc=&buffer_InfoNodeScc[currentNode-buffer_primoNodo];

      #ifdef DEBUGGING_MODE
      printf("OLD InfoNodeScc scc=%ld,type=%c,size=%ld,[CurrentNode=%ld]",auxInfoNodeScc->scc,auxInfoNodeScc->sccType,auxInfoNodeScc->sizeScc,currentNode);
      #endif

      //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      //se sovrascrivo una label gia' assegnata (escluso il caso dei tubes) incremento il num di ripetizioni
      //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      if ( auxInfoNodeScc->scc<0 && (auxInfoNodeScc->sccType==labelScc_NULL) ){
        numNodi_TipoSccAssegnatoSccNonValide++;//scc non valida - prima assegmazione
      }

      else if ( auxInfoNodeScc->scc<0 )
        numNodi_TipoSccAssegnatoPiuVolte_SccNonValide++;//scc non valida - ulteriori assegnazioni oltre la prima

      else if ( (auxInfoNodeScc->sccType!=labelScc_SCC) && !(auxInfoNodeScc->sccType==labelScc_TENDRILS && label==labelScc_TUBES) )
        numNodi_TipoSccAssegnatoPiuVolte_SccValide++;//scc valida - ulteriori assegnazioni oltre la prima, escluso il caso tubes/tendrils
      //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      //in ogni caso modifico la label
      auxInfoNodeScc->sccType=label;

      #ifdef DEBUGGING_MODE
      printf("NEW InfoNodeScc scc=%ld,type=%c,size=%ld,[CurrentNode=%ld]",auxInfoNodeScc->scc,auxInfoNodeScc->sccType,auxInfoNodeScc->sizeScc,currentNode);
      #endif
    }

    //salva l'ultimo buffer
    //il prec buffer lo riscrive
    if (buffer_dim>0){

       #ifdef DEBUGGING_MODE
       printf("\nSCRITTURA BUFFER FINALE ==>> ");
       for (int i=0;i<buffer_dim;i++){
         printf("[%ld] %ld,%c || ",buffer_primoNodo+i, buffer_InfoNodeScc[i].scc,buffer_InfoNodeScc[i].sccType);
       }
       printf("\n");
       #endif

       if (fOUT.seek(posSeekBufferCorrente)<0){perror("Program error 1:");FINE;}
       if (fOUT.write(buffer_InfoNodeScc, sizeof(infoNodeScc), buffer_dim )<buffer_dim){perror("Program error 2:");FINE;}
       fOUT.flush();

    }

    t.forcePrint("%c [%6.2f%c] Scanning %ld nodes ",(char)13,100.0,'%', contaNodi);
    printf("\n");

    if (numNodi_TipoSccAssegnatoSccNonValide>0 || numNodi_TipoSccAssegnatoPiuVolte_SccNonValide>0 || numNodi_TipoSccAssegnatoPiuVolte_SccValide>0)
      printf("Warning:\n");

    if (numNodi_TipoSccAssegnatoSccNonValide>0 || numNodi_TipoSccAssegnatoPiuVolte_SccNonValide>0){
      printf(" [x] %ld label assigments to nodes with scc number -1 (not assigned in any scc) divided in: \n",numNodi_TipoSccAssegnatoSccNonValide+numNodi_TipoSccAssegnatoPiuVolte_SccNonValide);
      printf("    -> %ld SINGLE label assigments to invalid sccs \n",numNodi_TipoSccAssegnatoSccNonValide);
      printf("    -> %ld MULTIPLE label assigments to invalid sccs (trying to assign more than 1 scc type to a node)\n",numNodi_TipoSccAssegnatoPiuVolte_SccNonValide);
    }

    if (numNodi_TipoSccAssegnatoPiuVolte_SccValide>0)
      printf(" [x] %ld MULTIPLE label assigments to nodes with valid scc value\n",numNodi_TipoSccAssegnatoPiuVolte_SccValide);
    fflush(stdout);

    mf.close();
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/









/**********************************************************************************************************
***********************************************************************************************************
  Etichetta ciascun nodo con le info <scc,sccType,sccSize> secondo il formato infoNodeScc e in base
  ai files su SCC e ai set li arricchisce di informazioni
***********************************************************************************************************
**********************************************************************************************************/
void etichettaNodi_infoNodeScc(
        char* sccINFO, char* sccSUCC,
        char* grafoINFO,
        char* INset,char* OUTset, char* DISCset, char*TENDRILSset, char* TUBESset,
        char* temp,
        long MAX_DIM_FILE,long memoria,
        long minSccSize, char*outputSccRegion){



    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      crea il file in cui per ogni nodo ho la scc in cui si trova (labelInfoNOdeScc)
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    char tempLabelInfoNodeScc[300];//file che contiene le scc x ogni nodo come lista di long
    sprintf(tempLabelInfoNodeScc,"%s_tempLabelInfoNodeScc",temp);

    multiFileWriter out  = multiFileWriter(tempLabelInfoNodeScc,MAX_DIM_FILE);//succ file
    PRIVATE_create_LabelInfoNodeScc_Multifile(sccINFO, sccSUCC,out, minSccSize, 0);
    printf("\n");
    out.close();



    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      crea il file in cui per ogni nodo ho la scc in cui si trova (lista di scc)
      aggiugendo eventualmente i nodi non considerati in nessuna SCC
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    multiFile fOUT = multiFile(outputSccRegion,MAX_DIM_FILE);//file delle label
    if ( !fOUT.estOpen() ){perror("Impossible create output file");FINE};

    long numGraphNodes = getTotalSizeMultiFiles(grafoINFO)/(long long)sizeof(infoNodo);
    PRIVATE_trasforma_labelInfoNodeScc2InfoNodeScc(tempLabelInfoNodeScc,temp, fOUT, numGraphNodes, memoria);
    printf("\n");

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      aggiunge label corretta per i nodi della regione specificata
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    #define NOME_PROC PRIVATE_aggiungiLabel_InfoNodeScc_NEW
    printf("Adding label %c for all nodes in the file %s\n",labelScc_IN,INset);fflush(stdout);
    NOME_PROC(fOUT ,INset       ,temp ,labelScc_IN       ,memoria);
    printf("\n");

    printf("Adding label %c for all nodes in the file %s\n",labelScc_OUT,OUTset);fflush(stdout);
    NOME_PROC(fOUT ,OUTset      ,temp ,labelScc_OUT      ,memoria);
    printf("\n");

    printf("Adding label %c for all nodes in the file %s\n",labelScc_DISC,DISCset);fflush(stdout);
    NOME_PROC(fOUT ,DISCset     ,temp ,labelScc_DISC     ,memoria);
    printf("\n");

    printf("Adding label %c for all nodes in the file %s\n",labelScc_TENDRILS,TENDRILSset);fflush(stdout);
    NOME_PROC(fOUT ,TENDRILSset ,temp ,labelScc_TENDRILS ,memoria);
    printf("\n");

    printf("Adding label %c for all nodes in the file %s\n",labelScc_TUBES,TUBESset);fflush(stdout);
    NOME_PROC(fOUT ,TUBESset    ,temp ,labelScc_TUBES    ,memoria);
    printf("\n");

    fOUT.seek(0);


    #ifdef PRINT_INFO_NODES_SCC
    long nodo=0;
    infoNodeScc auxInfoNodeScc;
    fprintf(stderr,"---- LABELED NODES ----\n");
    while (fOUT.read(&auxInfoNodeScc,sizeof(infoNodeScc),1)>0){
      fprintf(stderr,"node %2ld - scc %2ld - type %c - size %ld\n",nodo,auxInfoNodeScc.scc,auxInfoNodeScc.sccType,auxInfoNodeScc.sizeScc);
      nodo++;
    }
    fprintf(stderr,"--------------\n");
    #endif

    fOUT.close();
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/






/**********************************************************************************************************
***********************************************************************************************************
  Per ogni scc (posizione del record) ho la size e
  Tutte le scc sono considerate, e nel caso qualcuna non abbia nodi, si riporta solo se interna ad altre
  che hanno size>0
***********************************************************************************************************
**********************************************************************************************************/
void calcolaLabelSCC_InfoScc(
                     char* fileInfoNodeScc,//file contenente i record InfoNodeScc
                     char* temp,
                     long MAX_DIM_FILE,long available_RAM,
                     char* labelSccGraph//ritorna il file delle label
                     ){


    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      ordina i risultati per SCC in modo che xogni SCC conosco sia size e tipo
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    char tempExternalSort[300];      sprintf(tempExternalSort,"%s_tempExternalSort",    temp);
    char tempSortedInfoNodeScc[300]; sprintf(tempSortedInfoNodeScc,   "%s_tempSortedInfoNodeScc_PerScc",temp);
    printf("Sorting infoNodeScc elements (%0.2f KB) from file %s to %s ...\n",1.0*getTotalSizeMultiFiles(tempSortedInfoNodeScc)/1024, fileInfoNodeScc, tempSortedInfoNodeScc);
    fflush(stdout);
    externalSort_InfoNodeScc(fileInfoNodeScc, tempExternalSort, tempSortedInfoNodeScc,
                      false,(char*)"Sorting nodes",available_RAM,64*1024,4,
                      64*1024,10,30,false);



    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      crea un file label delle SCC nel formato infoSCC a partire dal file
      tempSortedInfoNodeScc
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    multiFileWriter fw= multiFileWriter(labelSccGraph,MAX_DIM_FILE);
    multiFileReader fr= multiFileReader(tempSortedInfoNodeScc,-1,true);
    infoNodeScc auxInfoNodeScc;

    long numTotalNodesToRead=getTotalSizeMultiFiles(tempSortedInfoNodeScc)/sizeof(infoNodeScc);

    //var di gestione dati corrente SCC
    infoScc currentInfoScc;
    long currentScc_SCC=-1;
    currentInfoScc.sizeScc=0;
    currentInfoScc.sccType=labelScc_NULL;
    timer t=timer(2);
    long contaNodi=0;

    printf("\nCreate label grapfSCCs file %s from %s\n", labelSccGraph , tempSortedInfoNodeScc);

    //mentre ci sono ancora info sugli SCC
    while (fr.read(&auxInfoNodeScc,sizeof(infoNodeScc),1)>0){
      //stampa
      contaNodi++;
      t.print("%c [%6.2f%c] Searching SCCs info (%ld scc found) ",(char)13,100.0*contaNodi/numTotalNodesToRead,'%', currentScc_SCC);

      //le scc <0 sono relative a nodi senza una scc assegnata e vengono ignorati !!!
      if (auxInfoNodeScc.scc<0) continue;

      //controlla se errore nel programma...
      if (auxInfoNodeScc.scc<currentScc_SCC){printf("Program error in calcolaLabelSCC_InfoScc(...)");FINE;}

      //se la SCC e' la stessa del passo precedente semplicemente aumento la size di 1
      else if (auxInfoNodeScc.scc==currentScc_SCC){/*printf("Same SCC\n");*/currentInfoScc.sizeScc++;}

      //se  una successiva, scrivo quella corrente e inserisco SCC vuote fino alla nuova letta...
      else {
        //scrivo la SCC corrente e passo alla successiva
        if (currentScc_SCC>=0){
          if (fw.write(&currentInfoScc,sizeof(infoScc),1)<1){perror("Write error on SCC lael file.");FINE;}
        }
        currentScc_SCC++;

        //mentre non arrivo alla SCC letta aggiungo scc vuote
        while (auxInfoNodeScc.scc>currentScc_SCC){
          currentInfoScc.sccType=labelScc_NULL;
          currentInfoScc.sizeScc=0;
          if (fw.write(&currentInfoScc,sizeof(infoScc),1)<1){perror("Write error on SCC label file.");FINE;}
          currentScc_SCC++;
        }

        //la nuova SCC e' quella correntemente letta
        currentInfoScc.sizeScc=1;
        currentInfoScc.sccType=auxInfoNodeScc.sccType;
      }

    }

    //controlla l'ultima SCC e la inserisco solo se non vuota
    if (currentInfoScc.sizeScc>0){
        if (fw.write(&currentInfoScc,sizeof(infoScc),1)<1){perror("Write error on SCC lael file.");FINE;}
        currentScc_SCC++;//una SCC in piu'
    }

    //stampa finale
    t.forcePrint("%c [%6.2f%c] Searching SCCs info (%ld scc found) \n",(char)13,100.0,'%', currentScc_SCC);

    fr.close();
    fw.close();

    // NOTA: l'ultima scc viene scritta solamente se esiste almeno un nodo che ne fa parte; questo garantisce
    //       che il num di SCC sia quello calcolato. SCC senza nodi si possono avere solo all'interno
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/







/**********************************************************************************************************
***********************************************************************************************************
  Funzioni di appoggio per la gestione delle sostituzioni dei nodi con le scc corrispondenti...
***********************************************************************************************************
**********************************************************************************************************/

/*
  Ad ogni nodo viene sostituito un valore negativo a partire dal valore -2. Ogni volta che si effettua una
  sostituzione di un nodo con la sua corrispondente SCC il valore sostituito puo' essere >=0 se la scc  valida
  oppure -1 se non e' valida (x questo il val -1 viene riservato)
*/

long trasformaNodo(long nodo){
  return -nodo - 2;
}

//un nodo e' trasfromato solo se assume un valore inferiore a -1 altrimenti o e' una scc corretta (>=0) o invalida (-1)
bool estNodoTrasformato(long nodo){
  if (nodo<-1) return true;
  else return false;//-1 resta valore speciale
}

//a partire dal nodo trasformato ricava quello di partenza (esclusione per -1 e per i valori che non risultano trasformati...)
long anti_trasformaNodo(long nodo){
  if (!estNodoTrasformato(nodo)) return nodo;
  else return -nodo - 2;
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/





/**********************************************************************************************************
***********************************************************************************************************
  Calcola il grafo delle SCC nel formato liste di archi, con l'aggiunta del file .label contentente x ogni
  scc, la size e il tipo
***********************************************************************************************************
**********************************************************************************************************/

// legge le prossime scc dal file in formato infoNodeScc fino ad un massimo di numElemBuffer elementi
// USATA internamente a calcolaGrafoSCC_NEW
long  leggiBufferSCC(multiFileReader& fInfoNodeScc,long* buffer,long numElemBuffer){

      //legge le info relative alle scc a partire dal nodoIniziale corrente
      long size=100;//dim buffer d'appoggio
      infoNodeScc* auxInfoNodeScc=new infoNodeScc[size];

      //variabili ausiliarie
      long numLetti=0;//num elem letti in totale
      long numLettiUltimoCiclo;//num elementi letti nell'ultimo ciclo

      do{
        //calcola il num di elementi da leggere (che ancora entrano nel buffer)
        long daLeggere=numElemBuffer-numLetti;
        if (daLeggere>size) daLeggere=size;

        //leggo quanti piu' elementi possibile
        if (daLeggere>0)
          numLettiUltimoCiclo = fInfoNodeScc.read(auxInfoNodeScc,sizeof(infoNodeScc),daLeggere);
        else
          numLettiUltimoCiclo = 0;

        //individuo la scc associata
        for (int i=0; i<numLettiUltimoCiclo; i++,numLetti++ ){
          buffer[numLetti]=auxInfoNodeScc[i].scc;//scc corrente relativa al nodo corrente
        }

      }
      while (numLettiUltimoCiclo>0);


      delete []auxInfoNodeScc;
      return numLetti;
}




/**********************************************************************************************************
***********************************************************************************************************
  Calcola il grafo delle SCC nel formato liste di archi, con l'aggiunta del file .label contentente x ogni
  scc, la size e il tipo
***********************************************************************************************************
**********************************************************************************************************/

//ritorna, se possibile, la scc associata al nodo nella var "dato".
//  Se "dato" e' una scc, ritorna lo stesso valore;
//  se "dato" e' un nodo che non fa parte del buffer, ritorna il nodo
//  se "dato" e' un nodo dell'intervallo, ritorna  -scc-2 (da -2 in sotto; -1 e' lasciato come scc non valida)
long PRIVATE_ottieni_sccTrasformato_da_nodo(long dato, long* hashing, long primoNodoBuffer, long ultimoNodoBuffer){

  //gia' SCC
  if (dato<0) return dato;
  //nodo non compreso nell'intervallo
  else if ( dato<primoNodoBuffer || dato>ultimoNodoBuffer ) return dato;
  //ritorna la scc (-1 se la scc non e' valida)
  else  return hashing[dato-primoNodoBuffer]<0 ? hashing[dato-primoNodoBuffer] : -2 - hashing[dato-primoNodoBuffer];

}

//ritorna, il valore non trasformato della scc a partire da quello trasformato
long PRIVATE_ottieni_sccNormale_da_sccTrasformata(long scc){

  if (scc==-1) return scc;
  else if (scc>=0) {
    printf("Program error in ottieni_sccNormale_da_sccTrasformata(...) -- value = %ld\n",scc);
    FINE;
    return scc;
  }
  else return -scc-2;
}


//  x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x
// x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x
//  x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x x
void calcolaGrafoSCC(
        char* sccINFO, char* sccSUCC,
        char* grafoINFO,char* grafoSUCC,
        char* INset,char* OUTset, char* DISCset, char*TENDRILSset, char* TUBESset,
        char* temp,
        long MAX_DIM_FILE,long available_RAM,
        long minSccSize, char* sccGraph){



    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      etichetta i nodi del grafo con le info sulle SCC
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    //calcola x ogni nodo la SCC corrispondente
    char tempInfoNodeScc[300];sprintf(tempInfoNodeScc,"%s_tempInfoNodeScc_PerNode",temp);
    etichettaNodi_infoNodeScc(sccINFO, sccSUCC,grafoINFO,
        INset,OUTset, DISCset, TENDRILSset, TUBESset,
        temp, MAX_DIM_FILE,available_RAM,minSccSize, tempInfoNodeScc);//ritorna le label per nodo. Si assicura che x tutti i nodi viene considerata la scc... (eventualmente -1)



    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      etichetta le varie SCC e crea il labelFile
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    printf("--------------------\n Creating SCC graph \n--------------------\n");fflush(stdout);

    //calcola per ogni scc le info che la caratterizzano
    char labelSccGraph[300];sprintf(labelSccGraph,"%s.labelSccs",sccGraph);
    calcolaLabelSCC_InfoScc(
                     tempInfoNodeScc,//file contenente i record InfoNodeScc
                     temp,
                     MAX_DIM_FILE,available_RAM,
                     labelSccGraph//ritorna il file delle label per ogni SCC
                     );


    #ifdef PRINT_INFO_NODES_SCC
    // stampa info per scc
    fprintf(stderr,"---- SCC ----\n");
    multiFileReader fr= multiFileReader(labelSccGraph,-1,false);
    infoScc currentInfoScc;
    long currentScc_SCC=0;
    while (fr.read(&currentInfoScc,sizeof(currentInfoScc),1)>0){
      fprintf(stderr,"scc %ld - size %ld - type %c\n",currentScc_SCC,currentInfoScc.sizeScc,currentInfoScc.sccType);
      currentScc_SCC++;
    }
    fr.close();
    fprintf(stderr,"-------------\n");
    #endif


    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      legge le SCC dei nodi a blocchi e scandisce tutto il file tempo e sost alle
      occorrenze dei nodi incontrate con le scc corrispondenti (se possibile)
      Per i nodi senza SCC si sostituisce -1
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    //conteggio num nodi e num scc
    long numeroSccTrovate=getTotalSizeMultiFiles(labelSccGraph)/sizeof(infoScc);
    long  numNodiGrafo      = getTotalSizeMultiFiles(grafoINFO)/sizeof(infoNodo);

    //file con le info sulle SCC associate ai nodi (x ogni nodo) ed elimino il file
    multiFileReader fInfoNodeScc=multiFileReader(tempInfoNodeScc,-1,true);

    //allocazione buffer di hashing
    long  sizeBufferHashing = (long)(available_RAM/sizeof(long))>numNodiGrafo ? numNodiGrafo : (available_RAM/sizeof(long));
    //sizeBufferHashing=3;

    long* hashing            = new long[sizeBufferHashing];
    if (sizeBufferHashing<=0){printf("Memory not enough\n");FINE;}

    //per ogni blocco di hashing
    long numNodiAnalizzati=0;
    long step=0;

    //conteggio degli archi
    long long archiInterniSCC=0;
    long long archiNonValidi =0;
    printf("\n");

    for (long numPrimoNodo=0;numPrimoNodo<numNodiGrafo;numPrimoNodo+=numNodiAnalizzati, step++){

      long numUltimoNodo= numPrimoNodo+sizeBufferHashing >= numNodiGrafo? numNodiGrafo-1 : numPrimoNodo+sizeBufferHashing;
      printf("STEP %ld - Reading nodes for hashing",step);

      //legge le prossime scc dal file in formato infoNodeScc fino ad un massimo di numElemBuffer elementi
      numNodiAnalizzati=leggiBufferSCC(fInfoNodeScc,hashing,sizeBufferHashing);
      numUltimoNodo=numPrimoNodo+numNodiAnalizzati-1;
      //printf("%cstep %ld - Read %ld nodes [%ld - %ld]",13,step,numNodiAnalizzati,numPrimoNodo,numUltimoNodo);

      #ifdef PRINT_INFO_NODES_SCC
      fprintf(stderr,"----- hashing ------");
      fprintf(stderr,"%4s --> %4s\n","nodo","scc");
      for (int i=0;i<numNodiAnalizzati;i++){
        fprintf(stderr,"%4ld --> %4ld\n",i+numPrimoNodo,hashing[i]);
      }
      fprintf(stderr,"--------------------\n");
      #endif

      //salta alla fine se non ho letto altri nodi
      if (numNodiAnalizzati==0) continue;

      //nome multifile corrente
      char nomeNuovoMultiFileStepCorrente[1000];
      sprintf(nomeNuovoMultiFileStepCorrente, "%s.edges_step%ld",temp,step);
      multiFileWriter scrittura=multiFileWriter(nomeNuovoMultiFileStepCorrente,MAX_DIM_FILE);

      //apertura file per la lettura degli archi precedenti
      accessoVirtualeArchi* lettura=NULL;
      if (step==0){
        //lettura dal file originale
        lettura=new accessoVirtualeArchi_IPS(grafoINFO,grafoSUCC);
        printf("%cSTEP %ld - Hashing nodes from IPS files %s, %s to %s  \n", 13, step, grafoINFO, grafoSUCC, nomeNuovoMultiFileStepCorrente);
      }
      else{
        //nome multifile corrente
        char nomePrecFile[1000];
        sprintf(nomePrecFile, "%s.edges_step%ld", temp, step-1);
        lettura = new accessoVirtualeArchi_EDGES (nomePrecFile);
        printf("%cSTEP %ld - Hashing nodes from EDGES file %s to %s  \n", 13, step, nomePrecFile, nomeNuovoMultiFileStepCorrente);
      }


      //mentre ci sono archi da leggere
      #define DIM_BUFFER_OUT  (1000)
      arco  bufferOut[DIM_BUFFER_OUT];
      long  posLibera_bufferOut=0;//dim buffer occupato
      arco* arcoCorrente;

      //contatori archi
      long long archiDaAnalizzare=lettura->getNumeroArchiTotali();
      long long numArchiAnalizzati=0;

      //contatore stampa
      timer t=timer(2);

      while (  (arcoCorrente=lettura->getNuovoArco())!=NULL ){
        numArchiAnalizzati++;
        //printf("arco %d %d \n",arcoCorrente->source,arcoCorrente->dest);

        #ifndef PRINT_INFO_NODES_SCC
        t.print("%c [%6.2f%c] Hashing %ld nodes [%ld-%ld] in %s",
                (char)13,100.0*numArchiAnalizzati/archiDaAnalizzare,'%',
                numUltimoNodo-numPrimoNodo+1, numPrimoNodo, numUltimoNodo, nomeNuovoMultiFileStepCorrente);
        #endif


        arco app;
        app.source = PRIVATE_ottieni_sccTrasformato_da_nodo (arcoCorrente->source , hashing, numPrimoNodo, numUltimoNodo );
        app.dest   = PRIVATE_ottieni_sccTrasformato_da_nodo (arcoCorrente->dest   , hashing, numPrimoNodo, numUltimoNodo );        

        #ifdef PRINT_INFO_NODES_SCC
        fprintf(stderr, "edge from %ld --> %ld   to  %ld --> %ld\n",arcoCorrente->source,arcoCorrente->dest, app.source, app.dest);
        #endif


        //controlla se l'arco tra le scc deve essere riportato o meno
        if (app.source==-1 || app.dest==-1) {archiInterniSCC++;continue;}//almeno una delle due scc non e' valida --> skip
        else if (app.source<0 && app.dest<0 && app.source==app.dest ) {archiNonValidi++;continue;}//non scrivo i link interni alle scc

        //se buffer pieno lo svuota
        if (posLibera_bufferOut>=DIM_BUFFER_OUT){
          if (scrittura.write(bufferOut,sizeof(arco), posLibera_bufferOut) != posLibera_bufferOut){perror("Program error n.1");FINE;}
          posLibera_bufferOut=0;
        }

        bufferOut[posLibera_bufferOut]=app;
        posLibera_bufferOut++;
      }

      //scrive gli ultimi dati
      if (posLibera_bufferOut>0){
        if (scrittura.write(bufferOut,sizeof(arco), posLibera_bufferOut) != posLibera_bufferOut){perror("Program error n.2");FINE;}
        posLibera_bufferOut=0;
      }

      //chiusura file
      scrittura.close();
      lettura->close();
      delete[] lettura;

      t.forcePrint("%c [%6.2f%c] Hashing %ld nodes [%ld-%ld] in %s\n",
                (char)13,100.0*numArchiAnalizzati/archiDaAnalizzare,'%',
                numUltimoNodo-numPrimoNodo+1, numPrimoNodo, numUltimoNodo, nomeNuovoMultiFileStepCorrente);

      //printf("%cstep %ld - [%6.2f%c] Hashing %ld nodes in %s \n\n",(char)13, step, 100.0, '%', numUltimoNodo-numPrimoNodo+1,nomeNuovoMultiFileStepCorrente);

      //elimina il file temporaneo dello step precedente se esiste
      if (step>0){
        char nomePrecFile[1000];
        sprintf(nomePrecFile, "%s.edges_step%ld", temp, step-1);
        multiFile::removeMultiFile(nomePrecFile);
      }

      printf("\n");

    }

    delete[] hashing;
    fInfoNodeScc.close();

    //CONVERSIONE FINALE

    //nome ultimo file step precedente
    char nomeMultiFileIn[1000];
    sprintf(nomeMultiFileIn, "%s.edges_step%ld",temp,step-1);
    multiFileReader f_in=multiFileReader(nomeMultiFileIn,-1,true);//elimina il file alla fine

    //nome file output
    char nomeMultiFileOut[1000];
    sprintf(nomeMultiFileOut, "%s.edges",sccGraph);
    multiFileWriter f_out=multiFileWriter(nomeMultiFileOut,MAX_DIM_FILE);

    printf("STEP %ld - Create scc graph from %s to %s\n",step,nomeMultiFileIn,nomeMultiFileOut);
    long letti;
    long long archiScritti=0;
    arco  buf[DIM_BUFFER_OUT];

    #ifdef PRINT_INFO_NODES_SCC
    fprintf(stderr,"----- SCC graph in edges format -----\n");
    #endif

    long long archiTotali=getTotalSizeMultiFiles(nomeMultiFileIn)/sizeof(arco);
    timer t=timer(2);

    while ( (letti=f_in.read(buf,sizeof(arco),DIM_BUFFER_OUT))>0){
      for (int i=0;i<letti;i++){
        buf[i].source=PRIVATE_ottieni_sccNormale_da_sccTrasformata(buf[i].source);
        buf[i].dest=  PRIVATE_ottieni_sccNormale_da_sccTrasformata(buf[i].dest);

        #ifdef PRINT_INFO_NODES_SCC
        fprintf(stderr,"%ld %ld \n",buf[i].source,buf[i].dest);
        #endif

      }
      if (f_out.write(buf,sizeof(arco),letti)<letti){perror ("Program error n.4");FINE;}
      archiScritti+=letti;

      t.print("%c [%6.2f%c] Converting graph to edges (%lld edges) ",(char)13,100.0*archiScritti/archiTotali,'%', archiScritti);
    }

    t.forcePrint("%c [%6.2f%c] Converting graph to edges (%lld edges)\n",(char)13,100.0,'%', archiScritti);


    #ifdef PRINT_INFO_NODES_SCC
    fprintf(stderr,"-----   -----\n");
    #endif

    arco appo;appo.source=appo.dest=-numeroSccTrovate;
    if (f_out.write(&appo,sizeof(arco),1)<1){perror ("Program error n.5");FINE;}

    f_in.close();
    f_out.close();

    printf("\nReport:\n >> %lld loops\n >> %lld invalid edges\n >> %lld valid edges (edges of Scc graph)\n",archiInterniSCC,archiNonValidi,archiScritti);
    printf("Sccs found : %ld (nodes of Scc graph)\n",numeroSccTrovate);
    fflush(stdout);

}

/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/



#include <map>
using namespace std;

/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/
void  calcolaDistribuzione( char* sccINFO,  char* sccSUCC,  char* grafoINFO,
             char* INset,  char* OUTset,  char* DISCset,  char* TENDRILSset,
             char* TUBESset,  char* temp, long MAX_DIM_FILE, long memoria,  char* report){

  typedef map<long, long> mappa;
  mappa m;//mappa vuota

  FILE* f=fopen(report,"w");
  if (f==NULL){perror("");FINE;}


  //==================================================
  multiFileReader mf = multiFileReader(sccINFO,-1);
  const long n=1000;
  infoTree buffer[n];
  long letti;

  map<long, long>::iterator p;
  
  while ( (letti=mf.read(buffer,sizeof(infoTree),N))>0){

    for (long i=0;i<letti;i++){
       long sizeSccCorrente= buffer[i].numberOfChildren;
       p=m.find(sizeSccCorrente);
       if (p==m.end())
         m[sizeSccCorrente]=1;
       else
         m[sizeSccCorrente]++;
    }        
  }
  
  mf.close();
    

  fprintf(f,"%s\t%s\n","size","num sccs");    
  for (map<long, long>::iterator i=m.begin(); i!=m.end(); i++){
    fprintf(f,"%ld\t%ld\n",i->first,i->second);    
  }

  cout<<"Report written in file "<<report<<"\n";


  long long numNodiGrafo=  (getTotalSizeMultiFiles(grafoINFO)/sizeof(infoNodo));
  long long in=            (getTotalSizeMultiFiles(INset)/sizeof(long));
  long long out=           (getTotalSizeMultiFiles(OUTset)/sizeof(long));
  long long tendrils=      (getTotalSizeMultiFiles(TENDRILSset)/sizeof(long));
  long long tubes=         (getTotalSizeMultiFiles(TUBESset)/sizeof(long));
  long long disc=          (getTotalSizeMultiFiles(DISCset)/sizeof(long));
  long long core=numNodiGrafo-in-out-tendrils-disc;
    
  fprintf(f,"\n");
  fprintf(f,"Report Bow-tie regions:\n");
  fprintf(f," - Number of graph nodes  :\t%lld\n",numNodiGrafo);
  fprintf(f,"\n");
  fprintf(f," - CORE region size       :\t%lld\n",core);
  fprintf(f," - IN region size         :\t%lld\n",in);
  fprintf(f," - OUT region size        :\t%lld\n",out);
  fprintf(f," - TENDRILS region size   :\t%lld\n",tendrils);
  fprintf(f," - TUBES(*) region size   :\t%lld\n",tubes);
  fprintf(f," - DISC region size       :\t%lld\n",disc);
  fprintf(f,"(*) TUBES is a subset of tendrils\n");
  
  fprintf(stdout,"\n");
  fprintf(stdout,"Report Bow-tie regions:\n");
  fprintf(stdout," - Number of graph nodes  :\t%lld\n",numNodiGrafo);
  fprintf(stdout,"\n");
  fprintf(stdout," - CORE region size       :\t%lld\n",core);
  fprintf(stdout," - IN region size         :\t%lld\n",in);
  fprintf(stdout," - OUT region size        :\t%lld\n",out);
  fprintf(stdout," - TENDRILS region size   :\t%lld\n",tendrils);
  fprintf(stdout," - TUBES(*) region size   :\t%lld\n",tubes);
  fprintf(stdout," - DISC region size       :\t%lld\n",disc);
  fprintf(stdout,"(*) TUBES is a subset of tendrils\n");

  fclose(f);
  
            
}
/**********************************************************************************************************
**********************************************************************************************************/
/**********************************************************************************************************
**********************************************************************************************************/


