
#include "funzioni.h"

void stampaHelp(int arg,char*argv[]){
/*		cout<< "This utility labels each graph node with the following informations:\n";
    cout<< " number of SCC, location of the SCC (IN/OUT/DISC/TENDRILS), size of SCC \n";*/
    cout<< "This utility uses 3 switches: \n";
		cout<< "\n";
    cout<< "1) "<<argv[0]<<" memory -LABEL INset%d OUTset%d TENDRILSset%d TUBESset%d DISCset%d scc%d graph%d Temp%d \n";    
    cout<< "   to create a label file (<graph%d>.labelNode) where for each node the following information are stored: \n";
    cout<< "   <number of scc, scc type, scc size> (infoNodeScc internal type name).\n";
    cout<< "\n";
    cout<<" 2) "<<argv[0]<<" memory -GRAPH INset%d OUTset%d TENDRILSset%d TUBESset%d DISCset%d scc%d graph%d Temp%d minimumSize sccGraph%d\n";
    cout<< "   to create the SCC graph in edges format (<output%d>.edges). Moreover <output%d>.labelSccs is created with the following information for each scc (vertex): <scc type, scc size> (infoScc internal type name) where size >=minimumSize.\n";
    cout<<"\n";
    cout<<" 3) "<<argv[0]<<" memory -REPORT INset%d OUTset%d TENDRILSset%d TUBESset%d DISCset%d scc%d graph%d Temp%d ReportFile_TXT \n";
    cout<< "   to generate a general report that contains the SCCs distribution and the size of each Bow-Tie region.\n";
    cout<<"\n";
    cout<< "parameters:\n";
    cout<< " memory           available memory in MB\n";
    cout<< " INset%d          list of nodes in the IN region of Bow Tie(as long)\n";
    cout<< " OUTset%d         list of nodes in the OUT region of Bow Tie(as long)\n";
    cout<< " TENDRILSset%d    list of nodes in the TENDRILS region of Bow Tie(as long)\n";
    cout<< " TUBESset%d       list of nodes in the TUBES (subset of TENDRILS) region of Bow Tie(as long)\n";
    cout<< " DISCset%d        list of nodes in the DISC region of Bow Tie(as long)\n";
    cout<< " scc%d            scc multifile where are stored all scc of the graph (in tree format)\n";
    cout<< " graph%d          input graph multifile\n";
    cout<< " temp%d           temporary multifile\n";
    cout<< " minimumSize      minimum size of valid scc (else are reported as -1)\n";
    cout<< " sccGraph%d       output multifile in edge format\n";
    cout<< " ReportFile_TXT   output report filename\n";
    cout<< " \n";
    cout<< " All nodes not part of some specified set (IN/OUT/TUBES/TENDRILS/DISC) are assigned to the CORE region\n";
    arg=0;argv=NULL;
}





/**********************************************************************************************************
***********************************************************************************************************
                                                     main
***********************************************************************************************************
**********************************************************************************************************/
int main(int arg, char *argv[])
{

    stampaLicenza(arg,argv);
    if (arg<11) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");
    long memoria=(long)atof(argv[1])*1024*1024;
    if (memoria<=1024*1024){cout<<"Memory parameter error or memory not enough!";EXIT(0);}

    char* param       = argv[2];
    char* INset       = argv[3];
    char* OUTset      = argv[4];
    char* TENDRILSset = argv[5];
    char* TUBESset    = argv[6];
    char* DISCset     = argv[7];
    char* scc         = argv[8];
    char* grafo       = argv[9];
    char* temp        = argv[10];
    

    char grafoINFO[300]; strcpy (grafoINFO,grafo);strcat(grafoINFO,".info");
    char grafoPRED[300]; strcpy (grafoPRED,grafo);strcat(grafoPRED,".pred");
    char grafoSUCC[300]; strcpy (grafoSUCC,grafo);strcat(grafoSUCC,".succ");
    char sccINFO  [300]; strcpy (sccINFO,scc);strcat(sccINFO,".infotree");//info file of SCC
    char sccSUCC  [300]; strcpy (sccSUCC,scc);//SCC files (list of sccs)


    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       controllo correttezza nomi/size dei files
    - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    long MAX_DIM_FILE=1*1024*1024*1024;//1*1024*1024*1024;

    //controllo nome multifile corretto dei files di input
    for (int i=3;i<=10;i++){
       if (!estMultiFileName(argv[i])){
           char app[300];sprintf(app,"Multifile name %s must contain the substring '%s'",argv[i], "%d");
           STAMPA_ERRORE_PARAMETRI(argv[0],app);
       }
    }

    //controllo sizes dei files di input
    if (getTotalSizeMultiFiles(grafoINFO)<=0){cout<<"The input graph is empty ("<<grafoINFO<<")\n";FINE};
    if (getTotalSizeMultiFiles(grafoPRED)<=0){cout<<"The input graph is empty ("<<grafoPRED<<")\n";FINE};
    if (getTotalSizeMultiFiles(grafoSUCC)<=0){cout<<"The input graph is empty ("<<grafoSUCC<<")\n";FINE};    
    if (getTotalSizeMultiFiles(sccINFO)<=0){cout<<"The multiFile "<<sccINFO<<" is empty\n";FINE};
    if (getTotalSizeMultiFiles(sccSUCC)<=0){cout<<"The multiFile "<<sccSUCC<<" is empty\n";FINE};

    
    /************************************************************************************************/
//    long minSccSize=minimumSize;

    
    if (strcasecmp(param,"-LABEL")==0){
        //-------------------------------------------------
        //  preleva i parametri non comuni
        //-------------------------------------------------
        if (arg!=11) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");
        char* outputGraph = grafo;//file finale che per ogni nodo contiene SCC e regione
        long minSccSize=0;
        //-------------------------------------------------

        //-------------------------------------------------
        //  avvia elaborazione
        //-------------------------------------------------
        cout<<"Labeling graph nodes, considering all SCCs\n";
        char app[300];sprintf(app,"%s.labelNodes",outputGraph);
        etichettaNodi_infoNodeScc(sccINFO, sccSUCC,grafoINFO,
            INset,OUTset, DISCset, TENDRILSset, TUBESset,
            temp, MAX_DIM_FILE,memoria,minSccSize, app);
        //-------------------------------------------------
    }
    else if (strcasecmp(param,"-GRAPH")==0){
        //-------------------------------------------------
        //  preleva i parametri non comuni
        //-------------------------------------------------
        if (arg!=13) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");
        long  minSccSize = (long)atof(argv[11]);
        char* outputGraph = argv[12];//file finale che per ogni nodo contiene SCC e regione

        if (!estMultiFileName(outputGraph)){
          char app[300];sprintf(app,"Multifile name %s must contain the substring '%s'",outputGraph, "%d");
          STAMPA_ERRORE_PARAMETRI(argv[0],app);
        }
        //-------------------------------------------------

        //-------------------------------------------------
        //  avvia elaborazione
        //-------------------------------------------------
        char app[1000];sprintf(app," Creating SCC graph, considering all SCCs with size >=%ld\n",minSccSize);
        for (long i=0;i<(long)strlen(app);i++,printf("-"));printf("\n%s",app);
        for (long i=0;i<(long)strlen(app);i++,printf("-"));printf("\n");
        
        calcolaGrafoSCC(
          sccINFO, sccSUCC,
          grafoINFO,grafoSUCC,
          INset,OUTset, DISCset, TENDRILSset, TUBESset,
          temp,
          MAX_DIM_FILE,memoria,
          minSccSize, outputGraph);
        //-------------------------------------------------
    }
    else if (strcasecmp(param,"-REPORT")==0){
        //-------------------------------------------------
        //  preleva i parametri non comuni
        //-------------------------------------------------
        if (arg!=12) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");
        char* outputGraph = argv[11];//file finale che per ogni nodo contiene SCC e regione
        //-------------------------------------------------

        //-------------------------------------------------
        //  avvia elaborazione
        //-------------------------------------------------
        cout<<"Reporting SCCs...\n";
        calcolaDistribuzione(sccINFO, sccSUCC,grafoINFO,
            INset,OUTset, DISCset, TENDRILSset, TUBESset,
            temp, MAX_DIM_FILE, memoria, outputGraph);
        //-------------------------------------------------
    }
    else{printf("Incorrect switch\n");FINE;}

    EXIT(0);
}
/**********************************************************************************************************
**********************************************************************************************************/

