#include "BufferLetturaBits.h"
#include <stdlib.h>

BufferLetturaBits::~BufferLetturaBits()
{
  if(buf_input != NULL)
    delete buf_input;
  
  if(file_input != NULL)
    fclose(file_input);
}



BufferLetturaBits::BufferLetturaBits(const char * nome_file_bits, unsigned long num_elem, unsigned long dim_memoria_buffer)
{
  init(nome_file_bits, num_elem, dim_memoria_buffer, NULL); //non specifichiamo un buffer che ci e' stato passsato, lo facciamo riallocare
}



BufferLetturaBits::BufferLetturaBits(const char * nome_file_bits, unsigned long num_elem, unsigned long dim_memoria_buffer, unsigned char * punt_buffer_passato)
{
  init(nome_file_bits, num_elem, dim_memoria_buffer, punt_buffer_passato);
}



//ritorna false se c'e' stato un errore, true se OK
void BufferLetturaBits::init(const char * nome_file, unsigned long num_el, unsigned long dim_memoria_buffer, unsigned char * punt_buffer_passato)
{
  numero_elementi_vettore_bits= num_el;

  file_input= NULL;
  buf_input= NULL;

  file_input= fopen(nome_file, "r");
  if(file_input == NULL) //errore nell'aprire il file
  {
    fprintf(stderr, "BufferLetturaBits: Memory allocation error.\n");
    abort();
  }

  if(punt_buffer_passato == NULL)
    buf_input= new BufferLetturaFile<unsigned char>(file_input, dim_memoria_buffer);
  else //usiamo il buffer che ci e' stato passato
    buf_input= new BufferLetturaFile<unsigned char>(file_input, dim_memoria_buffer, punt_buffer_passato);

  //inizializziamo le variabili usate per scorrere il file:
  indice_ultimo_bit_letto= 0; //questo valore iniziale in pratica permettere di leggere tutti i bit, a partire dal primo che ha indice zero
  num_byte_letti= 0;
}



bool BufferLetturaBits::get(unsigned long indice_bit_richiesto, long & errore)
{
  unsigned long byte_richiesto= indice_bit_richiesto / 8;
  unsigned long pos_nel_byte= indice_bit_richiesto % 8;

  errore= 1; //errore per default
  if((indice_bit_richiesto >= numero_elementi_vettore_bits) || (indice_bit_richiesto < indice_ultimo_bit_letto))
    return true; //valore arbitrario

  unsigned long distanza= byte_richiesto + 1 - num_byte_letti;

  if(distanza > 0) //se invece (distanza == 0) ho gia' pronto il byte che mi serve
  {
    bool ok;
    if(distanza > 1) //c'e' almeno un byte di mezzo (che non mi interessa)
    {
      ok= buf_input->skip(distanza - 1);
      if(!ok)
        return true; //valore arbitrario
    }

    ok= (bool) buf_input->get(ultimo_byte_letto);
    if(!ok)
      return true; //valore arbitrario

    num_byte_letti += distanza;
  }
  
  //a questo punto supponiamo che 'ultimo_byte_letto' sia il byte contenente il bit richiesto:  
  unsigned long mask= 1 << pos_nel_byte;
  errore= 0; //ristabiliamo la situazione OK
  indice_ultimo_bit_letto= indice_bit_richiesto;
  return (ultimo_byte_letto & mask); //testiamo il bit richiesto
}


