#include "ListaStatica.h"
#include "nodi_passivi.h"
#include <stdio.h>


const char MSG_NODO_FUORI_RANGE[]= "Error: index if out of range for successor list nodes";


//******
extern long long tot_tempo_elimina_marche;
static clock_t  tag01, tag02; //******variabili che servono solo a prendere i tempi (sono globali solo per non ridichiararle ogni volta)





//alloca il vettore dei puntatori ed il vettore degli adiacenti. I due vettori sono memorizzati in maniera contigua (nello stesso array di memoria che viene effettivamente allocato) in modo da poter rendere disponibile con un apposito swap l'intera memoria allocata (per utilizzarla temporaneamente per altri scopi)
ListaStatica::ListaStatica(unsigned long n_nodi, unsigned long max_ad)
{
  size_vett_punt= n_nodi + 1; //una posizione in piu' per un puntatore fittizio che serve a indicare la fine della lista di adiacenza in modo da maneggiare l'ultimo nodo uniformemente rispetto agli altri
  max_adiacenti= max_ad;
  
  memoria_allocata_lista= new unsigned long[size_vett_punt + max_adiacenti];
  vettore_punt= memoria_allocata_lista; //occupa la prima parte della memoria allocata
  vettore_adiacenze= memoria_allocata_lista + size_vett_punt; //occupa la parte restante della memoria allocata

  for(unsigned long i=0; i<size_vett_punt; i++)
    vettore_punt[i]= 0;    
}


ListaStatica::~ListaStatica()
{
  delete [] memoria_allocata_lista;
}


void ListaStatica::print_adiacenti()
{
  for(unsigned long i=0; i<size_vett_punt-1 ; i++)
  {
    if(i<size_vett_punt-2)
      printf("* Node %lu (%lu neibourgh): ", i, degree_of(i));
    else
      printf("# Node %lu (%lu neibourgh): ", i, degree_of(i));

    if(degree_of(i)>0)
    {
      if(vettore_adiacenze[vettore_punt[i] + 0] == marca_X_adiac)
        printf("X");
      else
        printf("%lu", vettore_adiacenze[vettore_punt[i] + 0]);

      for(unsigned long j=1; j<degree_of(i); j++) //inseriamo solo marche
        if(vettore_adiacenze[vettore_punt[i] + j] == marca_X_adiac)
          printf(", X");
        else
          printf(", %lu", vettore_adiacenze[vettore_punt[i] + j]);
    }
    printf("\n");
  }
}


bool ListaStatica::ricerca_sequenziale_adiacente(unsigned long nodo_src, unsigned long nodo_dest, unsigned long max_adiac_da_esaminare)
{
  for(unsigned long i=vettore_punt[nodo_src] ; i < vettore_punt[nodo_src] + max_adiac_da_esaminare; i++)
    if( vettore_adiacenze[i] == nodo_dest )
      return true;
  
  return false;
}



void ListaStatica::elimina_marche()
{
  unsigned long * punt_pos_inserimento= vettore_adiacenze;
  unsigned long * punt_elem_corrente= vettore_adiacenze;
  unsigned long * fine_lista_corrente;
  unsigned long nodo_successivo= 1;
  
  #ifdef VERBOSE_DEBUG
  printf("\n Marks remove:\n"); print_adiacenti();
  #endif

  tag01= clock();

//  unsigned long * fine_liste= vettore_adiacenze + vettore_punt[size_vett_punt-1];
//  while( *punt_elem_corrente != marca_X_adiac)
//  {
//    punt_elem_corrente++;
//
//    if(punt_elem_corrente >= fine_liste)
//      return; //--> usciamo perche' non c'era nemmeno una marca da eliminare
//  }
//  //se siamo qui e' stata trovata una marca
//
//  punt_pos_inserimento= punt_elem_corrente; //la posizione della prima marca
//  punt_elem_corrente++; //subito oltre la prima marca
//
//  unsigned long indice_prima_marca= punt_pos_inserimento - vettore_adiacenze;
//  while(vettore_punt[nodo_successivo] <= indice_prima_marca)
//  {
//    nodo_successivo++;
//  }
//  fine_lista_corrente= vettore_adiacenze + vettore_punt[nodo_successivo];
  
    
  fine_lista_corrente= vettore_adiacenze + vettore_punt[1];
  
  do
  {
    while(punt_elem_corrente == fine_lista_corrente)
    {
      vettore_punt[nodo_successivo]= punt_pos_inserimento - vettore_adiacenze; //"chiudiamo" la lista corrente

      if(nodo_successivo == size_vett_punt - 1) //abbiamo appena processato e "chiuso" l'ultima lista: possiamo uscire
        return;

      nodo_successivo++;

      fine_lista_corrente= vettore_adiacenze + vettore_punt[nodo_successivo];
    }

    if( *punt_elem_corrente != marca_X_adiac)
    {
      *punt_pos_inserimento= *punt_elem_corrente;
      punt_pos_inserimento++;
    }

    punt_elem_corrente++;
  } while( 1 );


  tag02= clock();  tot_tempo_elimina_marche += tag02 - tag01;
  
  #ifdef VERBOSE_DEBUG
  printf("\n after mark remove:\n"); print_adiacenti();
  #endif  
}



//versione lenta: sposta tutti gli adiacenti piu' a destra
bool ListaStatica::inserisci_elemento(unsigned long nodo_src, unsigned long nodo_dest)
{
  accertati_che( (capienzaMemoriaInutilizzata() >= 1), "Si vogliono inserire nella lista di adiacenza piu' elementi di quanto non sia possibile"); //se si tenta di inserire piu' nodi adiacenti di quanti ne possono ancora entrare nel vettore --> errore
  
  if(ricerca_sequenziale_adiacente(nodo_src, nodo_dest, degree_of(nodo_src))) //non inserito perche' c'e' gia'
    return false;
  //else..continuiamo con l'inserimento
  unsigned long salva_size_adiac= degree_of(nodo_src);

  unsigned long larg_intervallo= vettore_punt[size_vett_punt-1] - vettore_punt[nodo_src+1];
  unsigned long conta_down= larg_intervallo - 1;
  for(unsigned long conta_up=0; conta_up < larg_intervallo ; conta_up++)
  {
    vettore_adiacenze[ vettore_punt[nodo_src+1] + conta_down + 1] = vettore_adiacenze[ vettore_punt[nodo_src+1] + conta_down];
    conta_down--; //conta nello stesso intervallo di 'conta_up', ma nell'ordine inverso. N.B. non ho messo direttamente un 'for' decrescente perche' non so bene come si comporta appena arriva allo zero, deve decrementare, e il numero e' unsigned (temo che vada in loop)    
  }

  for(unsigned long i=nodo_src + 1; i<size_vett_punt; i++)
    vettore_punt[i] = vettore_punt[i] + 1;

  vettore_adiacenze[ vettore_punt[nodo_src] + salva_size_adiac] = nodo_dest;

  return true;
}



//aggiunge il numero richiesto di posizioni alla fine della lista di adiac. del nodo specificato
void ListaStatica::aggiungi_spazio_alla_lista_del_nodo(unsigned long nodo_src, unsigned long spazio_da_riservare)
{
  accertati_che( (capienzaMemoriaInutilizzata() >= spazio_da_riservare), "Si vogliono inserire nella lista di adiacenza piu' elementi di quanto non sia possibile"); //se si tenta di inserire piu' nodi adiacenti di quanti ne possono ancora entrare nel vettore --> errore

  unsigned long copy_from= vettore_punt[nodo_src+1]; //qua iniziano gli elementi da copiare
  unsigned long larg_intervallo= vettore_punt[size_vett_punt-1] - copy_from;

  copia_avanti(copy_from, larg_intervallo, spazio_da_riservare);
  
  for(unsigned long i=nodo_src + 1; i<size_vett_punt; i++)
    vettore_punt[i] += spazio_da_riservare;
}


//rimuovere il numero richiesto di posizioni in fondo alla lista del nodo specificato
void ListaStatica::accorcia_lista_del_nodo(unsigned long nodo_src, unsigned long spazio_da_togliere)
{
  accertati_che( (degree_of(nodo_src) >= spazio_da_togliere), "Si vogliono togliere dalla lista di adiacenza del nodo `%lu' piu' elementi di quanto non sia possibile", nodo_src);

  if(nodo_src < size_vett_punt - 2) //se non e' l'ultimo nodo della lista di adiacenza
    sposta_indietro_liste(spazio_da_togliere, nodo_src + 1);
  else
    vettore_punt[size_vett_punt-1] -= spazio_da_togliere;
}



//copia 'larg_intervallo' locazioni di memoria, partendo da 'copy_from', 'di_quanto_spostare' posizioni piu' avanti
//poiche' il blocco sorgente e destinazione possono sovrapporsi, gli elementi vengono copiati a partire dagli ultimi
void ListaStatica::copia_avanti(unsigned long copy_from, unsigned long larg_intervallo, unsigned long di_quanto_spostare)
{
  unsigned long copy_to= copy_from + di_quanto_spostare; //l'indice a partire dal quale verranno copiato gli elementi da trasferire

  unsigned long conta_down= larg_intervallo - 1;
  for(unsigned long conta_up=0; conta_up < larg_intervallo ; conta_up++)
  {
    vettore_adiacenze[ copy_to + conta_down ] = vettore_adiacenze[ copy_from + conta_down];
    conta_down--; //conta nello stesso intervallo di 'conta_up', ma nell'ordine inverso. N.B. non ho messo direttamente un 'for' decrescente perche' non so bene come si comporta appena arriva allo zero, deve decrementare, e il numero e' unsigned (temo che vada in loop)
  }
}



/* sposta indietro del numero richiesto di posizioni tutti gli adiacenti delle liste dei nodi da 'nodo_from' a 'nodo_to' (compresi)
   N.B. Vengono sovrascritte le posizioni della lista di adiac. del nodo che precede 'nodo_from', che si suppone non siano utilizzate
*/

void ListaStatica::sposta_indietro_liste(unsigned long di_quanto_spostare, unsigned long nodo_from, unsigned long nodo_to)
{
  unsigned long copy_from, copy_to, largh_intervallo;

  accertati_che( (nodo_from<size_vett_punt-1)&&(nodo_to<size_vett_punt-1), MSG_NODO_FUORI_RANGE);

  //shiftiamo a SX tutti i nodi delle liste tra i due nodi "from" e "to" (compresi)
  copy_from= vettore_punt[nodo_from]; //inizio intervallo adiacenti da copiare (estremo incluso)
  accertati_che( (nodo_from > 0)&&(copy_from - di_quanto_spostare >= vettore_punt[nodo_from-1]) , "Passati dei parametri non validi a sposta_indietro_liste()");

  copy_to= copy_from - di_quanto_spostare; //dobbiamo spingere gli elementi selezionati alla sinistra della quantita' di posizioni specificata
  largh_intervallo= vettore_punt[nodo_to + 1] - copy_from;

  for(unsigned long i=0; i<largh_intervallo; i++)
    vettore_adiacenze[ copy_to + i ] =  vettore_adiacenze[ copy_from + i ]; //shiftiamo a SX i nodi nell'intervallo richiesto

  for(unsigned long indice= nodo_from; indice <= nodo_to ; indice++) //aggiustiamo i puntatori delle liste di adiacenza shiftate.
    vettore_punt[indice] -= di_quanto_spostare; //gli adiacenti copiati sono stati mossi all'estremita' SX del buco, quindi sono stati shiftati di tante posizioni quanto era la dimensione del buco
}



// questa versione, rispetto alla precedente con 3 argomenti, sposta tutte le liste da quella del nodo specificato in poi
void ListaStatica::sposta_indietro_liste(unsigned long di_quanto_spostare, unsigned long nodo_from)
{
  sposta_indietro_liste(di_quanto_spostare, nodo_from, size_vett_punt-2); //sposta tutte le liste da quella del nodo specificato in poi N.B. l'ultimo nodo e' size_vett_punt-2 = n_nodi + 1 - 2 = n_nodi - 1
  vettore_punt[size_vett_punt-1] -= di_quanto_spostare; //"chiudiamo" la fine della lista di adiacenza (eliminiamo gli elementi rimasti in fondo all'ultima lista)
}



void ListaStatica::riversa_lista_in_pila_DX_SX(unsigned long nodo_src, SwappedStack<unsigned long> * pila)
{
  unsigned long * punt_adiac;
  unsigned long size_lista_adiac= degree_of(nodo_src);

  punt_adiac= vettore_adiacenze + vettore_punt[nodo_src + 1] - 1; //(!)

  for(unsigned long i=0; i<size_lista_adiac ; i++, punt_adiac--)
    pila->push( *punt_adiac );
}



void ListaStatica::riversa_lista_in_pila_DX_SX(unsigned long nodo_src, SwappedStack<unsigned long> * pila, unsigned long quanti_inserirne)
{
  unsigned long * punt_adiac;
  assert(quanti_inserirne<=degree_of(nodo_src));

  punt_adiac= vettore_adiacenze + vettore_punt[nodo_src + 1] - 1; //(!)

  for(unsigned long i=0; i<quanti_inserirne ; i++, punt_adiac--)
    pila->push( *punt_adiac );
}



void ListaStatica::riversa_lista_in_pila_SX_DX(unsigned long nodo_src, SwappedStack<unsigned long> * pila)
{
  unsigned long * punt_adiac;
  unsigned long size_lista_adiac= degree_of(nodo_src);

  punt_adiac= vettore_adiacenze + vettore_punt[nodo_src];

  for(unsigned long i=0; i<size_lista_adiac ; i++, punt_adiac++)
    pila->push( *punt_adiac );
}



void ListaStatica::riversa_lista_in_pila_SX_DX(unsigned long nodo_src, SwappedStack<unsigned long> * pila, unsigned long quanti_inserirne)
{
  unsigned long * punt_adiac;
  assert(quanti_inserirne<=degree_of(nodo_src));

  punt_adiac= vettore_adiacenze + vettore_punt[nodo_src];

  for(unsigned long i=0; i<quanti_inserirne ; i++, punt_adiac++)
    pila->push( *punt_adiac );
}



inline unsigned long ListaStatica::degree_of(unsigned long nodo)
{
  accertati_che( (nodo<size_vett_punt-1), MSG_NODO_FUORI_RANGE);
  return ( vettore_punt[nodo+1] - vettore_punt[nodo] ) ;
}



unsigned long & ListaStatica::get_adiac(unsigned long nodo, unsigned long indice)
{
  accertati_che( (nodo<size_vett_punt-1), MSG_NODO_FUORI_RANGE);
  accertati_che( (indice<degree_of(nodo)), "Il nodo non ha tanti adiacenti");
  return ( vettore_adiacenze[ vettore_punt[nodo] + indice ] ) ;
}



//Attenzione: il riferimento restituito e' valido finche' non vengono fatte modifiche alla lista di adiacenza
unsigned long * ListaStatica::get_punt_mem_libera()
{
  return vettore_adiacenze + vettore_punt[size_vett_punt-1]; //punta subito dopo l'ultimo elemento nelle liste di adiacenza
}



//ritorna lo spazio (in numero di elementi, cioe' di unsigned long) inutilizzato nel vettore della lista di adiacenza
//Attenzione: la dimensione restituita e' valida finche' non vengono fatte modifiche alla lista di adiacenza
unsigned long ListaStatica::capienzaMemoriaInutilizzata()
{
  return max_adiacenti - ( vettore_punt[size_vett_punt-1] - vettore_punt[0] );
}



//ritorna il numero di elementi totali nelle liste di adiacenza, senza distinguere tra marche e adiacenti normali
unsigned long ListaStatica::numero_adiacenti_totali()
{
  return ( vettore_punt[size_vett_punt-1] - vettore_punt[0] );
}



//aumenta lo spazio riservato ad una serie di liste di adiacenza, della quantita' specificata.
//'vettore_spazi' ha tanti elementi quanti sono complessivamente i nodi nelle liste di adiacenza. Ogni elemento indica quanto spazio aggiungere alla lista di quel nodo.
//'tot_spazio_da_aggiungere' e 'num_liste_da_espandere' sono due info. implicite nel vettore: la somma totale degli spazi da aggiungere (i.e. la somma di tutti gli elementi del vettore), ed il numero di nodi distinti a cui vanno ampliate le liste (i.e. il # di elementi non nulli del vettore)
//si deve specificare se aggiungere lo spazio alle liste interessate all'inizio o in coda.
//Il metodo utilizzato e' questo: se un certo nodo 'u' ha N nodi swappati tra le liste di adiacenza precedono la sua, allora la lista di adiacenza di 'u' andra' spostata in avanti di N posizioni. Per ricavare il # di swappati prima di ciascuna lista consideriamo il loro # totale, procediamo all'indietro e decrementiamo del # di swappati finora incontrati.
void ListaStatica::aumenta_spazio_liste(unsigned long * vettore_spazi, unsigned long tot_spazio_da_aggiungere, unsigned long num_liste_da_espandere, bool inserimento_inizio_lista)
{
  unsigned long nodo_src, last_nodo_src;

  accertati_che(capienzaMemoriaInutilizzata()>=tot_spazio_da_aggiungere, "Errore in aumenta_spazio_liste(): si vorrebbe aggiungere troppo spazio alle liste di adiacenza");
  
  vettore_punt[size_vett_punt-1] += tot_spazio_da_aggiungere; //aggiungiamo in fondo alla lista lo spazio che complessivamente si vuole aggiungere (allarghiamo il puntatore fittizio di fine lista)
  
  //inizializziamo le variabili utilizzate:  
  last_nodo_src= size_vett_punt - 1; //cioe' indica subito oltre l'ultimo nodo (e' il puntatore fittizio finale)
  nodo_src= size_vett_punt - 2; //iniziamo dall'ultimo nodo delle liste di adiac. e procediamo all'indietro
    
  if(inserimento_inizio_lista) //aggiungiamo lo spazio all'inizio della lista
    while(num_liste_da_espandere > 0) //N.B. il parametro 'num_liste_da_espandere' viene decrementato
    {
      //troviamo il prossimo nodo che ha figli swappati (per ipotesi ce n'e' almeno 1):
      while(vettore_spazi[nodo_src] == 0)
        nodo_src--;
      //a questo punto sono posizionato su un elemento non nullo del vettore, me lo garantisce il fatto che e' (num_liste_da_espandere > 0)

      unsigned long spazio_corrente= vettore_spazi[nodo_src];

      //spostiamo le liste di adiac. da quella corrente da espandere (compresa) all'ultima espansa (esclusa); notare che e' (last_nodo_src > nodo_src) per via dell'ordine decrescente di visita
      unsigned long copy_from= vettore_punt[nodo_src]; //dobbiamo spostare a partire dalla prima lista che segue quella da espandere corrente
      unsigned long larg_intervallo= vettore_punt[last_nodo_src] - copy_from - tot_spazio_da_aggiungere; //non dobbiamo ricopiare le locazioni inutilizzate presenti nell'attuale "buco"
      copia_avanti(copy_from, larg_intervallo, tot_spazio_da_aggiungere); //compattiamo i nodi delle liste interessate verso la fine del vettore

      //aggiorniamo i puntatori delle liste appena spostate:
      for(unsigned long i=nodo_src; i<last_nodo_src; i++)
        vettore_punt[i] += tot_spazio_da_aggiungere;

      //aggiustiamo la lista di 'nodo_src', i.e. quella corrente da espandere:
      vettore_punt[nodo_src] -= spazio_corrente; //aumentiamo la dim. della lista dello spazio richiesto (nella parte iniziale)
      tot_spazio_da_aggiungere -= spazio_corrente; //aggiorniamo lo spazio che va ancora aggiunto

      //aggiorniamo le altre variabili:
      last_nodo_src= nodo_src;
      nodo_src--; //devo poter passare ad un altra posizione non nulla del vettore nella prossima esecuzione del loop
      num_liste_da_espandere--;
    }//fine blocco while  
  else //aggiungiamo lo spazio alla fine della lista
    while(num_liste_da_espandere > 0) //N.B. il parametro 'num_liste_da_espandere' viene decrementato
    {
      //troviamo il prossimo nodo che ha figli swappati (per ipotesi ce n'e' almeno 1):
      while(vettore_spazi[nodo_src] == 0)
        nodo_src--;
      //a questo punto sono posizionato su un elemento non nullo del vettore, me lo garantisce il fatto che e' (num_liste_da_espandere > 0)

      unsigned long spazio_corrente= vettore_spazi[nodo_src];

      //spostiamo le liste di adiac. che seguono quella da espandere (se ve ne sono tra quella corrente e l'ultima espansa):
      if(last_nodo_src - nodo_src > 1) //notare che e' (last_nodo_src > nodo_src) per via dell'ordine decrescente di visita
      {
        unsigned long copy_from= vettore_punt[nodo_src + 1]; //dobbiamo spostare a partire dalla prima lista che segue quella da espandere corrente
        unsigned long larg_intervallo= vettore_punt[last_nodo_src] - copy_from - tot_spazio_da_aggiungere; //non dobbiamo ricopiare le locazioni inutilizzate presenti nell'attuale "buco"
        copia_avanti(copy_from, larg_intervallo, tot_spazio_da_aggiungere); //compattiamo i nodi delle liste interessate verso la fine del vettore

        //aggiorniamo i puntatori delle liste appena spostate:
        for(unsigned long i=nodo_src + 1; i<last_nodo_src; i++)
          vettore_punt[i] += tot_spazio_da_aggiungere;
      }

      //spostiamo e aggiustiamo la lista da espandere:
      if(num_liste_da_espandere > 1) //per l'ultima lista rimasta da espandere non e' necessario ne' spostarla, ne' aggiustare i puntatori
      {
        unsigned long copy_from= vettore_punt[nodo_src];
        unsigned long larg_intervallo= vettore_punt[nodo_src + 1] - copy_from - tot_spazio_da_aggiungere; //non dobbiamo ricopiare le locazioni inutilizzate presenti nell'attuale "buco"

        tot_spazio_da_aggiungere -= spazio_corrente; //(!) aggiorniamo qui lo spazio che resta da aggiungere alle liste (infatti questo valore serve nelle due istruzioni successive)
        copia_avanti(copy_from, larg_intervallo, tot_spazio_da_aggiungere); //spostando in avanti riserviamo in fondo lo spazio richiesto per questa lista
        vettore_punt[nodo_src] += tot_spazio_da_aggiungere; //aggiustiamo il puntatore in modo da lasciarsi alle spalle lo spazio richiesto per le altre liste
      }

      //aggiorniamo le altre variabili:
      last_nodo_src= nodo_src;
      nodo_src--; //devo poter passare ad un altra posizione non nulla del vettore nella prossima esecuzione del loop
      num_liste_da_espandere--;
    }//fine blocco while
    
}


