#ifndef LISTASTATICA_XX_H_
#define LISTASTATICA_XX_H_

#include "main.h"
#include <assert.h>

class ListaStatica
{
  public:

    static const unsigned long marca_X_adiac= MAX_ULONG; //costante usata per marcare le posizioni nella lista di adiacenza corrispondenti a nodi gia' usati, locazioni che sono quindi di nuovo disponibili


    ListaStatica(unsigned long n_nodi, unsigned long max_ad);

    ~ListaStatica();

    void print_adiacenti();

    //le 4 seguenti sono operazioni "a basso livello": be careful!
    unsigned long * vettore_riferimenti() { return vettore_punt; };
    unsigned long * vettore_adiacenti() { return vettore_adiacenze; };
    unsigned long get_indice_lista_adiacenza(unsigned long nodo) {
                                                                    accertati_che( (nodo<size_vett_punt-1), "Valore fuori range per il nodo passato a get_indice_lista_adiacenza()");
                                                                    return vettore_punt[nodo];
                                                                    };
    unsigned long get_size_vett_punt() { return size_vett_punt; };
    
    //Attenzione: il riferimento restituito e' valido finche' non vengono fatte modifiche alla lista di adiacenza
    unsigned long * get_punt_mem_libera();

    unsigned long get_max_adiacenti() { return max_adiacenti; };
    
    //Attenzione: la dimensione restituita e' valida finche' non vengono fatte modifiche alla lista di adiacenza
    unsigned long capienzaMemoriaInutilizzata();
    
    //ritorna il numero di elementi totali nelle liste di adiacenza, senza distinguere tra marche e adiacenti normali
    unsigned long numero_adiacenti_totali();
                    
    bool ricerca_sequenziale_adiacente(unsigned long nodo_src, unsigned long nodo_dest, unsigned long max_adiac_da_esaminare);

    void elimina_marche();

    bool inserisci_elemento(unsigned long nodo_src, unsigned long nodo_dest);

    unsigned long degree_of(unsigned long nodo);

    unsigned long & get_adiac(unsigned long nodo, unsigned long indice);

    void aggiungi_spazio_alla_lista_del_nodo(unsigned long nodo_src, unsigned long spazio_da_aggiungere);
    void accorcia_lista_del_nodo(unsigned long nodo_src, unsigned long spazio_da_togliere);
    void sposta_indietro_liste(unsigned long di_quanto_spostare, unsigned long nodo_from, unsigned long nodo_to);
    void sposta_indietro_liste(unsigned long di_quanto_spostare, unsigned long nodo_from);
    
    void riversa_lista_in_pila_SX_DX(unsigned long nodo_src, SwappedStack<unsigned long> * pila);
    void riversa_lista_in_pila_SX_DX(unsigned long nodo_src, SwappedStack<unsigned long> * pila, unsigned long quanti_inserirne);
    void riversa_lista_in_pila_DX_SX(unsigned long nodo_src, SwappedStack<unsigned long> * pila);
    void riversa_lista_in_pila_DX_SX(unsigned long nodo_src, SwappedStack<unsigned long> * pila, unsigned long quanti_inserirne);

    void aumenta_spazio_liste(unsigned long * vettore_spazi, unsigned long tot_spazio_da_aggiungere, unsigned long num_liste_da_espandere, bool aggiungi_spazio_alla_fine);
            
  private:
    unsigned long * memoria_allocata_lista; //memoria allocata complessivamente per ospitare i due vettori: dei puntatori e degli adiacenti
    unsigned long * vettore_punt; //puntatori alle liste di adiac. dei singoli nodi
    unsigned long * vettore_adiacenze; //contiene i nodi adiacenti
    unsigned long  size_vett_punt; //dimensione del vettore <vettore_punt>
    unsigned long  max_adiacenti; //dimensione del vettore <vettore_adiacenze>

    void copia_avanti(unsigned long copy_from, unsigned long larg_intervallo, unsigned long di_quanto_spostare);
        
    ListaStatica & operator=(const ListaStatica &); //disabilitiamo l'operatore di assegnazione (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
    ListaStatica(const ListaStatica &); //disabilitazione del costruttore di copia (altrimenti andrebbe gestito opportunamente essendoci della memoria allocata dinamicamente)
};

#endif
