/*

  Ernesto Ricci
   19/3/2003
*/



#include <getopt.h>
#include <time.h>
#include "main.h"
#include <stdarg.h>
#include <stdlib.h>

#include "../common/scc_utl/VettoreBits.h"
#include "ListaStatica.h"
#include "nodi_passivi.h"
#include "output_risultati.h"
#include "riordinamento_liste.h"
#include "BufferLetturaBits.h"


//nomi dei file temporanei:
#ifdef USA_DIR_CORR_COME_TMP
  const char * nome_mfile_max_reached= "tmp_max_reached%d";
  const char * nome_mfile_alberoDFS_prec="prevDFStree%d";
  const char * nome_mfile_Inizio_foresta_prec= "tmp_dati_Inizio_foresta%d";
  const char * nome_mfile_Fine_foresta_prec= "tmp_dati_Fine_foresta%d";
  const char * nome_mfile_nodi_swap_passivi_InizioForesta= "tmp_passivi_swappati_InizioForesta%d";
  const char * nome_mfile_info_swap_passivi_InizioForesta= "tmp_info_passivi_swappati_InizioForesta%d";  
  const char * nome_mfile_nodi_swap_passivi_FineForesta= "tmp_passivi_swappati_FineForesta%d";
  const char * nome_mfile_info_swap_passivi_FineForesta= "tmp_info_passivi_swappati_FineForesta%d";
  const char * nome_mfile_appoggio_radici= "tmp_appoggio_radici%d";
  const char * nome_file_bits_passivi_prec= "tmp_nodi_passivi"; //file
  const char * nome_file_bits_padri_esclusi= "tmp_nodi_padri_esclusi"; //file
#else
  const char * nome_mfile_max_reached= "/tmp/tmp_max_reached%d";
  const char * nome_mfile_alberoDFS_prec="/tmp/prevDFStree%d";
  const char * nome_mfile_Inizio_foresta_prec= "/tmp/tmp_dati_Inizio_foresta%d";
  const char * nome_mfile_Fine_foresta_prec= "/tmp/tmp_dati_Fine_foresta%d";
  const char * nome_mfile_nodi_swap_passivi_InizioForesta= "/tmp/tmp_passivi_swappati_InizioForesta%d";
  const char * nome_mfile_info_swap_passivi_InizioForesta= "/tmp/tmp_info_passivi_swappati_InizioForesta%d";
  const char * nome_mfile_nodi_swap_passivi_FineForesta= "/tmp/tmp_passivi_swappati_FineForesta%d";
  const char * nome_mfile_info_swap_passivi_FineForesta= "/tmp/tmp_info_passivi_swappati_FineForesta%d";
  const char * nome_mfile_appoggio_radici= "/tmp/tmp_appoggio_radici%d";
  const char * nome_file_bits_passivi_prec= "/tmp/tmp_nodi_passivi"; //file
  const char * nome_file_bits_padri_esclusi= "/tmp/tmp_nodi_padri_esclusi"; //file
#endif




unsigned long numero_nodi; //numero nodi del grafo
unsigned long numero_archi; //numero archi del grafo

unsigned long archi_caricabili_alla_volta; //quanti archi(<-->adiacenti) posso caricare ogni volta accanto a quelli della foresta DFS. Non puo' essere minore di N (=numero_nodi)
unsigned long minimo_archi_caricabili_alla_volta; //mantiene il valore iniziale di 'archi_caricabili_alla_volta' e non viene piu' modificato

long numINFO; //numero parti multifile dei nodi
long numADIAC; //numero parti multifile predecessori/successori
long numFileOrd; //numero parti multifile ordinamento delle radici DFS
long parti_mfile_alberoDFS_prec= 0; //numero parti multifile dell'albero salvato nella fase precedente

//opzioni riga di comando:
char grafoINFO[1000];
char * grafoADIAC; //grafo dei successori o dei predecessori a seconda dei casi
char * nome_mfile_output;
long long memoria_max= -1; //memoria massima cha il programma puo' usare
bool visita_avanti; //seguiamo gli archi in avanti o all'indietro?
bool output_nodo_quando_scoperto; //diamo in output un nodo appena questo viene scoperto dalla visita o appena questo viene completato (i.e. diventa nero)
bool esegui_comunque; //flag per eseguire incondizionatamente la visita anche se la memoria necessaria ?maggiore di quella disponibile
char * nome_mfile_ordinamento= NULL; //nome del file che stabilisce l'ordine in cui cercare nuovi nodi di partenza per le visite degli alberi DFS. Significativo solo se modo_visita_alberi = DFS_file_ord
unsigned long size_mfile_ordinamento= 0;
unsigned long nodo_start= 0; //unico nodo da considerare come radice di alberi DFS: variabile significativa solo se se modo_visita_alberi = DFS_single
treeDFS modo_visita_alberi; //stabilisce la modalita' di esplorazione degli alberi DFS
bool ordine_inverso_output; //in pratica dice di scrivere in ordine contrario il file di output con i nodi: a partire dall'ultimo nodo scoperto o a partire dall'ultimo completato

ListaStatica * liste_adiacenza= NULL;
VettoreBits * nodi_passivi= NULL;

struct Arco * blocco_riservato_buffer; //(!) trucco: riserviamo la parte finale del vettore degli adiacenti ai vari buffers che possiamo usare contemporaneamente alla lista di adiac. piena, in modo da non sovrascriverlo nel riempirlo, ma in maniera tale che lo si comprende nella memoria libera della lista di adiacenza che viene sfruttata in molte parti
unsigned long elementi_riservati_buffer; //elementi allocati dalla lista di adiacenza, ma che non verranno riempiti di adiacenti in modo da avere sempre un buffer minimo con la memoria libera della lista di adiacenza

unsigned long num_archi_albero_DFS_precedente= 0; //quanti archi (quanti adiacenti) aveva l'albero (foresta) DFS precedente
unsigned long valore_test_modifiche_albero_prec; //variabile per euristica che scopre se la foresta e' cambiata, senza vedere gli archi. Memorizza l'indirizzo iniziale della lista di adiac. del nodo "mediano" (i.e. con identificatore N/2)
bool foresta_prec_salvata;

//puntatori agli oggetti per la lettura degli archi:
multiFileReader * mf_info= NULL;
multiFileReader * mf_adiac= NULL;
BufferLetturaMultifile<struct grafoInfo> * buf_info= NULL;
BufferLetturaMultifile<unsigned long> * buf_adiac= NULL;


//informazioni per la scansione circolare dei multifile di input:
unsigned long successori_restanti_info_corrente= 0;  //info. usata da leggi_arco() e inizializzata da rewind_multifile()
unsigned long info_scandite_nel_round= 0; //info. usata da leggi_arco() e inizializzata da rewind_multifile()
unsigned long archi_restanti_nel_file;  //info. usata da carica_prossimi_archi() e inizializzata da rewind_multifile()
unsigned long num_adiacenti_info_corrente;


//variabili usate per stabilire la versione di visita DFS interna da invocare, in modo da sfruttare possibilmente l'accesso sequenziale alle radici:
enum visiteDFS {DFS_memoria, DFS_radici_crescenti, DFS_file_radici};
visiteDFS old_tipo_visita;
const char * old_nome_mfile_radici= NULL;
long old_numFile_radici;
unsigned long num_radici_salvate_su_file;
unsigned long old_archi_caricabili;

unsigned long ultimo_nodo_leftmost_path; //variabile usata per aggiungere al vettore dei bits passivi i nodi del cammino piu' a SX prima di ogni caricamento di archi. Indica il primo dei nodi sul cammino piu' a SX che sono da rendere passivi


//variabili per le "statistiche":
unsigned long num_controlli_evitati= 0;
unsigned long foreste_differenti= 0;
unsigned long num_foreste_salvate= 0;
unsigned long num_confronti_incerti_foreste= 0;
//unsigned long conta_nodi_swappabili_aggiuntivi= 0; //******
unsigned long conta_scritture_radici_su_disco= 0;
unsigned long conta_visite_DFS_file= 0;
unsigned long conta_visite_DFS_cheap= 0;
unsigned long conta_visite_DFS_memoria= 0;
unsigned long conta_spazio_guadagnato_radici= 0;
unsigned long conta_nodi_ordinati_con_buffer_corto= 0;


//variabili per cronometrare i tempi d'esecuzione di varie funzioni
long long  tot_tempo_caricamento_archi= 0;
long long  tot_tempo_dfs_internal= 0;
long long  tot_tempo_elimina_marche= 0;
long long  tot_tempo_ordinamento= 0;
long long  tot_calcolo_pesi= 0;
long long  tot_tempo_salvataggio_foresta= 0;
long long  tot_tempo_controllo_modifiche= 0;
long long  tot_calcolo_passivi_NPE= 0;
long long  tot_calcolo_passivi_InitialPart= 0;
long long  tot_calcolo_passivi_FinalPart= 0;
long long  tot_max_preord_calcolo_passivi_NPE;
long long  tot_numero_preordine_NPE;
clock_t  tag1, tag2; //variabili che servono solo a prendere i tempi (sono globali solo per non ridichiararle ogni volta)



//variabili usate per uscire prima ottenendo un risultato "approssimato"
bool richiesta_controlc= false; //mem. una richiesta di uscire con CTRL+C
time_t tempo_inizio;
bool specificato_max_tempo;
bool specificato_max_iterazioni;
bool specificato_max_rounds;
unsigned long tempo_max;
unsigned long iterazioni_max;
unsigned long rounds_max;



/*
void chiudi_tutto(int par) //intercetta la richiesta di chiudere forzatamante il programma da parte di assert() e libera la memoria uscendo ordinatamente
{
  printf("\n Deallocazione della memoria e preparazione all'uscita forzata\n");
  libera_memoria();
  ABORT();
}
*/


//scrive un messaggio ed esce dopo liberando le strutture di memoria impegnate
//per quanto riguarda i parametri usa la stessa sintassi di printf:  printf(const char *formato, ...)
void gestione_errore(const char *formato, ...)
{
  va_list ap;
  //stampiamo il messaggio d'errore ed usciamo:
  fprintf(stderr,"\n");
  va_start (ap, formato); //inizializziamo la struttura 'ap'
  vfprintf (stderr, formato, ap); //scriviamo il messaggio d'errore con tutti i parametri eventuali
  va_end (ap); //chiudiamo con la variabile 'ap'
  fprintf(stderr,"\n");
  perror("");
  
  libera_memoria();
  ABORT();
}


/*
void richiesta_chiusura(int par)
{
  richiesta_controlc= true;
  fprintf(stderr, "\nRicevuto CTRL+C\nLa richiesta di chiusura sara' accolta alla fine dell'iterazione corrente\n Attenzione! Un'ulteriore pressione di CTRL+C interrompera' il programma immediatamente\n");
  signal(SIGINT, SIG_DFL);
} */


bool uscita_anticipata(unsigned long num_iterazioni, unsigned long num_rounds)
{
  bool risposta= false;

  if( specificato_max_iterazioni && (num_iterazioni >= iterazioni_max) )
  {
#ifdef VERBOSE_MODE
    printf("--> Raggiunto il numero minino previsto di iterazioni (%ld iterazioni)\n", iterazioni_max);
#endif
    risposta= true;
  }

  if(specificato_max_rounds && (num_rounds >= rounds_max))
  {
#ifdef VERBOSE_MODE
    printf("--> Raggiunto il numero minino previsto di rounds (%ld rounds)\n", rounds_max);
#endif
    risposta= true;
  }

  if(specificato_max_tempo)
  {
    time_t tempo_adesso;
    time(&tempo_adesso);
    if(difftime(tempo_adesso, tempo_inizio)/60 >= tempo_max) //calcoliamo i minuti trascorsi
    {
#ifdef VERBOSE_MODE
      printf("--> Raggiunto il tempo massimo specificato (%ld minuti)\n", tempo_max);
#endif
      risposta= true;
    }
  }

  if(richiesta_controlc)
  {
#ifdef VERBOSE_MODE
    printf("--> Uscita per richiesta d'interruzione dell'utente\n");
#endif
    risposta= true;
  }

  return risposta;
}




void indiceProgressione(const char* testo,long perc1,long perc2){

  /*======================== STAMPA MESSAGGI ===================================================================*/
  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  if (perc1<0) perc1=0;
  else if (perc1>100) perc1=100;

  if (perc2<0) perc2=0;
  else if (perc2>100) perc2=100;

  perc2=(long) (20.0*perc2/100);
  //indicatore
  char st[50];

  strcpy(st,"____________________");
  for (long i=0;(i<20)&&(i<perc2);i++ )
     st[i]='=';

  //limite a 30 la lunghezza del testo
  char s[55];
  long i=0;
  for (i=0; (testo[i]!=0)&&(i<40);i++)
    s[i]=testo[i];
  for (; (i<40);i++)
    s[i]=' ';
  s[40]=0;


	printf("\r [%c] %3ld%% %20s %s\r",cont,perc1,st, s);
  fflush(stdout);
  /*============================================================================================================*/
}



void eraseMultiFile(const char* nomeMulti){

    long conta=0;

    FILE* fp;
    char nomeFile[1000];

    while(true){
      sprintf(nomeFile,nomeMulti,conta);
      fp=fopen(nomeFile,"r");
      if (fp!=NULL){
          conta++;
          fclose(fp);
          remove(nomeFile);
      }
      else
          return;
    }
}



void carica_mfile_in_memoria(const char * nome_mfile, unsigned long parti_mfile, void * vettore_memoria, unsigned long size_tot_mfile)
{
  unsigned long ris;
  multiFileReader mf_input((char *)nome_mfile, parti_mfile);
  ris= mf_input.read(vettore_memoria, 1, size_tot_mfile);
  accertati_che( (ris == size_tot_mfile), MSG_ERRORE_LETTURA_MFILE, nome_mfile);
  mf_input.close();
}



unsigned long salva_su_mfile(const char * nome_mfile, void * vettore_memoria, unsigned long size_memoria)
{
  unsigned long parti_mfile;
  eraseMultiFile(nome_mfile); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mfile_out((char*)nome_mfile, 512*1024*1024);

  unsigned long ris= mfile_out.write(vettore_memoria, 1, size_memoria);
  accertati_che( (ris==size_memoria), MSG_ERRORE_SCRITTURA_MFILE, nome_mfile);

  parti_mfile= mfile_out.close();
  return parti_mfile;
}



void alloca_memoria()
{
  elementi_riservati_buffer= DIM_MEMORIA_RISERVATA_BUFFER / sizeof(unsigned long);

  archi_caricabili_alla_volta= numero_nodi ; //dobbiamo avere almeno altri N posti nella lista di adiacenza per poter lavorare

  long long memoria_prevista_tot= ((long long) numero_nodi + 1)*sizeof(unsigned long) + (numero_nodi + archi_caricabili_alla_volta)*sizeof(unsigned long) + DIM_MEMORIA_RISERVATA_BUFFER + numero_nodi/8 ; //approssimativamente la memoria richiesta per i vettori delle liste di adiacenza (un vettore di "puntatori", un vettore con gli adiacenti grande almeno 2*numero_nodi), il vettore di bit, la parte in memoria della pila ed i buffers.
        
  if(memoria_prevista_tot > memoria_max) //sforiamo anche con la richiesta minima di memoria
  {
    printf("Warning:memory not enough (required at least %ld bytes)\n", memoria_prevista_tot);
    if( !esegui_comunque )
    {
      char car;
      printf("Continue?(y/n)\n");

      car= getchar();
      if((car != 'y') && (car != 'Y'))
      {
        printf("Aborted by user.\n");
        ABORT();
      }
    }
  }
#ifndef TEST_MEMORIA_MIN
  else //memoria_prevista_tot <= memoria_max : sfruttiamo anche la memoria aggiuntiva
    if( archi_caricabili_alla_volta < numero_archi) //con N posti (oltre agli N adiacenti fissi per la foresta DFS) non possiamo caricare tutti gli archi, sfruttiamo piu' memoria per caricarne un numero maggiore alla volta. Se invece fosse N > numero_archi, dobbiamo comunque allocare posto gli N adiacenti fissi della foresta DFS piu' altri N, altrimenti non si puo' lavorare
    {
      long long mem_surplus= memoria_max - memoria_prevista_tot; //max archi caricabili in piu'
      unsigned long surplus_adiacenti= mem_surplus/sizeof(unsigned long); //quanti altri adiacenti entrerebbero in memoria oltre agli N + N di base.

      if(archi_caricabili_alla_volta + surplus_adiacenti > numero_archi) //la memoria in piu' e' anche troppa: allochiamo solo quella che serve
        surplus_adiacenti= numero_archi - archi_caricabili_alla_volta; //quante ulteriori posizioni occorrono per poter caricare tutti gli adiacenti in una sola volta

      archi_caricabili_alla_volta += surplus_adiacenti;
      memoria_prevista_tot += ((long long) surplus_adiacenti)*sizeof(unsigned long);
    }
#endif

  accertati_che( (archi_caricabili_alla_volta>=numero_nodi), "Errore in alloca_memoria(): previsto spazio per meno di 2*N adiacenti"); //giusto per garantire al 100% questa cosa
  
  liste_adiacenza= new ListaStatica(numero_nodi + 1, numero_nodi + archi_caricabili_alla_volta + elementi_riservati_buffer); //un nodo in piu' per ospitare anche il "super root"
  nodi_passivi= new VettoreBits(numero_nodi); //non serve il bit per il "super root"
  accertati_che( (liste_adiacenza != NULL)&&(nodi_passivi != NULL),"Errore nell'allocazione della memoria.");
  
  blocco_riservato_buffer= (struct Arco * )  (liste_adiacenza->vettore_adiacenti() + numero_nodi + archi_caricabili_alla_volta);

  minimo_archi_caricabili_alla_volta= archi_caricabili_alla_volta; //salviamo il valore originale: 'archi_caricabili_alla_volta' puo' aumentare swappando nuovi passivi
  
#ifdef VERBOSE_MODE
  printf("\n> Memoria allocata: circa %lld KB", memoria_prevista_tot/1024);
  printf("\n> Percentuale di archi caricabile in memoria: %.1f%%\n\n", archi_caricabili_alla_volta >= numero_archi  ?  100.0f  :  100.0f * archi_caricabili_alla_volta/numero_archi);
#endif  
}



void libera_memoria()
{
  if(liste_adiacenza != NULL)
  {
    delete liste_adiacenza;
    liste_adiacenza= NULL;
  }
  if(nodi_passivi != NULL)
  {
    delete nodi_passivi;
    nodi_passivi= NULL;
  }  
}




//---------------------------------------------------------------------------------------------------


void leggi_albero_precedente(unsigned long * memoria_predecessori);




void alloca_buffers_lettura_IPS(unsigned long * inizio_memoria_buffers, unsigned long dimensione_memoria_tot_buffers)
{
  struct grafoInfo * blocco_buffer_info;
  unsigned long * blocco_buffer_adiac;
  unsigned long size_blocco_buffer_info, size_blocco_buffer_adiac;

  size_blocco_buffer_adiac= dimensione_memoria_tot_buffers / 2; //fifty fifty
  size_blocco_buffer_adiac -= (size_blocco_buffer_adiac % sizeof(unsigned long)); //usiamo multipli di 4 byte
  size_blocco_buffer_info= dimensione_memoria_tot_buffers - size_blocco_buffer_adiac; //cio' che resta va all'altro buffer

  blocco_buffer_adiac= inizio_memoria_buffers;
  blocco_buffer_info= (struct grafoInfo *) (inizio_memoria_buffers + size_blocco_buffer_adiac/sizeof(unsigned long)); //occhio! e' una somma tra puntatori

  mf_info= new multiFileReader(grafoINFO, numINFO); //apertura dei multifile di input
  mf_adiac= new multiFileReader(grafoADIAC, numADIAC);
  accertati_che( (mf_info!=NULL)&&(mf_adiac!=NULL),"Impossibile allocare gli oggetti multifile");

  buf_info= new BufferLetturaMultifile<struct grafoInfo>(mf_info, size_blocco_buffer_info, blocco_buffer_info);
  buf_adiac= new BufferLetturaMultifile<unsigned long>(mf_adiac, size_blocco_buffer_adiac, blocco_buffer_adiac);  
}


void libera_buffers_lettura_IPS()
{
  if(buf_info != NULL)
  {
    delete buf_info;
    buf_info= NULL;
  }
  if(buf_adiac != NULL)
  {
    delete buf_adiac;
    buf_adiac= NULL;
  }
  if(mf_info != NULL)
  {
    mf_info->close();
    delete mf_info;
    mf_info= NULL;
  }
  if(mf_adiac != NULL)
  {
    mf_adiac->close();
    delete mf_adiac;
    mf_adiac= NULL;
  }
}



//ci posizioniamo con i buffer dove eravamo l'ultima volta facendo dei salti nei multifile
void posiziona_buffers_lettura()
{
  bool ok;
  unsigned long adiacenti_da_saltare;

  if(info_scandite_nel_round == 0) //non occorre fare nulla, siamo gia' all'inizio
    return;

  //se era info_scandite_nel_round=y, vuol dire che avevamo letto gia' y info
  ok= buf_info->skip(info_scandite_nel_round); //saltiamo le prime y info
  accertati_che( ok, MSG_ERRORE_SEEK, grafoINFO);

  adiacenti_da_saltare= numero_archi - archi_restanti_nel_file;
  if(adiacenti_da_saltare > 0)
  {
    ok= buf_adiac->skip(adiacenti_da_saltare);
    accertati_che( ok, MSG_ERRORE_SEEK, grafoADIAC);
  }
}



void leggi_info()
{
  struct grafoInfo info;

  bool ris= buf_info->get(info); //leggiamo l'info del nodo successivo
  accertati_che( (ris), MSG_ERRORE_LETTURA_MFILE, grafoINFO);

  if(visita_avanti)
  {
    successori_restanti_info_corrente= info.outDeg;
    num_adiacenti_info_corrente= info.outDeg;
  }
  else
  {
    successori_restanti_info_corrente= info.inDeg;
    num_adiacenti_info_corrente= info.inDeg;
  }

  info_scandite_nel_round++; //abbiamo appena esaminato un altro record nel multifile INFO
}



//si suppone che venga chiamata solo se c'e' ancora almeno un arco da leggere
void leggi_arco(struct Arco & arc)
{
  bool ris;

  //cerchiamo il primo nodo che ha nodi adiacenti (continuiamo con l'ultimo considerato se ha ancora qualche arco da usare)
  while(successori_restanti_info_corrente == 0) //sarebbe piu' giusto dire adiacenti restanti perche' possiamo lavorare sia su SUCC che su PRED
    leggi_info();
  // ora successori_restanti_info_corrente e' > 0, siamo posizionati sul nodo possessore dell'adiacente che stiamo per leggere dal multifile PRED/SUCC

  ris= buf_adiac->get(arc.dest); //scriviamo la destinazione dell'arco prelevando l'informazione direttamente dalla prossima posizione nel vettore PRED/SUCC
  accertati_che( (ris), MSG_ERRORE_LETTURA_MFILE, grafoADIAC);
  successori_restanti_info_corrente--; //ne abbiamo appena considerato uno
  archi_restanti_nel_file--;
  arc.source= info_scandite_nel_round - 1; //'info_scandite_nel_round' parte da 1, mentre l'ID dei nodi parte da 0
  assert(arc.source<numero_nodi);
  assert(arc.dest<numero_nodi);  
}



//dealloca e rialloca gli oggetti per la lettura e riposizionandosi all'inizio
void rewind_archi(unsigned long * inizio_memoria_buffers, unsigned long dimensione_memoria_tot_buffers)
{
  libera_buffers_lettura_IPS(); //deallochiamo gli oggetti multifile e buffer lettura

  successori_restanti_info_corrente= 0;
  info_scandite_nel_round= 0;
  archi_restanti_nel_file= numero_archi;
  //non serve a inizializzare num_adiacenti_info_corrente che verranno ricaricate al primo accesso

  alloca_buffers_lettura_IPS(inizio_memoria_buffers, dimensione_memoria_tot_buffers);
}



inline unsigned long leggi_adiacente()
{
  unsigned long nodo;

  assert(successori_restanti_info_corrente>0);
  bool ris= buf_adiac->get(nodo); //scriviamo la destinazione dell'arco prelevando l'informazione direttamente dalla prossima posizione nel vettore PRED/SUCC
  accertati_che( (ris), MSG_ERRORE_LETTURA_MFILE, grafoADIAC);
  successori_restanti_info_corrente--;
  archi_restanti_nel_file--;
  return nodo;
}



inline bool is_padre_da_escludere(unsigned long nodo, BufferLetturaBits * buffer_padri_da_escludere)
{
  if(numero_padri_da_escludere==0)
    return false;
    
  assert(buffer_padri_da_escludere!=NULL);

  long err;
  bool risposta= buffer_padri_da_escludere->get(nodo, err);
  accertati_che(!err, "Errore nell'accedere all'oggetto BufferVettoreBits in is_padre_da_escludere()");

  assert(!risposta || is_passivo(nodo));
    
  return risposta;
}



//N.B. 'max_archi_da_leggere' deve essere <= archi_restanti_nel_file
unsigned long seleziona_prossimo_nodo_sorgente(unsigned long max_archi_da_leggere, unsigned long * archi_scanditi, BufferLetturaBits * buffer_padri_da_escludere)
{
  assert(max_archi_da_leggere <= archi_restanti_nel_file);

  //N.B. 'max_archi_da_leggere' viene decrementato man mano, finche' non diventa zero
  *archi_scanditi= 0;

  while(max_archi_da_leggere > 0)
  {  
    if(successori_restanti_info_corrente == 0) //primo motivo per ignorare il nodo corrente: non ha figli (o non ne rimaggono da leggere in questa iterazione). Questo e' anche il caso in cui ci troviamo dopo l'inizializzazione (partiamo da un ipotetico nodo "-1" con 0 figli)
    {
      leggi_info();
    }
    else
    {
      unsigned long nodo_corrente= info_scandite_nel_round - 1; // (info_scandite_nel_round - 1) e non 'info_scandite_nel_round' indica il nodo sorgente corrente

      //verifichiamo se il nodo corrente puo' essere ignorato in base ad altri due criteri:
      if((num_adiacenti_info_corrente == liste_adiacenza->degree_of(nodo_corrente)) || is_padre_da_escludere(nodo_corrente, buffer_padri_da_escludere))//se il nodo corrente ha gia' in memoria l'intera sua lista di adiacenza contenuta su disco ; se e' un padre i cui figli sicuramente non sono utili, allora lo saltiamo
      {
        unsigned long lung_salto= MIN(successori_restanti_info_corrente, max_archi_da_leggere);

        bool ok= buf_adiac->skip(lung_salto);
        accertati_che( ok, MSG_ERRORE_SEEK, grafoADIAC);

        successori_restanti_info_corrente -= lung_salto; //e' come se avessi letto gli adiacenti che ho appena saltato
        archi_restanti_nel_file -= lung_salto; //(!) dobbiamo emulare quello che farebbe leggi_adiacente()
        max_archi_da_leggere -= lung_salto;
        *archi_scanditi += lung_salto;
      }
      else
        break; //il nodo corrente e' sfuggito a tutti i controlli che potevano permetterci di saltarlo, percio' esso va elaborato e quindi ci fermiamo
        
    }//fine blocco else
    
  }//fine blocco while
  
  return (info_scandite_nel_round - 1); //l'ultimo nodo sorgente considerato
}



void set_bits_nodi_adiacenti(unsigned long nodo_src, bool passivo, unsigned long num_adiacenti, BufferAppoggio<unsigned long> * buffer_valori_bits)
{
  if(num_adiacenti==0)
    return;

  unsigned long * punt_adiac= liste_adiacenza->vettore_adiacenti() + liste_adiacenza->get_indice_lista_adiacenza(nodo_src);
  
  if( passivo ) //i suoi figli possono essere gia' passivi, dobbiamo stare attenti a non perdere informazioni
  {
    unsigned long LW_corrente= 0, segna_bit= 1;

    for(unsigned long i=0 ; i < num_adiacenti; i++, punt_adiac++)
    {
      bool val_prec= set_passivo( *punt_adiac );
      
      if(val_prec)
        LW_corrente |= segna_bit; //se valeva zero lascio lo zero nel bit corrente della LW

      if(segna_bit == 0x80000000) //1000...0b ovvero e' il MSB, i.e. ho completato 32 bit
      {
        buffer_valori_bits->depositaValore(LW_corrente);
        LW_corrente= 0;
        segna_bit= 1; //ripartiamo dal LSB
      }
      else
        segna_bit <<= 1; //shifto a SX di 1
        
    }

    if(segna_bit != 1) //se non e' il LSB ho messo almeno un bit nella LW corrente
      buffer_valori_bits->depositaValore(LW_corrente);      
  }
  else
    for(unsigned long i=0 ; i < num_adiacenti; i++, punt_adiac++)
    {
      bool ris= set_passivo( *punt_adiac );
//      accertati_che(!ris, "Errore: nodo non passivo (%lu) con figlio passivo (%lu)\n", nodo_src, *punt_adiac);
      assert(!ris);
    }
}



void reset_bits_nodi_adiacenti(unsigned long nodo_src, bool passivo, unsigned long num_adiacenti, BufferAppoggio<unsigned long> * buffer_valori_bits)
{
  if(num_adiacenti==0)
    return;
    
  unsigned long * punt_adiac= liste_adiacenza->vettore_adiacenti() + liste_adiacenza->get_indice_lista_adiacenza(nodo_src);

  if( passivo ) //i suoi figli possono essere gia' passivi, dobbiamo stare attenti a non perdere informazioni
  {
    unsigned long LW_corrente, segna_bit;

    buffer_valori_bits->recuperaValore(LW_corrente); //per ipotesi so che c'e' almeno un adiacente --> almento una LW salvata nel buffer
    segna_bit= 1;
    
    for(unsigned long i=0 ; i < num_adiacenti; i++, punt_adiac++)
    {
      if( !(LW_corrente & segna_bit) ) //se aveva il valore 1 non lo resetto perche' va bene cosi'
        reset_passivo( *punt_adiac );

      if(segna_bit == 0x80000000) //1000...0b ovvero e' il MSB, i.e. ho completato 32 bit
      {
        if(i + 1 < num_adiacenti) //se ci sono ancora adiacenti da considerare
          buffer_valori_bits->recuperaValore(LW_corrente);
          
        segna_bit= 1; //ripartiamo dal LSB
      }
      else
        segna_bit <<= 1; //shifto a SX di 1
        
    }
  }
  else
    for(unsigned long i=0 ; i < num_adiacenti; i++, punt_adiac++)
    {
      reset_passivo( *punt_adiac );
    }
}



void leggi_serie_archi(unsigned long max_archi_da_caricare, unsigned long max_archi_da_leggere, unsigned long * archi_caricati, unsigned long * archi_letti, BufferAppoggio<unsigned long> * buffer_appoggio_bits, BufferLetturaBits * buffer_padri_da_escludere)
{
  bool passivo_nodo_src;
  unsigned long nodo_src, last_nodo_src, size_lista_originale, ultimi_archi_letti;
  unsigned long quanti_letti= 0, quanti_caricati= 0;
  bool prima_iterazione_loop= true;
  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
  // N.B. max_archi_da_caricare e max_archi_da_leggere vengono decrementati man mano

  do
  {
    nodo_src= seleziona_prossimo_nodo_sorgente(max_archi_da_leggere, &ultimi_archi_letti, buffer_padri_da_escludere);

    quanti_letti += ultimi_archi_letti;
    max_archi_da_leggere -= ultimi_archi_letti;
    if(max_archi_da_leggere == 0)
      break; //esce direttamente dal ciclo

    size_lista_originale= liste_adiacenza->degree_of(nodo_src); //salviamo questa info. prima di sporcarla, o con aggiungi_spazio_alla_lista_del_nodo() o con sposta_indietro_liste()

    if(prima_iterazione_loop) //(!) invece di mettere prima del loop una iterazione a parte per il caso del primo nodo considerato
    {
      liste_adiacenza->aggiungi_spazio_alla_lista_del_nodo(nodo_src, max_archi_da_caricare);

      prima_iterazione_loop= false; //questo blocco di codice va eseguito una volta soltanto
    }
    else
      liste_adiacenza->sposta_indietro_liste(max_archi_da_caricare, last_nodo_src + 1, nodo_src);


    passivo_nodo_src= is_passivo(nodo_src); //lo salviamo, cosi' evitiamo di leggere due volte lo stesso valore dal vettore di bit

    set_bits_nodi_adiacenti(nodo_src, passivo_nodo_src, size_lista_originale, buffer_appoggio_bits); //gli archi verso i nodi passivi sono gia' da ignorare: adesso settiamo anche i bit dei nodi presenti nella lista di adiac. in memoria di 'nodo_src', cosi' con un unico controllo sul vettore di bits verifichiamo etrambe le cose

    unsigned long * punt_inserimento_adiac= vettore_adiacenti + liste_adiacenza->get_indice_lista_adiacenza(nodo_src) + size_lista_originale;
    unsigned long vincolo_esterno= MIN(max_archi_da_caricare, max_archi_da_leggere); //va ricalcolato perche' entrambi i parametri passati li decrementiamo man mano

    if(successori_restanti_info_corrente <= vincolo_esterno) //anche caricando l'intera viene rispettato il vincolo esterno
    {
      unsigned long ultimi_caricati= 0;
      unsigned long iterazioni_loop= successori_restanti_info_corrente; //poiche' 'successori_restanti_info_corrente' varia ad ogni lettura di un nuovo adiacente, lo salvo in una "costante" (il valore originale serve anche per dopo la fine del loop)

      for(unsigned long i=0 ; i < iterazioni_loop; i++)
      {
        unsigned long adiac= leggi_adiacente();

        if( !is_passivo(adiac) )
        {
          *punt_inserimento_adiac= adiac;
          punt_inserimento_adiac++;
          ultimi_caricati++;
        }
      }

      quanti_letti += iterazioni_loop; //(!) occhio che adesso successori_restanti_info_corrente e' = zero
      max_archi_da_leggere -= iterazioni_loop;
      quanti_caricati += ultimi_caricati;
      max_archi_da_caricare -= ultimi_caricati;
    }
    else //occorre vedere ad ogni inserimento di adiacenti se sono ancora soddisfatti i vincoli esterni
    {
      unsigned long ultimi_scanditi= 0, ultimi_caricati= 0;
      unsigned long vincolo_costante= MIN(successori_restanti_info_corrente, max_archi_da_leggere); //poiche' deve essere: (ultimi_scanditi < A) && (ultimi_scanditi < B), e siccome A e B sono 2 costanti, mi risparmio un confronto ad ogni iterazione del loop

      while( (ultimi_scanditi < vincolo_costante) && (ultimi_caricati < max_archi_da_caricare))
      {
        unsigned long adiac= leggi_adiacente();

        if( !is_passivo(adiac) )
        {
          *punt_inserimento_adiac= adiac;
          punt_inserimento_adiac++;
          ultimi_caricati++;
        }

        ultimi_scanditi++;
      }

      quanti_letti += ultimi_scanditi;
      max_archi_da_leggere -= ultimi_scanditi;
      quanti_caricati += ultimi_caricati;
      max_archi_da_caricare -= ultimi_caricati;
    }

    reset_bits_nodi_adiacenti(nodo_src, passivo_nodo_src, size_lista_originale, buffer_appoggio_bits); //ripristiniamo il contenuto vettore dei nodi passivi, cosi' come era prima di farne l'unione con i nodi figli di 'nodo_src' gia' presenti in memoria

    last_nodo_src= nodo_src;
  }
  while((max_archi_da_caricare > 0) && (max_archi_da_leggere > 0));


  if(max_archi_da_caricare > 0) //se e' rimasto un "buco" lo eliminiamo
    if(!prima_iterazione_loop) //(!) se siamo usciti in anticipo nella prima iterazione non era stato creato nessun "buco"
      liste_adiacenza->sposta_indietro_liste(max_archi_da_caricare, last_nodo_src + 1); // ricompattiamo spostando tutte le liste oltre il buco indietro della dimensione del buco, ed accorciando l'intera lista della dimensione del "buco"

  *archi_caricati= quanti_caricati;
  *archi_letti= quanti_letti;
}




void predisponi_vettore_bits()
{
  ripristina_nodi_passivi_se_necessario(); //ci serve il contenuto integro del vettore dei bits dei nodi passivi

  unsigned long nodo_SX= ultimo_nodo_leftmost_path;
  set_passivo(nodo_SX);
  while(liste_adiacenza->degree_of(nodo_SX) > 0)
  {
    nodo_SX= liste_adiacenza->get_adiac(nodo_SX, 0);

    set_passivo(nodo_SX);
  }
  
  is_dirty_vettore_bits= true; //tanto non comporta un ulteriore caricamento da disco del vettore di bits, infatti i bits in memoria verrebbero comunque sporcati dalla successiva visita DFS in memoria  
}



void carica_prossimi_archi(unsigned long max_archi_da_caricare, unsigned long max_archi_da_provare, unsigned long * archi_scanditi, unsigned long * archi_caricati=NULL)
{
  unsigned long dim_memoria_per_buffers_lettura, capienza_appoggio_bits;
  unsigned long * punt_memoria_per_buffers_lettura, * punt_memoria_appoggio_bits;
  unsigned char * punt_memoria_buff_padri_esclusi;
  unsigned long quanti_provati, quanti_caricati, ultimi_archi_caricati, ultimi_archi_provati;
  BufferLetturaBits * buff_padri_da_escludere= NULL;
  
  const unsigned long DIM_MEMORIA_APPOGGIO_BITS= 512; //quanta memoria riservare al buffer di appoggio per salvare i bits
  const unsigned long DIM_MEMORIA_BUFFER_PADRI_ESCLUSI= DIM_MEMORIA_RISERVATA_BUFFER / 400; //facendo le proporzioni con gli accessi ai file dei nodi e delle info, per fare una cosa "equa" questo buffer dovrebbe essere circa 200 volte piu' piccolo di ciascuno dei due principali.
  
#ifdef VERBOSE_MODE
  printf(" loading edges in memory...\n"); fflush(stdout);  
#endif
  tag1= clock();


  predisponi_vettore_bits(); //ricarica il vettore di bits dei nodi passivi e fa delle aggiunte provvisorie, escludendo dal caricamento tutti i nodi sul cammino piu' a SX

  quanti_provati= 0;
  quanti_caricati= 0;

  //allochiamo la memoria che e' riservata per i buffers, sia per i buffers per leggere i file del grafo, che per gli altri due ausiliari:
  accertati_che( (liste_adiacenza->capienzaMemoriaInutilizzata()>=elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  capienza_appoggio_bits= DIM_MEMORIA_APPOGGIO_BITS/sizeof(unsigned long);
  accertati_che( (capienza_appoggio_bits + DIM_MEMORIA_BUFFER_PADRI_ESCLUSI/sizeof(unsigned long) < 0.05f*elementi_riservati_buffer), "Attenzione: la memoria destinata ai buffers ausiliari e' troppa: %lu e %lu elementi su %lu totali", capienza_appoggio_bits, DIM_MEMORIA_BUFFER_PADRI_ESCLUSI/sizeof(unsigned long), elementi_riservati_buffer);

  if(numero_padri_da_escludere > 0) //in questo caso ci serve un terzo buffer (lettura sequenziale vettore di bit su file)
  {
    dim_memoria_per_buffers_lettura= (elementi_riservati_buffer - capienza_appoggio_bits)*sizeof(unsigned long) - DIM_MEMORIA_BUFFER_PADRI_ESCLUSI;
    dim_memoria_per_buffers_lettura -= (dim_memoria_per_buffers_lettura % sizeof(unsigned long)); //rendiamo multiplo di 4
    
    punt_memoria_per_buffers_lettura= (unsigned long *) blocco_riservato_buffer;
    punt_memoria_appoggio_bits= punt_memoria_per_buffers_lettura + dim_memoria_per_buffers_lettura/sizeof(unsigned long); //(!) aritmetica dei puntatori
    punt_memoria_buff_padri_esclusi= (unsigned char *) (punt_memoria_appoggio_bits + capienza_appoggio_bits); //(!) aritmetica dei puntatori
    
    buff_padri_da_escludere= new BufferLetturaBits( nome_file_bits_padri_esclusi, numero_nodi, DIM_MEMORIA_BUFFER_PADRI_ESCLUSI, punt_memoria_buff_padri_esclusi);
  }
  else
  {
    dim_memoria_per_buffers_lettura= (elementi_riservati_buffer - capienza_appoggio_bits)*sizeof(unsigned long);

    punt_memoria_per_buffers_lettura= (unsigned long *) blocco_riservato_buffer;
    punt_memoria_appoggio_bits= punt_memoria_per_buffers_lettura + (elementi_riservati_buffer - capienza_appoggio_bits); //(!) aritmetica dei puntatori  
  }
  
  BufferAppoggio<unsigned long> buffer_appoggio_bits(capienza_appoggio_bits*sizeof(unsigned long), punt_memoria_appoggio_bits); //la dimensione del buffer va specificata in byte

  alloca_buffers_lettura_IPS( punt_memoria_per_buffers_lettura, dim_memoria_per_buffers_lettura);

  posiziona_buffers_lettura();

  unsigned long limite_max_lettura= MIN( max_archi_da_provare, archi_restanti_nel_file); //ne leggiamo abbastanza pochi da soddisfare il vincolo piu' stringente

  leggi_serie_archi(max_archi_da_caricare, limite_max_lettura, &ultimi_archi_caricati, &ultimi_archi_provati, &buffer_appoggio_bits, buff_padri_da_escludere);
  assert(ultimi_archi_provati<=limite_max_lettura);
  quanti_provati += ultimi_archi_provati;
  quanti_caricati += ultimi_archi_caricati;  
  max_archi_da_provare -= ultimi_archi_provati;
  max_archi_da_caricare -= ultimi_archi_caricati;
  

  if((max_archi_da_caricare > 0) && (max_archi_da_provare > 0)) //se non ho soddisfatto nessuna delle sue condizioni specificate dal chiamante, allora abbiamo raggiunto la fine del file degli archi: "riavvolgiamo" il file degli archi ricominciando a leggerlo dall'inizio
  {
    //facciamo il rewind:
    accertati_che( (liste_adiacenza->capienzaMemoriaInutilizzata()>=elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
    rewind_archi(punt_memoria_per_buffers_lettura, dim_memoria_per_buffers_lettura); //dealloca e rialloca gli oggetti per la lettura e riposizionandosi all'inizio

    if(numero_padri_da_escludere > 0) //bisogna fare il rewind anche del buffer di lettura dei bit sul file
    {
      delete buff_padri_da_escludere; //cancelliamo l'oggetto vecchio
      buff_padri_da_escludere= new BufferLetturaBits( nome_file_bits_padri_esclusi, numero_nodi, DIM_MEMORIA_BUFFER_PADRI_ESCLUSI, punt_memoria_buff_padri_esclusi);
    }
    
    leggi_serie_archi(max_archi_da_caricare, max_archi_da_provare, &ultimi_archi_caricati, &ultimi_archi_provati, &buffer_appoggio_bits, buff_padri_da_escludere);
    assert(ultimi_archi_provati<=max_archi_da_provare);
    quanti_provati += ultimi_archi_provati;
    quanti_caricati += ultimi_archi_caricati;

    assert((max_archi_da_provare==ultimi_archi_provati)||(max_archi_da_caricare==ultimi_archi_caricati));
  }

    
  libera_buffers_lettura_IPS(); //deallochiamo gli oggetti multifile e buffer lettura

  if(numero_padri_da_escludere > 0) //deallochiamo (se necessario) il buffer di lettura dei bits dei padri da escludere
    delete buff_padri_da_escludere;
    
  printf(MSG_DONE);    
 tag2= clock();  tot_tempo_caricamento_archi += tag2 - tag1;

  *archi_scanditi= quanti_provati;
  
  if(archi_caricati != NULL)
    *archi_caricati= quanti_caricati;

  //usciamo perche' abbiamo caricato oppure solo scandito abbastanza archi
}



void caricamento_primi_archi(unsigned long quanti_archi_da_caricare)
{
  accertati_che( (liste_adiacenza->capienzaMemoriaInutilizzata()>=quanti_archi_da_caricare+elementi_riservati_buffer), "errore in caricamento_primi_archi(): non c'e' abbastanza spazio");

  //accediamo alle informazioni "private" della classe lista di adiacenza:
  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
  unsigned long * const vettore_riferimenti= liste_adiacenza->vettore_riferimenti();
  unsigned long size_vett_punt= liste_adiacenza->get_size_vett_punt();

  tag1= clock();
  
  //caricamento degli adiacenti:
  carica_mfile_in_memoria(grafoADIAC, numADIAC, vettore_adiacenti, sizeof(unsigned long)*quanti_archi_da_caricare);

  //lettura delle info sui nodi e aggiustamento dei puntatori nelle liste di adiac. in memoria:
  struct grafoInfo info;
  unsigned long tot_info_lette, tot_adiacenti_considerati;
  multiFileReader mf_lettura_info(grafoINFO, numINFO);
  BufferLetturaMultifile<struct grafoInfo> buf_lettura_info(&mf_lettura_info, DIM_MEMORIA_RISERVATA_BUFFER, (struct grafoInfo *) blocco_riservato_buffer);

  tot_info_lette= 0;
  tot_adiacenti_considerati= 0;

  do
  {
    vettore_riferimenti[tot_info_lette]= tot_adiacenti_considerati;

    bool ok= buf_lettura_info.get(info);
    accertati_che( ok, MSG_ERRORE_LETTURA_MFILE, grafoINFO);

    if(visita_avanti)
      tot_adiacenti_considerati += info.outDeg;    
    else
      tot_adiacenti_considerati += info.inDeg;

    tot_info_lette++;
  } while(tot_adiacenti_considerati < quanti_archi_da_caricare);

  
  for(  ; tot_info_lette < size_vett_punt; tot_info_lette++) //"chiudiamo" la lista di adiacenza aggiustando i puntatori che rimangono
    vettore_riferimenti[tot_info_lette]= quanti_archi_da_caricare;

  mf_lettura_info.close();

  tag2= clock();  tot_tempo_caricamento_archi += tag2 - tag1;
  
#ifdef VERBOSE_DEBUG
  printf("\ndopo caricamento primi archi:\n"); liste_adiacenza->print_adiacenti();
#endif  
}





/* semplice euristica per vedere se ci sono cambiamenti nella foresta DFS (che permette di evitare, per scoprirlo, di caricare da file i predecessori salvati).
   Legge l'indice in cui inizia la lista di adiacenza del nodo "intermedio" nella lista di adiacenza. Se il valore salvato nell'iterazione precedente non coincide con quello appena letto, sicuramente ci sono state delle modifiche.
*/
unsigned long quick_test_modifiche_albero()
{
  return liste_adiacenza->get_indice_lista_adiacenza(numero_nodi/2);
}



//salva su file le informazioni predecessore[n] per ogni nodo 'n' dell'albero corrente, in modo da salvarne sinteticamente la struttura per fare confronti successivi
//va invocata quando le liste di adiacenza contengono l'albero corrente, senza marche "X" tra gli adiacenti.
//L'albero complessivo (foresta DFS) ha (al piu', se facciamo una DFS "completa") N+1 nodi, quindi scriviamo N predecessori. Il predecessore del nodo "super root" non serve (non esiste) e non viene usato.
//sfruttiamo la memoria che avanza alle liste di adiacenza, che e' N perche' l'albero ha appunto N archi, mentre la lista di adiacenza ha spazio 2*N
//specifichiamo da quale nodo iniziare a calcolare i predec., assumendo che per quelli prima vadano bene i valori gia' presenti
void salva_albero_corrente(unsigned long * memoria_predecessori)
{
  unsigned long valore_test_old= valore_test_modifiche_albero_prec;
  
  num_archi_albero_DFS_precedente= liste_adiacenza->numero_adiacenti_totali(); //salviamo il numero di nodi nella foresta DFS attuale
  valore_test_modifiche_albero_prec= quick_test_modifiche_albero();

  if(valore_test_old != valore_test_modifiche_albero_prec) //se ci sono stati cambiamenti e' probabile che ce ne saranno anche la volta prossima, percio' il file con i predecessori della foresta precedente probabilmente non servira'
  {
    foresta_prec_salvata= false;
    return;
  }  

  printf(" Saving current tree on file...\n"); fflush(stdout);
    
//  if(liste_adiacenza->numero_adiacenti_totali() < numero_nodi) //se dei nodi non hanno un padre (e non compaiono nella foresta) a causa dell'inizializzazione con un solo nodo o comunque con meno di N nodi
//  {
//    //inizializzazione della memoria che deve contenere i predecessori: mettiamo tutte marche (MAX_ULONG) in modo che poi si puo' capire facilmente quale nodo non ha predecessori, cioe' non appartiene alla foresta DFS. Cio' e' necessario per poter implementare la visita di un singolo albero (con radice 'nodo_start'), in quanto in tal caso la foresta puo' avere meno di N nodi:
//    for(unsigned long i=0; i < numero_nodi; i++) //per tutti i nodi tranne il "super root" inizializzo il relativo predecessore
//      memoria_predecessori[i]= MAX_ULONG; //usiamo al solito come marca il massimo numero rappresentabile con gli unsigned long
//
//    nodo_src= 0; //la scrittura dei predecessori deve ripartire dall'inizio, perche' abbiamo appena cancellato tutto
//    ind_adiac= 0;
//  }

  //creazione del vettore dei PI[n] (predecessore di n) per ogni nodo n che non sia il "super root". Usiamo la memoria libera della lista di adiacenza (N locazioni perche' l'albero ha N archi --> 2N - N = N)
  for( unsigned long nodo_src= 0; nodo_src <= NODO_SUPER_ROOT; nodo_src++) //va compreso anche il "super root" perche' ci interessano anche i suoi adiacenti
  {
    unsigned long degree= liste_adiacenza->degree_of(nodo_src);

    for( unsigned long ind_adiac= 0; ind_adiac<degree ; ind_adiac++)
    {
      unsigned long nodo_dest= liste_adiacenza->get_adiac(nodo_src, ind_adiac); //compreso tra 0 e N-1 perche' il "super root" non ha mai archi entranti
      memoria_predecessori[nodo_dest]= nodo_src; //memorizziamo che 'nodo_dest' e' stato trovato nell'esplorazione DFS seguendo un arco proveniente da 'nodo_src'. Ovvero PI[nodo_dest]= nodo_src
    }
  }

  //salvataggio su disco:
  parti_mfile_alberoDFS_prec= salva_su_mfile(nome_mfile_alberoDFS_prec, memoria_predecessori, sizeof(unsigned long)*numero_nodi); //salvataggio delle N ('numero_nodi') posizioni di memoria libera appena scritte: se l'albero ha N nodi scriviamo un predecessore per ogni nodo

  foresta_prec_salvata= true;  
  num_foreste_salvate++;

  printf(MSG_DONE);
}



void leggi_albero_precedente(unsigned long * memoria_predecessori)
{
  accertati_che( (parti_mfile_alberoDFS_prec>0), "L'albero della fase precedente non e' stato salvato correttamente: il numero delle parti registrato non e' valido");
  assert( getTotalSizeMultiFiles((char*)nome_mfile_alberoDFS_prec)/sizeof(unsigned long) == numero_nodi);

  carica_mfile_in_memoria(nome_mfile_alberoDFS_prec, parti_mfile_alberoDFS_prec, memoria_predecessori, sizeof(unsigned long)*numero_nodi);
}



/*
  ritorna true se l'albero corrente e' diverso dall'albero della fase precedente. Cio' avviene se vi e' almeno un arco di un albero che non compare nell'altro
  Vediamo se tutti gli archi dell'albero precedente ci sono anche nell'albero corrente, e se il numero di archi e' lo stesso. In tal modo i due alberi coincidono, a meno dell'ordinamento tra gli archi che escono da uno stesso nodo sorgente (cioe' a meno dell'ordinamento delle liste di adiacenza).
  Poiche' pero' noi ordiniamo i sottoalberi in base al loro peso, e poiche' essendo uguali gli archi saranno uguali i sottoalberi e i rispettivi pesi, gli alberi coincideranno completamente.
  PRECondizioni: poiche' usiamo ricerca_binaria_adiacente() le liste di adiacenza devono essere state gia' ordinate con ordina_adiacenti_in_modo_crescente()

*/  
bool controlla_modifiche_albero(unsigned long * memoria_albero_precedente)
{
  printf(" Check update of tree...\n");

  //non ci sono differenze tra l'albero vecchio e quello nuovo se e solo se: l'albero nuovo contiene tutti i predecessori dell'albero vecchio e il numero dei predecessori nei due alberi e' lo stesso. Viene fatto un controllo rapido per indovinare se la foresta DFS e' cambiata, senza leggere il file dal disco
  if( quick_test_modifiche_albero() != valore_test_modifiche_albero_prec )
  {
    num_controlli_evitati++;
    foreste_differenti++;
    
    printf(MSG_DONE);    
    return true; //se il numero di archi non e' lo stesso i due alberi differiscono sicuramente
  }

  if(!foresta_prec_salvata) //non abbiamo salvato precedentemente la foresta vecchia, per cui non possiamo dire se coincide effettivamente con la nuova: restituiamo di si' per non sbagliare
  {
    num_confronti_incerti_foreste++;
    return true;
  }

  leggi_albero_precedente(memoria_albero_precedente); //carichiamo l'albero vecchio nel vettore passato

  //consideriamo tutti gli archi dell'albero corrente: ciascuno di essi deve essere presente anche nell'albero vecchio (memorizzato pero' nel formato di predecessori dei nodi dell'albero)
  for(unsigned long nodo_src=0; nodo_src <= NODO_SUPER_ROOT; nodo_src++) //va compreso anche il "super root" perche' ci interessano anche i suoi adiacenti
  {
    unsigned long degree= liste_adiacenza->degree_of(nodo_src);

    for(unsigned long ind_adiac=0; ind_adiac<degree ; ind_adiac++)
    {
      unsigned long nodo_dest= liste_adiacenza->get_adiac(nodo_src, ind_adiac); //compreso tra 0 e N-1 perche' il "super root" non ha mai archi entranti

      if( memoria_albero_precedente[nodo_dest] != nodo_src) //se 'nodo_dest' aveva un diverso predecessore nell'albero vecchio (o non lo aveva proprio, se questa e' una marca), i due alberi differiscono
      {
        foreste_differenti++;
        
        printf(MSG_DONE);        
        return true;
      }
    }
  }

  printf(MSG_DONE);
  return false;
}



//versione con riversamento nella pila e senza ricerca binaria (sperimentalmente sembra _la_piu'_veloce_)
void internal_DFS(VettoreBits * nodi_bianchi)
{
  printf(" Internal DFS (NO backtracking)...\n"); fflush(stdout);

  tag1= clock();
  
  //nodi_bianchi indica i nodi che sono bianchi. N.B. per il "super root" non serve il bit perche' non ha archi entranti.
  nodi_bianchi->reset(true); //Inizialmente i nodi sono tutti bianchi: inizializziamo il vettore con tutti uno

  accertati_che( (liste_adiacenza->capienzaMemoriaInutilizzata()>=elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  SwappedStack<unsigned long> pila( liste_adiacenza->capienzaMemoriaInutilizzata()*sizeof(unsigned long), liste_adiacenza->get_punt_mem_libera());

  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();

  unsigned long numero_alberi= liste_adiacenza->degree_of(NODO_SUPER_ROOT); //ogni arco che parte dal nodo "super root" individua un albero DFS
  for(unsigned long indice_albero_corrente=0; indice_albero_corrente<numero_alberi ; indice_albero_corrente++)
  {
    long errore;
    bool esplorare_albero;

    unsigned long & rif_nodo_radice= liste_adiacenza->get_adiac(NODO_SUPER_ROOT, indice_albero_corrente); //riferimento usato per leggere e manipolare il nodo adiacente corrente.

    esplorare_albero= nodi_bianchi->get(rif_nodo_radice, errore); //vedo se il nodo e' bianco (i.e. non sono ancora stati espansi i suoi successori).
    accertati_che( (errore==0), MSG_ERRORE_BITS);

    if(!esplorare_albero) //nodo radice non-bianco
      rif_nodo_radice= ListaStatica::marca_X_adiac; //marchiamo cosi' sappiamo che l'arco corrispondente non fa parte dell'albero DFS
    else //l'arco corrispondente viene usato nella costruzione della nuova foresta DFS e quindi l'adiacente non va cancellato
    {
#ifdef OPT_PRIMO_CAMMINO
      unsigned long indice_nella_lista;
      unsigned long nodo_corrente= rif_nodo_radice;
      unsigned long degree;//= liste_adiacenza->degree_of(nodo_corrente);
      bool bianco= true;
      while(bianco && ((degree= liste_adiacenza->degree_of(nodo_corrente)) > 0))
      {
        nodi_bianchi->set(nodo_corrente, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
        accertati_che( (errore==0), MSG_ERRORE_BITS);

        indice_nella_lista= liste_adiacenza->get_indice_lista_adiacenza(nodo_corrente) + degree - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

        for(unsigned long i=0; i < degree - 1; i++)
        {
          unsigned long figlio;

          figlio= vettore_adiacenti[indice_nella_lista];

          bool da_inserire= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
          accertati_che( (errore==0), MSG_ERRORE_BITS);

          if(da_inserire) //se non e' bianco e' inutile inserirlo
          {
            pila.push( indice_nella_lista); //locazione di appartenenza del nodo nel vettore degli adiacenti
            pila.push( figlio ); //nodo
          }
          else
            vettore_adiacenti[indice_nella_lista]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

          indice_nella_lista--;
        }

        nodo_corrente= vettore_adiacenti[indice_nella_lista];
//        degree= liste_adiacenza->degree_of(nodo_corrente);
        bianco= nodi_bianchi->get(nodo_corrente, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
        accertati_che( (errore==0), MSG_ERRORE_BITS);
      }

      if(!bianco)
      {
        vettore_adiacenti[indice_nella_lista]= ListaStatica::marca_X_adiac; //lo cancelliamo e basta: i suoi adiacenti non ci interessano
      }
      else //e' bianco, quindi deve avere degree==0  --> dobbiamo solo renderlo non-bianco
      {
        nodi_bianchi->set(nodo_corrente, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
        accertati_che( (errore==0), MSG_ERRORE_BITS);
      }
      
#else

      nodi_bianchi->set(rif_nodo_radice, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
      accertati_che( (errore==0), MSG_ERRORE_BITS);

     //mettiamo i dati per i nodi figli nella pila:
      unsigned long size_lista_nodo_radice= liste_adiacenza->degree_of(rif_nodo_radice);
      unsigned long indice_nella_lista_nodoradice= liste_adiacenza->get_indice_lista_adiacenza(rif_nodo_radice) + size_lista_nodo_radice - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

      for(unsigned long i=0; i < size_lista_nodo_radice ; i++)
      {
        bool bianco;
        unsigned long figlio;

        figlio= vettore_adiacenti[indice_nella_lista_nodoradice];

        bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
        accertati_che( (errore==0), MSG_ERRORE_BITS);

        if(bianco) //se non e' bianco e' inutile inserirlo
        {
          pila.push( indice_nella_lista_nodoradice ); //locazione di appartenenza del nodo nel vettore degli adiacenti
          pila.push( figlio ); //nodo
        }
        else
          vettore_adiacenti[indice_nella_lista_nodoradice]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

        indice_nella_lista_nodoradice--; //dobbiamo scandire la lista in senso contrario
      }
#endif

      while( !pila.isEmpty() )
      {
        bool ris;
        unsigned long pos_vett_adiacenti, nodo_top;

        //rimuoviamo l'arco che affiora sulla pila
        ris= pila.pop( nodo_top ); //l'arco compare in ordine inverso rispetto a quando inserito: quindi questa e' il nodo, poi viene la posizione
        accertati_che( ris, MSG_ERRORE_PILA);
        ris= pila.pop( pos_vett_adiacenti );
        accertati_che( ris, MSG_ERRORE_PILA);

        ris= nodi_bianchi->get(nodo_top, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
        accertati_che( (errore==0), MSG_ERRORE_BITS);

        if(ris) //se e' bianco allora ne espandiamo la lista dei successori sulla pila ; inoltre dobbiamo ricordarci dell'arco con cui l'abbiamo raggiunto, che fa parte dell'albero DFS. Se invece era non-bianco, allora lo abbiamo gia' espanso precedentemente e non dobbiamo farlo ancora
        {
#ifdef OPT_PRIMO_CAMMINO        
          unsigned long indice_nella_lista;
          unsigned long nodo_corrente= nodo_top;
          unsigned long degree; //= liste_adiacenza->degree_of(nodo_corrente);
          bool bianco= true;
          while(bianco && ((degree= liste_adiacenza->degree_of(nodo_corrente)) > 0))
          {
            nodi_bianchi->set(nodo_corrente, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
            accertati_che( (errore==0), MSG_ERRORE_BITS);

            indice_nella_lista= liste_adiacenza->get_indice_lista_adiacenza(nodo_corrente) + degree - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

            for(unsigned long i=0; i < degree - 1; i++)
            {
              unsigned long figlio;

              figlio= vettore_adiacenti[indice_nella_lista];

              bool da_inserire= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
              accertati_che( (errore==0), MSG_ERRORE_BITS);

              if(da_inserire) //se non e' bianco e' inutile inserirlo
              {
                pila.push( indice_nella_lista); //locazione di appartenenza del nodo nel vettore degli adiacenti
                pila.push( figlio ); //nodo
              }
              else
                vettore_adiacenti[indice_nella_lista]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

              indice_nella_lista--;
            }

            nodo_corrente= vettore_adiacenti[indice_nella_lista];
//            degree= liste_adiacenza->degree_of(nodo_corrente);
            bianco= nodi_bianchi->get(nodo_corrente, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
            accertati_che( (errore==0), MSG_ERRORE_BITS);
          }

          if(!bianco)
          {
            vettore_adiacenti[indice_nella_lista]= ListaStatica::marca_X_adiac; //lo cancelliamo e basta: i suoi adiacenti non ci interessano
          }
          else //e' bianco, quindi deve avere degree==0  --> dobbiamo solo renderlo non-bianco
          {
            nodi_bianchi->set(nodo_corrente, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
            accertati_che( (errore==0), MSG_ERRORE_BITS);
          }
          
#else

          nodi_bianchi->set(nodo_top, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
          accertati_che( (errore==0), MSG_ERRORE_BITS);

          //passiamo ad esaminare la sua lista di adiacenza:
          unsigned long size_lista_nodotop= liste_adiacenza->degree_of(nodo_top);
          unsigned long indice_nella_lista_nodo_top= liste_adiacenza->get_indice_lista_adiacenza(nodo_top) + size_lista_nodotop - 1;

          //inseriamo gli adiacenti bianchi nella pila e quindi cancelliamo con delle "X" l'intera lista di adiacenza appena esaminata: notare che l'ordine di priorita' delle liste di adiac. e' da SX a DX. Devo inserire nella pila in ordine contrario in modo che, rileggendo gli elementi nell'ordine inverso prodotto dalla pila, l'ordine con cui esamino gli archi sia alla fine quello voluto
          for(unsigned long i=0; i < size_lista_nodotop ; i++) //inseriamo gli archi che portano a nodi inesplorati e relativi a 'nodo_dest' ('nodo_dest' e' destinazione dell'arco appena estratto dalla pila, ma sorgente di quelli che andremo adesso a inserire): gli archi vanno inseriti allo stesso modo, prima src e poi dst. Inoltre vanno inseriti prima gli adiacenti alla fine della lista (quelli piu' a DX), poiche' la pila inverte l'ordine una volta che li estraggo
          {
             bool bianco;
             unsigned long figlio;

             figlio= vettore_adiacenti[indice_nella_lista_nodo_top];

             bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
             accertati_che( (errore==0), MSG_ERRORE_BITS);

             if(bianco)
             {
               pila.push( indice_nella_lista_nodo_top ); //locazione di appartenenza del nodo nel vettore degli adiacenti
               pila.push( figlio ); //nodo
             }
             else
               vettore_adiacenti[indice_nella_lista_nodo_top]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

             indice_nella_lista_nodo_top--;
          }
#endif

        }
        else
        { //l'arco corrispondente NON fa parte del'albero DFS e lo dobbiamo cancellare:          
          vettore_adiacenti[pos_vett_adiacenti]= ListaStatica::marca_X_adiac; 
        }
        
      } //fine del blocco while: la pila adesso e' vuota

    } //fine del blocco else

  } //file del ciclo for

  
#ifdef VERBOSE_MODE
  printf(" Remove marks...\n"); fflush(stdout); //*****
#endif
  liste_adiacenza->elimina_marche(); //riarrangiamo le liste di adiacenza in modo da eliminare le marche scritte durante la visita. In questo modo otteniamo la lista di adiacenza di un albero (la nuova foresta DFS con il "super root" padre di tutti gli alberi DFS)

  tag2= clock();  tot_tempo_dfs_internal += tag2 - tag1;
  conta_visite_DFS_memoria++;
  
  printf(MSG_DONE);
}



//le radici degli alberi DFS non sono presenti in memoria, ma per ipotesi le radici compaiono in ordine crescente di Id
void internal_DFS_con_radici_crescenti(VettoreBits * nodi_bianchi)
{
  printf(" Internal DFS (roots in increase order)...\n"); fflush(stdout);

  
  tag1= clock();
  
  //nodi_bianchi indica i nodi che sono bianchi. N.B. per il "super root" non serve il bit perche' non ha archi entranti.
  nodi_bianchi->reset(true); //Inizialmente i nodi sono tutti bianchi: inizializziamo il vettore con tutti uno

  accertati_che( (liste_adiacenza->capienzaMemoriaInutilizzata()>=elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  SwappedStack<unsigned long> pila( liste_adiacenza->capienzaMemoriaInutilizzata()*sizeof(unsigned long), liste_adiacenza->get_punt_mem_libera());

  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();

  unsigned long radice_corrente, conta_radici;
  conta_radici= 0;
  
  bool found= nodi_bianchi->findNext1(radice_corrente, 0);
  
  while(found)
  {
    long errore;
    bool esplorare_albero;

    esplorare_albero= nodi_bianchi->get(radice_corrente, errore); //vedo se il nodo e' bianco (i.e. non sono ancora stati espansi i suoi successori).
    assert(!errore);
    
    if(esplorare_albero)
    {
      //(!) non resettiamo il bit della radice corrente, perche' questa info. ci servira' poi per capire quali nodi sono delle radici
      //mettiamo i dati per i nodi figli nella pila:
      unsigned long size_lista_nodo_radice= liste_adiacenza->degree_of(radice_corrente);
      unsigned long indice_nella_lista_nodoradice= liste_adiacenza->get_indice_lista_adiacenza(radice_corrente) + size_lista_nodo_radice - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

      for(unsigned long i=0; i < size_lista_nodo_radice ; i++) //trattiamo a parte la lista del nodo radice, non lo mettiamo direttamente nella pila perche' esso non va marcato
      {
        bool bianco;
        unsigned long figlio;

        figlio= vettore_adiacenti[indice_nella_lista_nodoradice];

        if(figlio > radice_corrente)
        {
          bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
          accertati_che( (errore==0), MSG_ERRORE_BITS);
        }
        else //non puo' essere bianco perche' 'radice_corrente' e' il primo nodo (in ordine crescente di Id, i.e. il piu' "piccolo") ad essere bianco
          bianco= false;
          
        if(bianco) //se non e' bianco e' inutile inserirlo
        {
          pila.push( indice_nella_lista_nodoradice ); //locazione di appartenenza del nodo nel vettore degli adiacenti
          pila.push( figlio ); //nodo
        }
        else
          vettore_adiacenti[indice_nella_lista_nodoradice]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

        indice_nella_lista_nodoradice--; //dobbiamo scandire la lista in senso contrario
      }

      while( !pila.isEmpty() )
      {
        bool ris;
        unsigned long pos_vett_adiacenti, nodo_top;

        //rimuoviamo l'arco che affiora sulla pila
        ris= pila.pop( nodo_top ); //l'arco compare in ordine inverso rispetto a quando inserito: quindi questa e' il nodo, poi viene la posizione
        accertati_che( ris, MSG_ERRORE_PILA);
        ris= pila.pop( pos_vett_adiacenti );
        accertati_che( ris, MSG_ERRORE_PILA);

        if(nodo_top > radice_corrente)
        {
          ris= nodi_bianchi->get(nodo_top, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
          accertati_che( (errore==0), MSG_ERRORE_BITS);
        }
        else //non puo' essere bianco perche' 'radice_corrente' e' il primo nodo (in ordine crescente di Id, i.e. il piu' "piccolo") ad essere bianco
          ris= false;
        

        if(ris) //se e' bianco allora ne espandiamo la lista dei successori sulla pila ; inoltre dobbiamo ricordarci dell'arco con cui l'abbiamo raggiunto, che fa parte dell'albero DFS. Se invece era non-bianco, allora lo abbiamo gia' espanso precedentemente e non dobbiamo farlo ancora
        {
          nodi_bianchi->set(nodo_top, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
          accertati_che( (errore==0), MSG_ERRORE_BITS);

          //passiamo ad esaminare la sua lista di adiacenza:
          unsigned long size_lista_nodotop= liste_adiacenza->degree_of(nodo_top);
          unsigned long indice_nella_lista_nodo_top= liste_adiacenza->get_indice_lista_adiacenza(nodo_top) + size_lista_nodotop - 1;

          //inseriamo gli adiacenti bianchi nella pila e quindi cancelliamo con delle "X" l'intera lista di adiacenza appena esaminata: notare che l'ordine di priorita' delle liste di adiac. e' da SX a DX. Devo inserire nella pila in ordine contrario in modo che, rileggendo gli elementi nell'ordine inverso prodotto dalla pila, l'ordine con cui esamino gli archi sia alla fine quello voluto
          for(unsigned long i=0; i < size_lista_nodotop ; i++) //inseriamo gli archi che portano a nodi inesplorati e relativi a 'nodo_dest' ('nodo_dest' e' destinazione dell'arco appena estratto dalla pila, ma sorgente di quelli che andremo adesso a inserire): gli archi vanno inseriti allo stesso modo, prima src e poi dst. Inoltre vanno inseriti prima gli adiacenti alla fine della lista (quelli piu' a DX), poiche' la pila inverte l'ordine una volta che li estraggo
          {
             bool bianco;
             unsigned long figlio;

             figlio= vettore_adiacenti[indice_nella_lista_nodo_top];

             if(figlio > radice_corrente)
             {
               bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
               accertati_che( (errore==0), MSG_ERRORE_BITS);
             }
             else
               bianco= false;
             
             if(bianco)
             {
               pila.push( indice_nella_lista_nodo_top ); //locazione di appartenenza del nodo nel vettore degli adiacenti
               pila.push( figlio ); //nodo
             }
             else
               vettore_adiacenti[indice_nella_lista_nodo_top]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

             indice_nella_lista_nodo_top--;
          }

        }
        else
        { //l'arco corrispondente NON fa parte del'albero DFS e lo dobbiamo cancellare:
          vettore_adiacenti[pos_vett_adiacenti]= ListaStatica::marca_X_adiac;
        }

      } //fine del blocco while: la pila adesso e' vuota

      conta_radici++;
    } //fine if

    found= nodi_bianchi->findNext1(radice_corrente);
  } //file del ciclo while delle radici

  
#ifdef VERBOSE_MODE
  printf(" Remove all marks...\n"); fflush(stdout); //*****
#endif
  liste_adiacenza->elimina_marche(); //riarrangiamo le liste di adiacenza in modo da eliminare le marche scritte durante la visita. In questo modo otteniamo la lista di adiacenza di un albero (la nuova foresta DFS con il "super root" padre di tutti gli alberi DFS)


  liste_adiacenza->aggiungi_spazio_alla_lista_del_nodo(NODO_SUPER_ROOT, conta_radici); //aggiungiamo alle liste di adiacenza lo spazio per inserire le radici della foresta DFS

  //cerchiamo i nodi ancora bianchi, ovvero true e quindi non marcati durante la visita: essi sono le radici, quindi le inseriamo in memoria come figli del nodo "super root".
  unsigned long * punt_cella_nodo_radice= vettore_adiacenti + liste_adiacenza->get_indice_lista_adiacenza(NODO_SUPER_ROOT);
  found= nodi_bianchi->findNext1(radice_corrente, 0);
  assert(found);
  (*punt_cella_nodo_radice)= radice_corrente;
  punt_cella_nodo_radice++;

  for(unsigned long i=1; i < conta_radici; i++)
  {
    found= nodi_bianchi->findNext1(radice_corrente); //calcoliamo la radice corrente (il primo nodo non-marcato che incontriamo)
    assert(found);
    (*punt_cella_nodo_radice)= radice_corrente;
    punt_cella_nodo_radice++;
  }

#ifdef VERBOSE_DEBUG
printf("\ndopo ripristino radici:\n"); liste_adiacenza->print_adiacenti();
#endif

  tag2= clock();  tot_tempo_dfs_internal += tag2 - tag1;
  conta_visite_DFS_cheap++;
  
  assert(liste_adiacenza->numero_adiacenti_totali() == numero_nodi - archi_caricabili_alla_volta + minimo_archi_caricabili_alla_volta);
  printf(MSG_DONE);
}



void aggiungi_radici_alle_liste(VettoreBits * nodi_normali, const char * nome_mfile, unsigned long numParti_mfile, unsigned long quante_radici_nel_file, unsigned long conta_radici)
{
  unsigned long * punt_cella_nodo_radice_src, * punt_cella_nodo_radice_dest;
  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
  
  assert(liste_adiacenza->degree_of(NODO_SUPER_ROOT)==0); //*****
  liste_adiacenza->aggiungi_spazio_alla_lista_del_nodo(NODO_SUPER_ROOT, quante_radici_nel_file); //aggiungiamo alla lista di adiacenza del "super root" lo spazio massimo che puo' servire, poi eventualmente la accorciamo. N.B. per ipotesi c'e' abbastanza memoria per caricare tutte le radici che sono sul file

  punt_cella_nodo_radice_src= vettore_adiacenti + liste_adiacenza->get_indice_lista_adiacenza(NODO_SUPER_ROOT);

  carica_mfile_in_memoria(nome_mfile, numParti_mfile, punt_cella_nodo_radice_src, sizeof(unsigned long)*quante_radici_nel_file); //carichiamo tutte le potenziali radici della foresta

  unsigned long radici_ricopiate= 0;

  while(radici_ricopiate < conta_radici) //questo loop evita di ricopiare su se stesse le celle delle prime radici incontrate
  {
    long errore;
    bool is_radice= !( nodi_normali->get( *punt_cella_nodo_radice_src, errore) );
    assert(!errore);

    if(is_radice)
    {
      radici_ricopiate++;
      punt_cella_nodo_radice_src++;
    }    
    else
    {
      punt_cella_nodo_radice_dest= punt_cella_nodo_radice_src; //impostiamo il puntatore alla cella destinazione della copia
      punt_cella_nodo_radice_src++;
      break; //--> usciamo da questo loop
    }    
  }

  while(radici_ricopiate < conta_radici)
  {
    long errore;
    bool is_radice= !( nodi_normali->get( *punt_cella_nodo_radice_src, errore) );
    assert(!errore);

    if(is_radice)
    {
      (*punt_cella_nodo_radice_dest)= (*punt_cella_nodo_radice_src);
      punt_cella_nodo_radice_dest++;
      radici_ricopiate++;
    }
    
    punt_cella_nodo_radice_src++;
  }
  
  assert(punt_cella_nodo_radice_src - vettore_adiacenti - liste_adiacenza->get_indice_lista_adiacenza(NODO_SUPER_ROOT) <= quante_radici_nel_file);
  
  liste_adiacenza->accorcia_lista_del_nodo(NODO_SUPER_ROOT, quante_radici_nel_file - conta_radici); //togliamo lo spazio in piu'
}



//le radici degli alberi DFS non sono presenti in memoria, ma per ipotesi le radici compaiono in ordine crescente di Id
void internal_DFS_con_file_radici(VettoreBits * nodi_bianchi, const char * nome_mfile, unsigned long numParti_mfile, unsigned long quante_radici_nel_file)
{
  printf(" Internal DFS (file of tree roots)...\n"); fflush(stdout);

  tag1= clock();
  
  assert(quante_radici_nel_file > 0);

  unsigned long radice_corrente, conta_radici, radici_rimaste_nel_file, capienza_mem_libera, capienza_blocco_buffer_roots, capienza_blocco_pila;
  unsigned long * blocco_buffer_roots, * blocco_pila;
  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
    
  //nodi_bianchi indica i nodi che sono bianchi. N.B. per il "super root" non serve il bit perche' non ha archi entranti.
  nodi_bianchi->reset(true); //Inizialmente i nodi sono tutti bianchi: inizializziamo il vettore con tutti uno  

  //"allocazione" della memoria per gli oggetti usati:
  capienza_mem_libera= liste_adiacenza->capienzaMemoriaInutilizzata();  
  accertati_che( (capienza_mem_libera >= elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  capienza_blocco_buffer_roots=(unsigned long) (0.04f * capienza_mem_libera); //lasciamo un 25esimo della memoria al buffer lettura delle radici
  capienza_blocco_pila= capienza_mem_libera - capienza_blocco_buffer_roots;
  blocco_pila= liste_adiacenza->get_punt_mem_libera();
  blocco_buffer_roots= blocco_pila + capienza_blocco_pila; //(!) aritmetica dei puntatori..
    
  SwappedStack<unsigned long> pila( sizeof(unsigned long)*capienza_blocco_pila, blocco_pila);

  multiFileReader mf_input_roots((char *)nome_mfile, numParti_mfile);
  BufferLetturaMultifile<unsigned long> buf_input_roots(&mf_input_roots, sizeof(unsigned long)*capienza_blocco_buffer_roots, blocco_buffer_roots);
  

  conta_radici= 0;
  radici_rimaste_nel_file= quante_radici_nel_file;

  while(radici_rimaste_nel_file > 0)
  {
    long errore;
    bool esplorare_albero;

    do //cerchiamo la prima radice non marcata
    {
      bool ok= buf_input_roots.get(radice_corrente);
      accertati_che( ok, MSG_ERRORE_LETTURA_MFILE, nome_mfile);
      radici_rimaste_nel_file--;
      
      esplorare_albero= nodi_bianchi->get(radice_corrente, errore); //vedo se il nodo e' bianco (i.e. non sono ancora stati espansi i suoi successori).
      assert( !errore );    
    } while( !esplorare_albero && radici_rimaste_nel_file);

    if(esplorare_albero)
    {
      nodi_bianchi->set(radice_corrente, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
      accertati_che( (errore==0), MSG_ERRORE_BITS);
    
      //mettiamo i dati per i nodi figli nella pila:
      unsigned long size_lista_nodo_radice= liste_adiacenza->degree_of(radice_corrente);
      unsigned long indice_nella_lista_nodoradice= liste_adiacenza->get_indice_lista_adiacenza(radice_corrente) + size_lista_nodo_radice - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

      for(unsigned long i=0; i < size_lista_nodo_radice ; i++)
      {
        bool bianco;
        unsigned long figlio;

        figlio= vettore_adiacenti[indice_nella_lista_nodoradice];

        bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
        accertati_che( (errore==0), MSG_ERRORE_BITS);

        if(bianco) //se non e' bianco e' inutile inserirlo
        {
          pila.push( indice_nella_lista_nodoradice ); //locazione di appartenenza del nodo nel vettore degli adiacenti
          pila.push( figlio ); //nodo
        }
        else
          vettore_adiacenti[indice_nella_lista_nodoradice]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

        indice_nella_lista_nodoradice--; //dobbiamo scandire la lista in senso contrario
      }

      while( !pila.isEmpty() )
      {
        bool ris;
        unsigned long pos_vett_adiacenti, nodo_top;

        //rimuoviamo l'arco che affiora sulla pila
        ris= pila.pop( nodo_top ); //l'arco compare in ordine inverso rispetto a quando inserito: quindi questa e' il nodo, poi viene la posizione
        accertati_che( ris, MSG_ERRORE_PILA);
        ris= pila.pop( pos_vett_adiacenti );
        accertati_che( ris, MSG_ERRORE_PILA);

        ris= nodi_bianchi->get(nodo_top, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
        accertati_che( (errore==0), MSG_ERRORE_BITS);

        if(ris) //se e' bianco allora ne espandiamo la lista dei successori sulla pila ; inoltre dobbiamo ricordarci dell'arco con cui l'abbiamo raggiunto, che fa parte dell'albero DFS. Se invece era non-bianco, allora lo abbiamo gia' espanso precedentemente e non dobbiamo farlo ancora
        {
          nodi_bianchi->set(nodo_top, false, errore); //azzeriamo il bit corrispondente, il nodo e' diventato non-bianco
          accertati_che( (errore==0), MSG_ERRORE_BITS);

          //passiamo ad esaminare la sua lista di adiacenza:
          unsigned long size_lista_nodotop= liste_adiacenza->degree_of(nodo_top);
          unsigned long indice_nella_lista_nodo_top= liste_adiacenza->get_indice_lista_adiacenza(nodo_top) + size_lista_nodotop - 1;

          //inseriamo gli adiacenti bianchi nella pila e quindi cancelliamo con delle "X" l'intera lista di adiacenza appena esaminata: notare che l'ordine di priorita' delle liste di adiac. e' da SX a DX. Devo inserire nella pila in ordine contrario in modo che, rileggendo gli elementi nell'ordine inverso prodotto dalla pila, l'ordine con cui esamino gli archi sia alla fine quello voluto
          for(unsigned long i=0; i < size_lista_nodotop ; i++) //inseriamo gli archi che portano a nodi inesplorati e relativi a 'nodo_dest' ('nodo_dest' e' destinazione dell'arco appena estratto dalla pila, ma sorgente di quelli che andremo adesso a inserire): gli archi vanno inseriti allo stesso modo, prima src e poi dst. Inoltre vanno inseriti prima gli adiacenti alla fine della lista (quelli piu' a DX), poiche' la pila inverte l'ordine una volta che li estraggo
          {
             bool bianco;
             unsigned long figlio;

             figlio= vettore_adiacenti[indice_nella_lista_nodo_top];

             bianco= nodi_bianchi->get(figlio, errore); //vedo se il nodo dest e' bianco (i.e. non sono ancora stati espansi i suoi successori). Puo' accadere che, anche se i nodi vengono inseriti solo se bianchi, poi vengano estratti che non sono piu' bianchi: cio' avviene se dopo che il nodo dest e' stato inserito insieme ad altri adiacenti, venisse incontrato in qualche altro modo e visitato
             accertati_che( (errore==0), MSG_ERRORE_BITS);

             if(bianco)
             {
               pila.push( indice_nella_lista_nodo_top ); //locazione di appartenenza del nodo nel vettore degli adiacenti
               pila.push( figlio ); //nodo
             }
             else
               vettore_adiacenti[indice_nella_lista_nodo_top]= ListaStatica::marca_X_adiac; //cancelliamo la lista di adiac. del nodo radice corrente

             indice_nella_lista_nodo_top--;
          }

        }
        else
        { //l'arco corrispondente NON fa parte del'albero DFS e lo dobbiamo cancellare:
          vettore_adiacenti[pos_vett_adiacenti]= ListaStatica::marca_X_adiac;
        }

      } //fine del blocco while: la pila adesso e' vuota

      conta_radici++;
    } //fine if

  } //file del ciclo while delle radici

  mf_input_roots.close();
  
#ifdef VERBOSE_MODE
  printf(" removing marks...\n"); fflush(stdout); //*****
#endif
  liste_adiacenza->elimina_marche(); //riarrangiamo le liste di adiacenza in modo da eliminare le marche scritte durante la visita. In questo modo otteniamo la lista di adiacenza di un albero (la nuova foresta DFS con il "super root" padre di tutti gli alberi DFS)

  /* a questo punto il contenuto dei bits e':
     <false> per tutti i nodi raggiunti nella visita (radici comprese)
     <true> per tutti i nodi non raggiunti (perche' swappati)
     Percio' se marchiamo true tutti i nodi che hanno un padre, otteniamo:
     <false> solo per le radici
  */
  unsigned long * punt_nodo_dst= vettore_adiacenti; //punta al primo adiacente nelle liste di tutti i nodi
  unsigned long tot_adiacenti= liste_adiacenza->numero_adiacenti_totali();
  assert(tot_adiacenti==numero_nodi - (archi_caricabili_alla_volta - minimo_archi_caricabili_alla_volta) - conta_radici);
  for(unsigned long i=0; i < tot_adiacenti; i++)
  {
    long errore;
    nodi_bianchi->set( *punt_nodo_dst, true, errore);
    punt_nodo_dst++;
    assert( !errore );
  }

  aggiungi_radici_alle_liste(nodi_bianchi, nome_mfile, numParti_mfile, quante_radici_nel_file, conta_radici);

#ifdef VERBOSE_DEBUG
printf("\ndopo ripristino radici:\n"); liste_adiacenza->print_adiacenti();
#endif

  tag2= clock();  tot_tempo_dfs_internal += tag2 - tag1;
  conta_visite_DFS_file++;
  
  assert(liste_adiacenza->numero_adiacenti_totali() == numero_nodi - archi_caricabili_alla_volta + minimo_archi_caricabili_alla_volta);
  printf(MSG_DONE);
}



void prima_DFS_interna(bool visita_unica)
{  
  switch(modo_visita_alberi)
  {
    case DFS_single: //un solo albero DFS e' specificato: quello del nodo iniziale fornito dall'utente.
      accertati_che( false, "DFS_single not supported"); //*****
      break;

    case DFS_ascending: //consideriamo in ordine crescente tutti i nodi per iniziare nuovi alberi DFS
      internal_DFS_con_radici_crescenti(nodi_passivi);
      
      num_radici_salvate_su_file= numero_nodi; //(!) in pratica e' come se avessimo disponibile un file con le radici nell'ordine: 0,1,2, ..., N-1. Serve impostarlo per poter effettuare correttamente la scelta della prossima visita DFS interna.
      old_tipo_visita= DFS_radici_crescenti;
      break;

    case DFS_file_ord: //gli adiacenti del nodo "super root", nel loro ordine, vengono letti dal file di ordinamento specificato dall'utente. Le radici degli alberi DFS verranno cercate sulla base di tale ordinamento.
      num_radici_salvate_su_file= size_mfile_ordinamento / sizeof(unsigned long);
      accertati_che(num_radici_salvate_su_file==numero_nodi, "ERROR 1: File di ordinamento con meno di N nodi non ancora supportato"); //*****
      old_nome_mfile_radici= nome_mfile_ordinamento;
      old_numFile_radici= numFileOrd;
      old_tipo_visita= DFS_file_radici;
            
      internal_DFS_con_file_radici(nodi_passivi, nome_mfile_ordinamento, numFileOrd, num_radici_salvate_su_file);
      break;

    default:
      accertati_che( false, "Error: visit flag not valid"); //se siamo qui e' sempre per errore
      break;
  }

  old_archi_caricabili= archi_caricabili_alla_volta;
    
  is_dirty_vettore_bits= true; //abbiamo modificato con la visita il contenuto del vettore di bit

      
  if(visita_unica)
    return;
  //else..


  unsigned long * mem_riciclabile= liste_adiacenza->get_punt_mem_libera(); //riceviamo un puntatore alla memoria libera all'interno delle liste di adiacenza. Sono piu' di N elementi perche' in questo momento rappresentiamo un albero di N+1 nodi--> N adiacenti, e non ci sono marche. Il puntatore resta valido finche' non si modificano le liste di adiacenza
  unsigned long capienza_mem_libera= liste_adiacenza->capienzaMemoriaInutilizzata();
  accertati_che( (capienza_mem_libera>=archi_caricabili_alla_volta+elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
    
 tag1= clock();
  salva_albero_corrente(mem_riciclabile); //salviamo l'albero per l'iterazione successiva, specificandogli l'array (grande N) dove crearsi le informazioni da scrivere su disco.
 tag2= clock();  tot_tempo_salvataggio_foresta += tag2 - tag1;

  //ordinamento dei sottoalberi della foresta DFS attuale (sfruttato dalla fase successiva e anche da salva_risultati() ! Se i risultati non vengono salvati con lo stesso ordine degli archi con cui si e' elaborato il grafo, allora non sarebbero corretti poiche' possono comparire archi di attraversameto in avanti)
 tag1= clock();
  calcola_pesi_sottoalberi(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi); //calcoliamo, usando lo spazio delle liste di adiacenza attualmente non utilizzato (memorizzano un albero di N nodi, dunque N archi --> N posizioni libere), quanti discendenti ha ciascun nodo
 tag2= clock();  tot_calcolo_pesi += tag2 - tag1;
#ifdef CHECK_PESI
  controlla_pesi(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi);
#endif

  //la funzione di ordinamento usa il vettore dei bit passivi, percio' lo ripristiniamo se il suo contenuto e' stato modificato nel frattempo:

 tag1= clock();
  mergesort_adiacenti_per_peso(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi); //ordiniamo le liste di adiacenza in modo tale che i nodi con piu' discendenti ricevano maggiore priorita' nella lista (i.e. vengono messi prima)
 tag2= clock(); tot_tempo_ordinamento += tag2 - tag1;

#ifdef VERBOSE_DEBUG
  printf("\ndopo ordinamento in base a pesi:\n"); liste_adiacenza->print_adiacenti();
#endif
}



void rimozione_file_temporanei()
{
  eraseMultiFile(nome_mfile_alberoDFS_prec);
  eraseMultiFile(nome_mfile_Inizio_foresta_prec);
  eraseMultiFile(nome_mfile_Fine_foresta_prec);
  eraseMultiFile(nome_mfile_nodi_swap_passivi_InizioForesta);
  eraseMultiFile(nome_mfile_info_swap_passivi_InizioForesta);
  eraseMultiFile(nome_mfile_nodi_swap_passivi_FineForesta);
  eraseMultiFile(nome_mfile_info_swap_passivi_FineForesta);
  eraseMultiFile(nome_mfile_appoggio_radici);
  remove(nome_file_bits_passivi_prec);
  remove(nome_file_bits_padri_esclusi);
}



visiteDFS scelta_tipo_visita_DFS(const char * * nome_mfile_radici, unsigned long * numFile_radici, unsigned long * quante_radici_su_disco)
{
  unsigned long degree_SR;
  visiteDFS risposta;

  degree_SR= liste_adiacenza->degree_of(NODO_SUPER_ROOT);

  if(degree_SR < PERC_MINIMA_RADICI_OUT * numero_nodi) //se supponiamo che non valga la pena accedere al disco per guadagnare cosi' poco spazio
  {
    risposta= DFS_memoria;

    //il resto non interessa, ma per sicurezza lo assegnamo lo stesso: (infatti, una volta che il numero delle radici e' sceso al di sotto di questa soglia, non puo' piu' risalire e dunque la visita resta sempre di tipo 'DFS_memoria')
    (*nome_mfile_radici)= NULL;
    (*numFile_radici)= 0;
    (*quante_radici_su_disco)= 0;
  }
  else
  if((degree_SR < num_radici_salvate_su_file) && (archi_caricabili_alla_volta > old_archi_caricabili)) //in questo caso siamo costretti ad aggiornare il file perche' una delle radici che sono sul file e che mancano all'appello potrebbero essere state swappate: noi non siamo in grado di dire, leggendo il solo file vecchio, quali radici sono quelle swappate
  {
    risposta= DFS_file_radici;

    unsigned long * punt_memoria_da_salvare= liste_adiacenza->vettore_adiacenti() + liste_adiacenza->get_indice_lista_adiacenza(NODO_SUPER_ROOT);
    (*numFile_radici)= salva_su_mfile( nome_mfile_appoggio_radici, punt_memoria_da_salvare, sizeof(unsigned long)*degree_SR); //salviamo tutti gli adiacenti del "super root", ovvero tutte le radici attuali della foresta
    (*nome_mfile_radici)= nome_mfile_appoggio_radici;
    (*quante_radici_su_disco)= degree_SR;
    
    conta_scritture_radici_su_disco++; //*****
  }
  else //(degree_SR == num_radici_salvate_su_file) || (archi_caricabili_alla_volta == old_archi_caricabili)
  {
    assert((degree_SR == num_radici_salvate_su_file) || (archi_caricabili_alla_volta == old_archi_caricabili));

    risposta= old_tipo_visita; //ribadiamo la visita precedente e tutti i suoi vecchi parametri

    (*nome_mfile_radici)= old_nome_mfile_radici;
    (*numFile_radici)= old_numFile_radici;
    (*quante_radici_su_disco)= num_radici_salvate_su_file;
  }

  
  assert((risposta==DFS_memoria)||(old_tipo_visita!=DFS_memoria));

  //salviamo i parametri di questa visita:
  old_tipo_visita= risposta;
  old_nome_mfile_radici= *nome_mfile_radici;
  old_numFile_radici= *numFile_radici;
  num_radici_salvate_su_file= *quante_radici_su_disco;

  old_archi_caricabili= archi_caricabili_alla_volta;
  
  return risposta;
}



/*'prima_invocazione' indica se e' la prima volta che chiamiamo questa funzione (invocazione di questa funzione: sinonimo di "fase")
  ritorna true se ci sono modifiche rispetto all'invocazione precedente, false altrimenti.
  Se 'prima_invocazione'=true viene restituito sempre true, senza fare il confronto con la foresta di inizializzazione.
  Se 'solo_la_visita' viene passato =true, allora si suppone che basti una sola visita DFS perche' tutti gli archi del grafo sono gia' in memoria. In questo caso non viene fatto il riordinamento in base al peso, che rovinerebbe il risultato della visita, gia' definitivo.
  PRECondizioni: Si suppone che siano gia' stati caricati i nuovi archi prima di invocare questa funzione
                 E' gia' stata calcolata una prima foresta, e salvata con salva_albero_corrente()
*/
bool calcola_nuova_foresta(unsigned long max_archi_da_scandire, unsigned long * quanti_scanditi, unsigned long * quanti_caricati)
{
  unsigned long bonus_radici, numFile_radici, quante_radici_su_disco;
  const char * nome_mfile_radici= NULL; //punta alla stringa del nome
  visiteDFS visita_da_effettuare; //indica quale versione della visita DFS interna usare

  visita_da_effettuare= scelta_tipo_visita_DFS(&nome_mfile_radici, &numFile_radici, &quante_radici_su_disco);

  if(visita_da_effettuare == DFS_memoria)
    bonus_radici= 0;
  else //possiamo togliere temporaneamente dalla memoria le radici, che hanno la caratteristica favorevole di venire usate sequenzialmente durante la visita DFS interna. In questo modo quadagnamo dello spazio in piu' per caricare ulteriori archi in memoria
  {
    bonus_radici= liste_adiacenza->degree_of(NODO_SUPER_ROOT);
    liste_adiacenza->accorcia_lista_del_nodo(NODO_SUPER_ROOT, bonus_radici);    
  }
  
  carica_prossimi_archi(archi_caricabili_alla_volta + bonus_radici, max_archi_da_scandire, quanti_scanditi, quanti_caricati);

  somma_perc_caricamento_archi += ((float) *quanti_scanditi) / numero_archi;
  numero_caricamenti_archi_nel_round++;
  
  conta_spazio_guadagnato_radici += bonus_radici;//*****

#ifdef VERBOSE_MODE
  printf("\narchi scanditi nel caricamento: %lu (%.1f%%)\n", *quanti_scanditi, 100.0f * (*quanti_scanditi) / numero_archi);//*****
#endif

  if((*quanti_scanditi)==numero_archi) //*****
    printf("*** Letto un intero round durante il caricamento! ***\n");
  
  #ifdef VERBOSE_DEBUG
  printf("\ndopo il caricamento degli archi:\n"); liste_adiacenza->print_adiacenti();
  #endif


  switch(visita_da_effettuare)
  {
    case DFS_memoria:
      internal_DFS(nodi_passivi);
      break;

    case DFS_radici_crescenti: //consideriamo in ordine crescente tutti i nodi per iniziare nuovi alberi DFS
      internal_DFS_con_radici_crescenti(nodi_passivi);
      break;

    case DFS_file_radici: //gli adiacenti del nodo "super root", nel loro ordine, vengono letti da file
      internal_DFS_con_file_radici(nodi_passivi, nome_mfile_radici, numFile_radici, quante_radici_su_disco);
      break;

    default:
      accertati_che( false, "Errore: modo di visita DFS interna non valido"); //se siamo qui e' sempre per errore
      break;
  }

  is_dirty_vettore_bits= true; //abbiamo modificato il contenuto del vettore di bit

  #ifdef VERBOSE_DEBUG
  printf("\ndopo la DFS interna:\n"); liste_adiacenza->print_adiacenti();
  #endif

  assert(liste_adiacenza->numero_adiacenti_totali() == numero_nodi - archi_caricabili_alla_volta + minimo_archi_caricabili_alla_volta);

  
  //controllo dei cambiamenti subiti dall'albero rispetto all'invocazione (fase) precedente:
  if( (*quanti_scanditi) == numero_archi) //se nell'ultima visita effettuata abbiamo elaborato in pratica un intero round la foresta individutata e' quella finale (poiche' sono stati tenuti in conto tutti gli archi (significativi) in questa visita, essa non potra' piu cambiare in seguito al caricamento di altri archi)
    return false; //e' necessario forzare l'uscita, perche' il controllo della foresta si basa sull'euristica seguente: finche' e' evidente che la foresta e' cambiata dall'ultima volta essa non viene salvata; quando potenzialmente coincide con la precedente, la foresta corrente viene salvata su file per essere confrontata la foresta dell'iterazione successiva - a questo punto e' possibile fare un confronto esatto, non approssimato, e la coincidenza permette di terminare l'algoritmo. Forzando qui l'uscita, ci risparmiamo due iterazioni (con evntuale calcolo passivi) inutili.
  
  bool modifiche;  
  unsigned long capienza_mem_libera= liste_adiacenza->capienzaMemoriaInutilizzata();
  accertati_che( (capienza_mem_libera>=archi_caricabili_alla_volta+elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  unsigned long * mem_riciclabile= liste_adiacenza->get_punt_mem_libera(); //riceviamo un puntatore alla memoria libera all'interno delle liste di adiacenza. Sono piu' di N elementi perche' in questo momento rappresentiamo un albero di N+1 nodi--> N adiacenti, e non ci sono marche. Il puntatore resta valido finche' non si modificano le liste di adiacenza

  tag1= clock();
  modifiche= controlla_modifiche_albero(mem_riciclabile); //vediamo se la foresta salvata su file nella fase precedente coincide con quella calcolata in questa invocazione della funzione
  tag2= clock();  tot_tempo_salvataggio_foresta += tag2 - tag1;

  if(modifiche) //se non ci sono state modifiche l'albero e' gia' nell'ordine corretto e non va riordinato
  {
    tag1= clock();
    salva_albero_corrente(mem_riciclabile); //salviamo l'albero complessivo (foresta DFS) per l'iterazione successiva (siamo costretti ad aggiornare il file perche' la foresta DFS e' cambiata). Passiamo l'array (grande N) dove crearsi le informazioni da scrivere su disco
    tag2= clock();  tot_tempo_salvataggio_foresta += tag2 - tag1;

    //ordinamento dei sottoalberi della foresta DFS attuale (sfruttato dalla fase successiva e anche da salva_risultati() ! Se i risultati non vengono salvati con lo stesso ordine degli archi con cui si e' elaborato il grafo, allora non sarebbero corretti poiche' possono comparire archi di attraversameto in avanti)
    tag1= clock();
    calcola_pesi_sottoalberi(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi); //calcoliamo, usando lo spazio delle liste di adiacenza attualmente non utilizzato (memorizzano un albero di N nodi, dunque N archi --> N posizioni libere), quanti discendenti ha ciascun nodo
    tag2= clock();  tot_calcolo_pesi += tag2 - tag1;

    #ifdef CHECK_PESI
    controlla_pesi(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi);
    #endif

    //la funzione di ordinamento usa il vettore dei bit passivi, percio' lo ripristiniamo se il suo contenuto e' stato modificato nel frattempo:

    tag1= clock();
    mergesort_adiacenti_per_peso(mem_riciclabile, capienza_mem_libera - numero_nodi, mem_riciclabile + numero_nodi); //ordiniamo le liste di adiacenza in modo tale che i nodi con piu' discendenti ricevano maggiore priorita' nella lista (i.e. vengono messi prima)
    tag2= clock();  tot_tempo_ordinamento += tag2 - tag1;

    #ifdef CHECK_PASSIVI //*******
    controlla_passivi("Check nodi passivi dopo l'ordinamento");
    #endif
    #ifdef VERBOSE_DEBUG
    printf("\ndopo ordinamento in base a pesi:\n"); liste_adiacenza->print_adiacenti();
    #endif
  }

  return modifiche;
}



void semiext_DFS()
{
  unsigned long ultimi_selezionati;
  unsigned long long totale_archi_scanditi= 0, totale_archi_selezionati= 0;
  unsigned long numero_round_effettuati= 0;
  bool changes;
  unsigned long num_iterazioni= 1;
  bool interruzione_visita= false;

  successori_restanti_info_corrente= 0;
  info_scandite_nel_round= 0;
  archi_restanti_nel_file= numero_archi;

  time(&tempo_inizio); //salviamo il tempo corrente
  richiesta_controlc= false;
//  signal(SIGINT, richiesta_chiusura);

  //allochiamo la lista di adiacenza ed il vettore di bit:
  alloca_memoria();

  init_modulo_nodi_passivi(); //inizializziamo le variabili che riguardano il calcolo dei nodi passivi. N.B. oiche' inizializziamo il vettore dei passivi, la sua memoria deve essere gia' stata allocata

  foresta_prec_salvata= false;
  valore_test_modifiche_albero_prec= (unsigned long) -1; //N.B. questo valore iniziale influisce solo sul "quick test avanzato", nel determinare se la prima foresta calcolata verra' salvata su disco. Cio' non influenza la correttezza dell'algoritmo, ma siccome e' probabile che venga rilevata una modifica alla foresta con il quick test successivo, e' meglio evitare la scrittura su disco. Basta inizializzare con un valore che la funzione valore_test_modifiche_albero_prec() non puo' mai restituire.

#ifdef VERBOSE_DEBUG
printf("\ndopo l'inizializzazione:\n"); liste_adiacenza->print_adiacenti();
#endif

  if(numero_archi <= archi_caricabili_alla_volta + numero_nodi) // si fa con un solo passaggio, e' in pratica quasi come la DFS standard (i nodi passivi non vengono nemmeno calcolati)
  {
    //printf("\n\n * Iterazione 1/1:\n\n");
    printf("Iteraton 1/1:\n");

    caricamento_primi_archi(numero_archi);
    prima_DFS_interna(true); //prima e unica visita (niente riordinamento delle liste in base ai pesi)

    //solo per le statistiche:
    totale_archi_selezionati= totale_archi_scanditi= numero_archi;
  }
  else // M > archi_caricabili_alla_volta: gli archi vanno letti in piu' passate da N archi alla volta
  {
    unsigned long archi_scanditi_senza_cambiamenti; //, archi_scanditi_da_ultimo_calcolo_passivi;
    printf("Iteraton 0:\n");

    #ifndef DISABLE_ROOTS
    caricamento_primi_archi(archi_caricabili_alla_volta + numero_nodi); //nella prima iterazione possiamo cariacare fino a 3N archi, perche' possiamo posticipare alla fine della prima DFS interna il caricamento delle radici
    prima_DFS_interna(false); //viene fatto anche il riordinamento delle liste in base ai pesi

    totale_archi_scanditi= totale_archi_selezionati= archi_caricabili_alla_volta + numero_nodi;
    #else
    //*****
    caricamento_primi_archi(archi_caricabili_alla_volta); //nella prima iterazione possiamo cariacare fino a 3N archi, perche' possiamo posticipare alla fine della prima DFS interna il caricamento delle radici
    prima_DFS_interna(false); //viene fatto anche il riordinamento delle liste in base ai pesi
    totale_archi_scanditi= totale_archi_selezionati= archi_caricabili_alla_volta;
    #endif
    
    archi_scanditi_senza_cambiamenti= totale_archi_scanditi;

    ultimo_nodo_leftmost_path= liste_adiacenza->get_adiac(NODO_SUPER_ROOT, 0);
    interruzione_visita= uscita_anticipata(num_iterazioni, totale_archi_scanditi/numero_archi);

    while( (archi_scanditi_senza_cambiamenti < numero_archi) && !interruzione_visita ) //se ho gia' esaminato una volta tutti gli archi e non ho avuto cambiamenti nella foresta, allora questa non cambiera' piu'
    {
      unsigned long archi_scanditi_ultima_volta;

#ifndef DISABLE_PASSIVES
      gestione_ricerca_nodi_passivi(totale_archi_scanditi); //esegue il calcolo di nuovi nodi passivi (ad ogni round, se necessario). N.B. anticipando qui questo controllo evito di effettuare il calcolo dei nodi passivi se sto uscendo perche' ho finito
#endif

      printf("Iteration %ld:\n", num_iterazioni);
      changes= calcola_nuova_foresta(numero_archi - archi_scanditi_senza_cambiamenti, &archi_scanditi_ultima_volta, &ultimi_selezionati);
      totale_archi_scanditi += archi_scanditi_ultima_volta;
      totale_archi_selezionati += ultimi_selezionati;
      
      if(changes)
        archi_scanditi_senza_cambiamenti= 0; //riazzeriamo il conteggio quando scopriamo delle modifiche
      else
        archi_scanditi_senza_cambiamenti += archi_scanditi_ultima_volta;

      num_iterazioni++;

      if(totale_archi_scanditi / numero_archi > numero_round_effettuati)
      {

        printf("\n Completed round %lu\n", numero_round_effettuati);
        numero_round_effettuati= totale_archi_scanditi / numero_archi;
      }

      interruzione_visita= uscita_anticipata(num_iterazioni, totale_archi_scanditi/numero_archi);
    }
  }
  
  if(interruzione_visita)
  {
#ifdef VERBOSE_MODE
    printf("Warning: saving an approximation of visit\n");
#endif

    if(changes) //l'ultima volta la foresta e' stata riordinata in base al peso, quindi va rimessa in maniera dfs-like
    {
#ifdef VERBOSE_MODE      
      printf("\n> Ripristino della foresta \"approssimata\"\n");
#endif
      internal_DFS(nodi_passivi); //ricalcoliamo l'ultima foresta
    }
  }

#ifdef CHECK_PASSIVI
  controlla_passivi("Check nodi passivi FINE");
#endif

#ifdef VERBOSE_MODE
  printf(" Ripristino dei nodi swappati..."); fflush(stdout);
#endif

  ripristina_passivi_swappati(); //ripristina tutti i nodi della foresta che erano stati swappati su disco perche' della parte finale stabile.
  printf(MSG_DONE);
#ifdef VERBOSE_DEBUG
  printf("\ndopo il ripristino dei nodi Swappati:\n"); liste_adiacenza->print_adiacenti();
#endif

  salva_risultati(nome_mfile_output); //scrive l'ultima e definitiva foresta DFS calcolata

  extern unsigned long conta_calcoli_NPE; //*****

#ifdef VERBOSE_MODE
  printf("\n- Visita completata.\n\n> Sono stati letti in tutto %llu archi\n> Il rapporto \"r\" tra archi scanditi ed archi totali e' %.2f\n> Sono stati selezionati in tutto %llu archi \n", totale_archi_scanditi, ((float)totale_archi_scanditi)/numero_archi, totale_archi_selezionati);
  printf("- Numero di alberi: %lu\n", liste_adiacenza->degree_of(NODO_SUPER_ROOT));
  printf("> Archi letti complessivamente*: %lu\n", totale_archi_scanditi + numero_archi*conta_calcoli_NPE);
#endif

tot_tempo_dfs_internal -= tot_tempo_elimina_marche; //*****

#ifdef VERBOSE_MODE
  fprintf(stdout, "\n- Tempi impiegati:\n  Visita DFS internal: %.0f msec \n  Eliminazione marche: %.0f msec\n  Calcolo pesi: %.0f msec \n  Ordinamento: %.0f msec \n  Caricamento archi: %0.f msec \n  Salvataggio foresta: %0.f msec \n  Confronto foresta: %0.f msec \n  Calcolo nodi passivi (NPE): %0.f msec \n  Calcolo nodi passivi (Initial Part): %0.f msec \n  Calcolo nodi passivi (Final Part): %0.f msec \n    Parziali NPE> max_reached: %0.f msec  ;  # preord: %0.f msec\n", ((float)tot_tempo_dfs_internal/CLOCKS_PER_SEC)*1000.0, ((float)tot_tempo_elimina_marche/CLOCKS_PER_SEC)*1000.0, ((float)tot_calcolo_pesi/CLOCKS_PER_SEC)*1000.0, ((float)tot_tempo_ordinamento/CLOCKS_PER_SEC)*1000.0, ((float)tot_tempo_caricamento_archi/CLOCKS_PER_SEC)*1000.0, ((float)tot_tempo_salvataggio_foresta/CLOCKS_PER_SEC)*1000.0, ((float)tot_tempo_controllo_modifiche/CLOCKS_PER_SEC)*1000.0, ((float)tot_calcolo_passivi_NPE/CLOCKS_PER_SEC)*1000.0, ((float)tot_calcolo_passivi_InitialPart/CLOCKS_PER_SEC)*1000.0, ((float)tot_calcolo_passivi_FinalPart/CLOCKS_PER_SEC)*1000.0, ((float)tot_max_preord_calcolo_passivi_NPE/CLOCKS_PER_SEC)*1000.0, ((float)tot_numero_preordine_NPE/CLOCKS_PER_SEC)*1000.0);
#endif


  libera_memoria();

  //cancellazione dei file temporanei:
  rimozione_file_temporanei();
}



void print_flags()
{
  printf("\n");
#ifdef TEST_MEMORIA_MIN
  printf("Flag TEST_MEMORIA_MIN\n");
#endif
#ifdef USA_DIR_CORR_COME_TMP
  printf("Flag USA_DIR_CORR_COME_TMP\n");
#endif
#ifdef DISABLE_PASSIVES
  printf("Flag DISABLE_PASSIVES\n");
#endif
#ifdef CHECK_PASSIVI
  printf("Flag CHECK_PASSIVI\n");
#endif
#ifdef CHECK_PESI
  printf("Flag CHECK_PESI\n");
#endif
#ifdef OPT_PRIMO_CAMMINO
  printf("Flag OPT_PRIMO_CAMMINO\n");
#endif
#ifdef NDEBUG
  printf("Flag NDEBUG\n");
#endif
#ifdef DISABLE_ROOTS
  printf("Flag DISABLE_ROOTS\n");
#endif
#ifdef NPE_MAGIC
  printf("Flag NPE_MAGIC\n");
#endif
#ifdef NPEMAGIC_SIMPLE
  printf("Flag NPEMAGIC_SIMPLE\n");
#endif

}



void stampaHelp(int arg,char** args){
  arg=0;
  printf("Utility that performs a dfs of a graph using a semiexternal algorithm\n");
  printf("For each node at least (12 + 1/8) bytes are required\n");
  printf("\nUsage:\n");
  printf("%s <memory> <inputGraphMultiFile%%> <outputSccMultifile%%> [Options]... \n", args[0]);
  printf("<memory>:                 available memory in MB\n");
  printf("<inputGraphMultiFile%%>:  input graph in IPS format\n");
  printf("<outputSccMultifile%%>:   output dfs forest in adjacency list format\n");
  printf("\nOptions:\n"
         "-b, -backward      to execute a backward visit\n"
         "-f, -output_f      nodes are listed in the order they are found the first time instead of the order they finish their DFS calls\n"
         "-I, -inv_ord_out   reverse order of nodes (from last to first) \n"
         "    -force         executes the visit even if the main memory is not enough\n"
         "-n, -sourceNode=<N>       performs a DFS starting from node <N>\n"
         "-o, -file_ord=<fileOrd%%> visit the DFS forest following the nodes order specified in <fileOrd%%>\n\n"
         "-r, -roundsMax=<num>       numbers of rounds after which the utility stops\n"
         "-t, -iterationsMax=<num>   max number of iterations to do\n"
         "-m, -minutesMax=<minuti>   max elapsed time after which the utility stops\n"
         "\n");
         
/*  printf("Per default la visita esplora tutti gli alberi DFS in ordine\n"
         "crescente di nodo sorgente\n\n");*/
         
  printf("\n%% implies that multifile name must contains %%d\n");
}


//uscita dal programma per errore
#define ERRORE_PAR(STRINGA) {fprintf(stderr,"%s\nUse -h for help\n",STRINGA);EXIT(1);}


int main(int argc,char* argv[])
{
    stampaLicenza(argc,argv);

//    signal(SIGABRT, chiudi_tutto); //permette di uscire ordinatamente in seguito ad una assert() fallita o ad altro errore
    
    //poniamo i valori di default per i flag
    visita_avanti= true;
    esegui_comunque= false;
    output_nodo_quando_scoperto= true;
    ordine_inverso_output= false;
    modo_visita_alberi= DFS_ascending;
    specificato_max_tempo= false;
    specificato_max_iterazioni= false;
    specificato_max_rounds= false;
    
    int modalita_scelte= 0; //bitmask che serve per capire se abbiamo usato piu' opzioni contrastati
    int temp;
    
    struct option long_options[] =
    {
      {"backward", no_argument, NULL, 'b'},
      {"output_f", no_argument, NULL, 'f'},
      {"inv_ord_out", no_argument, NULL, 'I'},
      {"force", no_argument, NULL, '!'},
      {"sourceNode",  required_argument, NULL, 'n'},
      {"file_ord",  required_argument, NULL, 'o'},
      {"roundsMax",  required_argument, NULL, 'r'},
      {"iterationsMax",  required_argument, NULL, 't'},
      {"minutesMax",  required_argument, NULL, 'm'},      
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index= 0;

    int car;

    do //esaminiamo le opzioni
    {
      car= getopt_long_only(argc, argv, "bfIn:o:r:t:m:", long_options, &option_index);

      switch(car)
      {
        case 'b':
            visita_avanti= false;
          break;

        case 'f':
            output_nodo_quando_scoperto= false;
          break;

        case 'I':
            ordine_inverso_output= true;
          break;
          
        case '!':
            esegui_comunque= true;
          break;

        case 'n':
          accertati_che(false, "option not implemented.."); //*****
          modo_visita_alberi= DFS_single;
          nodo_start= atol( optarg );
          modalita_scelte |= 1;
          break;

        case 'o':
          modo_visita_alberi= DFS_file_ord;
          nome_mfile_ordinamento= optarg;
          if(!estMultiFileName(nome_mfile_ordinamento))
            ERRORE_PAR("Multifile name <fileOrd> must contains '%d' ");

          numFileOrd= contaNumFilesMultiFile(nome_mfile_ordinamento);
          if(numFileOrd<=0)
            ERRORE_PAR("<fileOrd> is empty");

          size_mfile_ordinamento= getTotalSizeMultiFiles(nome_mfile_ordinamento);
          if(size_mfile_ordinamento % sizeof(unsigned long))
            ERRORE_PAR("file dimension of <fileOrd> must by divisible by 4");
          
          modalita_scelte |= 2;
          break;

        case 'r':
          specificato_max_rounds= true;
          temp= atol(optarg);
          if(temp < 1)
            ERRORE_PAR("number of rounds must be >= 1");
          rounds_max= temp;
          break;

        case 't':
          specificato_max_iterazioni= true;
          temp= atol(optarg);
          if(temp < 1)
            ERRORE_PAR("number of iterations must be >= 1");
          iterazioni_max= temp;
          break;

        case 'm':
          specificato_max_tempo= true;
          temp= atol(optarg);
          if(temp < 1)
            ERRORE_PAR("time must be >= 1");
          tempo_max= temp;
          break;

                                                                                
        case -1: //fine delle opzioni, ignoriamo il valore
          break;

        case '?': //parametro sconosciuto: lo ignoriamo, tanto il messaggio d'errore lo stampa getopt()
          EXIT(1);
          break;

        default:
          fprintf(stderr, "Parameters error\n");
          EXIT(1);
      }

    } while(car != -1);

    if(modalita_scelte == 3) //vede se ci sono due opzioni che chiedono modalita' di visita diverse
    {
      fprintf(stderr, "-file_ord and -sourceNode can't be used together \n");
      EXIT(1);
    }

    if( argc - optind !=  3) //contiamo quanti parametri che non sono opzioni restano: sono tutti i parametri dall'indice "optind" incluso in poi
      ERRORE_PAR("Invalid number of parameters.\n");

    //optind indica il primo dei parametri non-opzioni
    memoria_max= atol(argv[optind + 0]) * ((long long) 1024)*1024; //salva convertendo da MB a byte
    char * grafo=         argv[optind + 1];
    nome_mfile_output= argv[optind + 2];
    sprintf(grafoINFO,"%s.info",grafo);
    char grafoPRED[1000]; sprintf(grafoPRED,"%s.pred",grafo);
    char grafoSUCC[1000]; sprintf(grafoSUCC,"%s.succ",grafo);

    //controllo dei multifile passati come argomento:
    if (!estMultiFileName(grafo)) ERRORE_PAR("Input graph name must contains '%d' ");
    if (!estMultiFileName(nome_mfile_output)) ERRORE_PAR("Output multifile must contains '%d' ");

    numINFO= contaNumFilesMultiFile(grafoINFO);
    if (numINFO<=0)   		ERRORE_PAR("Empty graph");

    if(visita_avanti)
    {
      grafoADIAC= grafoSUCC;
      numADIAC= contaNumFilesMultiFile(grafoSUCC);
      if(numADIAC<=0)
        ERRORE_PAR("SUCC files not found for the input graph");
    }
    else //visita in backward
    {
      grafoADIAC= grafoPRED;
      numADIAC= contaNumFilesMultiFile(grafoPRED);
      if(numADIAC<=0)
        ERRORE_PAR("PRED files not found for the input graph");
    }


    long long nnodi= getTotalSizeMultiFiles(grafoINFO)/sizeof(struct grafoInfo);
    long long narchi= getTotalSizeMultiFiles(grafoADIAC)/sizeof(long);
    accertati_che( (nnodi < MAX_ULONG), "Too many nodes to use unsigned long");
    accertati_che( (narchi < MAX_ULONG), "too many edges to use unsigned long");

    numero_nodi= (unsigned long) nnodi;
    numero_archi= (unsigned long) narchi;

#ifdef VERBOSE_MODE    
    printf("\n> Il grafo ha %lu nodi e %lu archi.\n", numero_nodi, numero_archi);
    printf("visita_avanti:%d, output_nodo_quando_scoperto:%d, modo_visita_alberi (DFS_single|DFS_ascending|DFS_file_ord):%d\n", visita_avanti, output_nodo_quando_scoperto, modo_visita_alberi);
    printf("\nnome_mfile_ordinamento: %s,  nodo_start:%ld \n esegui_comunque=%d\n", nome_mfile_ordinamento, nodo_start, esegui_comunque);
#endif


    accertati_che( ((modo_visita_alberi != DFS_single) || (nodo_start < numero_nodi)), "Il nodo di partenza <%l> e' fuori dal range consentito per questo grafo [0-%l].", nodo_start, numero_nodi-1);

#ifdef VERBOSE_MODE
    print_flags(); //per vedere che versione stiamo usando
#endif

    semiext_DFS();

#ifdef VERBOSE_MODE
    print_flags(); //per vedere che versione stiamo usando
#endif

//******
#ifdef VERBOSE_MODE
  printf("\n * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n");
  printf(" controlli foreste evitati: %lu / %lu  ;  foreste salvate:%lu  ;  controlli nulli:%lu\n\n", num_controlli_evitati, foreste_differenti, num_foreste_salvate, num_confronti_incerti_foreste);
  print_statistiche_passivi();
  printf(" - Radici - Num. scritture su disco: %lu  ;  Spazio guadagnato:%lu \n   # Visite DFS:  memoria=%lu, ordine_cresc=%lu, con file radici=%lu\n", conta_scritture_radici_su_disco, conta_spazio_guadagnato_radici, conta_visite_DFS_memoria, conta_visite_DFS_cheap, conta_visite_DFS_file);
  printf(" Sono state ordinate %lu liste usando il buffer d'appoggio\n", conta_nodi_ordinati_con_buffer_corto);
  printf(" * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n");
#endif

                
    EXIT(0);
}

