#ifndef MAIN_H
#define MAIN_H

#include <stdio.h>
#include <assert.h>
#include "comuni.h"
#include "../common/scc_utl/BufferLetturaMultifile.h"
#include "../common/scc_utl/BufferScritturaMultifile.h"
#include "../common/scc_utl/VettoreBits.h"
#include "../common/scc_utl/BufferAppoggio.h"
#include "../common/scc_utl/SwappedStack.h"


//#define USA_DIR_CORR_COME_TMP //usa la dir. corrente per i file temporanei e non /tmp

//#define NDEBUG //flag usato anche da assert, se definito indica che non stiamo sviluppando e non serve il controllo per il debugging
//flags per il debugging:
//  #define TEST_MEMORIA_MIN //alloca per la lista di adiac. la memoria minima per funzionare, anche se ce ne fosse di piu'
//  #define VERBOSE_DEBUG //per il debugging: mostra passo passo le modifiche alla lista di adiacenza
//  #define DISABLE_PASSIVES //salta il calcolo dei nodi passivi
//  #define DISABLE_ROOTS //esegue soltanto DFS in memoria
//  #define CHECK_PASSIVI
//  #define CHECK_PESI
//#define VERBOSE_MODE



//implementazioni alternative delle funzioni:

#define NPE_MAGIC //versione di calcola_max_preordine_raggiunto() che ricicla il vettore di input per l'output; le prestazioni in termini di passivi calcolati sono le stesse (o migliori) della versione che ricorre al disco per salvare l'output
//#define NPEMAGIC_SIMPLE //versione di NPE_MAGIC che invece di impostare la soglia al massimo valore associato ai figli, l'imposta al valore dell'ultimo nodo, che dovrebbe avere probabilmente il massimo valore tra i figli
#define OPT_PRIMO_CAMMINO  
  
  

//definizione costanti:

const unsigned long MAX_ULONG= (unsigned long)((long)-1); //e' 0xFFFFFFFFh, e viene usato come valore speciale

#define NODO_SUPER_ROOT numero_nodi //il nodo fittizio, radice di tutti gli alberi della foresta DFS
#define INTERVALLO_REFRESH_INDICATORE ((unsigned long) (0.10f*numero_nodi)) //percentuale di nodi da processare prima di aggiornare la progress bar (usato nelle operazioni che scandiscono sequenzialmente tutti i nodi del grafo)

const unsigned long DIM_MEMORIA_RISERVATA_BUFFER= 2*1024*1024; //memoria riservata per i vari buffers che utilizzeremo di volta in volta, come ad esempio per caricare nuovi archi in memoria. Valori alti migliorano gli accessi al disco, ma tolgono memoria destinabile a caricare archi, rallentando cosi' la convergenza dell'algoritmo

//******
#ifndef DISABLE_ROOTS
const float PERC_MINIMA_RADICI_OUT= 0.05f; //decide quante visite vengono fatte con le radici mantenute in memoria: mettendo la percentuale al 100% vengono fatte soltanto visite DFS con le radici in memoria (tranne per quella iniziale che e' indipendente). Usata da scelta_tipo_visita_DFS(). Valori bassi migliorano la convergenza dell'algoritmo, ma i valori troppo bassi provocano scritture su disco con miglioramenti trascurabili alla convergenza
#else
const float PERC_MINIMA_RADICI_OUT= 1.0f; //******
#endif



const char EXT_INFOALBERO[]=".infotree"; //estensione da aggiungere al file di output con le informazioni sugli alberi DFS, rispetto al file di output principale
const char MSG_ERRORE_PILA[]= "Read error - stack";
const char MSG_ERRORE_BITS[]= "Error during access to the bit vector";
const char MSG_ERRORE_SALVA_BITS[]= "Write error during saving of bit vector";
const char MSG_ERRORE_LETTURA_BITS[]= "Read error during loading of bit vector";
const char MSG_ERRORE_SEEK[]= "Seek error on multifile %s";
const char MSG_ERR_INS_ROOTS2[]= "Internal adjacency list error (invalid number of elements)";//.Errore: la lista di adiacenza del nodo \"super root\" non contiene il numero di elementi previsto
const char MSG_ERR_MEM_LIBERA_RISERVATA[]= "Internal adjacency list error (memory not enough)";;//"Errore: la memoria libera della lista di adiacenza e' meno dello spazio minimo garantito";
const char MSG_ERRORE_LETTURA_MFILE[]= "Read error on multifile %s.";
const char MSG_ERRORE_SCRITTURA_MFILE[]= "Write error on multifile %s.";
const char MSG_DONE[]= "";      
  
struct  grafoInfo{
  long inDeg;
  long outDeg;
  long long puntIn;
  long long puntOut;
};


struct Arco{
    unsigned long source;
    unsigned long dest;
};


struct InfoAlbero { //informazioni relative a ciascun albero DFS visitato, da salvare nel file .infotree dell'output
  long num_nodi_albero; //dimensione dell'albero
  long punt; //indice del primo nodo dell'albero nel file di output principale
};


struct SwapPassiviInfo{
  unsigned long nodoPadre;
  unsigned long numFigliSwappati;
};
    
    

//funzioni template: vanno definite qui se   le vogliamo usare in piu' file
    
template <class T>
T MIN(T a, T b)
{
  if(a<b)
    return a;
  else
    return b;
}

template <class T>
T MAX(T a, T b)
{
  if(a>b)
    return a;
  else
    return b;
}

template <class T>
T MIN(T a, T b, T c)
{
  T minimo= a;

  if(b<minimo)
    minimo= b;

  if(c<minimo)
    minimo= c;

  return minimo;
}



enum treeDFS { DFS_single, DFS_ascending, DFS_file_ord};



extern treeDFS modo_visita_alberi; //*****

extern unsigned long numero_nodi; //numero nodi del grafo
extern unsigned long numero_archi; //numero archi del grafo

extern unsigned long archi_caricabili_alla_volta; //quanti archi(<-->adiacenti) posso caricare ogni volta accanto a quelli della foresta DFS. Non puo' essere minore di N (=numero_nodi)
extern unsigned long minimo_archi_caricabili_alla_volta; //mantiene il valore iniziale di 'archi_caricabili_alla_volta' e non viene piu' modificato
extern unsigned long elementi_riservati_buffer;

extern char grafoINFO[1000];
extern char * grafoADIAC; //grafo dei successori o dei predecessori a seconda dei casi
extern bool visita_avanti; //seguiamo gli archi in avanti o all'indietro?
extern long numINFO; //numero parti multifile dei nodi
extern long numADIAC; //numero parti multifile predecessori/successori

extern unsigned long elementi_riservati_buffer; //elementi allocati dalla lista di adiacenza, ma che non verranno riempiti di adiacenti in modo da avere sempre un buffer minimo con la memoria libera della lista di adiacenza

extern BufferLetturaMultifile<struct grafoInfo> * buf_info;
extern BufferLetturaMultifile<unsigned long> * buf_adiac;

extern VettoreBits * nodi_passivi;




//definizione della macro per verificare se ci sono errori. Il primo parametro e' la condizione da verificare e che deve risultare vera, il 2o, 3o, ecc. sono passati a printf per stampare i vari messaggi
#define accertati_che(a, resto...)  \
  if(!(a))                               \
  {  gestione_errore( resto ); }




void indiceProgressione(const char* testo,long perc1,long perc2);        
void libera_memoria();
void gestione_errore(const char *formato, ...);
void alloca_buffers_lettura_IPS(unsigned long * inizio_memoria_buffers, unsigned long dimensione_memoria_tot_buffers);
void libera_buffers_lettura_IPS();
void eraseMultiFile(const char* nomeMulti);
void carica_mfile_in_memoria(const char * nome_mfile, unsigned long parti_mfile, void * vettore_memoria, unsigned long size_tot_mfile);
unsigned long salva_su_mfile(const char * nome_mfile, void * vettore_memoria, unsigned long size_memoria);

#endif
