#include "main.h"
#include "ListaStatica.h"
#include "output_risultati.h"


static BufferScritturaMultifile<unsigned long> * buf_out= NULL;
static BufferScritturaMultifile<struct InfoAlbero> * buf_infotree= NULL;
static unsigned long num_nodi_output= 0;
static bool abilita_scoperta; //indica se l'evento che produce l'output e' la scoperta o la terminazione del nodo. Dipende da che tipo di output vogliamo: nodi scoperti/completati ; ordine normale/inverso


//per vedere le variabili definite altrove:
extern ListaStatica * liste_adiacenza;
extern bool output_nodo_quando_scoperto;
extern bool ordine_inverso_output;




//funzione invocata da output_albero() al momento della scoperta di un nuovo nodo
inline void scoperta_nodo(unsigned long nodo)
{
  if(abilita_scoperta)
  {
    buf_out->put(nodo);
    num_nodi_output++; //aggiorniamo il conteggio complessivo
  }
}



//funzione invocata da output_albero() quando un nodo viene completato
inline void fine_nodo(unsigned long nodo)
{
  if( !abilita_scoperta ) //allora un nodo va scritto in output quando terminato
  {
    buf_out->put(nodo);
    num_nodi_output++; //aggiorniamo il conteggio complessivo
  }
}



//funzione invocata da salva_risultati(): visitiamo in avanti (i.e. le liste di adiacenza sono esaminate da SX a DX) l'albero con nodo radice 'radice' e lo diamo in output su disco
//notare che essendo un albero non occorre il vettore di bit per i nodi bianchi-non bianchi
void output_albero_avanti(unsigned long radice, SwappedStack<unsigned long> * pila)
{
  bool ok;
  struct InfoAlbero infotree;
  
  infotree.punt= num_nodi_output;  

  scoperta_nodo(radice);
  pila->push(radice); //inizializziamo lo stack con il solo nodo radice (e con l'indice del suo figlio, ma concettualmente e' come un unico record)
  pila->push(0); //mettiamo uno zero: dobbiamo partire dal primo figlio del nodo radice


  while( !pila->isEmpty() )
  {
    unsigned long indice_next, nodo_top, nodo_figlio;

    ok= pila->pop(indice_next); //informazione prossimo nodo adiacente da considerare relativamente al nodo in cima allo stack
    accertati_che(ok, MSG_ERRORE_PILA);    
    ok= pila->readTop(nodo_top); //consideriamo il nodo in cima
    accertati_che(ok, MSG_ERRORE_PILA);    

    if(indice_next < liste_adiacenza->degree_of(nodo_top)) //vediamo se il nodo affiorante ha (altri) figli da visitare
    {
      nodo_figlio= liste_adiacenza->get_adiac( nodo_top, indice_next); //il figlio corrente del nodo_top

      //gestione scoperta del nodo:      
      scoperta_nodo(nodo_figlio);

      indice_next++; //ci dobbiamo ricordare di passare al figlio successivo di 'nodo_top'
      pila->push(indice_next); //rimette il contatore aggiornato per il nodo affiorante sulla pila

      pila->push(nodo_figlio); //mettiamo il nodo trovato sullo stack: prima il suo ID, poi uno zero come indice del prossimo figlio da considerare
      pila->push(0); //mettiamo uno zero: esaminando successivamente i figli del nodo scoperto dovremo iniziare dal primo
    }
    else //sono stati considerati tutti gli adiacenti
    {
      ok= pila->pop(nodo_top); //leggiamo il nodo in cima, appena finito, e lo togliamo dalla pila
      accertati_che(ok, MSG_ERRORE_PILA);

      fine_nodo(nodo_top); //gestiamo la terminazione del nodo
    }

  } //fine blocco while

  //a questo punto e' terminata la visita dell'albero DFS e possiamo scrivere le informazioni relative:
  infotree.num_nodi_albero= num_nodi_output - infotree.punt; //calcoliamo anche la dim. dell'albero
  buf_infotree->put(infotree);
}



//funzione invocata da salva_risultati(): visitiamo all'indietro (i.e. le liste di adiacenza sono esaminate da DX a SX) l'albero con nodo radice 'radice' e lo diamo in output su disco
//notare che essendo un albero non occorre il vettore di bit per i nodi bianchi-non bianchi
void output_albero_indietro(unsigned long radice, SwappedStack<unsigned long> * pila)
{
  bool ok;
  struct InfoAlbero infotree;

  infotree.punt= num_nodi_output;

  scoperta_nodo(radice);
  pila->push(radice); //inizializziamo lo stack con il solo nodo radice (e con l'indice del suo figlio, ma concettualmente e' come un unico record)
  pila->push(liste_adiacenza->degree_of(radice)); //mettiamo il suo degree: rimangono cioe' ancora tutti i suoi figli da esaminare

  
  while( !pila->isEmpty() )
  {
    unsigned long indice_next, nodo_top;

    ok= pila->pop(indice_next); //informazione prossimo nodo adiacente da considerare relativamente al nodo in cima allo stack
    accertati_che(ok, MSG_ERRORE_PILA);

    if(indice_next > 0) //vediamo se il nodo affiorante ha (altri) figli da visitare
    {
      ok= pila->readTop(nodo_top); //consideriamo il nodo in cima
      accertati_che(ok, MSG_ERRORE_PILA);

      indice_next--; //(!) dobbiamo PRIMA decrementare (l'indice sulla pila rappr. il numero di nodi restanti all'inizio della lista di adiac. del nodo padre)
      unsigned long nodo_figlio= liste_adiacenza->get_adiac( nodo_top, indice_next); //il figlio corrente del nodo_top
      
      //gestione scoperta del nodo:
      scoperta_nodo(nodo_figlio);
      
      pila->push(indice_next); //rimette il contatore aggiornato per il nodo affiorante sulla pila

      pila->push(nodo_figlio); //mettiamo il nodo trovato sullo stack: prima il suo ID, poi il # di suoi adiacenti che rimangono da considerare (tutti)
      pila->push(liste_adiacenza->degree_of(nodo_figlio)); //mettiamo il suo degree: rimangono cioe' ancora tutti i suoi figli da esaminare
    }
    else //indice_next==0, ovvero il nodo affiorante sullo stack e' appena finito
    {
      ok= pila->pop(nodo_top); //leggiamo il nodo in cima, appena finito, e lo togliamo dalla pila
      accertati_che(ok, MSG_ERRORE_PILA);

      fine_nodo(nodo_top); //gestiamo la terminazione del nodo
    }

  } //fine blocco while

  //a questo punto e' terminata la visita dell'albero DFS e possiamo scrivere le informazioni relative:
  infotree.num_nodi_albero= num_nodi_output - infotree.punt; //calcoliamo anche la dim. dell'albero
  buf_infotree->put(infotree);
}



//salviamo su disco i risultati della visita (i nodi incontrati nell'ordine di scoperta o di fine visita)
//notare che poiche' visitiamo un albero non servono marche
//PRECondizioni: le liste di adiacenza contengono la foresta DFS (senza marche) e sono gia' ordinate per peso decrescente dei nodi: sarebbe sbagliato usare un diverso ordinamento degli archi rispetto a quello con cui e' stato ottenuta la foresta DFS corrente. Invertendo l'ordine di due nodi nella lista di adiacenza si puo' avere un arco di attraversamento in avanti che rende sbagliata la foresta
void salva_risultati(const char * nome_mfile_output)
{
  printf(" Saving DFS forest...\n"); fflush(stdout);
  unsigned long dimensione_mem_libera, dim_mem_stack, dim_mem_buffer_nodi, dim_mem_buffer_info;
  unsigned long * punt_mem_stack, * punt_mem_buffer_nodi, * punt_mem_buffer_info;

  //calcoliamo la memoria libera:
  accertati_che((liste_adiacenza->capienzaMemoriaInutilizzata()==minimo_archi_caricabili_alla_volta+elementi_riservati_buffer), MSG_ERR_MEM_LIBERA_RISERVATA);
  dimensione_mem_libera= sizeof(unsigned long) * liste_adiacenza->capienzaMemoriaInutilizzata(); //ogni elemento e' un long

  dim_mem_stack= dimensione_mem_libera/2; //diamo il 50% allo stack e i buffers si spartiscono a meta' il restante 50%
  dim_mem_buffer_nodi= dimensione_mem_libera/4;
  dim_mem_buffer_info= dimensione_mem_libera/4;
  punt_mem_stack= liste_adiacenza->get_punt_mem_libera();
  punt_mem_buffer_nodi= punt_mem_stack + (dim_mem_stack / sizeof(unsigned long)); //aritmetica dei puntatori..
  punt_mem_buffer_info= punt_mem_buffer_nodi + (dim_mem_buffer_nodi / sizeof(unsigned long)); //aritmetica dei puntatori..

  //apertura dei multifile in cui salvare le informazioni sulla visita:
  eraseMultiFile(nome_mfile_output); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mfile_out((char*)nome_mfile_output, 512*1024*1024);  
  buf_out= new BufferScritturaMultifile<unsigned long>(&mfile_out, dim_mem_buffer_nodi, punt_mem_buffer_nodi); //lo creiamo dinamicamente perche' tanto ci si accede con un puntatore in questo modulo

  //creiamo il nome del multifile con le info. sugli alberi
  int lung_nome= strlen(nome_mfile_output) + strlen(EXT_INFOALBERO);
  char * nome_mfile_infotree= new char[lung_nome + 1]; //un carattere in piu' per il NULL finale
  sprintf(nome_mfile_infotree, "%s%s", nome_mfile_output, EXT_INFOALBERO); //concateniamo il nome del file di output principale con un'estensione, ottenendo il nome del file con le info. sugli alberi

  eraseMultiFile(nome_mfile_infotree); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mfile_infotree(nome_mfile_infotree, 512*1024*1024);
  buf_infotree= new BufferScritturaMultifile<struct InfoAlbero>(&mfile_infotree, dim_mem_buffer_info, (struct InfoAlbero *) punt_mem_buffer_info);

  //allocazione dello stack:
  SwappedStack<unsigned long> pila( dim_mem_stack, punt_mem_stack);


  //esplorazione dei singoli alberi DFS:

  bool esegui_visita_albero_in_avanti;
  
  if(ordine_inverso_output) //se l'output lo voglio nell'ordine contrario devo invertire i due parametri ("dualita" dei due tipi di visite: scoperta-avanti da' lo stesso output, solo nell'ordine inverso, di terminazione-indietro ; analogamente terminazione-avanti e' equivalente, se non fosse per l'ordine invertito, a scoperta-indietro
  {
    esegui_visita_albero_in_avanti= false;
    abilita_scoperta= !output_nodo_quando_scoperto;
  }
  else //impostazioni di default, quando l'ordine di output e' quello consueto (dal primo nodo scoperto/dal primo nodo completato):
  {
    esegui_visita_albero_in_avanti= true;
    abilita_scoperta= output_nodo_quando_scoperto;
  }

  num_nodi_output= 0; //inizializziamo il conteggio dei nodi scritti, aggiornato dalle funzioni chiamate
  
  unsigned long numero_alberi= liste_adiacenza->degree_of(NODO_SUPER_ROOT); //ogni arco che parte dal nodo "super root" individua un albero DFS
    
  if(esegui_visita_albero_in_avanti)
    for(unsigned long i=0; i<numero_alberi ; i++)
    {
      unsigned long radice= liste_adiacenza->get_adiac(NODO_SUPER_ROOT, i); //consideriamo la radice dell'i-esimo albero DFS

      output_albero_avanti(radice, &pila); //visitiamo l'albero e lo diamo in output su disco
    }
  else
    for(unsigned long i=numero_alberi; i > 0 ; i--) //'i' va da 'numero_alberi' a 1
    {
      unsigned long radice= liste_adiacenza->get_adiac(NODO_SUPER_ROOT, i - 1); //consideriamo le radicie in ordine contrario

      output_albero_indietro(radice, &pila); //visitiamo l'albero e lo diamo in output su disco
    }
  
  //chiusura dei multifile:
  buf_out->chiudi(); //N.B. i buffer devono essere chiusi prima dei rispettivi multifile
  buf_infotree->chiudi();
  mfile_out.close();
  mfile_infotree.close();

  delete buf_out;
  delete buf_infotree;
  delete [] nome_mfile_infotree; //cancelliamo la stringa dalla memoria

//printf("num_nodi_output=%lu, numero_nodi=%lu\n", num_nodi_output, numero_nodi); //*****
  assert(num_nodi_output == numero_nodi); //******
//  accertati_che( (modo_visita_alberi != DFS_ascending) || (num_nodi_output == numero_nodi), "Errore: non sono stati scritti su disco tutti i nodi previsti");

  printf(MSG_DONE);
}



