#include "main.h"
#include "riordinamento_liste.h"
#include "ListaStatica.h"
#include "nodi_passivi.h"



//per vedere le variabili definite altrove:
extern ListaStatica * liste_adiacenza;
extern unsigned long conta_nodi_ordinati_con_buffer_corto; //*****



#ifdef CHECK_PESI //*****
//versione con riversamento liste in pila
void calcola_pesi_sottoalberi_OLD(unsigned long * vettore_pesi, unsigned long capienza_memoria_extra, unsigned long * memoria_extra)
{
  SwappedStack<unsigned long> pila(capienza_memoria_extra*sizeof(unsigned long), memoria_extra); //contiene coppie di nodi corrispondenti ad archi da esplorare (<'nodo sorgente', 'nodo destinazione'>), piu' la terna <MARCA_FINE_NODO, 'nodo sorgente', 'nodo destinazione'>   N.B. memorizziamo anche il padre di ogni nodo nella pila perche' con le liste di adiacenza non sarebbe altrimenti facile risalirvi
  const unsigned long marca_fine_nodo= MAX_ULONG; //valore speciale messo nello stack per capire che questo "nodo" e il successivo hanno un significato particolare: il nodo che segue ha appena terminato la sua lista di adiacenza percio' il suo peso e' definitivo e puo' essere aggiunto al padre

  printf(" Weight of subtrees (NO backtracking)..."); fflush(stdout);
  for(unsigned long i=0; i < numero_nodi; i++) //tutti i nodi tranne il "super root"
    vettore_pesi[i]= 0; //inizializziamo a 0 le informazioni sui pesi associate ai nodi (per default ho 0 discendenti).

  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();

  unsigned long numero_alberi= liste_adiacenza->degree_of(NODO_SUPER_ROOT); //ogni arco che parte dal nodo "super root" individua un albero DFS
  for(unsigned long indice_albero_corrente=0; indice_albero_corrente<numero_alberi ; indice_albero_corrente++)
  {
    unsigned long & rif_nodo_radice= liste_adiacenza->get_adiac(NODO_SUPER_ROOT, indice_albero_corrente); //riferimento usato per leggere e manipolare il nodo adiacente corrente.

   //mettiamo i dati per i nodi figli nella pila:
    unsigned long size_lista_nodo_radice= liste_adiacenza->degree_of(rif_nodo_radice);
    unsigned long indice_nella_lista_nodoradice= liste_adiacenza->get_indice_lista_adiacenza(rif_nodo_radice) + size_lista_nodo_radice - 1; //dobbiamo esaminare gli adiacenti in ordine contrario

    for(unsigned long i=0; i < size_lista_nodo_radice ; i++)
    {
      unsigned long figlio;

      figlio= vettore_adiacenti[indice_nella_lista_nodoradice];

      pila.push( rif_nodo_radice );
      pila.push( figlio );

      indice_nella_lista_nodoradice--; //dobbiamo scandire la lista in senso contrario
    }

    while( !pila.isEmpty() )
    {
      bool ris;
      unsigned long nodo_affiorante;

      ris= pila.readTop(nodo_affiorante); //legge senza rimuoverlo dalla pila il nodo affiorante
      accertati_che( ris, MSG_ERRORE_PILA);

      if(nodo_affiorante == marca_fine_nodo) //il nodo destinazione mem. subito dopo la marca e' stato completato con tutti i suoi sottoalberi
      {
        unsigned long nodo_src, nodo_dst;
        ris= pila.pop(nodo_affiorante); //eliminiamo la marca dalla pila
        accertati_che( ris, MSG_ERRORE_PILA);
        ris= pila.pop(nodo_dst); //l'arco compare in ordine inverso rispetto a quando inserito: quindi questa e' la destinazione, poi viene la sorgente
        accertati_che( ris, MSG_ERRORE_PILA);
        ris= pila.pop(nodo_src);
        accertati_che( ris, MSG_ERRORE_PILA);

        vettore_pesi[nodo_src] += (1 + vettore_pesi[nodo_dst]); //sommiamo il nodo figlio e tutti i discendenti del figlio
      }
      else //il nodo affiorante appena letto e' la destinazione di un arco i cui sottoalberi non sono ancora stati esplorati
      {
        pila.push(marca_fine_nodo); //cosi' quando lo ritiriamo fuori dalla pila sappiamo che l'arco (i.e. il nodo dest) che viene subito dopo e' stato completato con tutti i sottoalberi

        //mettiamo nella pila gli archi adiacenti al nodo appena letto dalla pila. Ricordare che 'nodo_affiorante' e' la destinazione dell'arco che era presente sulla cima della pila

        unsigned long size_lista_nodotop= liste_adiacenza->degree_of(nodo_affiorante);
        unsigned long indice_nella_lista_nodo_top= liste_adiacenza->get_indice_lista_adiacenza(nodo_affiorante) + size_lista_nodotop - 1;

        //inseriamo gli adiacenti bianchi nella pila e quindi cancelliamo con delle "X" l'intera lista di adiacenza appena esaminata: notare che l'ordine di priorita' delle liste di adiac. e' da SX a DX. Devo inserire nella pila in ordine contrario in modo che, rileggendo gli elementi nell'ordine inverso prodotto dalla pila, l'ordine con cui esamino gli archi sia alla fine quello voluto
        for(unsigned long i=0; i < size_lista_nodotop ; i++) //inseriamo gli archi che portano a nodi inesplorati e relativi a 'nodo_dest' ('nodo_dest' e' destinazione dell'arco appena estratto dalla pila, ma sorgente di quelli che andremo adesso a inserire): gli archi vanno inseriti allo stesso modo, prima src e poi dst. Inoltre vanno inseriti prima gli adiacenti alla fine della lista (quelli piu' a DX), poiche' la pila inverte l'ordine una volta che li estraggo
        {
           unsigned long figlio= vettore_adiacenti[indice_nella_lista_nodo_top];

           pila.push(nodo_affiorante); //sorgente dell'arco
           pila.push( figlio ); //nodo

           indice_nella_lista_nodo_top--;
        }
      }

    } //fine ciclo while pila

  } //fine ciclo for delle radici
  printf(MSG_DONE);
}
#endif



/* esplora il sottoalbero corrispondente al nodo affiorante dalla pila e contenente solo nodi non-passivi (poiche' il primo deve essere per ipotesi non-passivo).
*/
inline void pesa_sottoalbero(unsigned long * vettore_pesi, SwappedStack<unsigned long> * pila)
{
  bool ok;
  unsigned long nodo_cima, livello_iniziale_pila, conta_preord;

  conta_preord= 0;

  livello_iniziale_pila= pila->stackSize(); //salviamo il # di elementi di partenza (per ipotesi la pila puo' non contenere solo l'elemento da esplorare, ma questo e' solo il nodo affiorante)

  do
  {
    ok= pila->readTop(nodo_cima); //testiamo il nodo affiorante
    accertati_che( ok, MSG_ERRORE_PILA);

    if(nodo_cima == MAX_ULONG) //e' una marca?
    { //la marca indica che il nodo che segue e' stato completato
      ok= pila->pop(nodo_cima); //eliminiamo la marca dallo stack
      accertati_che( ok, MSG_ERRORE_PILA);
      ok= pila->pop(nodo_cima); //leggiamo (e togliamo) il nodo che e' appena stato completato
      accertati_che( ok, MSG_ERRORE_PILA);

      vettore_pesi[nodo_cima]= conta_preord - vettore_pesi[nodo_cima]; //ricaviamo il numero di discendenti che il nodo ha.
    }
    else //e' un nodo normale, che e' stato appena scoperto
    {
#ifdef OPT_PRIMO_CAMMINO
      unsigned long degree= liste_adiacenza->degree_of(nodo_cima);

      while(degree > 0)
      {
        vettore_pesi[nodo_cima]= conta_preord; //mem. in pratica il # di nodi che lo precedono nella visita in preordine del sottoalbero
        conta_preord++;

        pila->push(MAX_ULONG);
        liste_adiacenza->riversa_lista_in_pila_DX_SX(nodo_cima, pila); //mettiamo gli adiacenti nello stack

        nodo_cima= liste_adiacenza->get_adiac(nodo_cima, 0);
        degree= liste_adiacenza->degree_of(nodo_cima);
      }
      //l'ultimo nodo esaminato ha degree 0, quindi il suo peso e' 1, e lo abbiamo appena terminato
      vettore_pesi[nodo_cima]= 1;
      conta_preord++; //(!) dobbiamo contare anche questo nodo che e' appena stato completato
      ok= pila->pop(nodo_cima); //leggiamo (e togliamo) il nodo che e' appena stato completato
      accertati_che( ok, MSG_ERRORE_PILA);
#else
      vettore_pesi[nodo_cima]= conta_preord; //mem. in pratica il # di nodi che lo precedono nella visita in preordine del sottoalbero

      conta_preord++;

      pila->push(MAX_ULONG); //mettiamo una marca nella pila, che indica la fine della lista di adiacenza del nodo attuale, e dunque quando la rileggiamo dalla pila che il nodo che la segue e' appena stato completato
      liste_adiacenza->riversa_lista_in_pila_DX_SX(nodo_cima, pila); //mettiamo gli adiacenti nello stack
#endif
    }

  } while((unsigned long) pila->stackSize() >= livello_iniziale_pila); //usciamo quando e' stata conclusa l'elaborazione del sottoalbero del nodo affiorante inizialmente dalla pila

}



void calcola_pesi_sottoalberi(unsigned long * vettore_pesi, unsigned long capienza_memoria_extra, unsigned long * memoria_extra)
{
  SwappedStack<unsigned long> pila(capienza_memoria_extra*sizeof(unsigned long), memoria_extra); //contiene coppie di nodi corrispondenti ad archi da esplorare (<'nodo sorgente', 'nodo destinazione'>), piu' la terna <MARCA_FINE_NODO, 'nodo sorgente', 'nodo destinazione'>   N.B. memorizziamo anche il padre di ogni nodo nella pila perche' con le liste di adiacenza non sarebbe altrimenti facile risalirvi

  printf(" Weight of subtrees (NO backtracking)..."); fflush(stdout);

  ripristina_nodi_passivi_se_necessario();

  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
  unsigned long numero_alberi= liste_adiacenza->degree_of(NODO_SUPER_ROOT); //ogni arco che parte dal nodo "super root" individua un albero DFS
  unsigned long * punt_nodo_radice= vettore_adiacenti + liste_adiacenza->get_indice_lista_adiacenza(NODO_SUPER_ROOT);

  for(unsigned long i=0; i<numero_alberi ; i++, punt_nodo_radice++) //i figli del nodo super root vengono trattati come caso particolare per evitare di inserirli nella pila tutti assieme
  {
    if(!is_passivo(*punt_nodo_radice))
    { //nodo non passivo: la sua lista di adiac. va riordinata in base ai pesi, che bisogna percio' calcolare
      pila.push(*punt_nodo_radice); //inseriamo il nodo radice nella pila
      pesa_sottoalbero(vettore_pesi, &pila); //calcoliamo i pesi di tutti i nodi dell'albero della radice corrente
    } //a questo punto il nodo non-passivo e' stato rimosso dalla pila
    else
    { //in questo caso non ci interessa calcolare il peso dei sottoalberi, fino al prossimo nodo non-passivo continuiamo l'esplorazione dell'albero senza fare nulla
      unsigned long nodo_top;

      pila.push(*punt_nodo_radice);

      do
      {
        pila.readTop(nodo_top); //leggiamo adesso in 'nodo_top' il nodo affiorante (figlio di un passivo), senza toglierlo dalla pila

        if(!is_passivo(nodo_top))
        { //nodo non passivo: la sua lista di adiac. va riordinata in base ai pesi, che bisogna percio' calcolare
          pesa_sottoalbero(vettore_pesi, &pila); //calcoliamo i pesi di tutti i nodi del sottoalbero relativo a 'nodo_top'
        } //a questo punto il nodo non-passivo e' stato rimosso dalla pila
        else
        { //in questo caso non ci interessa calcolare il peso dei sottoalberi, fino al prossimo nodo non-passivo continuiamo l'esplorazione dell'albero senza fare nulla
#ifdef OPT_PRIMO_CAMMINO
          unsigned long nodo_corrente= nodo_top;
          unsigned long degree= liste_adiacenza->degree_of(nodo_corrente);
          bool passivo= true;

          pila.pop(nodo_top); //(!) rimuove dalla pila il nodo affiorante (ricopiandolo di nuovo in 'nodo_top')

          while(passivo && (degree > 0))
          {
            liste_adiacenza->riversa_lista_in_pila_DX_SX(nodo_corrente, &pila, degree - 1);

            nodo_corrente= liste_adiacenza->get_adiac(nodo_corrente, 0);
            degree= liste_adiacenza->degree_of(nodo_corrente);
            passivo= is_passivo(nodo_corrente);
          }

          if(degree > 0)
            pila.push(nodo_corrente); //lo mettiamo sulla pila per elaborarlo alla prossima iterazione del loop
          else
            if(!passivo) //se l'ultimo nodo esaminato era passivo, ma senza figli, lo ignoriamo ; se era non-passivo, allora il suo peso e' 1
              vettore_pesi[nodo_corrente]= 1;

#else
          pila.pop(nodo_top); //rimuove dalla pila il nodo affiorante (ricopiandolo di nuovo in 'nodo_top')

          liste_adiacenza->riversa_lista_in_pila_DX_SX(nodo_top, &pila); //consideriamo "ricorsivamente", mettendoli nella pila, i figli del nodo corrente:
#endif
        }
      } while( !pila.isEmpty() );
    }//fine blocco else delle radici passive
  }//fine ciclo for dei nodi radice

}




#ifdef CHECK_PESI //*****
#define size_bufferone_pesi 5*1024*1024
unsigned long bufferone_pesi[size_bufferone_pesi];

//*****serve per il debug
void controlla_pesi(unsigned long * vettore_pesi, unsigned long capienza_memoria_extra, unsigned long * memoria_extra)
{
  ripristina_nodi_passivi_se_necessario();
  calcola_pesi_sottoalberi_OLD(bufferone_pesi, capienza_memoria_extra, memoria_extra);

  for(unsigned long i=0; i<numero_nodi; i++)
    if(!is_passivo(i))
      accertati_che((bufferone_pesi[i] + 1 == vettore_pesi[i]), "Err. check pesi: il peso del nodo %lu viene %lu invece di %lu (versione `stabile')", i, vettore_pesi[i], bufferone_pesi[i]+1);
}
#endif



//---------------------------------------------------------------------------------------------



void Merge_buffer_corto(unsigned long v[], unsigned long indice_1, unsigned long indice_fine1, unsigned long indice_fine2, unsigned long * pesi, BufferAppoggio<unsigned long> * buffer_aux)
{
  unsigned long indice_2= indice_fine1 + 1;

  while( (indice_1 <= indice_fine1) && ( pesi[ v[indice_1] ]  >=  pesi[ v[indice_2] ] ) ) //vediamo gli elementi all'estrema SX che sono gia' al loro posto
  {
    indice_1++;
  }

  if(indice_1 > indice_fine1)
    return; //le due parti della lista erano gia' tra loro ben ordinate

  unsigned long punto_inserimento= indice_1; //ci ricordiamo a partire da dove dovremo inserire gli elementi che verranno messi nel buffer: quello che si trova a SX di questo punto e' gia' al suo posto

  while((indice_1 <= indice_fine1) && (indice_2 <= indice_fine2)) //finche' restano elementi in entrambe le sottoliste copiamo gli elementi piu' piccoli nel buffer
    if( pesi[ v[indice_1] ] >= pesi[ v[indice_2] ] )
    {
      buffer_aux->depositaValore( v[indice_1] );
      indice_1++;
    }
    else
    {
      buffer_aux->depositaValore( v[indice_2] );
      indice_2++;
    }

  assert( buffer_aux->getSize() == (indice_1 - punto_inserimento) + (indice_2 - (indice_fine1 + 1)) );

  if(indice_1 > indice_fine1) //parte a SX terminata, quello che resta alla fine della parte DX e' gia' al suo posto
  {
    while(punto_inserimento < indice_2) //scriviamo al loro posto i nodi salvati nel buffer
    {
      unsigned long & rif_adiacente_da_scrivere= v[punto_inserimento];
      buffer_aux->recuperaValore(rif_adiacente_da_scrivere);
      punto_inserimento++;
    }
  }
  else
  if(indice_2 > indice_fine2) //parte a DX terminata: cio' che resta alla fine della prima parte va accodato alla fine della lista e in mezzo vanno caricati i nodi salvati nel buffer
  {
    unsigned long larghezza_intervallo= indice_fine1 - indice_1 + 1;  //quanti elementi restano da considerare nella parte 1 della lista
    unsigned long indice_copy_from= indice_1; //qua iniziano gli elementi da copiare
    unsigned long indice_copy_to= indice_copy_from + (indice_fine2 - indice_fine1); //l'indice a partire dal quale verranno copiato gli elementi da trasferire (notare che dobbiamo accodare gli elementi restanti nella parte 1 alla fine della lista, percio' trasliamo l'intervallo della dimensione della seconda parte della lista)

    memmove(v + indice_copy_to, v + indice_copy_from, sizeof(unsigned long)*larghezza_intervallo); //occhio a non usare memcpy perche' le aree da copiare overlappano

    while(punto_inserimento < indice_copy_to)  //scriviamo al loro posto i nodi salvati nel buffer
    {
      unsigned long & rif_adiacente_da_scrivere= v[ punto_inserimento ];
      buffer_aux->recuperaValore(rif_adiacente_da_scrivere);
      punto_inserimento++;
    }
  }

  assert(buffer_aux->getSize() == 0);
}


// Merge adjacent segments of size s.
inline void MergePass_buffer_corto(unsigned long x[], unsigned long s, unsigned long n, unsigned long vettore_pesi[], BufferAppoggio<unsigned long> * buffer_aux)
{
  unsigned long i= 0;
  while (i + 2 * s <= n)
  {
    // merge two adjacent segments of size s
    Merge_buffer_corto(x, i, i+s-1, i+2*s-1, vettore_pesi, buffer_aux);
    i = i + 2 * s;
  }

  // fewer than 2s elements remain
  if(i + s < n) //se invece ne restano meno di 'n' non devo fare nulla perche' sono gia' ordinati
    Merge_buffer_corto(x, i, i+s-1, n-1, vettore_pesi, buffer_aux);
}



inline void MergeSort_buffer_corto(unsigned long * vettore, unsigned long n, unsigned long * vettore_pesi, BufferAppoggio<unsigned long> * buffer_aux)
{
  unsigned long s= 1; //segment size
  while(s < n)
  {
    MergePass_buffer_corto(vettore, s, n, vettore_pesi, buffer_aux);
    s += s;
  }

  conta_nodi_ordinati_con_buffer_corto++; //*****
}



#ifdef CHECK_PESI //*****
bool check_ordinamento_ok(unsigned long nodo_src, unsigned long degree, unsigned long * vettore_pesi)
{
  for(unsigned long j=degree-1; j>=1 ; j--)
  {
    //consideriamo due elementi successivi: quello alla posizione 'j'-1 ("prima") e quello alla posizione 'j' ("dopo")
    unsigned long & rif_nodo_prima= liste_adiacenza->get_adiac(nodo_src, j-1);
    unsigned long & rif_nodo_dopo= liste_adiacenza->get_adiac(nodo_src, j);

    if( vettore_pesi[rif_nodo_dopo] > vettore_pesi[rif_nodo_prima] ) //se l'elemento "dopo" e' piu' pesante dell'elemento "prima" li scambiamo
      return false;
  }

  return true;
}
#endif



// Merge c[l:m] and c]m:r] -----> d[l:r].
inline void Merge_standard(unsigned long c[], unsigned long d[], unsigned long l, unsigned long m, unsigned long r, unsigned long pesi[])
{
  unsigned long i= l; // cursor for first segment
  unsigned long j= m + 1; // cursor for second
  unsigned long k = l; // cursor for result

  // merge until i or j exits its segment
  while((i <= m) && (j <= r))
    if( pesi[ c[i] ] >= pesi[ c[j] ] )
      d[k++] = c[i++];
    else
      d[k++] = c[j++];

  // take care of left overs
  if(i > m)
  {
    for( ; j <= r; j++)
      d[k++] = c[j];
  }
  else
    for( ; i <= m; i++)
      d[k++] = c[i];
}


// Merge adjacent segments of size s.
void MergePass(unsigned long x[], unsigned long y[], unsigned long s, unsigned long n, unsigned long vettore_pesi[])
{
  unsigned long i= 0;
  while (i + 2 * s <= n)
  {
    // merge two adjacent segments of size s
    Merge_standard(x, y, i, i+s-1, i+2*s-1, vettore_pesi);
    i = i + 2 * s;
  }

  // fewer than 2s elements remain
  if(i + s < n)
    Merge_standard(x, y, i, i+s-1, n-1, vettore_pesi);
  else
    for (unsigned long j = i; j <= n-1; j++) // copy last segment to y
      y[j] = x[j];
}


//per ordinare un vettore di interi in base al peso (crescente) degli elementi specificato nell'apposito vettore: si specifica in vettore da ordinare, un altro vettore "gemello" d'appoggio, la dimensione dei due vettori, il vettore dei pesi
// Sort a[0:n-1] using merge sort.
inline void MergeSort_iterativo_doppio_vettore(unsigned long * vettore, unsigned long * aux, unsigned long n, unsigned long * vettore_pesi)
{
  unsigned long s= 1; //segment size
  while(s < n)
  {
    MergePass(vettore, aux, s, n, vettore_pesi); // merge from a to b
    s += s;
    MergePass(aux, vettore, s, n, vettore_pesi); // merge from b to a
    s += s;
  }
}



void mergesort_adiacenti_per_peso(unsigned long * vettore_pesi, unsigned long capienza_vettore_extra, unsigned long * memoria_extra)
{
  BufferAppoggio<unsigned long> buffer_aux(capienza_vettore_extra*sizeof(unsigned long), memoria_extra); //la dimensione del buffer va specificata in byte
  unsigned long conta_attesa_refresh= INTERVALLO_REFRESH_INDICATORE;
  unsigned long * const vettore_adiacenti= liste_adiacenza->vettore_adiacenti();
  
  ripristina_nodi_passivi_se_necessario(); //ci serve il contenuto integro del vettore dei bits dei nodi passivi

  for(unsigned long nodo_src=0; nodo_src < NODO_SUPER_ROOT ; nodo_src++) //escludiamo dall'ordinamento il "super root" che in pratica e' sempre passivo
  {
    unsigned long degree= liste_adiacenza->degree_of(nodo_src);

    //ordiniamo la lista di adiacenza del nodo corrente:
    if(degree > 1) //altrimenti non c'e' nulla da ordinare
      if( !is_passivo(nodo_src) ) //per i nodi passivi non bisogna ordinare
      {
        unsigned long * const puntatore_lista_adiacenza_nodo= vettore_adiacenti + liste_adiacenza->get_indice_lista_adiacenza(nodo_src);
      
        if(capienza_vettore_extra >= degree) //possiamo usare un secondo vettore nel merge sort
          MergeSort_iterativo_doppio_vettore(puntatore_lista_adiacenza_nodo, memoria_extra, degree, vettore_pesi);
        else
          MergeSort_buffer_corto(puntatore_lista_adiacenza_nodo, degree, vettore_pesi, &buffer_aux);
      }
    
    if(++conta_attesa_refresh >= INTERVALLO_REFRESH_INDICATORE)
    {
      conta_attesa_refresh= 0;
#ifdef VERBOSE_MODE
      indiceProgressione("Ordinamento in base ai pesi", (long) (100.0*nodo_src/NODO_SUPER_ROOT), (long) (100.0*nodo_src/NODO_SUPER_ROOT));
#endif
    }
  }

#ifdef VERBOSE_MODE
  indiceProgressione("Ordinamento in base ai pesi", (long) 100.0, (long) 100.0); printf("\n\n");
#endif
}


