/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file subgraph.cpp :source code of subgraphrand utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include "comuni.h"
#include "config.h"

/*
struttura dati che rappresenta un nodo all'interno del file .info del grafo
Viene usata tipicamente per la semplice lettura del file .info
*/

#ifndef INFONODO_H
#define INFONODO_H

class infoNodo {
public:
  long inDeg;
  long outDeg;
  long long puntIn;
  long long puntOut;
};
#endif




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//                           ordina il vettore di elementi di tipo T
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
template <class T>
  //ordina il vettore da inizio a fine ESTREMI COMPRESI
  void quickSort2(T* v,long inizio,long fine){

    if (fine>inizio){
      long nSplit=partition(v,inizio,fine);

      /*----------OTTIMIZZA ----------*/
      long ordinato=TRUE;
      for (long i=inizio;(i<fine)&&(ordinato);i++)
          ordinato= v[i]<=v[i+1];
      if (ordinato) return; //se gia' ordinato salto il processo
      /*------------------------------*/

      quickSort2(v,inizio,nSplit-1);
      quickSort2(v,nSplit+1,fine);
    }
  }






/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//          associato al quick sort
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/

template <class T>
  long partition(T* szArray,long nLower, long nUpper){
// Pivot with first element

    long nLeft = nLower + 1;
    T szPivot = szArray[nLower];
    long nRight = nUpper;

    // Partition array elements
    T szSwap;//var di appoggio per lo scambio
    while (nLeft <= nRight)
    {
        // Find item out of place
        bool finito=false;
        while (nLeft <= nRight && !finito)
            if (szArray[nLeft]<=szPivot)	nLeft = nLeft + 1;

            else                        	finito=true;


        finito=false;
        while (nLeft <= nRight && !finito){
            if (szArray[nRight]> szPivot)

            		nRight = nRight - 1;
            else
            		finito=true;
        }

        // Swap values if necessary
        if (nLeft < nRight)
        {
            szSwap = szArray[nLeft];
            szArray[nLeft] = szArray[nRight];
            szArray[nRight] = szSwap;
            nLeft = nLeft + 1;
            nRight = nRight - 1;
        }
    }

    // Move pivot element
    szSwap = szArray[nLower];
    szArray[nLower] = szArray[nRight];
    szArray[nRight] = szSwap;
    return nRight;

  }



long rimuoviDuplicati(long* v, long size){
  long dim=1;
  for (long i=dim;i<size;i++){
      if (v[dim-1]!=v[i]){
         v[dim]=v[i];
         dim++;
      }
  }
  return dim;
}



// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################




long ricBin(long* v, long primoCompreso,long ultimoCompreso,long elem){
  if (primoCompreso<ultimoCompreso){
      long medio=(primoCompreso+ultimoCompreso)/2;
      if (v[medio]<elem)        return ricBin(v,medio+1,ultimoCompreso,elem);
      else                      return ricBin(v,primoCompreso,medio,elem);
  }
  else
      if (primoCompreso==ultimoCompreso)
          if (v[primoCompreso]==elem) return primoCompreso;
          else  return -1;
      else                                      return -1;//non trovato

}





long ricercaBinaria(long*nodiScelti,long size,long nodoDaCercare){
  return ricBin(nodiScelti, 0,size-1,nodoDaCercare);
}







// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
long normalizzaLancio(unsigned long lancio, long inferiore, long superiore){
    long diff=superiore-inferiore+1;
    if (diff<0) return -1;
    if (diff==0) return inferiore;
    return inferiore+lancio%diff;
}
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################
// ##########################################################################################################################################################



// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
long scegliNodiUniformemente(long* nodiScelti, long numNodiSubGrafo, long numNodiGrafoTot,long probScelta,bool completaNodi){
  long conta=0;

  //nodi rimasti tra cui scegliere
  long rimastiTraCuiScegliere=numNodiGrafoTot;
  long rimastiDaScegliere=numNodiSubGrafo;

  for (int i=0;i<numNodiGrafoTot && conta<numNodiSubGrafo;  i++,rimastiTraCuiScegliere--)
      //controlla se scegliere il nodo ( o se abilitato, completa il num di nodi)
      if ( ( (float)rand()/RAND_MAX<probScelta ) || (probScelta==100.0) || ((rimastiTraCuiScegliere<=rimastiDaScegliere)&&(completaNodi)) ){
          nodiScelti[conta++]=i;
          rimastiDaScegliere--;
      }

  return conta;

}


// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


long scegliNodiPerIndegree(char* grafoInfo,long* nodiScelti, long numNodiSubGrafo, long numNodiGrafoTot,long long numArchi, char* messaggio){

  //nota che il numArchi e' anche la somma totale degli indegree di tutti i nodi ....

  long posDisponibile=0;
  long numFilesInfo=contaNumFilesMultiFile(grafoInfo);

  printf("%c%sthrowing generation%c",13,messaggio,13);fflush(stdout);
  //il buffer viene inizializzato con i lanci da fare
  for (long i=0;i<numNodiSubGrafo;i++){
      nodiScelti[i]=normalizzaLancio(rand(),1,numArchi+numNodiGrafoTot);//prendo 1 se il nodo ha indeg=0
  }

  printf("%c%sthrowing sorting   %c",13,messaggio,13);fflush(stdout);
  quickSort2(nodiScelti,0,numNodiSubGrafo-1);  //ordina i lanci

  //nome del file temporaneo
  char app[1000]; strcpy(app,grafoInfo);strcat(app,".tempoFileRand.step0");

  //apre il file temporaneo in cui scaricare i lanci ordinati e vi copia i lanci ordinati
  multiFileWriter fTempo=multiFileWriter(app,DIM_MAX);
  fTempo.write(nodiScelti,sizeof(long),numNodiSubGrafo);
  long numFilesTempoCreati=fTempo.close();

  //ed esegue il merge con i precedenti
  bufferedThreadFileInputStream fRand=bufferedThreadFileInputStream(app,numFilesTempoCreati,256*1024,6,true);
  bufferedThreadFileInputStream fInfo=bufferedThreadFileInputStream(grafoInfo,numFilesInfo,256*1024,6,false);

  infoNodo inf;
  long lancio;//lancio corrente


  long long indegPrecCorrente=0;//indegree dei nodi predecenti
  if (fInfo.read(&inf,sizeof(inf),1)<=0){cout<<"s";FINE;}
  inf.inDeg++;//aggiungo 1 all'indegree per calcolare il valore usato nei lanci (valore reale + 1)
  long nodo=0;
  long long conta=1000000;

  long pos=0;
  char* a="|/-\\";


  while (fRand.read(&lancio,sizeof(long),1)>0){


     //controlla se ci sono ancora nodi da prendere e se no esce
     if (posDisponibile==numNodiSubGrafo) break;

     while (( lancio>(indegPrecCorrente+inf.inDeg) ) ){
          indegPrecCorrente+=inf.inDeg;
          if (fInfo.read(&inf,sizeof(inf),1)<=0){cout<<"Unexpected fatal error in reading";FINE;}
          inf.inDeg++;//aggiungo 1 all'indegree per calcolare il valore usato nei lanci (valore reale + 1)
          nodo++;//leggo il nodo successivo

          if (conta++>50000){
            printf("%c%s ricerca [%c] %ld nodi    %c",13,messaggio,a[pos],nodo+1,13);fflush(stdout);
            conta=0;
            pos=(pos+1)%4;
          }
     }

     bool accettaNodo=false;
     if (posDisponibile>0){
         if (nodiScelti[posDisponibile-1]!=nodo)  accettaNodo=true;
     }
     else
          accettaNodo=true;

     if (accettaNodo){
         nodiScelti[posDisponibile++]=nodo;
     }

  }


  //chiude tutti i files
  fInfo.close();
  fRand.close();


  //cancella i files temporanei se ne e' rimasto qualcuno
  strcpy(app,grafoInfo);strcat(app,".tempoFileRand.step0");
  for (int i=0;i< numFilesTempoCreati;i++){
      char a[1000];
      sprintf(a,app,i);
      remove(a);
  }

//  printf("%c%s terminato step %ld\n",13,messaggio,step);fflush(stdout);

  return posDisponibile;

}


// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>



long leggiAffiorante(multiFileReader& f){
   long app;
   if (f.read(&app,sizeof(long),1)>0) return app;
   return -1;
}


long mergeFiles(char* nome1,long nFiles1,char* nome2,long nFiles2,long* v, long size){

  multiFileReader f1 = multiFileReader(nome1,nFiles1,true);
  multiFileReader f2 = multiFileReader(nome2,nFiles2,true);

//  cout<<nome1<<" - "<<nFiles1<<" / "<<nome2<<" - "<<nFiles2<<"\n";

  long primo1=leggiAffiorante(f1);
  long primo2=leggiAffiorante(f2);
  long posLibera=0;


  while (primo1>=0 || primo2>=0){

//      cout<<"\n"<<primo1<<"  "<<primo2<<"";fflush(stdout);
      if (posLibera>=size){cout<<"Unexpected fatal error: vector space not enough...";FINE;}

      if (primo1<0){//finito il file 1
         v[posLibera++]=primo2;
         primo2=leggiAffiorante(f2);
      }
      else
          if (primo2<0){//finito il file 2
             v[posLibera++]=primo1;
             primo1=leggiAffiorante(f1);
          }
          else{ //ci sono ancora tutti e due i files
              if (primo1<=primo2){//minimo nel 1 file
                  v[posLibera++]=primo1;
                  primo1=leggiAffiorante(f1);
              }
              else{//minimo nel 2 file
                  v[posLibera++]=primo2;
                  primo2=leggiAffiorante(f2);
              }
          }
//      cout<<" --> "<<v[posLibera-1]<<" (dim "<<posLibera<<")\n";fflush(stdout);
  }

  f1.close();
  f2.close();

  return posLibera;//num elementi inseriti
}


// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
// <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>




void stampaHelp(int arg,char* args[]){
  cout<<"This routine, given a multifile (in the .ips format) in input,representing a graph G=(V,E), creates the subgraph G' induced by a random selection of the nodes. It follows the command line usage:\n\n";
  cout<<"subgraphrand memory -RU|-RI maxIterations nameInputMultifile%d nameOutputFile%d numNodesInSubgraph seed\n";
  cout<<"memory          memory available (MB)\n";
  cout<<"-RU|-RI         specify whether to select the nodes of the subgraph with a uniform (-RU) or indegree based (-RI) distribution probability\n";
  cout<<"maxIterations   specify the maximum number of iterations to randomly extract the set V'; we suggest not to exceed the value 10\n";
  cout<<"nameInputMultifile%d   name of the multifile of the input graph (in ips format)\n";
  cout<<"nameOutputMultifile%d  name of the multifile of the output graph (in edges format)\n";
  cout<<"numNodesInSubgraph     the (expected) number of nodes of the subgraph\n";
  cout<<"seed                   the seed for the random number generator\n";

}


int main(int arg,char* args[]){

  stampaLicenza(arg,args);
  if (arg!=8) STAMPA_ERRORE_PARAMETRI(args[0],"Incorrect parameters.");

  long  m                = (long)atof(args[1])*1024*1024;
  char* tipo             = args[2];
  long  numIterazioniMax = atol(args[3]);
  
  char grafoInfo[300];strcpy(grafoInfo,args[4]);strcat(grafoInfo,".info");
  char grafoPred[300];strcpy(grafoPred,args[4]);strcat(grafoPred,".pred");
  char grafoSucc[300];strcpy(grafoSucc,args[4]);strcat(grafoSucc,".succ");

/*  char* grafoInfo       = args[4];
  char* grafoPred       = args[5];
  char* grafoSucc       = args[6];
*/  
  char* multiFileOut    = args[5];
  long  numNodiSubGrafo = atol(args[6]);
  long  seed            = atol(args[7]);
  bool  completaNodi  ;


  if (numIterazioniMax>0)   completaNodi=true;
  else                      completaNodi=false;

  if (!estMultiFileName(grafoInfo))     STAMPA_ERRORE_PARAMETRI(args[0],"INFO multifile name must contain the substring '%d' ");
  if (!estMultiFileName(grafoPred))     STAMPA_ERRORE_PARAMETRI(args[0],"PRED multifile name must contain the substring '%d' ");
  if (!estMultiFileName(grafoSucc))     STAMPA_ERRORE_PARAMETRI(args[0],"SUCC multifile name must contain the substring '%d' ");
  if (!estMultiFileName(multiFileOut))  STAMPA_ERRORE_PARAMETRI(args[0],"Output multifile name must contain the substring '%d' ");

  setMemoriaDisponibile(m);

  //************************ CONTROLLO FILES DI INPUT ***************************
  if (getTotalSizeMultiFiles(grafoInfo)<=0)  STAMPA_ERRORE_PARAMETRI(args[0],"The	input graph is empty\n");
  if (getTotalSizeMultiFiles(grafoPred)<=0)  STAMPA_ERRORE_PARAMETRI(args[0],"The input graph is empty\n");
  if (getTotalSizeMultiFiles(grafoSucc)<=0)  STAMPA_ERRORE_PARAMETRI(args[0],"The input graph is empty\n");

  long numInfo=contaNumFilesMultiFile(grafoInfo);
  long numSucc=contaNumFilesMultiFile(grafoSucc);

  //calcolo il numero di nodi
  long long totSizeInfo=0;
  for (int i=0;i<numInfo;i++){
      char a[1000];
      sprintf(a,grafoInfo,i);
      struct stat buf;
      stat(a, &buf);
      totSizeInfo+=buf.st_size;
  }
  long numNodiGrafoIniziale=totSizeInfo/sizeof(infoNodo);
  if (numNodiGrafoIniziale<numNodiSubGrafo){cout<<"Subgraph nodes number must be lower or equal to graph nodes number (it consist of "<<numNodiGrafoIniziale<<" nodes )\n;";FINE;}

  //calcola il num di archi
  long long totSizeSucc=0;
  for (int i=0;i<numSucc;i++){
      char a[1000];
      sprintf(a,grafoSucc,i);
      struct stat buf;
      stat(a, &buf);
      totSizeSucc+=buf.st_size;
  }
  long numArchiTotIniziali=totSizeSucc/sizeof(long);
  cout<<"Nodes number of input graph = "<<numNodiGrafoIniziale<< "\nEdges number of input graph "<<numArchiTotIniziali<<"\n";

  //controllo memoria
  if (getMemoriaDisponibile()/sizeof(long)<numNodiSubGrafo){cout<<"Memory not enough: required at least "<<numNodiSubGrafo*sizeof(long)<<" bytes";FINE;}

  //scelgo max numNodiGrafoIniziale nodi
  long* nodiScelti=new long[numNodiSubGrafo];
  long  numNodiScelti;

  char msg[1000];
  if (strcasecmp(tipo,"-RU")==0){
      cout<<"Nodes choice of subgraph ..... ";fflush(stdout);
      //prob con cui scelgo il nodo  (tra 0 e 1000)
      long probSceltaNodo=(long)(1000.0*numNodiSubGrafo/numNodiGrafoIniziale);

      srand(seed);
      numNodiScelti = scegliNodiUniformemente (nodiScelti,numNodiSubGrafo,numNodiGrafoIniziale,probSceltaNodo,completaNodi);
      sprintf(msg,"Nodes uniform choice - seed %ld",seed);
      cout<<"OK\n";fflush(stdout);
  }
  else if (strcasecmp(tipo,"-RI")==0){

      long step=0;
      long iter=0;
      srand(seed);

      //calcola il primo insieme di nodi
      numNodiScelti=0;
      printf("\n(*) Running: nodes choice\n");
      printf("+----------------------------------------------+\n");
      printf("| %s tot  | num nodes|lackingNodes|added nodes  |\n","%");
      printf("+--------+----------+------------+-------------+-----------------------------\n");
      char msg[1000];

      sprintf(msg,"%c|%6.2f %s|%10ld| %10ld | %10ld  |",13,100.0*numNodiScelti/numNodiSubGrafo,"%",
           numNodiScelti,numNodiSubGrafo-numNodiScelti,
           numNodiScelti);printf("%s",msg);
      fflush(stdout);

      //primo calcolo
      numNodiScelti=scegliNodiPerIndegree(grafoInfo,nodiScelti,numNodiSubGrafo,numNodiGrafoIniziale,numArchiTotIniziali, msg);


      sprintf(msg,"%c|%6.2f %s|%10ld| %10ld | %10ld  |",13,100.0*numNodiScelti/numNodiSubGrafo,"%",

           numNodiScelti,numNodiSubGrafo-numNodiScelti,numNodiScelti);printf("%s step %ld ...OK             \n",msg,step++);
      fflush(stdout);

      while( (numNodiScelti<numNodiSubGrafo) && (++iter<numIterazioniMax) && completaNodi){
              long numNodiSceltiInizio=numNodiScelti;
              sprintf(msg,"%c|%6.2f %s|%10ld| %10ld | %10ld  |",13,100.0*numNodiScelti/numNodiSubGrafo,"%",
                 numNodiScelti,numNodiSubGrafo-numNodiScelti,(long)0);printf("%s",msg);


              //file temporaneo per il salvataggio dei nodi scelti precedentemente
              char app1[1000];strcpy(app1,multiFileOut);strcat(app1,".tempoNodiScelti_base");
              multiFileWriter fTemp=multiFileWriter(app1,DIM_MAX);
              fTemp.write(nodiScelti,sizeof(long),numNodiScelti);
              long nFiles1=fTemp.close();

              //nodi rimasti
              long rimastiLiberi=numNodiSubGrafo-numNodiScelti;



              //salvataggio nodi blocco corrente (la sola parte mancante)
              long numNodiScelti_Nuovi=scegliNodiPerIndegree(grafoInfo,nodiScelti,rimastiLiberi,numNodiGrafoIniziale,numArchiTotIniziali, msg);
              char app2[1000];strcpy(app2,multiFileOut);strcat(app2,".tempoNodiScelti_nuovo");
              fTemp=multiFileWriter(app2,DIM_MAX);
              fTemp.write(nodiScelti,sizeof(long),numNodiScelti_Nuovi);
              long nFiles2=fTemp.close();

              printf("%s merging...   ",msg);
              //merge dei due files (e cancellazione dei files stessi) e rimozione dei duplicati
              numNodiScelti=mergeFiles(app1,nFiles1,app2,nFiles2,nodiScelti,numNodiSubGrafo);
              numNodiScelti=rimuoviDuplicati(nodiScelti,numNodiScelti);

              //stampa
              long ultimiNodiAggiunti=numNodiScelti-numNodiSceltiInizio ;
              sprintf(msg,"%c|%6.2f %s|%10ld| %10ld | %10ld  |",13,100.0*numNodiScelti/numNodiSubGrafo,"%",
									numNodiScelti,numNodiSubGrafo-numNodiScelti,
                 ultimiNodiAggiunti);
							printf("%s step %ld ...OK                  \n",msg,step++);
              fflush(stdout);

      }
      printf("+--------+----------+------------+-------------+-----------------------------\n");
      fflush(stdout);

  }
  else
      STAMPA_ERRORE_PARAMETRI(args[0],"Random choice type not valid");
  cout<<" "<<numNodiScelti<<" nodes selected\n\n";fflush(stdout);











// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD
// CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD CVD








/*  for (int i=0;i<numNodiScelti;i++){
      cout<<"  "<<nodiScelti[i];
  }

*/

//  EXIT(0);





  multiFileReader finfo=multiFileReader (grafoInfo,numInfo);
  multiFileReader fsucc=multiFileReader (grafoSucc,numSucc);
  bufferedThreadFileOutputStream farchi=bufferedThreadFileOutputStream(multiFileOut,DIM_MAX,DIM_BUFFER,NUM_BUFFER);

  infoNodo inf;
  long nodo=0;

  #define SIZESUCC (1000)
  long succ[SIZESUCC];
  long long numArchiScritti=0;
  long long numArchiIncontrati=0;


  printf("\n(*) Running: subgraph generation\n");
  printf("+-------------------------------------------------------------------+\n");
  printf("| %stot |  edges found     | written edges ( in %s )|   filesize      |\n","%","%");
  printf("+------+------------------+-----------------------+-----------------+\n");
  long long conta=40000000;


  while ( finfo.read(&inf,sizeof(infoNodo),1)>0){

    //legge i successori
    long nSucc=inf.outDeg;//num successori da leggere
    long rimasti=nSucc;//num nodi succ rimasti da leggere

//    cout<<nodo<<") succ "<<rimasti<<"\n";
    while (rimasti>0){
        //calcolo dim blocco da leggere
        long lettura=rimasti;
        if (rimasti>SIZESUCC)  lettura=SIZESUCC;

        //leggo i successori e li scalo dai rimasti, controllando se ci sono tutti
        long letti=fsucc.read(succ,sizeof(long),lettura);
//        cout<<"  |-->letti "<<letti<<"\n";
        rimasti-=letti;
        if (letti!=lettura){cout<<"Unexpected fatal error n.1 (node "<<nodo<<")";FINE;}

        //controllo se il nodo sorgente va preso e se si analizzo i successori
        long sorgente=ricercaBinaria(nodiScelti,numNodiScelti,nodo);
        for (long i=0;i<letti;i++){
                numArchiIncontrati++;
                //vedo se il destinatario va preso e se si' allora genero un arco
                long destinatario=ricercaBinaria(nodiScelti,numNodiScelti,succ[i]);
                if (sorgente>=0 && destinatario>=0){
                    arco a;
                    a.source=sorgente;
                    a.dest=destinatario;
                    farchi.write(&a,sizeof(a),1);
                    numArchiScritti++;
                }
        }
    }
    nodo++;

    if (conta++>30000 ||numArchiIncontrati>=numArchiTotIniziali-2){
          long inc=0;if (numArchiIncontrati==0) inc=1;
          printf("|%5.1f%s|      %10lld  |%14lld (%5.1f%s)| %13.2f MB|%c",
              100.0*numArchiIncontrati/numArchiTotIniziali,"%", numArchiIncontrati,numArchiScritti,
              100.0*numArchiScritti/(numArchiIncontrati+inc),"%",1.0*numArchiScritti*sizeof(arco)/1024/1024  ,13)   ;
          fflush(stdout);
          conta=0;
    }





  }



  printf("\n+-------------------------------------------------------------------+\n\n");

  
  farchi.close();

  finfo.close();
  fsucc.close();
  delete[]nodiScelti;

  /* STAMPA RAPPORTO ESECUZIONE */
  cout<<"\n";
//	fprintf(stdout,"Grafo in formato lista di archi a 32bit (%s)\n",msg);
//	fprintf(stdout,"Nome multiFiles grafo.....................%s\n",multiFileOut);
//	fprintf(stdout,"Numero files grafo creati (da 0 a MAX-1)..%ld\n",numFilesCreati);
	fprintf(stdout,"Edges number .............%lld\n",numArchiScritti );
	fprintf(stdout,"Filesize  ................%lld\n",numArchiScritti*sizeof(arco) );		
  EXIT(0);
}
