/*                                                                   
                                                                     
*********************************************************************
*********************************************************************

  file main.cpp :source code of dimacs2edges utility

  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   

/*
  Questo programma converte un grafo dal formato INFO-PRED-SUCC al formato DIMACS.
  Utilizza le funzioni di libreria standard "fgets", "sscanf" per analizzare il file di input,
  che e' costituito da righe di testo. L'output avviene con un oggetto della classe
  multiFileWriter usando un semplice buffer per le scritture.

  Ernesto Ricci
  27/11/2002
*/


#include <stdlib.h>

#include "comuni.h"


//max dim. blocchi multifile in output= 256 MB
#define DIM_MAX 256*1024*1024


const long capienza_buffer_out= 1024;
struct arco buffer_out[capienza_buffer_out]; //buffer per gli archi da scrivere
long next_pos= 0; //punta sempre alla prossima posizione nel buffer di scrittura

long n_nodi=0;
long n_archi;

FILE * file_in= NULL; //handler file lettura
multiFileWriter mf_out; //scrittore multifile

const long capienza_buffer_input= 1024; //assumiamo che le righe di testo siano di lunghezza inferiore a 1024 caratteri
char buffer_input[capienza_buffer_input]; //buffer per leggere una riga di testo alla volta


long numero_riga= 0; //indica la riga corrente
long numero_archi= 0; //indica il numero di archi gia' convertiti





void stampaHelp(int arg,char* argv[]){
    cout<<"This utility is not designed for stand alone use. It's used as a part of others utilities.\n";
    cout<<"\n";
		cout<< "This utility converts a dimacs format file in 'extended' edge format multifile \n";
		cout<< " \n";
		cout<< "Usage: "<<eliminaPath(argv[0])<<" memory -savesource|-deletesource nameInputMultifile%d nameOutputFile\n";
		cout<< " \n";
		cout<< "memory                     memory available(MB)\n";
		cout<< "-savesource|-deletesource  specify whether to delete or keep the input file\n";
		cout<< "nameInputFile              name of the file of the input graph (dimacs format)\n";
		cout<< "nameOutputMultifile%d      name of the output multifile ('extended' edges format)\n";
}



void scrivi_arco(struct arco * edge)
{
  if(next_pos >= capienza_buffer_out) //lo spazio nel buffer e' finito
  {
    mf_out.write( buffer_out, sizeof(struct arco), capienza_buffer_out); //svuotiamo il buffer riversandolo su disco
    next_pos= 0; //ci riposizioniamo all'inizio
  }

  buffer_out[next_pos]= *edge; //scrittura dell'elemento nel buffer
  next_pos++; //incremento dell'indice

  numero_archi++; //incr. conteggio archi convertiti
}


void  leggi_linea()
{
  int ret;
  char tipo;

  numero_riga++;

  ret= sscanf(buffer_input, "%c", &tipo); //vediamo di che tipo di riga si tratta
  if(ret == 0) //riga vuota: la saltiamo
    return;

  switch(tipo) //N.B. Dalle specifiche i caratteri identificativi del tipo di riga sono lowercase
  {
    case 'c': //commento
      printf("comment:%s\n", strchr(buffer_input, 'c') + 1); //scrive il resto della riga che segue il carattere identificatore del tipo di riga
    break;

    case 'p': //informazioni sul grafo
      
      ret= sscanf(buffer_input, "p edge %ld %ld", &n_nodi, &n_archi );
      if(ret == 2)
        printf("Graph information: %ld nodes and %ld edges\n\n", n_nodi, n_archi);
    break;

    case 'e': //definizione di un arco
      struct arco edge;
      ret= sscanf(buffer_input, "e %ld %ld", &(edge.source), &(edge.dest) );
      if(ret != 2)
        printf("Row #%ld not valid ", numero_riga);
      else
      {
         (edge.source)--; //decrementiamo i valori letti perche' DIMACS identifica il primo nodo con 1 e non con 0.
         (edge.dest)--;
         scrivi_arco(&edge);
      }
    break;
  }

}


void creaFileArchi( char * grafoDimacs, char * grafoArchi, bool cancellaSource)
{
  int iterazioni= 0;

  next_pos= 0; //inizializza la posizione della prima scrittura nel buffer_out
  mf_out= multiFileWriter(grafoArchi, DIM_MAX); //multifile output

  file_in= fopen(grafoDimacs, "r"); //grafo input
  check(file_in,(char*)"Unexcepted error: opening file error (%s)\n", grafoDimacs);

  long dim_file_in= getFileSize(grafoDimacs);
  long byte_letti= 0;

  printf("\nRunning: conversion dimacs --> edges\n\n");

  //leggiamo una riga di testo alla volta nel buffer di input (con fgets) e quindi la analizziamo
  //assumiamo che le righe di testo siano di lunghezza inferiore a 1024 caratteri: una riga piu' lunga di cosi' potrebbe non venire interpretata correttamente
  while( fgets(buffer_input, capienza_buffer_input, file_in) != NULL )
  {
    byte_letti+= strlen(buffer_input); //calcoliamo i byte letti con l'ultima riga

    leggi_linea(); //esaminiamo la linea appena messa nel buffer

    if (iterazioni++ >= 100000) {
        iterazioni=0;
        printf("%c[%5.2f%%] dimacs --> edges",13,100.0*byte_letti/dim_file_in);
        fflush(stdout);
    }

  }

  printf("%c[%5.2f%%] dimacs --> edges", 13, 100.0);
  fflush(stdout);
  printf("\n\nNumber of nodes: %ld\nNumber of edges: %ld\n", numero_riga, numero_archi);

  if(next_pos > 0) //passa il resto su disco se necessario
    mf_out.write( buffer_out, sizeof(struct arco), next_pos); //scriviamo cio' che resta nel buffer

  fclose(file_in);


  //aggiuge la label relativa al num di nodi...
  arco aux;
  aux.source=aux.dest=-n_nodi;
  if (mf_out.write(&aux,sizeof(arco),1)<1){perror("Write error");FINE;}
  PRINT_ISOLATED_NODES_SPECIFICATION;
  
  
  mf_out.close();

  if(cancellaSource) //elimino il file sorgente se richiesto
    remove(grafoDimacs);
}





int main(int argc,char** argv)
{
    long arg=argc;
    stampaLicenza(arg,argv);

    if (arg!=5) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");
    long cancellaSource;
    if (strcasecmp(argv[2],"-saveSource")==0)  cancellaSource=FALSE;
    else    if (strcasecmp(argv[2],"-deleteSource")==0)  cancellaSource=TRUE;
            else  STAMPA_ERRORE_PARAMETRI(argv[0],"option -deleteSource or -saveSource not found");

    if (!estMultiFileName(argv[4])) STAMPA_ERRORE_PARAMETRI(argv[0],"Output multifile name must contain the substring '%d' ");


    creaFileArchi( argv[3], argv[4], cancellaSource );

    EXIT(0);
}

