/*

  Ernesto Ricci
   7/4/2003

   
*/


#include <getopt.h>
#include "main.h"
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include "comuni.h"
#include "../common/scc_utl/VettoreBits.h"
#include "semi_ext.h"


//opzioni riga di comando:
long long memoria_max= -1; //memoria massima cha il programma puo' usare
char * nome_mfile_nodi_sottografo;
unsigned long numero_nodi; //numero nodi del grafo originale
char * nome_mfile_nodi_input;
char * nome_mfile_nodi_output;
bool esegui_comunque; //flag per eseguire incondizionatamente la visita anche se la memoria necessaria ?maggiore di quella disponibile
bool nodi_input_come_bits;
bool escludi_nodi_specificati;

//numero parti dei vari multifile di input:
long numPartiNodiInput;
long numPartiNodiSottografo;


//puntatori alle altre strutture in memoria:
unsigned long * vettore_Id_nodi= NULL;

unsigned long numero_nodi_selezionati; //il numero di nodi nel sottografo

bool full_memory; //indica se usiamo tutta la memoria che ci serve (4 byte per ogni nodo del sottografo), oppure se usiamo l'approccio semi-external





/*
void chiudi_tutto(int par) //intercetta la richiesta di chiudere forzatamante il programma da parte di assert() e libera la memoria uscendo ordinatamente
{
  printf("\n Deallocazione della memoria e preparazione all'uscita forzata\n");
  libera_memoria();
  ABORT();
}
   */


//scrive un messaggio ed esce dopo liberando le strutture di memoria impegnate
//per quanto riguarda i parametri usa la stessa sintassi di printf:  printf(const char *formato, ...)
void gestione_errore(const char *formato, ...)
{
  va_list ap;
  //stampiamo il messaggio d'errore ed usciamo:
  fprintf(stderr,"\n");
  va_start (ap, formato); //inizializziamo la struttura 'ap'
  vfprintf (stderr, formato, ap); //scriviamo il messaggio d'errore con tutti i parametri eventuali
  va_end (ap); //chiudiamo con la variabile 'ap'
  fprintf(stderr,"\n");
  perror("");

  libera_memoria();
  ABORT();
}



void indiceProgressione(const char* testo)
{
  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  printf("\r [%c] %s\r", cont, testo);
  fflush(stdout);
}


void indiceProgressione(const char* testo,long perc1,long perc2){

  /*======================== STAMPA MESSAGGI ===================================================================*/
  static char cont='-';

  switch(cont){
  case '-': cont='\\';break;
  case '\\': cont='|';break;
  case '|': cont='/';break;
  case '/': cont='-';break;
  }

  if (perc1<0) perc1=0;
  else if (perc1>100) perc1=100;

  if (perc2<0) perc2=0;
  else if (perc2>100) perc2=100;

  perc2=(long) (20.0*perc2/100);
  //indicatore
  char st[50];

  strcpy(st,"");
/*  strcpy(st,"____________________");
  for (long i=0;(i<20)&&(i<perc2);i++ )
     st[i]='=';*/

  //limite a 30 la lunghezza del testo
  char s[55];
  long i=0;
  for (i=0; (testo[i]!=0)&&(i<40);i++)
    s[i]=testo[i];
  for (; (i<40);i++)
    s[i]=' ';
  s[40]=0;


	printf("\r [%c] %3ld%% %20s %s\r",cont,perc1,st, s);
  fflush(stdout);
  /*============================================================================================================*/
}




void eraseMultiFile(const char* nomeMulti){

    long conta=0;

    FILE* fp;
    char nomeFile[1000];

    while(true){
      sprintf(nomeFile,nomeMulti,conta);
      fp=fopen(nomeFile,"r");
      if (fp!=NULL){
          conta++;
          fclose(fp);
          remove(nomeFile);
      }
      else
          return;
    }
}



//N.B. nel caso full-memory ignoriamo nel calcolo della memoria occupata quella per il vettore di bits gia' allocato. Infatti questo viene usato solo temporaneamente per riempire il vettore di conversione e poi deallocato. Anche quando viene usato assieme al vettore di lookup, l'accesso e' comunque sequenziale e possiamo fare affidamento alla gestione della cache del sistema operativo (che puo' tra l'altro avvalersi della memoria destinata ai buffers e non ancora utilizzata)
void alloca_memoria()
{
  const char * ERR_ALLOCAZIONE_MEM= "Memory allocation error.";
  long long memoria_impiegata; //usata solo per visualizzare il dato all'utente
  
  //allochiamo e carichiamo il vettore di bits:
  nodi_da_includere= new VettoreBits(numero_nodi); //assumiamo di avere sempre 1 bit per ogni nodo del grafo di partenza
  accertati_che( (nodi_da_includere != NULL), ERR_ALLOCAZIONE_MEM);

  imposta_vettore_bits(); //carica il vettore di bits con i valori appropriati e calcola 'numero_nodi_selezionati', che e' il # di nodi nel sottografo  
//printf("> numero nodi nel sottografo: %ld\n", numero_nodi_selezionati); //*****

  long long memoria_ideale= ((long long) numero_nodi_selezionati)*sizeof(unsigned long) + DIM_BUFFER_LETTURASCRITTURA; //approssimativamente la memoria richiesta

  //calcoliamo in 'capienza_vettore_lookup' la dimensione del vettore per la conversione:
  if(memoria_ideale <= memoria_max) //tanta memoria: possiamo usare 4 byte per ciascun nodo del sottografo
  {
    full_memory= true; //per ricordarci che e' il caso full-mem
    capienza_vettore_lookup= numero_nodi_selezionati; //corrispondenza 1:1 tra nodi del sottografo e quelli dello stesso che sono rappresentati
    memoria_impiegata= memoria_ideale;
  }
  else //(memoria_ideale > memoria_max) --> dobbiamo usare un approccio semi-external
  {
    full_memory= false; //indica approccio semi-external
    capienza_vettore_lookup= (unsigned long) (PERC_MIN_RAPPRESENTANTI * numero_nodi_selezionati); //ci mettiamo dapprima nel caso della richiesta minima
    if(capienza_vettore_lookup == 0)
      capienza_vettore_lookup++;

    long long memoria_prevista_semiext= ((long long) capienza_vettore_lookup)*sizeof(unsigned long) + DIM_BUFFER_LETTURASCRITTURA + numero_nodi/8; //nel caso semi-external occorre anche un vettore di bit (un bit per ciascun nodo nel grafo di partenza), oltre ad un vettore parziale per la conversione ed ai buffers.
    memoria_impiegata= memoria_prevista_semiext;
    
    if(memoria_prevista_semiext <= memoria_max) //questa volta siamo dentro le specifiche di memoria o sforiamo ancora?
    {      
      capienza_vettore_lookup += (memoria_max - memoria_prevista_semiext) / sizeof(unsigned long); //tutta la memoria in piu' la usiamo per il vettore parziale di conversione.
      memoria_impiegata= memoria_max;
    }
    else //non e' garantito nemmeno il minimo per l'approccio semi-external: che facciamo? (richiesta all'utente)
    {
      printf("Warning: memory not enough(required at least %lld bytes).\n", memoria_prevista_semiext);
      if( !esegui_comunque )
      {
        char car;
        printf("Continue?(y/n)\n");

        car= getchar();
        if((car != 'y') && (car != 'Y'))
        {
          printf("Aborted by user.\n");
          ABORT();
        }
      }
    }
  }

#ifdef TEST_SEMIEXT //*****
  full_memory= false;
  capienza_vettore_lookup= (unsigned long)(0.357f * numero_nodi_selezionati);
  if(capienza_vettore_lookup == 0) capienza_vettore_lookup++;
#endif
  
  vettore_Id_nodi= new unsigned long[capienza_vettore_lookup]; //allochiamo un vettore un po' piu' piccolo
  accertati_che( (vettore_Id_nodi != NULL), ERR_ALLOCAZIONE_MEM);
  
#ifdef VERBOSE_MODE
  printf("\n> Memoria allocata: circa %lld KB\n", memoria_impiegata/1024); //allochiamo sempre tutta la memoria che ci viene messa a disposizione
  if(!full_memory)
    printf("> Risorse di memoria minime --> semi-external con un vettore di %lu elementi\n\n", capienza_vettore_lookup);
#endif
}



void libera_memoria()
{
  if(vettore_Id_nodi != NULL)
  {
    delete [] vettore_Id_nodi;
    vettore_Id_nodi= NULL;
  }

  if(nodi_da_includere != NULL)
  {
    delete nodi_da_includere;
    nodi_da_includere= NULL;
  }  
}



void riconverti_nodi()
{
  alloca_memoria(); //imposta inoltre il vettore con i bits
  
  if(full_memory)
  {
    calcola_elementi_vettore_lookup(); //calcola la traduzione dell'ID di ciascun nodo del sottografo, mettendola nella corrispondente cella del vettore
  
    delete nodi_da_includere; //non ci serve piu' questo vettore di bits, il vettore di interi contiene tutte le informazioni necessarie
    nodi_da_includere= NULL;

    scansione_nodi(false); //usiamo solo il vettore di lookup che e' completo
  }
  else //approccio semi-external
  {
    calcola_elementi_vettore_lookup(); //in questo caso calcola i valori di conversione solo per i rappresentanti.

    //in questo caso ci serve ancora il vettore di bits

    scansione_nodi(true); //usiamo il vettore di lookup parziale ed il vettore di bits
  }
      
  libera_memoria();

#ifdef VERBOSE_MODE
  printf("\n  Conversione ID nodi terminata\n");
#endif
}




void stampaHelp(int arg,char** args){
  arg=0;
  cout<<"This utility is not designed for stand alone use. It's used as a part of others utilities.\n";
  cout<<"\n";
  printf("A program that finds the IDs in the original graph starting from a re-hashed subgraph");
  printf("It requires at least %.2f bit of main memory for each node of subgraph and 1 bit for each node of original graph\n", 32*PERC_MIN_RAPPRESENTANTI);
  printf("\nUsage:\n\n");
  printf("%s [options]... <memory> <setOfSubgraphNodes%%> <numOfOriginalNodes> <inputListOfNodes%%> <outputListOfNodes%%>\n", args[0]);
  printf("<memory>:                available memory in MB\n");
  printf("<setOfSubgraphNodes%%>:  name of multifile that contains the nodes of the subgraph \n");
  printf("<numOfOriginalNodes>:    number of nodes in the original graph\n");
  printf("<inputListOfNodes%%>:    list of IDs of subgraph\n");
  printf("<outputListOfNodes%%>:   list of output nodes\n");

  printf("options:\n"
         "    -force        execute program even if the memory is not enough\n"
         "-b, -bits         specify that <setOfSubgraphNodes%%> if a vector of bits and not a list of nodes\n"
         "-e, -except       specify that <inputListOfNodes%%> contains nodes not in the subgraph \n"
         "\n");
         
  printf("%% implies that multifile name must contain %%d\n\n");
}


//uscita dal programma per errore
#define ERRORE_PAR(STRINGA) {fprintf(stderr,"%s\nUsare l'opzione -h per l'help\n",STRINGA);EXIT(1);}


int main(int argc,char* argv[])
{
    stampaLicenza(argc,argv);

    //signal(SIGABRT, chiudi_tutto); //permette di uscire ordinatamente in seguito ad una assert() fallita o ad altro errore

    //poniamo i valori di default per i flag
    esegui_comunque= false;
    nodi_input_come_bits= false;
    escludi_nodi_specificati= false;
    
    struct option long_options[] =
    {
      {"force", no_argument, NULL, '!'},
      {"bits", no_argument, NULL, 'b'},
      {"escludi", no_argument, NULL, 'e'},      
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index= 0;

    int car;

    do //esaminiamo le opzioni
    {
      car= getopt_long_only(argc, argv, "fbe", long_options, &option_index);

      switch(car)
      {
          
        case '!':
          esegui_comunque= true;
          break;

        case 'b':
          nodi_input_come_bits= true;
          break;

        case 'e':
          escludi_nodi_specificati= true;
          break;
                                                                                                                                                                                                                                                
        case -1: //fine delle opzioni, ignoriamo il valore
          break;

        case '?': //parametro sconosciuto: lo ignoriamo, tanto il messaggio d'errore lo stampa getopt()
          EXIT(1);
          break;

        default:
          fprintf(stderr, "Errore imprevisto nell'esaminare la riga di comando\n");
          EXIT(1);
      }

    } while(car != -1);

    if( argc - optind != 5) //contiamo quanti parametri che non sono opzioni restano: sono tutti i parametri dall'indice "optind" incluso in poi
      ERRORE_PAR("Numero di parametri errato.\n");

    //optind indica il primo dei parametri non-opzioni
    memoria_max=  atol(argv[optind + 0]) * ((long long)1024)*1024; //salva convertendo da MB a byte
    nome_mfile_nodi_sottografo= argv[optind + 1];
    numero_nodi=  atol(argv[optind + 2]);    
    nome_mfile_nodi_input= argv[optind + 3];
    nome_mfile_nodi_output= argv[optind + 4];

    //controllo dei multifile passati come argomento:
    if (!estMultiFileName(nome_mfile_nodi_sottografo)) ERRORE_PAR("Il nome del multifile di input <elencoNodiSottografo> deve contenere una substring '%d' ");        
    if (!estMultiFileName(nome_mfile_nodi_input)) ERRORE_PAR("Il nome del multifile di input <listaNodiInput> deve contenere una substring '%d' ");    
    if (!estMultiFileName(nome_mfile_nodi_output)) ERRORE_PAR("Il nome del multifile di output <listaNodiOutput> deve contenere una substring '%d' ");

    numPartiNodiSottografo= contaNumFilesMultiFile(nome_mfile_nodi_sottografo);
    if(numPartiNodiSottografo<=0)
      ERRORE_PAR("Errore nei parametri: numero files del multifile <elencoNodiSottografo>");
    
    numPartiNodiInput= contaNumFilesMultiFile(nome_mfile_nodi_input);
    if(numPartiNodiInput<=0)
      ERRORE_PAR("Errore nei parametri: numero files del multifile <elencoNodiInput>");
      
#ifdef VERBOSE_MODE
    printf("\n> Il grafo ha %lu nodi.\n", numero_nodi);
#endif

    riconverti_nodi();

#ifdef TEST_SEMIEXT
printf("\n * Flag TEST_SEMIEXT\n");
printf("\n");
#endif
    
    EXIT(0);
}

