#include "../common/scc_utl/VettoreBits.h"
#include "semi_ext.h"
#include "main.h"
VettoreBits * nodi_da_includere= NULL;
unsigned long capienza_vettore_lookup; //indica quanti nodi sono mappati direttamente nel vettore di conversione. Informazione significativa solo nell'approccio semiexternal.
unsigned long nodi_per_rappresentante; //rapporto tra il numero nodi del grafo di partenza e la capienza del vettore di lookup. C'e' poi la correzione in caso di resto, in modo che ad ogni nodo del grafo corrisponda effettivamente un nodo del vettore di lookup.




//passiamo da un nome di multifile ad un nome di file (il primo file, 0, del multifile, assumendo che sia un unico file)
//la stringa creata va poi esplicitamente deallocata
char * crea_nome_file_da_nome_multifile(const char * nome_multifile)
{
  int lung_nome;
  char * stringa_creata;
  lung_nome= strlen(nome_multifile) + 1; //un carattere in piu' per il NULL finale
  stringa_creata= new char[lung_nome];
  snprintf(stringa_creata, lung_nome, nome_multifile, 0);

  return stringa_creata;
}



//inizializza il vettore di bits in memoria leggendo il file contenente gia' un vettore di bits.
//inverte il vettore dopo il caricamento se c'e' l'opzione che dice di escludere i nodi specificati da file.
void carica_vettore_bits_da_file_bits()
{
  accertati_che(numPartiNodiInput==1, "Errore: assumiamo che il multifile di input con il vettore dei nodi da includere/escludere sia composto da un unico file");

  char * nome_file_bits= crea_nome_file_da_nome_multifile(nome_mfile_nodi_sottografo); //poiche' VettoreBits usa dei normali file per l'I/O

  bool ok= nodi_da_includere->caricaDaFile(nome_file_bits); //ripristiniamo il vettore di bits dei nodi passivi dal disco. Serve perche' la selezione degli archi evita di caricare quelli diretti verso nodi passivi
  accertati_che( ok, MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_sottografo);

  if(escludi_nodi_specificati) //i bit sono da intendersi invertiti
    nodi_da_includere->NOT( *nodi_da_includere );

  delete [] nome_file_bits;
}



void carica_vettore_bits_da_lista_nodi()
{
  unsigned long nodo;

  multiFileReader mf_nodi_sel(nome_mfile_nodi_sottografo, numPartiNodiSottografo);
  BufferLetturaMultifile<unsigned long> buffer_nodi_sel(&mf_nodi_sel, DIM_BUFFER_LETTURASCRITTURA);

  bool ris= buffer_nodi_sel.get(nodo);
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_sottografo);

  if(escludi_nodi_specificati)
  {
    nodi_da_includere->reset(true); //tutto a 1 per default
    while( ris )
    {
      accertati_che( (nodo < numero_nodi), "Letto un nodo fuori range dal file con i nodi da selezionare: %lu", nodo);
      long errore;
      nodi_da_includere->set(nodo, false, errore); //resetta (a false)
      accertati_che( (errore==0), MSG_ERRORE_BITS);

      ris= buffer_nodi_sel.get(nodo);
    }
  }
  else //i nodi indicati vanno inclusi
  {
    nodi_da_includere->reset(false); //azzera tutto
    while( ris )
    {
      accertati_che( (nodo < numero_nodi), "Letto un nodo fuori range dal file con i nodi da selezionare: %lu", nodo);
      long errore;
      nodi_da_includere->set(nodo, true, errore); //setta (a true)
      accertati_che( (errore==0), MSG_ERRORE_BITS);
//printf("* %lu\n", nodo); //*****
      ris= buffer_nodi_sel.get(nodo);
    }
  }

  mf_nodi_sel.close();
}



//sono gestiti da due funzioni separate i casi in cui il file contenga una lista di nodi, oppure sia un vettore di bit
void imposta_vettore_bits()
{
  unsigned long prossimo_nodo_sel;
  
#ifdef VERBOSE_MODE
  printf("\n  Caricamento dei nodi da selezionare..."); fflush(stdout);
#endif

  if(nodi_input_come_bits)
    carica_vettore_bits_da_file_bits();
  else
    carica_vettore_bits_da_lista_nodi();

  //calcoliamo il # di nodi nel sottografo (i.e. tutti quelli inclusi):
  numero_nodi_selezionati= 0;
  
  bool found= nodi_da_includere->findNext1(prossimo_nodo_sel, 0);
  while(found)
  {
    numero_nodi_selezionati++;
    found= nodi_da_includere->findNext1(prossimo_nodo_sel);
  }  

//  nodi_da_includere->print("Nodi selezionati");

#ifdef VERBOSE_MODE
  printf("fatto.\n");
#endif
}



//riempie il vettore di lookup con le corrispondenze per la conversione degli identificatori dei nodi.
//il vettore contiene meno di 'numero_nodi_selezionati' posizioni, percio' viene salvata il valore solo dei nodi "rappresentanti", scelti equidistanziati nell'intervallo dei nodi del _sottografo_
//la posizione i-esima del vettore contiene l'identificatore dell'i-esimo nodo selezionato nel grafo di partenza. Se ad es. il primo nodo ad essere selezionato nel grafo di partenza e' il 5, mentre il secondo e' il 7, allora la posizione 0 del vettore di lookup conterra' 5, e la posizione 1 conterra' 7 (supp. di essere nel caso: 1 nodo <-> 1 rappresentante).
void calcola_elementi_vettore_lookup()
{
  bool found;
  unsigned long prossimo_nodo_sel, indice_vettore_corrente, conta_modulo;

  nodi_per_rappresentante= numero_nodi_selezionati / capienza_vettore_lookup; //inizializziamo questa variabile del modulo
//printf("magick #: %.3f\n", ((float)nodi_per_rappresentante)*(numero_nodi_selezionati % capienza_vettore_lookup)/numero_nodi_selezionati - 0.5f);
  if(numero_nodi_selezionati % capienza_vettore_lookup) //se c'era resto nella divisione..
    nodi_per_rappresentante++; //..ogni rappresentante deve assumersi un nodo in piu' in modo che siano tutti coperti

//printf("nodi_per_rappresentante=%lu, capienza_vettore_lookup=%lu\n", nodi_per_rappresentante, capienza_vettore_lookup); //*****

  conta_modulo= nodi_per_rappresentante - 1; //forziamo il "Terminal Count" alla prima iterazione
  indice_vettore_corrente= 0;
  unsigned long test= 0;
  found= nodi_da_includere->findNext1(prossimo_nodo_sel, 0);  
  while(found)
  {
    conta_modulo++;
    if(conta_modulo == nodi_per_rappresentante)
    {
      vettore_Id_nodi[indice_vettore_corrente]= prossimo_nodo_sel;

      indice_vettore_corrente++;      
      conta_modulo= 0;
    }
    test++;
    found= nodi_da_includere->findNext1(prossimo_nodo_sel);
  }

  //N.B. a questo punto possono non essere stati scritti tutti gli elementi del vettore di lookup, gli ultimi. Questi pero' "avanzano" e non verranno mai letti.
#ifdef VERBOSE_MODE
  if(indice_vettore_corrente < capienza_vettore_lookup)
    printf("spreco nel vettore di lookup: %lu elementi (%.2f%%)\n", (capienza_vettore_lookup-indice_vettore_corrente), (100.0f*(capienza_vettore_lookup-indice_vettore_corrente))/capienza_vettore_lookup);
  #endif
//printf("test=%lu\n", test);
//printf("numero_nodi_selezionati=%lu\n", numero_nodi_selezionati);
//printf("%lu/%lu  ( indice_vettore_corrente / capienza_vettore_lookup)\n", indice_vettore_corrente, capienza_vettore_lookup); //*****    
//accertati_che( capienza_vettore_lookup==indice_vettore_corrente, "spreco nel vettore di lookup\n"); //****

//for(unsigned long i=0; i < capienza_vettore_lookup ; i++)
//  printf("indice %lu)  --> vettore_Id_nodi[i]=%lu \n", i, vettore_Id_nodi[i]);  
}



//legge dal vettore di bits se il nodo e' marcato come selezionato o no
bool is_selezionato(unsigned long nodo)
{
  long errore;

  bool risposta= nodi_da_includere->get(nodo, errore);
  accertati_che( (errore==0), MSG_ERRORE_BITS);

  return risposta;
}



//a partire dall'ID di un nodo nel sottografo produce l'ID corrispondente per quel nodo nel grafo di partenza
//PRECondizioni: il nodo da convertire deve essere uno di quelli selezionati (deve appartenere al sottografo)
unsigned long revert_ID_al_volo(unsigned long ID_sottografo)
{
  const char * ERR_FINDNEXT1= "Error in proc findNext1() and revert_ID_al_volo()";
  unsigned long ID_riconvertito, indice_rappresentante, rappresentante_del_nodo, nodo_selezionato;
  
  indice_rappresentante= ID_sottografo / nodi_per_rappresentante; //l'indice nel vettore di lookup
  rappresentante_del_nodo= indice_rappresentante * nodi_per_rappresentante; //N.B. rappresentante dei nodi del sottografo
  
  ID_riconvertito= vettore_Id_nodi[indice_rappresentante]; // 'rappresentante_del_nodo'-esimo nodo selezionato nel grafo di partenza, ovvero "traduzione" del nodo rappresentante del sottografo

  if(rappresentante_del_nodo < ID_sottografo) //se era (rappresentante_del_nodo == ID_sottografo) stavamo gia' a posto, perche' 'ID_riconvertito' era proprio il valore cercato. Invece adesso dobbiamo cercare nel grafo di partenza altri (ID_sottografo - rappresentante_del_nodo) nodi selezionati, a partire dal nodo 'ID_riconvertito'.
  {
    unsigned long nodo_riconvertito_corrente= rappresentante_del_nodo;    
    bool found= nodi_da_includere->findNext1(nodo_selezionato, ID_riconvertito + 1); //cerchiamo a partire dal nodo DOPO il rappresentante nel grafo originale (un rappresentante ha sicuramente un nodo corrispondente perche' e' tra quelli selezionati)
    accertati_che(found, ERR_FINDNEXT1);
    
    nodo_riconvertito_corrente++; //'nodo_selezionato' rappresenta la conversione dell'ID per il nodo corrente, che adesso e' quello subito dopo il rappresentante.

    while(nodo_riconvertito_corrente < ID_sottografo) //finche' non abbiamo disponibile la traduzione del nodo che ci interessa
    {
      nodo_riconvertito_corrente++;
      found= nodi_da_includere->findNext1(nodo_selezionato);
      accertati_che(found, ERR_FINDNEXT1);
    }

    ID_riconvertito= nodo_selezionato;
  }
  
  return ID_riconvertito;
}



void scansione_nodi(bool semi_ext)
{
  unsigned long elem_letti;
  const char * ERR_NODO_OUTOFRANGE= "Error: found a node not present in the subgraph: %lu";
  
  //allocazione memoria buffer In/Out:
  const unsigned long capienza_buffer= DIM_BUFFER_LETTURASCRITTURA/sizeof(unsigned long);
  unsigned long * buffer_IO= new unsigned long[capienza_buffer];

  //allocazione dei multifile:
  multiFileReader mf_lettura(nome_mfile_nodi_input, numPartiNodiInput);
  
  eraseMultiFile(nome_mfile_nodi_output); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mf_scrittura(nome_mfile_nodi_output, 800*1024*1024);

  do
  {
    //input da file:
    elem_letti= mf_lettura.read(buffer_IO, sizeof(unsigned long), capienza_buffer);

    //elaborazione dei dati letti:
    if(semi_ext)      
      for(unsigned long i=0; i<elem_letti; i++)
      {
        unsigned long nodo= buffer_IO[i];
        accertati_che( (nodo < numero_nodi_selezionati), ERR_NODO_OUTOFRANGE, nodo);

        buffer_IO[i]= revert_ID_al_volo(nodo); //convertiamo il valore da ID nel sottografo a ID nel grafo di partenza (usando l'apposita funzione che combina il vettore di lookup e il vettore di bits)
//        printf("%lu --> %lu\n", nodo, buffer_IO[i]); //*****        
      }
    else //il vettore di look up contiene la mappatura di tutti i nodi del sottografo, basta leggerli
      for(unsigned long i=0; i<elem_letti; i++)
      {
        unsigned long nodo= buffer_IO[i];
        accertati_che( (nodo < numero_nodi_selezionati), ERR_NODO_OUTOFRANGE, nodo);

        buffer_IO[i]= vettore_Id_nodi[nodo]; //in questo caso per convertire da ID nel sottografo a ID nel grafo di partenza basta leggere la cella corrispondente al nodo in questione nel vettore di lookup
//        printf("%lu --> %lu\n", nodo, buffer_IO[i]); //*****
      }
    
    //output su file dei dati convertiti:
    mf_scrittura.write(buffer_IO, sizeof(unsigned long), elem_letti);
    
  } while(elem_letti==capienza_buffer);
  
  //chiusura dei multifile:
  mf_lettura.close();
  mf_scrittura.close();

  delete [] buffer_IO;
}
