#include "full_mem.h"
#include "main.h"



void carica_inclusi_da_vettore_bit()
{
  unsigned long tot_byte, nodo;
  unsigned char byte_corrente;

  tot_byte= numero_nodi/8;
  if(numero_nodi % 8)
    tot_byte++;

  multiFileReader mf_nodi_sel(nome_mfile_nodi_input, numPartiNodiInput);
  BufferLetturaMultifile<unsigned char> buffer_nodi_sel(&mf_nodi_sel, memoria_per_buffers);

  nodo= 0;

  for(unsigned long i=0; i < tot_byte - 1 ; i++)
  {
    long ris= buffer_nodi_sel.get(byte_corrente);
    accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

    for(unsigned long j=0; j<8 ; j++)
    {
      bool bit= byte_corrente & 1;

      if(bit) //se e' selezionato il suo bit e' a 1
      {
        vettore_Id_nodi[nodo] = nodo; //se lo includiamo all'inizio ha proprio il suo ID
        numero_nodi_selezionati++;
      }

      byte_corrente = byte_corrente >> 1;
      nodo++;
    }
  }


  long ris= buffer_nodi_sel.get(byte_corrente); //c'e' almeno un byte ancora da leggere
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

  for(  ; nodo < numero_nodi ; nodo++)
  {
    bool bit= byte_corrente & 1;

    if(bit)
    {
      vettore_Id_nodi[nodo] = nodo; //se lo includiamo all'inizio ha proprio il suo ID
      numero_nodi_selezionati++;
    }

    byte_corrente = byte_corrente >> 1;
  }
  mf_nodi_sel.close();
}



void carica_esclusi_da_vettore_bit()
{
  unsigned long tot_byte, nodo;
  unsigned char byte_corrente;

  tot_byte= numero_nodi/8;
  if(numero_nodi % 8)
    tot_byte++;

  multiFileReader mf_nodi_sel(nome_mfile_nodi_input, numPartiNodiInput);
  BufferLetturaMultifile<unsigned char> buffer_nodi_sel(&mf_nodi_sel, memoria_per_buffers);

  nodo= 0;

  for(unsigned long i=0; i < tot_byte - 1 ; i++)
  {
    long ris= buffer_nodi_sel.get(byte_corrente);
    accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

    for(unsigned long j=0; j<8 ; j++)
    {
      bool bit= byte_corrente & 1;

      if(bit) //se e' ESCLUSO il suo bit e' a 1
      {
        vettore_Id_nodi[nodo] = MAX_ULONG; //lo escludiamo
        numero_nodi_selezionati--;
      }

      byte_corrente = byte_corrente >> 1;
      nodo++;
    }
  }

  long ris= buffer_nodi_sel.get(byte_corrente); //c'e' almeno un byte ancora da leggere
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

  for(  ; nodo < numero_nodi ; nodo++)
  {
    bool bit= byte_corrente & 1;

    if(bit)
    {
      vettore_Id_nodi[nodo] = MAX_ULONG; //lo escludiamo
      numero_nodi_selezionati--;
    }

    byte_corrente = byte_corrente >> 1;
  }

  mf_nodi_sel.close();
}



void carica_nodi_inclusi()
{

  printf("Loading nodes to be left in the graph\n"); fflush(stdout);
  unsigned long nodo;

  multiFileReader mf_nodi_sel(nome_mfile_nodi_input, numPartiNodiInput);
  BufferLetturaMultifile<unsigned long> buffer_nodi_sel(&mf_nodi_sel, memoria_per_buffers);

  bool ris= buffer_nodi_sel.get(nodo);
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

  while( ris )
  {
    accertati_che( (nodo < numero_nodi), "Letto un nodo fuori range dal file con i nodi da selezionare: %lu, alla pos. %lu", nodo, numero_nodi_selezionati);
    vettore_Id_nodi[nodo] = nodo; //se lo includiamo all'inizio ha proprio il suo ID
    numero_nodi_selezionati++;
    ris= buffer_nodi_sel.get(nodo);
  }

  mf_nodi_sel.close();

#ifdef VERBOSE_DEBUG
  printf("fatto.\n");
#endif
}



void carica_nodi_esclusi()
{
  printf("Loading nodes to be removed from graph\n"); fflush(stdout);
  unsigned long nodo;

  multiFileReader mf_nodi_sel(nome_mfile_nodi_input, numPartiNodiInput);
  BufferLetturaMultifile<unsigned long> buffer_nodi_sel(&mf_nodi_sel, memoria_per_buffers);

  bool ris= buffer_nodi_sel.get(nodo);
  accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, nome_mfile_nodi_input);

  while( ris )
  {
    accertati_che( (nodo < numero_nodi), "Letto un nodo fuori range dal file con i nodi da selezionare: %lu, alla pos. %lu", nodo, numero_nodi_selezionati);
    vettore_Id_nodi[nodo] = MAX_ULONG; //cosi' lo escludiamo
    numero_nodi_selezionati--;
    ris= buffer_nodi_sel.get(nodo);
  }

  mf_nodi_sel.close();

#ifdef VERBOSE_DEBUG
  printf("fatto.\n");
#endif
}



//calcola gli identificatori di ogni nodo incluso nel sottografo da creare
void imposta_vettore_nodi()
{
  if(escludi_nodi_specificati)
  {
    //all'inizio sono tutti da selezionare ed escludiamo solo quelli specificati:
    numero_nodi_selezionati= numero_nodi;
    for(unsigned long i=0; i<numero_nodi ; i++) //marchiamo i nodi non selezionati con il valore max della lword
      vettore_Id_nodi[i]= i;

    if(nodi_input_come_bits)
      carica_esclusi_da_vettore_bit();
    else
      carica_nodi_esclusi(); //leggo da file i nodi da ESCLUDERE ed imposto il vettore
  }
  else //includere i nodi specificati
  {
    //all'inizio escludiamo tutti i nodi, poi includiamo solo quelli specificati:
    numero_nodi_selezionati= 0;
    for(unsigned long i=0; i<numero_nodi ; i++) //marchiamo i nodi non selezionati con il valore max della lword
      vettore_Id_nodi[i]= MAX_ULONG;

    if(nodi_input_come_bits)
      carica_inclusi_da_vettore_bit();
    else
      carica_nodi_inclusi(); //leggo da file i nodi da INCLUDERE ed imposto il vettore
  }


  unsigned long nodi_esclusi_precedenti= 0;

  for(unsigned long i=0; i<numero_nodi ; i++)
  {
    if(vettore_Id_nodi[i] == MAX_ULONG) //e' marcato da escludere
      nodi_esclusi_precedenti++;
    else
      vettore_Id_nodi[i] -= nodi_esclusi_precedenti;
  }
}



unsigned long scansione_grafo()
{
  struct grafoInfo informazioni_nodo, informazioni_nodo_output;
  char * nome_mfile_output_info, * nome_mfile_output_pred, * nome_mfile_output_succ;
  unsigned long predecessori_dati_in_output, successori_dati_in_output;
  long conta_attesa_refresh= INTERVALLO_REFRESH_INDICATORE;
  const char messaggio_progress_bar[]= "Extracting subgraph";

  //creazione dei nomi dei multifile di output:
  nome_mfile_output_info= crea_nome_multifile_grafo(".info");
  nome_mfile_output_pred= crea_nome_multifile_grafo(".pred");
  nome_mfile_output_succ= crea_nome_multifile_grafo(".succ");

  //allocazione dei multifile e buffer di input:
  multiFileReader mf_lettura_info(grafoINFO, numINFO);
  multiFileReader mf_lettura_pred(grafoPRED, numPRED);
  multiFileReader mf_lettura_succ(grafoSUCC, numSUCC);

  BufferLetturaMultifile<struct grafoInfo> buffer_lettura_info(&mf_lettura_info, memoria_per_buffers/6);
  BufferLetturaMultifile<unsigned long> buffer_lettura_pred(&mf_lettura_pred, memoria_per_buffers/6);
  BufferLetturaMultifile<unsigned long> buffer_lettura_succ(&mf_lettura_succ, memoria_per_buffers/6);

  //allocazione dei multifile e buffer di output:
  eraseMultiFile(nome_mfile_output_info); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mf_scrittura_info(nome_mfile_output_info, 800*1024*1024);
  BufferScritturaMultifile<struct grafoInfo> buffer_scrittura_info(&mf_scrittura_info, memoria_per_buffers/6);

  eraseMultiFile(nome_mfile_output_pred); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mf_scrittura_pred(nome_mfile_output_pred, 800*1024*1024);
  BufferScritturaMultifile<unsigned long> buffer_scrittura_pred(&mf_scrittura_pred, memoria_per_buffers/6);

  eraseMultiFile(nome_mfile_output_succ); //per non ritrovarci nel nuovo multifile delle parti di un multifile vecchio
  multiFileWriter mf_scrittura_succ(nome_mfile_output_succ, 800*1024*1024);
  BufferScritturaMultifile<unsigned long> buffer_scrittura_succ(&mf_scrittura_succ, memoria_per_buffers/6);


  predecessori_dati_in_output= 0;
  successori_dati_in_output= 0;


  for(unsigned long nodo_corr=0; nodo_corr < numero_nodi ; nodo_corr++)
  {
    long ris= buffer_lettura_info.get(informazioni_nodo);
    accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, grafoINFO);

    unsigned long Id_nodo= vettore_Id_nodi[nodo_corr];

    if(Id_nodo == MAX_ULONG) //non scriviamo nulla in output in questo caso
    {
      bool ok;
      ok= buffer_lettura_pred.skip( informazioni_nodo.inDeg );
      accertati_che( ok, MSG_ERRORE_SEEK, grafoPRED);
      ok= buffer_lettura_succ.skip( informazioni_nodo.outDeg );
      accertati_che( ok, MSG_ERRORE_SEEK, grafoSUCC);
    }
    else //il nodo e' selezionato per il sottografo di output
    {
      informazioni_nodo_output.puntIn= predecessori_dati_in_output;
      informazioni_nodo_output.puntOut= successori_dati_in_output;

      //esaminiamo i predecessori:
      for(unsigned long i=0; i < informazioni_nodo.inDeg ; i++)
      {
        unsigned long nodo_adiacente, Id_adiacente;

        long ris= buffer_lettura_pred.get(nodo_adiacente);
        accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, grafoPRED);

        Id_adiacente= vettore_Id_nodi[nodo_adiacente];
        if(Id_adiacente != MAX_ULONG)
        {
          buffer_scrittura_pred.put( Id_adiacente );
          predecessori_dati_in_output++;
        }
      }

      //esaminiamo i successori:
      for(unsigned long i=0; i < informazioni_nodo.outDeg ; i++)
      {
        unsigned long nodo_adiacente, Id_adiacente;

        long ris= buffer_lettura_succ.get(nodo_adiacente);
        accertati_che( (ris), MSG_ERRORE_LETTURA_MULTIFILE, grafoSUCC);

        Id_adiacente= vettore_Id_nodi[nodo_adiacente];
        if(Id_adiacente != MAX_ULONG)
        {
          buffer_scrittura_succ.put( Id_adiacente );
          successori_dati_in_output++;
        }
      }

      informazioni_nodo_output.inDeg= predecessori_dati_in_output - informazioni_nodo_output.puntIn;
      informazioni_nodo_output.outDeg= successori_dati_in_output - informazioni_nodo_output.puntOut;
      buffer_scrittura_info.put( informazioni_nodo_output );
    } //fine blocco else

    if(++conta_attesa_refresh >= INTERVALLO_REFRESH_INDICATORE)
    {
      conta_attesa_refresh= 0;
      indiceProgressione(messaggio_progress_bar, (long)(100.0*nodo_corr/numero_nodi), (long)(100.0*nodo_corr/numero_nodi));
    }

  } //fine blocco ciclo for dei nodi

  indiceProgressione(messaggio_progress_bar, (long)100.0, (long)100.0); printf("\n");

  accertati_che( predecessori_dati_in_output==successori_dati_in_output, "Errore: il # di prec e' diverso dal # di successori" );

  //chiusura dei buffer scrittura e dei multifile:
  mf_lettura_info.close();
  mf_lettura_pred.close();
  mf_lettura_succ.close();

  buffer_scrittura_info.chiudi();
  buffer_scrittura_pred.chiudi();
  buffer_scrittura_succ.chiudi();

  mf_scrittura_info.close();
  mf_scrittura_pred.close();
  mf_scrittura_succ.close();

  //liberiamo la mem. per le stringhe dei nomi:
  delete [] nome_mfile_output_info;
  delete [] nome_mfile_output_pred;
  delete [] nome_mfile_output_succ;

  return predecessori_dati_in_output;
}
