/*                                                                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
  file succ2archi.cpp :source code of succ2edges utility
                                                                     
  Copyright  (C) 2003 S.Millozzi, L.Laura, S.Leonardi                
  email      (millozzi,laura,leon)@dis.uniroma1.it                   
                                                                     
*********************************************************************
*********************************************************************
                                                                     
* This program is free software; you can redistribute it and/or     *
* modify it under the terms of the GNU General Public License       *
* as published by the Free Software Foundation; either version 2    *
* of the License, or (at your option) any later version.            *
*                                                                   *
* This program is distributed in the hope that it will be useful,   *
* but WITHOUT ANY WARRANTY; without even the implied warranty of    *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
* GNU General Public License for more details.                      *
*                                                                   *
* You should have received a copy of the GNU General Public License *
* along with this program; if not, write to the Free Software       *
* Foundation, Inc., 59 Temple Place -                               *
* Suite 330, Boston, MA  02111-1307, USA.                           *
                                                                     
*********************************************************************
*********************************************************************
                                                                     
*/                                                                   
#include "comuni.h"
#include "config.h"
#include <stdio.h>




/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
//  stampa i parametri d'uso
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void stampaHelp(int arg,char* argv[]){
    cout<<"This utility is not designed for stand alone use. It's used as a part of others utilities.\n";
    cout<<"\n";
  cout<< "This utility converts a succ format multifile in the 'extended' edges format\n";
  cout<< " \n";
  cout<< "Usage: "<<eliminaPath(argv[0])<<" memory -savesource|-deletesource nameInputMultifile%d nameOutputMultifile%d\n";
  cout<< " \n";
  cout<< "memory                     memory available(MB)\n";
  cout<< "-savesource|-deletesource  specify whether to delete or keep the input file\n";
  cout<< "nameInputMultifile%d       name of the multifile of the input graph (succ format)\n";
  cout<< "nameOutputMultifile%d      name of the output multifile ('extended' edges format)\n";
}





/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
/*XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX*/
void stampaMbyteCorrentiArchi(const char* s,long long nodoCorrente ,long long byte){
  /*======================== STAMPA MESSAGGI ===================================================================*/
  cout<<(char)13<<s<<" "<<nodoCorrente<<" nodes [written "<<byte/1048576<<"MB]";
  fflush(stdout);
  /*============================================================================================================*/
}





//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// funzione che converte il grafo da una rappresentazione a liste di succ
// a 32 bit a rappr ad archi di 32 bit.
void convertiSuccessori2archi(char* multiFileInput,long numFilesIn,char* multiFileOut,char* nomeFileReport,bool cancellaFileIn){

  //apre il file di input e quello in output
  bufferedThreadFileInputStream f=bufferedThreadFileInputStream(multiFileInput,numFilesIn,DIM_FILES_THREAD,NUM_FILES_THREAD,cancellaFileIn);
  bufferedThreadFileOutputStream fout=bufferedThreadFileOutputStream(multiFileOut,MAX_DIM_FILES,DIM_FILES_THREAD,NUM_FILES_THREAD);

//  #define DIM_BUFFER_LETTURA_SUCCESSORI   (1000000)

//  long* buf=new long[DIM_BUFFER_LETTURA_SUCCESSORI];
//  arco* bufOut=new arco[DIM_BUFFER_LETTURA_SUCCESSORI];

  long nodo;
  long numSucc;
  long long nodiLetti=0;
  long long byteScritti=0;
  long long iterazioni=MAX_ITER;
  cout<<"(*) Conversion: succ --> edges\n";fflush(stdout);
  long nodoMax=0;

  //mentre ci sono ancora nodi da leggere
  while (f.read(&nodo,sizeof(long),1)>0){

     //carco il nodo massimo
     if (nodo>nodoMax) nodoMax=nodo;
     
     //leggo i successori e ne controllo la validita'
     if (f.read(&numSucc,sizeof(long),1)<0){cout<<"Unexpected fatal error: successors number not found";FINE;}
     //if (numSucc>DIM_BUFFER_LETTURA_SUCCESSORI){cout<<"Errore: numero successori troppo elevato per il nodo "<<nodo<<".Possibile perdita di sincronismo.";}

     //leggo i successori
     //if (f.read(buf,sizeof(long),numSucc)<numSucc){cout<<"Numero di successori del nodo "<<nodo<<" insufficienti";FINE;}

     //leggo i successori e creo gli archi
     for (long i=0;i<numSucc;i++){
        long app;
        arco ar;

        //leggo ogni successore
        if (f.read(&app,sizeof(long),1)<1){cout<<"Unexpected fatal error in successors numberfor node "<<nodo<<"";FINE;}


        //carco il nodo massimo
        if (app>nodoMax) nodoMax=app;
        
        //creo arco
        ar.source=nodo;
        ar.dest=app;

        //scrivo arco
        if (fout.write(&ar,sizeof(arco),1)<1){cout<<"Unexpected fatal error: unable to write file";FINE;}
     }

     //scrivo sul file di out
     //if (fout.write(bufOut,sizeof(arco),numSucc)<numSucc){cout<<"Errore nella scrittura degli archi";FINE;}

     byteScritti+=sizeof(arco)*numSucc;
     nodiLetti++;

     if (iterazioni++>MAX_ITER/5){
        iterazioni=0;
        stampaMbyteCorrentiArchi("", nodiLetti ,byteScritti);
     }
  }

  stampaMbyteCorrentiArchi("", nodiLetti ,byteScritti);



  //aggiuge la label relativa al num di nodi...
  long numNodi=nodoMax+1;
  arco aux;
  aux.source=aux.dest=-numNodi;
  if (fout.write(&aux,sizeof(arco),1)<1){perror("Write error");FINE;}
  PRINT_ISOLATED_NODES_SPECIFICATION;

  
  fout.close();
  fout.cancella();

  f.close();
  f.cancella();

  /* STAMPA RAPPORTO ESECUZIONE */
  cout<<"\n\n";
	fprintf(stdout,"Nodes number .............%lld\n",numNodi );
	fprintf(stdout,"Edges number .............%lld\n",byteScritti/sizeof(arco) );
	
}




//NOTA multifile = + file che rapresentano lo stesso file ma con nomi sequenziali
int main(int arg,char** argv)
{
     stampaLicenza(arg,argv);
     if (arg!=5) STAMPA_ERRORE_PARAMETRI(argv[0],"Incorrect parameters.");


     bool cancellaSource;
     if (strcasecmp(argv[2],"-saveSource")==0)  cancellaSource=FALSE;
     else   if (strcasecmp(argv[2],"-deleteSource")==0)  cancellaSource=TRUE;
            else  STAMPA_ERRORE_PARAMETRI(argv[0],"opzioni -deleteSource e -saveSource non indicate");


     char* nomeGrafo=argv[3];
     char* nomeOut  =argv[4];
     long num=contaNumFilesMultiFile(argv[3]);

  //************************ CONTROLLO FILES DI INPUT ***************************
  if (getTotalSizeMultiFiles(nomeGrafo)<=0){cout<<"Grafo di input vuoto o nomi files errati\n";FINE};


     if (!estMultiFileName(nomeGrafo)) STAMPA_ERRORE_PARAMETRI(argv[0],"Il nome del multifile di input deve contenere una substring '%d' ");
     if (!estMultiFileName(nomeOut))   STAMPA_ERRORE_PARAMETRI(argv[0],"Il nome del multifile finale deve contenere una substring '%d' ");
     if (num<=0)   STAMPA_ERRORE_PARAMETRI(argv[0],"Non e' stato trovato nessun file del grafo");

     char nomeReport[1000];
     strcpy(nomeReport,nomeOut);
     strcat(nomeReport,".succ2archi.report");
		 convertiSuccessori2archi(nomeGrafo,num,nomeOut,nomeReport,cancellaSource);
		
		 EXIT(0);
}

